package com.vmware.vcloud.sdk;

import com.vmware.vcloud.api.rest.schema.AbstractVAppType;
import com.vmware.vcloud.api.rest.schema.CreateSnapshotParamsType;
import com.vmware.vcloud.api.rest.schema.DeployVAppParamsType;
import com.vmware.vcloud.api.rest.schema.ObjectFactory;
import com.vmware.vcloud.api.rest.schema.ProductSectionListType;
import com.vmware.vcloud.api.rest.schema.ReferenceType;
import com.vmware.vcloud.api.rest.schema.SnapshotSectionType;
import com.vmware.vcloud.api.rest.schema.TaskType;
import com.vmware.vcloud.api.rest.schema.UndeployVAppParamsType;
import com.vmware.vcloud.api.rest.schema.ovf.ProductSectionType;
import com.vmware.vcloud.api.rest.schema.ovf.SectionType;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import com.vmware.vcloud.sdk.constants.UndeployPowerActionType;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.bind.JAXBElement;

public abstract class AbstractVapp<T extends AbstractVAppType> extends VcloudEntity<AbstractVAppType>
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  private List<ProductSectionType> productSections;
  private SnapshotSectionType snapshotSection;

  AbstractVapp(VcloudClient client, AbstractVAppType abstractvAppType)
  {
    super(client, abstractvAppType);
    sortOvfSectionsAndReferences();
  }

  public T getResource()
  {
    return (AbstractVAppType)super.getResource();
  }

  public Boolean isDeployed()
  {
    return getResource().isDeployed();
  }

  public List<ProductSectionType> getProductSections()
    throws VCloudException
  {
    return this.productSections;
  }

  public static List<ProductSectionType> getProductSections(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    ProductSectionListType productSectionListType = (ProductSectionListType)SdkUtil.get(client, ref.getHref() + "/productSections/", 200);

    return productSectionListType.getProductSection();
  }

  public SnapshotSectionType getSnapshotSection()
    throws VCloudException
  {
    if (this.snapshotSection == null) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.DATA_NOT_FOUND));
    }

    return this.snapshotSection;
  }

  public static SnapshotSectionType getSnapshotSection(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    return (SnapshotSectionType)SdkUtil.get(client, vappRef.getHref() + "/snapshotSection", 200);
  }

  public Task updateProductSections(List<ProductSectionType> productSections)
    throws VCloudException
  {
    String productSectionsURL = getReference().getHref() + "/productSections/";

    ProductSectionListType productSectionListType = new ProductSectionListType();
    productSectionListType.getProductSection().addAll(productSections);
    ObjectFactory objectFactory_v1_5 = new ObjectFactory();
    String productSectionsRequestString = JAXBUtil.marshal(objectFactory_v1_5.createProductSectionList(productSectionListType));

    return new Task(getVcloudClient(), (TaskType)SdkUtil.put(getVcloudClient(), productSectionsURL, productSectionsRequestString, "application/vnd.vmware.vcloud.productSections+xml", 202));
  }

  private void sortOvfSectionsAndReferences()
  {
    this.productSections = new ArrayList();
    List sections = getResource().getSection();

    Iterator it = sections.iterator();
    while (it.hasNext()) {
      JAXBElement sec = (JAXBElement)it.next();
      if ((sec.getValue() instanceof ProductSectionType))
        this.productSections.add((ProductSectionType)sec.getValue());
      else if ((sec.getValue() instanceof SnapshotSectionType)) {
        this.snapshotSection = ((SnapshotSectionType)sec.getValue());
      }
      logger.log(Level.INFO, ((SectionType)sec.getValue()).getClass().getName());
    }
  }

  public Task deploy(boolean powerOn, int lease, boolean forceCustomization)
    throws VCloudException
  {
    String actionUrl = getReference().getHref() + "/action/deploy";

    String content = createAbstractVappDeployParamBody(powerOn, lease, Boolean.valueOf(forceCustomization));

    return executeAbstractVappAction(getVcloudClient(), actionUrl, content, "application/vnd.vmware.vcloud.deployVAppParams+xml", 202);
  }

  public static Task deploy(VcloudClient client, ReferenceType ref, boolean powerOn, int lease, boolean forceCustomization)
    throws VCloudException
  {
    String url = ref.getHref() + "/action/deploy";

    String content = createAbstractVappDeployParamBody(powerOn, lease, Boolean.valueOf(forceCustomization));

    return executeAbstractVappAction(client, url, content, "application/vnd.vmware.vcloud.deployVAppParams+xml", 202);
  }

  private static String createAbstractVappDeployParamBody(boolean powerOn, int lease, Boolean forceCustomization)
  {
    ObjectFactory of = new ObjectFactory();
    DeployVAppParamsType deployParamsType = new DeployVAppParamsType();
    deployParamsType.setPowerOn(Boolean.valueOf(powerOn));
    deployParamsType.setDeploymentLeaseSeconds(Integer.valueOf(lease));
    deployParamsType.setForceCustomization(forceCustomization);
    JAXBElement deployParams = of.createDeployVAppParams(deployParamsType);

    String content = JAXBUtil.marshal(deployParams);
    return content;
  }

  private static String createVappUnDeployParamBody(UndeployPowerActionType undeployPowerAction)
  {
    ObjectFactory objectFactory = new ObjectFactory();
    UndeployVAppParamsType undeployVAppParamsType = new UndeployVAppParamsType();
    undeployVAppParamsType.setUndeployPowerAction(undeployPowerAction.value());

    JAXBElement deployParams = objectFactory.createUndeployVAppParams(undeployVAppParamsType);

    String content = JAXBUtil.marshal(deployParams);
    return content;
  }

  public Task undeploy(UndeployPowerActionType undeployPowerAction)
    throws VCloudException
  {
    String actionUrl = getReference().getHref() + "/action/undeploy";

    String content = createVappUnDeployParamBody(undeployPowerAction);
    return executeAbstractVappAction(getVcloudClient(), actionUrl, content, "application/vnd.vmware.vcloud.undeployVAppParams+xml", 202);
  }

  public static Task undeploy(VcloudClient client, ReferenceType ref, UndeployPowerActionType undeployPowerAction)
    throws VCloudException
  {
    String url = ref.getHref() + "/action/undeploy";

    String content = createVappUnDeployParamBody(undeployPowerAction);
    return executeAbstractVappAction(client, url, content, "application/vnd.vmware.vcloud.undeployVAppParams+xml", 202);
  }

  public Task powerOn()
    throws VCloudException
  {
    String actionUrl = getReference().getHref() + "/power/action/powerOn";

    return executeAbstractVappAction(getVcloudClient(), actionUrl, null, null, 202);
  }

  public static Task powerOn(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return executeAbstractVappAction(client, ref.getHref() + "/power/action/powerOn", null, null, 202);
  }

  public Task powerOff()
    throws VCloudException
  {
    String actionUrl = getReference().getHref() + "/power/action/powerOff";

    return executeAbstractVappAction(getVcloudClient(), actionUrl, null, null, 202);
  }

  public static Task powerOff(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return executeAbstractVappAction(client, ref.getHref() + "/power/action/powerOff", null, null, 202);
  }

  public Task reset()
    throws VCloudException
  {
    String actionUrl = getReference().getHref() + "/power/action/reset";

    return executeAbstractVappAction(getVcloudClient(), actionUrl, null, null, 202);
  }

  public static Task reset(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return executeAbstractVappAction(client, ref.getHref() + "/power/action/reset", null, null, 202);
  }

  public Task suspend()
    throws VCloudException
  {
    String actionUrl = getReference().getHref() + "/power/action/suspend";

    return executeAbstractVappAction(getVcloudClient(), actionUrl, null, null, 202);
  }

  public static Task suspend(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return executeAbstractVappAction(client, ref.getHref() + "/power/action/suspend", null, null, 202);
  }

  public Task discardSuspend()
    throws VCloudException
  {
    String actionUrl = getReference().getHref() + "/action/discardSuspendedState";

    return executeAbstractVappAction(getVcloudClient(), actionUrl, null, null, 202);
  }

  public static Task discardSuspend(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return executeAbstractVappAction(client, ref.getHref() + "/action/discardSuspendedState", null, null, 202);
  }

  public Task shutdown()
    throws VCloudException
  {
    String actionUrl = getReference().getHref() + "/power/action/shutdown";

    return new Task(getVcloudClient(), (TaskType)SdkUtil.post(getVcloudClient(), actionUrl, null, null, 202));
  }

  public static Task shutdown(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.post(client, ref.getHref() + "/power/action/shutdown", null, null, 202));
  }

  public Task reboot()
    throws VCloudException
  {
    String vappActionUrl = getReference().getHref() + "/power/action/reboot";

    return new Task(getVcloudClient(), (TaskType)SdkUtil.post(getVcloudClient(), vappActionUrl, null, null, 202));
  }

  public static Task reboot(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.post(client, ref.getHref() + "/power/action/reboot", null, null, 202));
  }

  public Task removeAllSnapshots()
    throws VCloudException
  {
    String vappActionUrl = getReference().getHref() + "/action/removeAllSnapshots";

    return new Task(getVcloudClient(), (TaskType)SdkUtil.post(getVcloudClient(), vappActionUrl, null, null, 202));
  }

  public static Task removeAllSnapshots(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.post(client, ref.getHref() + "/action/removeAllSnapshots", null, null, 202));
  }

  public Task revertToCurrentSnapshot()
    throws VCloudException
  {
    String vappActionUrl = getReference().getHref() + "/action/revertToCurrentSnapshot";

    return new Task(getVcloudClient(), (TaskType)SdkUtil.post(getVcloudClient(), vappActionUrl, null, null, 202));
  }

  public static Task revertToCurrentSnapshot(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.post(client, ref.getHref() + "/action/revertToCurrentSnapshot", null, null, 202));
  }

  public Task createSnapshot(String name, String description, boolean memory, boolean quiesce)
    throws VCloudException
  {
    CreateSnapshotParamsType createSnapshotParams = new CreateSnapshotParamsType();
    createSnapshotParams.setName(name);
    createSnapshotParams.setDescription(description);
    createSnapshotParams.setMemory(Boolean.valueOf(memory));
    createSnapshotParams.setQuiesce(Boolean.valueOf(quiesce));
    String reqXml = JAXBUtil.marshal(new ObjectFactory().createCreateSnapshotParams(createSnapshotParams));

    String actionUrl = getReference().getHref() + "/action/createSnapshot";

    return new Task(getVcloudClient(), (TaskType)SdkUtil.post(getVcloudClient(), actionUrl, reqXml, "application/vnd.vmware.vcloud.createSnapshotParams+xml", 202));
  }

  public static Task createSnapshot(VcloudClient client, ReferenceType ref, String name, String description, boolean memory, boolean quiesce)
    throws VCloudException
  {
    CreateSnapshotParamsType createSnapshotParams = new CreateSnapshotParamsType();
    createSnapshotParams.setName(name);
    createSnapshotParams.setDescription(description);
    createSnapshotParams.setMemory(Boolean.valueOf(memory));
    createSnapshotParams.setQuiesce(Boolean.valueOf(quiesce));
    String reqXml = JAXBUtil.marshal(new ObjectFactory().createCreateSnapshotParams(createSnapshotParams));

    return new Task(client, (TaskType)SdkUtil.post(client, ref.getHref() + "/action/createSnapshot", reqXml, "application/vnd.vmware.vcloud.createSnapshotParams+xml", 202));
  }

  public static Task delete(VcloudClient client, ReferenceType ref)
    throws VCloudException
  {
    return asyncDelete(client, ref.getHref());
  }

  public Task delete()
    throws VCloudException
  {
    String url = getReference().getHref();
    Task task = asyncDelete(getVcloudClient(), url);
    return task;
  }

  private static Task asyncDelete(VcloudClient client, String anyUrl)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.delete(client, anyUrl, 202));
  }

  static Task executeAbstractVappAction(VcloudClient client, String actionUrl, String content, String contentType, int statusCode)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.post(client, actionUrl, content, contentType, statusCode));
  }
}