package com.vmware.vcloud.sdk;

import com.vmware.vcloud.api.rest.schema.DiskType;
import com.vmware.vcloud.api.rest.schema.LinkType;
import com.vmware.vcloud.api.rest.schema.ObjectFactory;
import com.vmware.vcloud.api.rest.schema.OwnerType;
import com.vmware.vcloud.api.rest.schema.ReferenceType;
import com.vmware.vcloud.api.rest.schema.TaskType;
import com.vmware.vcloud.api.rest.schema.VmsType;
import com.vmware.vcloud.sdk.constants.BusSubType;
import com.vmware.vcloud.sdk.constants.BusType;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.bind.JAXBElement;

public class Disk extends VcloudEntity<DiskType>
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  private ReferenceType vdcReference;

  Disk(VcloudClient client, DiskType diskType)
  {
    super(client, diskType);
    sortDiskFilesAndReferences();
  }

  public static Disk getDiskByReference(VcloudClient client, ReferenceType diskRef)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + diskRef.getHref());

    return new Disk(client, (DiskType)getResourceByReference(client, diskRef));
  }

  public static Disk getDiskById(VcloudClient client, String vCloudId)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_VCLOUD_ID_MSG) + " - " + vCloudId);

    return new Disk(client, (DiskType)getEntityById(client, vCloudId, "application/vnd.vmware.vcloud.disk+xml"));
  }

  public Task updateDisk(String name, String description, ReferenceType storageProfileRef)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    String diskUpdateUrl = getReference().getHref();
    if (name != null)
      ((DiskType)getResource()).setName(name);
    if (description != null)
      ((DiskType)getResource()).setDescription(description);
    if (storageProfileRef != null)
      ((DiskType)getResource()).setStorageProfile(storageProfileRef);
    String diskRequestXmlString = JAXBUtil.marshal(of.createDisk((DiskType)getResource()));

    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.PUT_URL_MSG) + " - " + diskUpdateUrl);

    return new Task(getVcloudClient(), (TaskType)SdkUtil.put(getVcloudClient(), diskUpdateUrl, diskRequestXmlString, "application/vnd.vmware.vcloud.disk+xml", 202));
  }

  public Task delete()
    throws VCloudException
  {
    String diskUrl = getReference().getHref();
    return deleteDisk(getVcloudClient(), diskUrl);
  }

  public static Task delete(VcloudClient client, ReferenceType diskRef)
    throws VCloudException
  {
    return deleteDisk(client, diskRef.getHref());
  }

  private static Task deleteDisk(VcloudClient client, String DiskUrl)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.delete(client, DiskUrl, 202));
  }

  private void sortDiskFilesAndReferences()
  {
    for (LinkType diskLink : ((DiskType)getResource()).getLink())
      if ((diskLink.getRel().equals("up")) && (diskLink.getType().equals("application/vnd.vmware.vcloud.vdc+xml")))
      {
        this.vdcReference = diskLink;
      }
  }

  public ReferenceType getVdcReference()
    throws VCloudException
  {
    if (this.vdcReference != null)
      return this.vdcReference;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
  }

  public ReferenceType getOwner()
    throws VCloudException
  {
    return ((DiskType)getResource()).getOwner().getUser();
  }

  public BusType getBus()
  {
    return BusType.fromValue(((DiskType)getResource()).getBusType());
  }

  public BusSubType getBusSub()
  {
    return BusSubType.fromValue(((DiskType)getResource()).getBusSubType());
  }

  public static void changeOwner(VcloudClient client, ReferenceType diskRef, ReferenceType userReference)
    throws VCloudException
  {
    String url = diskRef.getHref() + "/owner";
    changeOwner(url, userReference, client);
  }

  public void changeOwner(ReferenceType userReference)
    throws VCloudException
  {
    String url = getReference().getHref() + "/owner";
    changeOwner(url, userReference, getVcloudClient());
  }

  private static void changeOwner(String url, ReferenceType userReference, VcloudClient client)
    throws VCloudException
  {
    String requestXML = null;
    ObjectFactory objectFactory = new ObjectFactory();
    OwnerType ownerType = new OwnerType();
    ownerType.setUser(userReference);
    JAXBElement changeOwnerParams = objectFactory.createOwner(ownerType);

    requestXML = JAXBUtil.marshal(changeOwnerParams);
    SdkUtil.put(client, url, requestXML, "application/vnd.vmware.vcloud.owner+xml", 204);
  }

  public List<ReferenceType> getAttachedVms()
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + getReference().getHref() + "/attachedVms");

    VmsType vmsType = (VmsType)SdkUtil.get(getVcloudClient(), getReference().getHref() + "/attachedVms", 200);

    List vmRefs = new ArrayList();
    for (ReferenceType vmRef : vmsType.getVmReference()) {
      vmRefs.add(vmRef);
    }
    return vmRefs;
  }
}