package com.vmware.vcloud.sdk;

import com.vmware.vcloud.api.rest.schema.FileType;
import com.vmware.vcloud.api.rest.schema.FilesListType;
import com.vmware.vcloud.api.rest.schema.LinkType;
import com.vmware.vcloud.api.rest.schema.MediaType;
import com.vmware.vcloud.api.rest.schema.ObjectFactory;
import com.vmware.vcloud.api.rest.schema.OwnerType;
import com.vmware.vcloud.api.rest.schema.ReferenceType;
import com.vmware.vcloud.api.rest.schema.TaskType;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

public class Media extends VcloudEntity<MediaType>
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  private Map<String, FileType> uploadFileMap;
  private Map<String, FileType> uploadedFileMap;
  private ReferenceType catalogItemReference;
  private ReferenceType vdcReference;

  public Media(VcloudClient client, MediaType mediaResource_v1_5)
  {
    super(client, mediaResource_v1_5);
    sortMediaFilesAndReferences_v1_5();
  }

  public static Media getMediaByReference(VcloudClient client, ReferenceType mediaRef)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + mediaRef.getHref());

    return new Media(client, (MediaType)getResourceByReference(client, mediaRef));
  }

  public static Media getMediaById(VcloudClient client, String vCloudId)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_VCLOUD_ID_MSG) + " - " + vCloudId);

    return new Media(client, (MediaType)getEntityById(client, vCloudId, "application/vnd.vmware.vcloud.media+xml"));
  }

  public Task updateMedia(MediaType mediaType)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    String mediaUpdateUrl = getReference().getHref();
    String mediaRequestXmlString = JAXBUtil.marshal(of.createMedia(mediaType));

    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.PUT_URL_MSG) + " - " + mediaUpdateUrl);

    return new Task(getVcloudClient(), (TaskType)SdkUtil.put(getVcloudClient(), mediaUpdateUrl, mediaRequestXmlString, "application/vnd.vmware.vcloud.media+xml", 202));
  }

  public Task delete()
    throws VCloudException
  {
    String mediaUrl = getReference().getHref();
    return deleteMedia(getVcloudClient(), mediaUrl);
  }

  public static Task delete(VcloudClient client, ReferenceType mediaRef)
    throws VCloudException
  {
    return deleteMedia(client, mediaRef.getHref());
  }

  private static Task deleteMedia(VcloudClient client, String mediaUrl)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.delete(client, mediaUrl, 202));
  }

  public Task deleteMedia()
    throws VCloudException
  {
    String mediaUrl = getReference().getHref();
    if (isPartOfCatalogItem())
    {
      CatalogItem.delete(getVcloudClient(), getCatalogItemReference());
    }
    return new Task(getVcloudClient(), (TaskType)SdkUtil.delete(getVcloudClient(), mediaUrl, 202));
  }

  public static Task deleteMedia(VcloudClient client, ReferenceType mediaRef)
    throws VCloudException
  {
    return deleteMediaWithCatalogItem(client, mediaRef);
  }

  private static Task deleteMediaWithCatalogItem(VcloudClient client, ReferenceType mediaRef)
    throws VCloudException
  {
    Media media = getMediaByReference(client, mediaRef);
    String mediaUrl = mediaRef.getHref();
    if (media.isPartOfCatalogItem())
    {
      CatalogItem.delete(client, media.getCatalogItemReference());
    }
    return new Task(client, (TaskType)SdkUtil.delete(client, mediaUrl, 202));
  }

  public Set<String> getUploadFileNames()
  {
    return this.uploadFileMap.keySet();
  }

  public Set<String> getUploadedFileNames()
  {
    return this.uploadedFileMap.keySet();
  }

  public void uploadFile(String fileName, InputStream is, long size)
    throws VCloudException
  {
    if (((MediaType)getResource()).getStatus().intValue() == 1) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.MEDIA_RESOLVED_MSG));
    }

    if (this.uploadFileMap.isEmpty()) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NO_FILES_TO_UPLOAD_MSG));
    }

    FileType file = (FileType)this.uploadFileMap.get(fileName);
    if (file != null)
      RestUtil.uploadFile(getVcloudClient(), file, is, size, 0L, 0L);
    else
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.FILE_NOT_FOUND_MSG));
  }

  public void uploadFile(String fileName, InputStream is, long startByte, long endByte)
    throws VCloudException
  {
    if (((MediaType)getResource()).getStatus().intValue() == 1) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.MEDIA_RESOLVED_MSG));
    }

    if (this.uploadFileMap.isEmpty()) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NO_FILES_TO_UPLOAD_MSG));
    }

    FileType file = (FileType)this.uploadFileMap.get(fileName);
    if (file != null) {
      RestUtil.uploadFile(getVcloudClient(), file, is, file.getSize().longValue(), startByte, endByte);
    }
    else
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.FILE_NOT_FOUND_MSG));
  }

  public HashMap<String, FileType> monitorUpload()
    throws VCloudException
  {
    Media media = getMediaByReference(getVcloudClient(), getReference());

    HashMap filesStatusMap = new HashMap();
    FilesListType filesType = ((MediaType)media.getResource()).getFiles();
    List files = filesType.getFile();

    for (FileType file : files) {
      filesStatusMap.put(file.getName(), file);
    }
    return filesStatusMap;
  }

  private void sortMediaFilesAndReferences_v1_5()
  {
    this.uploadFileMap = new HashMap();
    this.uploadedFileMap = new HashMap();
    FilesListType fileList = ((MediaType)getResource()).getFiles();
    if (fileList != null) {
      List files = fileList.getFile();

      Iterator fileIt = files.iterator();

      while (fileIt.hasNext()) {
        FileType file = (FileType)fileIt.next();
        String fileName = file.getName();
        if (file.getSize().compareTo(file.getBytesTransferred()) == 0)
          this.uploadedFileMap.put(fileName, file);
        else {
          this.uploadFileMap.put(fileName, file);
        }
      }
    }
    for (LinkType mediaLink : ((MediaType)getResource()).getLink()) {
      if ((mediaLink.getRel().equals("catalogItem")) && (mediaLink.getType().equals("application/vnd.vmware.vcloud.catalogItem+xml")))
      {
        this.catalogItemReference = mediaLink;
      }if ((mediaLink.getRel().equals("up")) && (mediaLink.getType().equals("application/vnd.vmware.vcloud.vdc+xml")))
      {
        this.vdcReference = mediaLink;
      }
    }
  }

  public ReferenceType getCatalogItemReference()
    throws VCloudException
  {
    if (this.catalogItemReference != null)
      return this.catalogItemReference;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
  }

  public ReferenceType getVdcReference()
    throws VCloudException
  {
    if (this.vdcReference != null)
      return this.vdcReference;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
  }

  public ReferenceType getOwner()
    throws VCloudException
  {
    return ((MediaType)getResource()).getOwner().getUser();
  }

  public boolean isPartOfCatalogItem()
    throws VCloudException
  {
    return this.catalogItemReference != null;
  }
}