package com.vmware.vcloud.sdk;

import com.vmware.vcloud.api.rest.schema.ContainerType;
import com.vmware.vcloud.api.rest.schema.QueryResultCatalogRecordType;
import com.vmware.vcloud.api.rest.schema.QueryResultDiskRecordType;
import com.vmware.vcloud.api.rest.schema.QueryResultMediaRecordType;
import com.vmware.vcloud.api.rest.schema.QueryResultRecordsType;
import com.vmware.vcloud.api.rest.schema.QueryResultVAppRecordType;
import com.vmware.vcloud.api.rest.schema.QueryResultVAppTemplateRecordType;
import com.vmware.vcloud.api.rest.schema.QueryResultVMRecordType;
import com.vmware.vcloud.api.rest.schema.ReferencesType;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import com.vmware.vcloud.sdk.constants.query.FormatType;
import com.vmware.vcloud.sdk.constants.query.QueryCatalogField;
import com.vmware.vcloud.sdk.constants.query.QueryDiskField;
import com.vmware.vcloud.sdk.constants.query.QueryField;
import com.vmware.vcloud.sdk.constants.query.QueryMediaField;
import com.vmware.vcloud.sdk.constants.query.QueryRecordType;
import com.vmware.vcloud.sdk.constants.query.QueryReferenceType;
import com.vmware.vcloud.sdk.constants.query.QueryVAppField;
import com.vmware.vcloud.sdk.constants.query.QueryVAppTemplateField;
import com.vmware.vcloud.sdk.constants.query.QueryVMField;
import com.vmware.vcloud.sdk.constants.query.SortType;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

public class QueryService
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  protected VcloudClient client;

  protected QueryService(VcloudClient client)
  {
    this.client = client;
  }

  public String buildQuery(FormatType formatType)
  {
    StringBuilder query = new StringBuilder();
    if (formatType != null) {
      query.append("&format=");

      query.append(formatType.value());
    }
    return query.toString();
  }

  public String buildQuery(String queryTypeValue)
    throws VCloudException
  {
    StringBuilder query = new StringBuilder();
    if (queryTypeValue == null) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.QUERY_TYPE_NOT_FOUND));
    }

    query.append("/query?type=");

    query.append(queryTypeValue);
    return query.toString();
  }

  public String buildQuery(QueryParams<? extends QueryField> queryParams)
    throws VCloudException
  {
    StringBuilder query = new StringBuilder();
    if (queryParams != null) {
      for (QueryField sortField : queryParams.getSortFields().keySet())
        if (((SortType)queryParams.getSortFields().get(sortField)).equals(SortType.SORT_ASC))
        {
          query.append("&sortAsc=");

          query.append(sortField.value());
        } else {
          query.append("&sortDesc=");

          query.append(sortField.value());
        }
      int metadatafieldSize;
      if ((queryParams.getFields().size() > 0) || (queryParams.getMetadataFields().size() > 0))
      {
        query.append("&fields=");

        int fieldSize = 0;
        for (QueryField field : queryParams.getFields()) {
          query.append(field.value());
          fieldSize++;
          if (fieldSize != queryParams.getFields().size()) {
            query.append(",");
          }
        }

        if ((fieldSize > 0) && (queryParams.getMetadataFields().size() > 0)) {
          query.append(",");
        }

        metadatafieldSize = 0;
        for (String metadataField : queryParams.getMetadataFields()) {
          query.append(metadataField);
          metadatafieldSize++;
          if (metadatafieldSize != queryParams.getMetadataFields().size())
          {
            query.append(",");
          }
        }
      }

      if (queryParams.getPage() != null) {
        query.append("&page=");

        query.append(queryParams.getPage());
      }
      if (queryParams.getOffset() != null) {
        query.append("&offset=");

        query.append(queryParams.getOffset());
      }
      if (queryParams.getPageSize() != null) {
        query.append("&pageSize=");

        query.append(queryParams.getPageSize());
      }
      if ((queryParams.getFilter() != null) && 
        (queryParams.getFilter().getFilterText() != null) && (!queryParams.getFilter().getFilterText().isEmpty()))
      {
        query.append("&filter=");

        query.append(queryParams.getFilter().getFilterText());
      }
    }

    return query.toString();
  }

  public <T extends Result> T executeQuery(String query)
    throws VCloudException
  {
    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.GET_URL_MSG)).append(" - ").append(query).toString());

    ContainerType resultType = (ContainerType)SdkUtil.get(this.client, query, 200);
    if ((resultType instanceof ReferencesType)) {
      return new ReferenceResult(this.client, (ReferencesType)resultType);
    }

    return new RecordResult(this.client, (QueryResultRecordsType)resultType);
  }

  public ReferenceResult queryReferences(QueryReferenceType queryReferenceType, QueryParams<? extends QueryField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append(buildQuery(queryReferenceType.value())).append(buildQuery(queryParams)).append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public RecordResult queryRecords(QueryRecordType queryRecordType, QueryParams<? extends QueryField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append(buildQuery(queryRecordType.value())).append(buildQuery(queryParams)).append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult queryIdRecords(QueryRecordType queryRecordType, QueryParams<? extends QueryField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append(buildQuery(queryRecordType.value())).append(buildQuery(queryParams)).append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public ReferenceResult queryReferences(QueryReferenceType queryReferenceType)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append(buildQuery(queryReferenceType.value())).append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public RecordResult queryRecords(QueryRecordType queryRecordType)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append(buildQuery(queryRecordType.value())).append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult queryIdRecords(QueryRecordType queryRecordType)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append(buildQuery(queryRecordType.value())).append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public ReferenceResult queryCatalogReferences()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/catalogs/query").append("?").append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public RecordResult<QueryResultCatalogRecordType> queryCatalogRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/catalogs/query").append("?").append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultCatalogRecordType> queryCatalogIdRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/catalogs/query").append("?").append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultCatalogRecordType> queryCatalogRecords(QueryParams<QueryCatalogField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/catalogs/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultCatalogRecordType> queryCatalogIdRecords(QueryParams<QueryCatalogField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/catalogs/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public ReferenceResult queryCatalogReferences(QueryParams<QueryCatalogField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/catalogs/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public ReferenceResult queryvAppReferences()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vApps/query").append("?").append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public RecordResult<QueryResultVAppRecordType> queryvAppRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vApps/query").append("?").append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultVAppRecordType> queryvAppIdRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vApps/query").append("?").append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultVAppRecordType> queryvAppRecords(QueryParams<QueryVAppField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vApps/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultVAppRecordType> queryvAppIdRecords(QueryParams<QueryVAppField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vApps/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public ReferenceResult queryvAppReferences(QueryParams<QueryVAppField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vApps/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public ReferenceResult queryVmReferences()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vms/query").append("?").append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public RecordResult<QueryResultVMRecordType> queryVmRecords(QueryParams<QueryVMField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vms/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultVMRecordType> queryVmIdRecords(QueryParams<QueryVMField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vms/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultVMRecordType> queryVmRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vms/query").append("?").append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultVMRecordType> queryVmIdRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vms/query").append("?").append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public ReferenceResult queryVmReferences(QueryParams<QueryVMField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vms/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public ReferenceResult queryvAppTemplateReferences()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vAppTemplates/query").append("?").append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public RecordResult<QueryResultVAppTemplateRecordType> queryvAppTemplateRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vAppTemplates/query").append("?").append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultVAppTemplateRecordType> queryvAppTemplateIdRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vAppTemplates/query").append("?").append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultVAppTemplateRecordType> queryvAppTemplateRecords(QueryParams<QueryVAppTemplateField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vAppTemplates/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultVAppTemplateRecordType> queryvAppTemplateIdRecords(QueryParams<QueryVAppTemplateField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vAppTemplates/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public ReferenceResult queryvAppTemplateReferences(QueryParams<QueryVAppTemplateField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/vAppTemplates/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public ReferenceResult queryMediaReferences()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/mediaList/query").append("?").append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public RecordResult<QueryResultMediaRecordType> queryMediaRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/mediaList/query").append("?").append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultMediaRecordType> queryMediaIdRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/mediaList/query").append("?").append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultMediaRecordType> queryMediaRecords(QueryParams<QueryMediaField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/mediaList/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultMediaRecordType> queryMediaIdRecords(QueryParams<QueryMediaField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/mediaList/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public ReferenceResult queryMediaReferences(QueryParams<QueryMediaField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/mediaList/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public ReferenceResult queryDiskReferences()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/disks/query").append("?").append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }

  public RecordResult<QueryResultDiskRecordType> queryDiskRecords(QueryParams<QueryDiskField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/disks/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultDiskRecordType> queryDiskIdRecords(QueryParams<QueryDiskField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/disks/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultDiskRecordType> queryDiskRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/disks/query").append("?").append(buildQuery(FormatType.RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public RecordResult<QueryResultDiskRecordType> queryDiskIdRecords()
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/disks/query").append("?").append(buildQuery(FormatType.ID_RECORD_VIEW)).toString();

    return (RecordResult)executeQuery(query);
  }

  public ReferenceResult queryDiskReferences(QueryParams<QueryDiskField> queryParams)
    throws VCloudException
  {
    String query = new StringBuilder().append(this.client.getVcloudApiURL()).append("/disks/query").append("?").append(buildQuery(queryParams)).append(buildQuery(FormatType.REFERENCE_VIEW)).toString();

    return (ReferenceResult)executeQuery(query);
  }
}