package com.vmware.vcloud.sdk;

import com.vmware.vcloud.api.rest.schema.ErrorType;
import com.vmware.vcloud.api.rest.schema.FileType;
import com.vmware.vcloud.api.rest.schema.LinkType;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import com.vmware.vcloud.sdk.exception.UnauthorizedAccessException;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.entity.StringEntity;

public class RestUtil
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");

  public static Response delete(VcloudClient client, String url)
    throws VCloudException
  {
    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.DELETE_URL_MSG)).append(" - ").append(url).toString());

    HttpDelete httpDelete = new HttpDelete(url);
    Response response = new Response();
    InputStream inputStream = null;
    try {
      httpDelete.setHeader("Accept", (String)VcloudClient.SUPPORTED_SDK_VERSIONS.get(client.getVcloudClientVersion()));

      httpDelete.setHeader(client.getVcloudTokenHeader(), client.getVcloudToken());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.DELETE_URL_MSG)).append(" - ").append(url).toString());

      HttpResponse httpResponse = client.getClient().execute(httpDelete);
      response.setResonseStatusCode(httpResponse.getStatusLine().getStatusCode());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_CODE_MSG)).append(" - ").append(response.getResponseStatusCode()).toString());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_MSG)).append(" - ").append(httpResponse.getStatusLine().getReasonPhrase()).toString());

      if (httpResponse.getEntity() != null)
        inputStream = httpResponse.getEntity().getContent();
      response.setResponseXml(getContent(inputStream));
      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.RESPONSE_MSG)).append(" - ").append(response.getResponseXml()).toString());
    }
    catch (UnsupportedEncodingException e)
    {
      logger.log(Level.SEVERE, e.getMessage());
      throw new VCloudRuntimeException(e);
    } catch (IOException e) {
      logger.log(Level.SEVERE, e.getMessage());
      throw new VCloudRuntimeException(e);
    } catch (RuntimeException e) {
      logger.log(Level.SEVERE, e.getMessage());
      httpDelete.abort();
      throw new VCloudRuntimeException(e);
    } finally {
      if (inputStream != null) {
        try {
          inputStream.close();
        } catch (IOException e) {
          throw new VCloudRuntimeException(e);
        }
      }
    }
    return response;
  }

  public static Response put(VcloudClient client, String url, String content, String contentType, String charset)
    throws VCloudException
  {
    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.PUT_URL_MSG)).append(" - ").append(url).toString());

    HttpPut httpPut = new HttpPut(url);
    Response response = new Response();
    InputStream inputStream = null;
    try {
      httpPut.setHeader("Accept", (String)VcloudClient.SUPPORTED_SDK_VERSIONS.get(client.getVcloudClientVersion()));

      httpPut.setHeader(client.getVcloudTokenHeader(), client.getVcloudToken());

      if ((content != null) && (content.length() > 0) && (contentType != null) && (contentType.length() > 0))
      {
        StringEntity req = new StringEntity(content, contentType, charset);

        logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.PUT_REQUEST_BODY)).append(" - ").append(content).toString());

        logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.PUT_CONTENT_TYPE)).append(" - ").append(contentType).toString());

        httpPut.setEntity(req);
      }
      HttpResponse httpResponse = client.getClient().execute(httpPut);
      response.setResonseStatusCode(httpResponse.getStatusLine().getStatusCode());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_CODE_MSG)).append(" - ").append(response.getResponseStatusCode()).toString());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_MSG)).append(" - ").append(httpResponse.getStatusLine().getReasonPhrase()).toString());

      if (httpResponse.getEntity() != null)
        inputStream = httpResponse.getEntity().getContent();
      response.setResponseXml(getContent(inputStream));
      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.RESPONSE_MSG)).append(" - ").append(response.getResponseXml()).toString());
    }
    catch (UnsupportedEncodingException e)
    {
      logger.log(Level.SEVERE, e.getMessage());
      throw new VCloudRuntimeException(e);
    } catch (IOException e) {
      logger.log(Level.SEVERE, e.getMessage());
      throw new VCloudRuntimeException(e);
    } catch (RuntimeException e) {
      logger.log(Level.SEVERE, e.getMessage());
      httpPut.abort();
      throw new VCloudRuntimeException(e);
    } finally {
      if (inputStream != null) {
        try {
          inputStream.close();
        } catch (IOException e) {
          throw new VCloudRuntimeException(e);
        }
      }
    }
    return response;
  }

  static String redirectURL(HttpResponse response)
  {
    String redirectedUrl = response.getHeaders("Location")[0].getValue();

    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.REDIRECTED_URL_MSG)).append(" - ").append(redirectedUrl).toString());

    return redirectedUrl;
  }

  static Response login(VcloudClient client, String url, Header authHeader)
    throws UnauthorizedAccessException, VCloudException
  {
    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.LOGIN_URL_MSG)).append(" - ").append(url).toString());

    HttpPost httpPost = new HttpPost(url);
    Response response = new Response();
    HttpResponse httpResponse = null;
    InputStream inputStream = null;
    try {
      httpPost.setHeader("Accept", (String)VcloudClient.SUPPORTED_SDK_VERSIONS.get(client.getVcloudClientVersion()));

      httpPost.setHeader(authHeader);

      httpResponse = client.getClient().execute(httpPost);
      if (httpResponse.getStatusLine().getStatusCode() == 302) {
        httpPost.setURI(new URI(redirectURL(httpResponse)));
        httpResponse = client.getClient().execute(httpPost);
      }
      if (httpResponse.getHeaders("x-vcloud-authorization").length > 0)
      {
        client.setVcloudToken(httpResponse.getHeaders("x-vcloud-authorization")[0].getValue());
      }

      response.setResonseStatusCode(httpResponse.getStatusLine().getStatusCode());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_CODE_MSG)).append(" - ").append(response.getResponseStatusCode()).toString());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_MSG)).append(" - ").append(httpResponse.getStatusLine().getReasonPhrase()).toString());

      if (httpResponse.getEntity() != null)
        inputStream = httpResponse.getEntity().getContent();
      response.setResponseXml(getContent(inputStream));
      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.RESPONSE_MSG)).append(" - ").append(response.getResponseXml()).toString());
    }
    catch (UnsupportedEncodingException e)
    {
      logger.log(Level.SEVERE, e.getMessage());
      throw new VCloudRuntimeException(e);
    } catch (IOException e) {
      logger.log(Level.SEVERE, e.getMessage());
      throw new VCloudRuntimeException(e);
    } catch (RuntimeException e) {
      logger.log(Level.SEVERE, e.getMessage());
      httpPost.abort();
      e.printStackTrace();
      throw new VCloudRuntimeException(e);
    } catch (URISyntaxException e) {
      logger.log(Level.SEVERE, e.getMessage());
      throw new VCloudRuntimeException(e);
    } finally {
      if (inputStream != null) {
        try {
          inputStream.close();
          if (httpResponse.getStatusLine().getStatusCode() == 401)
          {
            throw new UnauthorizedAccessException(httpResponse.getStatusLine().getReasonPhrase());
          }
        } catch (IOException e) {
          throw new VCloudRuntimeException(e);
        }
      }
    }
    return response;
  }

  public static Response post(VcloudClient client, String url, String content, String contentType, String charset)
    throws VCloudException
  {
    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.POST_URL_MSG)).append(" - ").append(url).toString());

    HttpPost httpPost = new HttpPost(url);
    Response response = new Response();
    InputStream inputStream = null;
    try {
      httpPost.setHeader("Accept", (String)VcloudClient.SUPPORTED_SDK_VERSIONS.get(client.getVcloudClientVersion()));

      httpPost.setHeader(client.getVcloudTokenHeader(), client.getVcloudToken());

      if ((content != null) && (content.length() > 0) && (contentType != null) && (contentType.length() > 0))
      {
        StringEntity req = new StringEntity(content, contentType, charset);

        logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.POST_REQUEST_BODY)).append(" - ").append(content).toString());

        logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.POST_CONTENT_TYPE)).append(" - ").append(contentType).toString());

        httpPost.setEntity(req);
      }
      HttpResponse httpResponse = client.getClient().execute(httpPost);
      response.setResonseStatusCode(httpResponse.getStatusLine().getStatusCode());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_CODE_MSG)).append(" - ").append(response.getResponseStatusCode()).toString());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_MSG)).append(" - ").append(httpResponse.getStatusLine().getReasonPhrase()).toString());

      if (httpResponse.getEntity() != null)
        inputStream = httpResponse.getEntity().getContent();
      response.setResponseXml(getContent(inputStream));
      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.RESPONSE_MSG)).append(" - ").append(response.getResponseXml()).toString());
    }
    catch (UnsupportedEncodingException e)
    {
      logger.log(Level.SEVERE, e.getMessage());
      throw new VCloudRuntimeException(e);
    } catch (IOException e) {
      logger.log(Level.SEVERE, e.getMessage());
      throw new VCloudRuntimeException(e);
    } catch (RuntimeException e) {
      logger.log(Level.SEVERE, e.getMessage());
      httpPost.abort();
      throw new VCloudRuntimeException(e);
    } finally {
      if (inputStream != null) {
        try {
          inputStream.close();
        } catch (IOException e) {
          throw new VCloudRuntimeException(e);
        }
      }
    }
    return response;
  }

  static Response getSupportedVersions(VcloudClient client, String url)
    throws VCloudException
  {
    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.GET_SUPPORTED_VERSIONS_URL_MSG)).append(" - ").append(url).toString());

    HttpGet httpGet = new HttpGet(url);
    Response response = new Response();
    InputStream inputStream = null;
    try {
      HttpResponse httpResponse = client.getClient().execute(httpGet);
      response.setResonseStatusCode(httpResponse.getStatusLine().getStatusCode());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_CODE_MSG)).append(" - ").append(response.getResponseStatusCode()).toString());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_MSG)).append(" - ").append(httpResponse.getStatusLine().getReasonPhrase()).toString());

      if (httpResponse.getEntity() != null)
        inputStream = httpResponse.getEntity().getContent();
      response.setResponseXml(getContent(inputStream));
      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.RESPONSE_MSG)).append(" - ").append(response.getResponseXml()).toString());
    }
    catch (IOException e)
    {
      logger.log(Level.SEVERE, null, e);
      throw new VCloudRuntimeException(e);
    } catch (RuntimeException e) {
      logger.log(Level.SEVERE, e.getMessage());
      httpGet.abort();
      throw new VCloudRuntimeException(e);
    } finally {
      if (inputStream != null)
        try {
          inputStream.close();
        } catch (IOException e) {
          throw new VCloudRuntimeException(e);
        }
    }
    return response;
  }

  public static Response get(VcloudClient client, String url)
    throws VCloudException
  {
    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.GET_URL_MSG)).append(" - ").append(url).toString());

    HttpGet httpGet = new HttpGet(url);
    Response response = new Response();
    InputStream inputStream = null;
    try {
      httpGet.setHeader("Accept", (String)VcloudClient.SUPPORTED_SDK_VERSIONS.get(client.getVcloudClientVersion()));

      httpGet.setHeader(client.getVcloudTokenHeader(), client.getVcloudToken());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.GET_URL_MSG)).append(" - ").append(url).toString());

      HttpResponse httpResponse = client.getClient().execute(httpGet);
      response.setResonseStatusCode(httpResponse.getStatusLine().getStatusCode());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_CODE_MSG)).append(" - ").append(response.getResponseStatusCode()).toString());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_MSG)).append(" - ").append(httpResponse.getStatusLine().getReasonPhrase()).toString());

      if (httpResponse.getEntity() != null)
        inputStream = httpResponse.getEntity().getContent();
      response.setResponseXml(getContent(inputStream));
      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.RESPONSE_MSG)).append(" - ").append(response.getResponseXml()).toString());
    }
    catch (IOException e)
    {
      logger.log(Level.SEVERE, null, e);
      throw new VCloudRuntimeException(e);
    } catch (RuntimeException e) {
      logger.log(Level.SEVERE, e.getMessage());
      httpGet.abort();
      throw new VCloudRuntimeException(e);
    } finally {
      if (inputStream != null)
        try {
          inputStream.close();
        } catch (IOException e) {
          throw new VCloudRuntimeException(e);
        }
    }
    return response;
  }

  private static String getContent(InputStream is)
    throws IOException
  {
    if (is != null) {
      StringBuilder sb = new StringBuilder();
      try
      {
        BufferedReader reader = new BufferedReader(new InputStreamReader(is, "UTF-8"));
        String line;
        while ((line = reader.readLine()) != null)
          sb.append(new StringBuilder().append(line).append("\n").toString());
      }
      finally {
        is.close();
      }
      return sb.toString();
    }
    return "";
  }

  public static void uploadFile(VcloudClient client, FileType file, InputStream is, long size, long startByte, long endByte)
    throws VCloudException
  {
    String fileURL = getDefaultUploadURL(file);
    uploadFile(client, file.getName(), fileURL, is, size, startByte, endByte);
  }

  private static void uploadFile(VcloudClient client, String fileName, String fileURL, InputStream is, long size, long startByte, long endByte)
    throws VCloudException
  {
    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.UPLOAD_FILE_SIZE_MSG)).append(" - ").append(fileName).append("\t").append(size).toString());

    HttpPut httpPut = new HttpPut(fileURL);
    httpPut.setHeader(client.getVcloudTokenHeader(), client.getVcloudToken());

    BufferedInputStream bufferedInputStream = new BufferedInputStream(is);
    InputStream inputStream = null;
    InputStreamEntity inputStreamRequestEntity = new InputStreamEntity(bufferedInputStream, size);

    if (endByte > startByte) {
      inputStreamRequestEntity = new InputStreamEntity(bufferedInputStream, endByte - startByte + 1L);

      httpPut.setHeader("Content-Range", new StringBuilder().append("bytes ").append(Long.toString(startByte)).append("-").append(Long.toString(endByte)).append("/").append(Long.toString(size)).toString());
    }

    inputStreamRequestEntity.setChunked(true);
    httpPut.setEntity(inputStreamRequestEntity);
    try
    {
      HttpResponse httpResponse = client.getClient().execute(httpPut);
      int status = httpResponse.getStatusLine().getStatusCode();
      if (httpResponse.getEntity() != null) {
        inputStream = httpResponse.getEntity().getContent();
      }
      if (status == 200) {
        logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.UPLOAD_COMPLETE_MSG)).append(" - ").append(fileName).toString());
      }
      else
      {
        logger.log(Level.SEVERE, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.UPLOAD_FAILED_MSG)).append(" - ").append(fileURL).toString());

        logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_MSG)).append(" - ").append(httpResponse.getStatusLine().getReasonPhrase()).toString());

        logger.log(Level.INFO, getContent(httpResponse.getEntity().getContent()));

        logger.log(Level.SEVERE, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.FILE_UPLOAD_FAILED_MSG)).append(" - ").append(fileURL).toString());

        if (inputStream != null)
        {
          throw new VCloudRuntimeException(getContent(inputStream));
        }

        throw new VCloudRuntimeException(new StringBuilder().append(httpResponse.getStatusLine().getStatusCode()).append("").toString());
      }
    }
    catch (IOException e) {
      e.printStackTrace();
      logger.log(Level.SEVERE, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.IO_EXCEPTION_UPLOADING_MSG)).append(" - ").append(fileURL).toString());

      throw new VCloudRuntimeException(e.getMessage());
    } catch (RuntimeException e) {
      logger.log(Level.SEVERE, e.getMessage());
      httpPut.abort();
      throw new VCloudRuntimeException(e);
    } finally {
      if (inputStream != null)
        try {
          inputStream.close();
        } catch (IOException e) {
          throw new VCloudRuntimeException(e);
        }
    }
  }

  private static String getDefaultUploadURL(FileType file)
  {
    String fileHref = null;

    List links = file.getLink();
    Iterator linkIt = links.iterator();
    while (linkIt.hasNext()) {
      LinkType ln = (LinkType)linkIt.next();
      if (ln.getRel().equals("upload:default")) {
        fileHref = ln.getHref();
        break;
      }
    }
    return fileHref;
  }

  public static InputStream downloadFile(VcloudClient client, String downloadURL)
    throws VCloudException
  {
    logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.DOWNLOAD_URL_MSG)).append(" - ").append(downloadURL).toString());

    HttpGet httpGet = new HttpGet(downloadURL);
    DownloadInputStream inputStream = null;
    int statusCode;
    try
    {
      httpGet.setHeader("Accept", (String)VcloudClient.SUPPORTED_SDK_VERSIONS.get(client.getVcloudClientVersion()));

      httpGet.setHeader(client.getVcloudTokenHeader(), client.getVcloudToken());

      HttpResponse httpResponse = client.getClient().execute(httpGet);
      statusCode = httpResponse.getStatusLine().getStatusCode();
      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_CODE_MSG)).append(" - ").append(statusCode).toString());

      logger.log(Level.INFO, new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.STATUS_MSG)).append(" - ").append(httpResponse.getStatusLine().getReasonPhrase()).toString());
      void tmp202_199 = new RestUtil(); tmp202_199.getClass(); inputStream = new DownloadInputStream(httpGet, httpResponse);
    }
    catch (IOException e)
    {
      logger.log(Level.SEVERE, null, e);
      throw new VCloudRuntimeException(e);
    } catch (RuntimeException e) {
      logger.log(Level.SEVERE, e.getMessage());
      httpGet.abort();
      throw new VCloudRuntimeException(e);
    }

    if (statusCode != 200) {
      if (inputStream != null)
      {
        ErrorType errType = (ErrorType)JAXBUtil.unmarshallResource(inputStream);
        throw new VCloudException(errType);
      }

      throw new VCloudException(new StringBuilder().append(SdkUtil.getI18nString(SdkMessage.DOWNLOAD_FAILED_MSG)).append(" - ").append(downloadURL).toString());
    }

    return inputStream;
  }

  private class DownloadInputStream extends InputStream
  {
    private InputStream inputStream = null;
    private HttpGet getMethod = null;

    public DownloadInputStream(HttpGet getMethod, HttpResponse httpResponse) throws IOException
    {
      this.getMethod = getMethod;
      if (httpResponse.getEntity() != null)
        this.inputStream = httpResponse.getEntity().getContent();
    }

    public void close()
      throws IOException
    {
      this.getMethod.abort();
    }

    public int read()
      throws IOException
    {
      return this.inputStream.read();
    }
  }

  public static class SdkStatusCode
  {
    public static final int SC_OK = 200;
    public static final int SC_CREATED = 201;
    public static final int SC_ACCEPTED = 202;
    public static final int SC_NO_CONTENT = 204;
    public static final int SC_MOVED_TEMPORARILY = 302;
    public static final int SC_UNAUTHORIZED = 401;
  }
}