package com.vmware.vcloud.sdk.admin;

import com.vmware.vcloud.api.rest.schema.GroupsListType;
import com.vmware.vcloud.api.rest.schema.ReferenceType;
import com.vmware.vcloud.api.rest.schema.ReferencesType;
import com.vmware.vcloud.api.rest.schema.UserType;
import com.vmware.vcloud.api.rest.schema.extension.EntityReferencesType;
import com.vmware.vcloud.api.rest.schema.extension.EntityRightsType;
import com.vmware.vcloud.api.rest.schema.extension.UserEntityRightsType;
import com.vmware.vcloud.sdk.JAXBUtil;
import com.vmware.vcloud.sdk.SdkUtil;
import com.vmware.vcloud.sdk.VCloudException;
import com.vmware.vcloud.sdk.VcloudClient;
import com.vmware.vcloud.sdk.VcloudEntity;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.bind.JAXBElement;

public class User extends VcloudEntity<UserType>
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  private HashMap<String, ReferenceType> groupRefsByName;

  User(VcloudClient client, UserType userType_v1_5)
  {
    super(client, userType_v1_5);
    sortReferences_v1_5();
  }

  private void sortReferences_v1_5()
  {
    this.groupRefsByName = new HashMap();

    GroupsListType groupList = ((UserType)getResource()).getGroupReferences();
    if (groupList != null)
      for (ReferenceType groupRef : groupList.getGroupReference())
        this.groupRefsByName.put(groupRef.getName(), groupRef);
  }

  public static User getUserByReference(VcloudClient client, ReferenceType userRef)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + userRef.getHref());

    return new User(client, (UserType)getResourceByReference(client, userRef));
  }

  public static User getUserById(VcloudClient client, String vCloudId)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_VCLOUD_ID_MSG) + " - " + vCloudId);

    return new User(client, (UserType)getEntityById(client, vCloudId, "application/vnd.vmware.admin.user+xml"));
  }

  public User updateUser(UserType userType)
    throws VCloudException
  {
    com.vmware.vcloud.api.rest.schema.ObjectFactory of = new com.vmware.vcloud.api.rest.schema.ObjectFactory();
    String userUpdateUrl = getReference().getHref();
    String userUpdateRequestXmlString = JAXBUtil.marshal(of.createUser(userType));

    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.PUT_URL_MSG) + " - " + userUpdateUrl);

    return new User(getVcloudClient(), (UserType)SdkUtil.put(getVcloudClient(), userUpdateUrl, userUpdateRequestXmlString, "application/vnd.vmware.admin.user+xml", 200));
  }

  public void delete()
    throws VCloudException
  {
    String userUrl = getReference().getHref();
    deleteUser(getVcloudClient(), userUrl);
  }

  public static void delete(VcloudClient client, ReferenceType userRef)
    throws VCloudException
  {
    deleteUser(client, userRef.getHref());
  }

  private static void deleteUser(VcloudClient client, String userUrl)
    throws VCloudException
  {
    SdkUtil.delete(client, userUrl, 204);
  }

  public void unlock()
    throws VCloudException
  {
    SdkUtil.post(getVcloudClient(), getReference().getHref() + "/action/unlock", null, null, 204);
  }

  public static void unlock(VcloudClient client, ReferenceType userRef)
    throws VCloudException
  {
    SdkUtil.post(client, userRef.getHref() + "/action/unlock", null, null, 204);
  }

  public ReferenceType getRoleReference()
    throws VCloudException
  {
    if (((UserType)getResource()).getRole() != null) {
      return ((UserType)getResource()).getRole();
    }
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
  }

  public HashMap<String, ReferenceType> getGroupRefsByName()
  {
    return this.groupRefsByName;
  }

  public ReferenceType getGroupRefByName(String groupName)
  {
    return (ReferenceType)this.groupRefsByName.get(groupName);
  }

  public Collection<ReferenceType> getGroupRefs()
  {
    return this.groupRefsByName.values();
  }

  public List<EntityRightsType> getEntityRights(List<ReferenceType> entityRefs)
    throws VCloudException
  {
    EntityReferencesType entityRefsType = new EntityReferencesType();
    List jaxbEntityRefs = entityRefsType.getReference();

    for (ReferenceType entityRef : entityRefs) {
      jaxbEntityRefs.add(new com.vmware.vcloud.api.rest.schema.ObjectFactory().createReference(entityRef));
    }
    String requestString = JAXBUtil.marshal(new com.vmware.vcloud.api.rest.schema.extension.ObjectFactory().createEntityReferences(entityRefsType));

    UserEntityRightsType userEntityRightsType = (UserEntityRightsType)SdkUtil.post(getVcloudClient(), ((UserType)getResource()).getHref() + "/entityRights", requestString, "application/vnd.vmware.admin.entityReferences+xml", 200);

    return userEntityRightsType.getEntityRights();
  }

  public List<ReferenceType> getGrantedRights()
    throws VCloudException
  {
    ReferencesType grantedRights = (ReferencesType)SdkUtil.get(getVcloudClient(), ((UserType)getResource()).getHref() + "/grantedRights", 200);

    List grantedRightReferences = new ArrayList();
    for (JAXBElement grantedRight : grantedRights.getReference())
    {
      grantedRightReferences.add(grantedRight.getValue());
    }
    return grantedRightReferences;
  }
}