package com.vmware.vcloud.sdk_v1_0;

import com.vmware.vcloud.api.rest.schema_v1_0.CatalogItemType;
import com.vmware.vcloud.api.rest.schema_v1_0.CatalogItemsType;
import com.vmware.vcloud.api.rest.schema_v1_0.CatalogType;
import com.vmware.vcloud.api.rest.schema_v1_0.LinkType;
import com.vmware.vcloud.api.rest.schema_v1_0.ObjectFactory;
import com.vmware.vcloud.api.rest.schema_v1_0.ReferenceType;
import com.vmware.vcloud.api.rest.schema_v1_0.RelationType;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.bind.JAXBElement;

public class Catalog extends VcloudEntity<CatalogType>
{
  private static Logger logger = Logger.getLogger(Catalog.class.getSimpleName());
  private ReferenceType orgReference;
  private HashMap<String, ReferenceType> catalogItemRefsByName;

  protected Catalog(VcloudClient client, CatalogType catalogType)
  {
    super(client, catalogType);
    sortItemRefs();
  }

  public static Catalog getCatalogByReference(VcloudClient client, ReferenceType catalogRef)
    throws VCloudException
  {
    String url = catalogRef.getHref();
    return getCatalog(client, url);
  }

  private static Catalog getCatalog(VcloudClient client, String url)
    throws VCloudException
  {
    logger.log(Level.INFO, "GET catalog " + url);
    Response response = RestUtil.get(client, url);
    CatalogType catalogType = null;

    if (response.isExpected(200))
      catalogType = (CatalogType)response.getResource();
    else {
      response.handleUnExpectedResponse();
    }
    Catalog catalog = new Catalog(client, catalogType);
    return catalog;
  }

  private void sortItemRefs()
  {
    this.catalogItemRefsByName = new HashMap();

    if (((CatalogType)getResource()).getCatalogItems() != null) {
      CatalogItemsType catalogItems = ((CatalogType)getResource()).getCatalogItems();

      List items = catalogItems.getCatalogItem();

      for (ReferenceType ref : items) {
        this.catalogItemRefsByName.put(ref.getName(), ref);
      }
    }

    for (LinkType catalogLink : ((CatalogType)getResource()).getLink())
      if ((catalogLink.getRel().equals(RelationType.UP)) && (catalogLink.getType().equals("application/vnd.vmware.vcloud.org+xml")))
      {
        this.orgReference = catalogLink;
      }
  }

  public CatalogItem addCatalogItem(CatalogItemType catalogItemType)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    JAXBElement catalogItemParams = of.createCatalogItem(catalogItemType);

    String requestBodyXml = JAXBUtil.marshal(catalogItemParams);
    String postURL = ((CatalogType)getResource()).getHref() + "/catalogItems";

    Response response = RestUtil.post(getVcloudClient(), postURL, requestBodyXml, "application/vnd.vmware.vcloud.catalogItem+xml", null);

    if (!response.isExpected(201))
    {
      response.handleUnExpectedResponse();
    }
    CatalogItemType newCatalogItemType = (CatalogItemType)response.getResource();

    CatalogItem catalogItem = new CatalogItem(getVcloudClient(), newCatalogItemType);

    return catalogItem;
  }

  public HashMap<String, ReferenceType> getCatalogItemRefsByName()
  {
    return this.catalogItemRefsByName;
  }

  public ReferenceType getCatalogItemRefByName(String name)
  {
    return (ReferenceType)this.catalogItemRefsByName.get(name);
  }

  public Collection<ReferenceType> getCatalogItemReferences()
  {
    return this.catalogItemRefsByName.values();
  }

  public ReferenceType getOrgReference()
    throws VCloudException
  {
    if (this.orgReference != null)
      return this.orgReference;
    throw new VCloudException("Reference Not Found");
  }
}