/*
 * Copyright (c) 2005-2007 Flamingo Kirill Grouchnikov. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  o Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *     
 *  o Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution. 
 *     
 *  o Neither the name of Flamingo Kirill Grouchnikov nor the names of 
 *    its contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission. 
 *     
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */
package org.jvnet.flamingo.wizard;

import java.awt.*;
import java.awt.event.*;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.*;

import org.jvnet.flamingo.common.ImagePanel;
import org.jvnet.flamingo.common.ProgressGlassPane;
import org.jvnet.lafwidget.layout.TransitionLayoutManager;

/**
 * The main class for wizard mode.
 * 
 * @author Kirill Grouchnikov
 */
public class WizardMainFrame extends JFrame {
	/**
	 * A list of wizard steps.
	 */
	protected final LinkedList<WizardStep> steps;

	/**
	 * The current wizard step
	 */
	protected WizardStep currStep;

	/**
	 * The title label of the current step
	 */
	private final JLabel titleLabel;

	/**
	 * The <code>Back</code> button of the wizard
	 */
	private final JButton backButton;

	/**
	 * The <code>Next</code> button of the wizard
	 */
	private final JButton nextButton;

	/**
	 * The wizard mode.
	 */
	private WizardMode mode;

	/**
	 * Wizard mode enum.
	 * 
	 * @author Kirill Grouchnikov.
	 */
	public enum WizardMode {
		/**
		 * Standalone mode. When the wizard is finished, the application closes.
		 */
		STANDALONE,

		/**
		 * Manual mode - all pages are empty.
		 */
		MANUAL,

		/**
		 * Semi-manual mode - pages are populated from the settings and the flow
		 * starts from the first page. Useful for rerunning the wizard based on
		 * some previous run.
		 */
		SEMI_MANUAL,

		/**
		 * Automatic mode - pages are populated from the settings and the finish
		 * listener is invoked.
		 */
		AUTOMATIC
	}

	/**
	 * The <code>Finish</code> button of the wizard
	 */
	protected final JButton finishButton;

	/**
	 * The <code>Quit</code> button of the wizard
	 */
	private final JButton quitButton;

	/**
	 * The main panel (contains title label, wizard screen controls and
	 * navigation buttons)
	 */
	private final JPanel mainPanel;

	/**
	 * The panel for wizard screen controls and wizard screen help message
	 */
	private final JPanel stepPanel;

	/**
	 * The image label.
	 */
	private final JLabel imageLabel;

	/**
	 * The help label.
	 */
	private final JLabel helpLabel;

	/**
	 * The GUI component of the current wizard step
	 */
	private Component currStepComponent;

	/**
	 * The width of the image panel on the left side of the frame
	 */
	private final int imageWidth;

	/**
	 * The width of the rest of the frame (without the image panel)
	 */
	private final int remainderWidth;

	/**
	 * The name of the directory of the last seen wizard step that involved
	 * dealing with directories. Is used to provide default locations for steps
	 * that choose file or directory as one of their user inputs.
	 */
	private String currentDirectoryName;

	/**
	 * Glass pane for making the wizard screen unresponsive to keyboard and
	 * mouse when XJC generation is running.
	 */
	private final ProgressGlassPane glassPane;

	/**
	 * Creates and initializes the wizard window.
	 * 
	 * @param title
	 *            Window title.
	 * @param width
	 *            Window width.
	 * @param height
	 *            Window height.
	 * @param mode
	 *            Wizard mode.
	 */
	protected WizardMainFrame(String title, int width, int height,
			WizardMode mode) {
		super(title);

		this.mode = mode;
		this.steps = new LinkedList<WizardStep>();

		// initialize GUI
		this.backButton = new JButton("Back");
		// this.backButton.setHasLeftArrow(true);
		this.nextButton = new JButton("Next");
		// this.nextButton.setHasRightArrow(true);
		this.finishButton = new JButton("Finish");
		this.quitButton = new JButton("Quit");
		this.backButton.setMargin(new Insets(0, 0, 0, 0));
		this.nextButton.setMargin(new Insets(0, 0, 0, 0));
		this.finishButton.setMargin(new Insets(0, 0, 0, 0));
		this.quitButton.setMargin(new Insets(0, 0, 0, 0));

		this.setLayout(new BorderLayout());

		this.imageWidth = width / 6;
		this.remainderWidth = width - this.imageWidth;

		this.imageLabel = new JLabel();
		this.imageLabel.setBorder(null);
		this.imageLabel
				.setPreferredSize(new Dimension(this.imageWidth, height));
		this.add(imageLabel, BorderLayout.WEST);

		this.mainPanel = new JPanel();
		this.mainPanel.setLayout(new BorderLayout());
		this.mainPanel.setOpaque(false);

		this.stepPanel = new ImagePanel(WizardImageCreator.getBackgroundImage(
				this.remainderWidth, height - 70, true));
		this.stepPanel.setOpaque(false);
		this.stepPanel.setLayout(new BorderLayout());
		this.helpLabel = new JLabel();
		this.stepPanel.add(this.helpLabel, BorderLayout.SOUTH);

		this.mainPanel.add(this.stepPanel, BorderLayout.CENTER);

		JPanel buttonPanel = new ImagePanel(WizardImageCreator
				.getBackgroundImage(this.remainderWidth, 40, true));
		buttonPanel.setPreferredSize(new Dimension(this.remainderWidth, 40));
		buttonPanel.setLayout(new BorderLayout());
		JPanel buttonsPanel = new JPanel();
		buttonsPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 10, 5));
		buttonsPanel.add(this.backButton);
		buttonsPanel.add(this.nextButton);
		buttonsPanel.add(this.finishButton);
		// set the quit button apart
		buttonsPanel.add(Box.createHorizontalStrut(20));
		buttonsPanel.add(this.quitButton);
		buttonsPanel.setOpaque(false);
		buttonPanel.setOpaque(false);
		buttonPanel.add(Box.createVerticalStrut(5), BorderLayout.NORTH);
		buttonPanel.add(buttonsPanel, BorderLayout.CENTER);

		// wire listeners
		this.backButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				back();
			}
		});
		this.nextButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				next();
			}
		});

		this.mainPanel.add(buttonPanel, BorderLayout.SOUTH);

		// create title label
		this.titleLabel = new JLabel();
		this.titleLabel
				.setPreferredSize(new Dimension(this.remainderWidth, 30));
		this.mainPanel.add(this.titleLabel, BorderLayout.NORTH);

		this.add(mainPanel, BorderLayout.CENTER);

		// create glass pane
		this.glassPane = new ProgressGlassPane(this.getRootPane());
		this.setGlassPane(this.glassPane);
		// this.glassPane.setMessage("Compiling generated classes");
		// this.glassPane.setVisible(true);

		// add on-close adapter
		this.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);

		this.setSize(width, height);

		this.setResizable(false);
	}

	/**
	 * Sets the current message that will shown by the glass pane of
	 * <code>this</code> panel while the listener on {@link #finishButton} is
	 * running.
	 * 
	 * @param message
	 *            The current message that will shown by the glass pane of
	 *            <code>this</code> panel while the listener on
	 *            {@link #finishButton} is running.
	 */
	public final void setWaitingMessage(String message) {
		this.glassPane.setMessage(message);
	}

	/**
	 * Sets a new image as a left-side panel.
	 */
	protected final void setImage() {
		Dimension d = this.getSize();
		this.imageLabel.setIcon(this.currStep.getStepIcon(this.imageWidth,
				(int) d.getHeight()));
	}

	/**
	 * Updates the wizard window based on the current step and the information
	 * that was suplied in it. For example, in the first step, the
	 * <code>Back</code> button is disabled; the <code>Next</code> button is
	 * enabled only if information in the corresponding wizard step is legal.
	 */
	public final void update() {
		// decide on the buttons
		boolean hasBackButton = (this.currStep != this.steps.getFirst());
		boolean hasNextButton = (this.currStep != this.steps.getLast());
		boolean hasFinishButton = (this.currStep == this.steps.getLast());
		boolean isNextStepEnabled = this.currStep.canProceed();

		this.backButton.setEnabled(false);
		this.nextButton.setEnabled(false);
		this.finishButton.setEnabled(false);
		if (hasBackButton) {
			this.backButton.setEnabled(true);
		}
		if (hasNextButton && isNextStepEnabled) {
			this.nextButton.setEnabled(true);
			this.getRootPane().setDefaultButton(this.nextButton);
		}
		if (hasFinishButton && isNextStepEnabled) {
			this.finishButton.setEnabled(true);
			this.getRootPane().setDefaultButton(this.finishButton);
		}

		// replace the step component
		if (this.currStepComponent != null) {
			this.currStepComponent.setVisible(false);
//			this.stepPanel.remove(this.currStepComponent);
		}
		this.currStepComponent = this.currStep.getComponent();
		this.currStepComponent.setVisible(true);
		this.stepPanel.add(this.currStepComponent, BorderLayout.CENTER);

		this.stepPanel.add(Box.createVerticalStrut(10), BorderLayout.NORTH);
		this.helpLabel.setIcon(WizardImageCreator.getMultiline(this.currStep
				.getDescription(), this.remainderWidth, Color.black,
				Color.white));

		this.titleLabel.setIcon(WizardImageCreator.getTitleImage(
				this.remainderWidth, 30, this.currStep.getTitle(), Color.black,
				Color.white));
		// without this line the components of 2+ step are not shown.
		// must be some bug in Swing ???
		this.titleLabel.setText("      " + this.currStep.getTitle());
		this.repaint();
	}

	/**
	 * Goes forward to the next wizard step.
	 */
	protected final void next() {
		if (this.currStep.isSupportsDirectory()) {
			this.currentDirectoryName = this.currStep.getCurrentDirectory();
		}

		// find next step
		Iterator<WizardStep> itStep = this.steps.iterator();
		boolean takeNext = false;
		boolean taken = false;
		while (itStep.hasNext()) {
			WizardStep step = itStep.next();
			if (takeNext) {
				this.currStep = step;
				taken = true;
				break;
			}
			if (step == this.currStep) {
				takeNext = true;
			}
		}
		if (!taken) {
			throw new IllegalStateException(
					"No next step when next was available");
		}

		if (this.currStep.isSupportsDirectory()) {
			this.currStep.setCurrentDirectory(this.currentDirectoryName);
		}

		// refresh the contents - will show the next window
		update();
		setImage();
	}

	/**
	 * Goes back to the previous wizard step.
	 */
	protected final void back() {
		if (this.currStep.isSupportsDirectory()) {
			this.currentDirectoryName = this.currStep.getCurrentDirectory();
		}

		// find previous step
		Iterator<WizardStep> itStep = this.steps.iterator();
		WizardStep prevStep = null;
		boolean taken = false;
		while (itStep.hasNext()) {
			WizardStep step = itStep.next();
			if (step == this.currStep) {
				this.currStep = prevStep;
				taken = true;
				break;
			}
			prevStep = step;

		}
		if ((!taken) || (this.currStep == null)) {
			throw new IllegalStateException(
					"No prev step when prev was available");
		}
		if (this.currStep.isSupportsDirectory()) {
			this.currStep.setCurrentDirectory(this.currentDirectoryName);
		}

		// refresh the contents - will show the previous window
		update();
		setImage();
	}

	/**
	 * Freezes the wizard frame making it unresponsive to mouse and keyboard.
	 * The only thing that can be done - move the frame around the screen.
	 */
	public final void freeze() {
		this.glassPane.setVisible(true);
	}

	/**
	 * Unfreezes the wizard frame making it responsive to mouse and keyboard.
	 */
	public final void unfreeze() {
		this.glassPane.setVisible(false);
	}

	/**
	 * Sets glass pane colors.
	 * 
	 * @param darkColor
	 *            Dark color.
	 * @param midColor
	 *            Mid color.
	 * @param lightColor
	 *            Light color.
	 */
	public void setGlassPaneColors(Color darkColor, Color midColor,
			Color lightColor) {
		this.glassPane.setGlassPaneColors(darkColor, midColor, lightColor);
	}

	/**
	 * Sets a listener on the <code>Quit</code> button.
	 * 
	 * @param quitListener
	 *            Listener for the <code>Quit</code> button.
	 */
	public void setQuitListener(final ActionListener quitListener) {
		this.quitButton.addActionListener(quitListener);
		this.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				quitListener.actionPerformed(new ActionEvent(e.getSource(), e
						.getID(), e.paramString()));
			}
		});
	}

	/**
	 * Sets a listener on the <code>Finish</code> button.
	 * 
	 * @param finishListener
	 *            Listener for the <code>Finish</code> button.
	 */
	public void setFinishListener(ActionListener finishListener) {
		this.finishButton.addActionListener(finishListener);
	}

	/**
	 * Returns wizard mode.
	 * 
	 * @return Wizard mode.
	 */
	public WizardMode getMode() {
		return mode;
	}

	/**
	 * Removes the focus from the component hierarchy.
	 * 
	 * @param comp
	 *            Component.
	 */
	private static void removeFocus(Component comp) {
		comp.setFocusable(false);
		if (comp instanceof Container) {
			Container cont = (Container) comp;
			for (int i = 0; i < cont.getComponentCount(); i++) {
				removeFocus(cont.getComponent(i));
			}
		}
	}

	/**
	 * Starts the wizard.
	 * 
	 * @param initialSettings
	 *            Wizard initial settings.
	 */
	public void start(WizardSettings initialSettings) {
		// fill info based on provided settings if necessary
		if ((mode == WizardMode.SEMI_MANUAL) || (mode == WizardMode.AUTOMATIC)) {
			for (WizardStep step : this.steps)
				step.importSettings(initialSettings);
		}

		this.currStep = this.steps.getFirst();
		this.currentDirectoryName = System.getProperty("user.dir");
		if (this.currStep.isSupportsDirectory()) {
			this.currStep.setCurrentDirectory(this.currentDirectoryName);
		}
		this.setImage();

		if (mode == WizardMode.AUTOMATIC) {
			WizardStep lastStep = steps.getLast();
			currStep = lastStep;
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					update();
					setImage();
					removeFocus(currStep.getComponent());
					SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							finishButton.doClick();
						}
					});
				}
			});
		}
		this.update();
	}

	/**
	 * Sets focusable state on the navigation buttons.
	 * 
	 * @param isFocusable
	 *            Indicates whether the navigation buttons should be focusable.
	 */
	public void setNavigationButtonsFocusable(boolean isFocusable) {
		this.backButton.setFocusable(isFocusable);
		this.nextButton.setFocusable(isFocusable);
		this.quitButton.setFocusable(isFocusable);
		this.finishButton.setFocusable(isFocusable);
	}
	
	/**
	 * Installs transition effects on the main panel. 
	 */
	public void installTransitionEffects() {
		TransitionLayoutManager.getInstance().track(this.stepPanel, true);
	}
}