/*
 * Copyright (c) 2005-2007 Flamingo Kirill Grouchnikov. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  o Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *     
 *  o Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution. 
 *     
 *  o Neither the name of Flamingo Kirill Grouchnikov nor the names of 
 *    its contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission. 
 *     
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */
package org.jvnet.flamingo.wizard;

import javax.swing.Icon;
import javax.swing.JComponent;

/**
 * A single wizard step contract.
 * 
 * @author Kirill Grouchnikov
 */
public interface WizardStep {
	/**
	 * Checks the legality of information provided by the user in this wizard
	 * step.
	 * 
	 * @return <code>true</code> if the information supplied was legal,
	 *         <code>false</code> otherwise.
	 */
	boolean canProceed();

	/**
	 * Updates the wizard settings based on the information provided by the user
	 * in this wizard step.
	 * 
	 * @param settings
	 *            The wizard settings.
	 */
	void exportSettings(WizardSettings settings);

	/**
	 * Updates the current wizard step based on the information in wizard
	 * settings.
	 * 
	 * @param settings
	 *            The wizard settings.
	 */
	void importSettings(WizardSettings settings);

	/**
	 * Returns the GUI component that reflects this wizard step.
	 * 
	 * @return The GUI component that reflects this wizard step.
	 */
	JComponent getComponent();

	/**
	 * Returns the title of this wizard step.
	 * 
	 * @return The title of this wizard step.
	 */
	String getTitle();

	/**
	 * Checks whether this step supports directory history.
	 * 
	 * @return <code>true</code> if this step supports work with file-system
	 *         directories (for history purposes), <code>false</code>
	 *         otherwise.
	 */
	boolean isSupportsDirectory();

	/**
	 * Gets the current directory of this wizard step.
	 * 
	 * @return The current directory of this wizard step.
	 * @throws UnsupportedOperationException
	 *             If this step does not support work with directory history.
	 */
	public String getCurrentDirectory() throws UnsupportedOperationException;

	/**
	 * Sets the current directory of this wizard step.
	 * 
	 * @param str
	 *            The new current directory of this wizard step. May be ignored
	 *            by the step if it already has some relevant information.
	 * @throws UnsupportedOperationException
	 *             If this step does not support work with directory history.
	 */
	public void setCurrentDirectory(String str)
			throws UnsupportedOperationException;

	/**
	 * Returns the description of this step.
	 * 
	 * @return The description of this step.
	 */
	public String getDescription();

	/**
	 * Returns the image that matches this step.
	 * 
	 * @param width
	 *            Image width.
	 * @param height
	 *            Image height.
	 * @return Step image of the specified dimension.
	 */
	public Icon getStepIcon(int width, int height);
}