package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LocalTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.NestedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticArgumentBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class AllocationExpression extends Expression
  implements InvocationSite
{
  public TypeReference type;
  public Expression[] arguments;
  public MethodBinding binding;
  protected MethodBinding codegenBinding;
  MethodBinding syntheticAccessor;
  public TypeReference[] typeArguments;
  public TypeBinding[] genericTypeArguments;
  public FieldDeclaration enumConstant;

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    checkCapturedLocalInitializationIfNecessary((ReferenceBinding)this.binding.declaringClass.erasure(), currentScope, flowInfo);

    if (this.arguments != null) {
      int i = 0; for (int count = this.arguments.length; i < count; i++)
        flowInfo = 
          this.arguments[i]
          .analyseCode(currentScope, flowContext, flowInfo)
          .unconditionalInits();
    }
    ReferenceBinding[] thrownExceptions;
    if ((thrownExceptions = this.binding.thrownExceptions).length != 0)
    {
      flowContext.checkExceptionHandlers(
        thrownExceptions, 
        this, 
        flowInfo.unconditionalCopy(), 
        currentScope);
    }
    manageEnclosingInstanceAccessIfNecessary(currentScope, flowInfo);
    manageSyntheticAccessIfNecessary(currentScope, flowInfo);

    return flowInfo;
  }

  public void checkCapturedLocalInitializationIfNecessary(ReferenceBinding checkedType, BlockScope currentScope, FlowInfo flowInfo) {
    if ((checkedType.isLocalType()) && 
      (!checkedType.isAnonymousType()) && 
      (!currentScope.isDefinedInType(checkedType))) {
      NestedTypeBinding nestedType = (NestedTypeBinding)checkedType;
      SyntheticArgumentBinding[] syntheticArguments = nestedType.syntheticOuterLocalVariables();
      if (syntheticArguments != null) {
        int i = 0; for (int count = syntheticArguments.length; i < count; i++) {
          SyntheticArgumentBinding syntheticArgument = syntheticArguments[i];
          LocalVariableBinding targetLocal;
          if (((targetLocal = syntheticArgument.actualOuterLocalVariable) != null) && 
            (targetLocal.declaration != null) && (!flowInfo.isDefinitelyAssigned(targetLocal)))
            currentScope.problemReporter().uninitializedLocalVariable(targetLocal, this);
        }
      }
    }
  }

  public Expression enclosingInstance() {
    return null;
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired) {
    int pc = codeStream.position;
    ReferenceBinding allocatedType = this.codegenBinding.declaringClass;

    codeStream.new_(allocatedType);
    boolean isUnboxing = (this.implicitConversion & 0x400) != 0;
    if ((valueRequired) || (isUnboxing)) {
      codeStream.dup();
    }

    if (this.type != null) {
      codeStream.recordPositionsFrom(pc, this.type.sourceStart);
    }
    else {
      codeStream.ldc(String.valueOf(this.enumConstant.name));
      codeStream.generateInlinedValue(this.enumConstant.binding.id);
    }

    if (allocatedType.isNestedType()) {
      codeStream.generateSyntheticEnclosingInstanceValues(
        currentScope, 
        allocatedType, 
        enclosingInstance(), 
        this);
    }

    generateArguments(this.binding, this.arguments, currentScope, codeStream);

    if (allocatedType.isNestedType()) {
      codeStream.generateSyntheticOuterArgumentValues(
        currentScope, 
        allocatedType, 
        this);
    }

    if (this.syntheticAccessor == null) {
      codeStream.invokespecial(this.codegenBinding);
    }
    else {
      int i = 0;
      int max = this.syntheticAccessor.parameters.length - this.codegenBinding.parameters.length;
      for (; i < max; 
        i++) {
        codeStream.aconst_null();
      }
      codeStream.invokespecial(this.syntheticAccessor);
    }
    if (valueRequired) {
      codeStream.generateImplicitConversion(this.implicitConversion);
    } else if (isUnboxing)
    {
      codeStream.generateImplicitConversion(this.implicitConversion);
      switch (postConversionType(currentScope).id) {
      case 7:
      case 8:
        codeStream.pop2();
        break;
      default:
        codeStream.pop();
      }
    }
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public TypeBinding[] genericTypeArguments()
  {
    return this.genericTypeArguments;
  }

  public boolean isSuperAccess() {
    return false;
  }

  public boolean isTypeAccess() {
    return true;
  }

  public void manageEnclosingInstanceAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo)
  {
    if ((flowInfo.tagBits & 0x1) != 0) return;
    ReferenceBinding allocatedTypeErasure = (ReferenceBinding)this.binding.declaringClass.erasure();

    if ((allocatedTypeErasure.isNestedType()) && 
      (currentScope.enclosingSourceType().isLocalType()))
    {
      if (allocatedTypeErasure.isLocalType()) {
        ((LocalTypeBinding)allocatedTypeErasure).addInnerEmulationDependent(currentScope, false);
      }
      else
      {
        currentScope.propagateInnerEmulation(allocatedTypeErasure, false);
      }
    }
  }

  public void manageSyntheticAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo)
  {
    if ((flowInfo.tagBits & 0x1) != 0) return;

    this.codegenBinding = this.binding.original();
    ReferenceBinding declaringClass;
    if ((this.codegenBinding.isPrivate()) && (currentScope.enclosingSourceType() != (declaringClass = this.codegenBinding.declaringClass)))
    {
      if (((declaringClass.tagBits & 0x10) != 0L) && (currentScope.compilerOptions().complianceLevel >= 3145728L))
      {
        this.codegenBinding.tagBits |= 1024L;
      } else {
        this.syntheticAccessor = ((SourceTypeBinding)declaringClass).addSyntheticMethod(this.codegenBinding, isSuperAccess());
        currentScope.problemReporter().needToEmulateMethodAccess(this.codegenBinding, this);
      }
    }
  }

  public StringBuffer printExpression(int indent, StringBuffer output) {
    if (this.type != null) {
      output.append("new ");
    }
    if (this.typeArguments != null) {
      output.append('<');
      int max = this.typeArguments.length - 1;
      for (int j = 0; j < max; j++) {
        this.typeArguments[j].print(0, output);
        output.append(", ");
      }
      this.typeArguments[max].print(0, output);
      output.append('>');
    }
    if (this.type != null) {
      this.type.printExpression(0, output);
    }
    output.append('(');
    if (this.arguments != null) {
      for (int i = 0; i < this.arguments.length; i++) {
        if (i > 0) output.append(", ");
        this.arguments[i].printExpression(0, output);
      }
    }
    return output.append(')');
  }

  public TypeBinding resolveType(BlockScope scope)
  {
    this.constant = Constant.NotAConstant;
    if (this.type == null)
    {
      this.resolvedType = scope.enclosingReceiverType();
    } else {
      this.resolvedType = this.type.resolveType(scope, true);

      if ((this.type instanceof ParameterizedQualifiedTypeReference)) {
        ReferenceBinding currentType = (ReferenceBinding)this.resolvedType;
        if (currentType == null) return null;

        while (((currentType.modifiers & 0x8) == 0) && 
          (!currentType.isRawType())) {
          if ((currentType = currentType.enclosingType()) == null) {
            ParameterizedQualifiedTypeReference qRef = (ParameterizedQualifiedTypeReference)this.type;
            for (int i = qRef.typeArguments.length - 2; i >= 0; i--) {
              if (qRef.typeArguments[i] != null) {
                scope.problemReporter().illegalQualifiedParameterizedTypeAllocation(this.type, this.resolvedType);
                break;
              }
            }
          }
        }
      }

    }

    if (this.typeArguments != null) {
      int length = this.typeArguments.length;
      boolean argHasError = false;
      this.genericTypeArguments = new TypeBinding[length];
      for (int i = 0; i < length; i++) {
        TypeReference typeReference = this.typeArguments[i];
        if ((this.genericTypeArguments[i] =  = typeReference.resolveType(scope, true)) == null) {
          argHasError = true;
        }
        if ((argHasError) && ((typeReference instanceof Wildcard))) {
          scope.problemReporter().illegalUsageOfWildcard(typeReference);
        }
      }
      if (argHasError) {
        return null;
      }

    }

    boolean argsContainCast = false;
    TypeBinding[] argumentTypes = Binding.NO_PARAMETERS;
    if (this.arguments != null) {
      boolean argHasError = false;
      int length = this.arguments.length;
      argumentTypes = new TypeBinding[length];
      for (int i = 0; i < length; i++) {
        Expression argument = this.arguments[i];
        if ((argument instanceof CastExpression)) {
          argument.bits |= 32;
          argsContainCast = true;
        }
        if ((argumentTypes[i] =  = argument.resolveType(scope)) == null) {
          argHasError = true;
        }
      }
      if (argHasError) {
        if ((this.resolvedType instanceof ReferenceBinding))
        {
          TypeBinding[] pseudoArgs = new TypeBinding[length];
          int i = length;
          do { pseudoArgs[i] = (argumentTypes[i] == null ? TypeBinding.NULL : argumentTypes[i]);

            i--; } while (i >= 0);

          this.binding = scope.findMethod((ReferenceBinding)this.resolvedType, TypeConstants.INIT, pseudoArgs, this);
          if ((this.binding != null) && (!this.binding.isValidBinding())) {
            MethodBinding closestMatch = ((ProblemMethodBinding)this.binding).closestMatch;

            if (closestMatch != null) {
              if (closestMatch.original().typeVariables != Binding.NO_TYPE_VARIABLES)
              {
                closestMatch = scope.environment().createParameterizedGenericMethod(closestMatch.original(), null);
              }
              this.binding = closestMatch;
              MethodBinding closestMatchOriginal = closestMatch.original();
              if (((closestMatchOriginal.isPrivate()) || (closestMatchOriginal.declaringClass.isLocalType())) && (!scope.isDefinedInMethod(closestMatchOriginal)))
              {
                closestMatchOriginal.modifiers |= 134217728;
              }
            }
          }
        }
        return this.resolvedType;
      }
    }
    if (this.resolvedType == null) {
      return null;
    }

    if ((this.type != null) && (!this.resolvedType.canBeInstantiated())) {
      scope.problemReporter().cannotInstantiate(this.type, this.resolvedType);
      return this.resolvedType;
    }
    ReferenceBinding allocationType = (ReferenceBinding)this.resolvedType;
    if (!(this.binding = scope.getConstructor(allocationType, argumentTypes, this)).isValidBinding()) {
      if (this.binding.declaringClass == null)
        this.binding.declaringClass = allocationType;
      scope.problemReporter().invalidConstructor(this, this.binding);
      return this.resolvedType;
    }
    if (isMethodUseDeprecated(this.binding, scope, true))
      scope.problemReporter().deprecatedMethod(this.binding, this);
    checkInvocationArguments(scope, null, allocationType, this.binding, this.arguments, argumentTypes, argsContainCast, this);

    return allocationType;
  }

  public void setActualReceiverType(ReferenceBinding receiverType)
  {
  }

  public void setDepth(int i)
  {
  }

  public void setFieldIndex(int i)
  {
  }

  public void traverse(ASTVisitor visitor, BlockScope scope) {
    if (visitor.visit(this, scope)) {
      if (this.typeArguments != null) {
        int i = 0; for (int typeArgumentsLength = this.typeArguments.length; i < typeArgumentsLength; i++) {
          this.typeArguments[i].traverse(visitor, scope);
        }
      }
      if (this.type != null) {
        this.type.traverse(visitor, scope);
      }
      if (this.arguments != null) {
        int i = 0; for (int argumentsLength = this.arguments.length; i < argumentsLength; i++)
          this.arguments[i].traverse(visitor, scope);
      }
    }
    visitor.endVisit(this, scope);
  }
}