package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class Argument extends LocalDeclaration
{
  private static final char[] SET = "set".toCharArray();

  public Argument(char[] name, long posNom, TypeReference tr, int modifiers)
  {
    super(name, (int)(posNom >>> 32), (int)posNom);
    this.declarationSourceEnd = ((int)posNom);
    this.modifiers = modifiers;
    this.type = tr;
    this.bits |= 1073741824;
  }

  public void bind(MethodScope scope, TypeBinding typeBinding, boolean used)
  {
    Binding existingVariable = scope.getBinding(this.name, 3, this, false);
    if ((existingVariable != null) && (existingVariable.isValidBinding())) {
      if (((existingVariable instanceof LocalVariableBinding)) && (this.hiddenVariableDepth == 0)) {
        scope.problemReporter().redefineArgument(this);
      } else {
        boolean isSpecialArgument = false;
        if ((existingVariable instanceof FieldBinding)) {
          if (scope.isInsideConstructor()) {
            isSpecialArgument = true;
          } else {
            AbstractMethodDeclaration methodDecl = scope.referenceMethod();
            if ((methodDecl != null) && (CharOperation.prefixEquals(SET, methodDecl.selector))) {
              isSpecialArgument = true;
            }
          }
        }
        scope.problemReporter().localVariableHiding(this, existingVariable, isSpecialArgument);
      }
    }

    if (this.binding == null) {
      this.binding = new LocalVariableBinding(this, typeBinding, this.modifiers, true);
    }
    scope.addLocalVariable(this.binding);
    resolveAnnotations(scope, this.annotations, this.binding);

    this.binding.declaration = this;
    this.binding.useFlag = (used ? 1 : 0);
  }

  public int getKind()
  {
    return 5;
  }

  public boolean isVarArgs() {
    return (this.type != null) && ((this.type.bits & 0x4000) != 0);
  }

  public StringBuffer print(int indent, StringBuffer output)
  {
    printIndent(indent, output);
    printModifiers(this.modifiers, output);
    if (this.annotations != null) printAnnotations(this.annotations, output);

    if (this.type == null)
      output.append("<no type> ");
    else {
      this.type.print(0, output).append(' ');
    }
    return output.append(this.name);
  }

  public StringBuffer printStatement(int indent, StringBuffer output)
  {
    return print(indent, output).append(';');
  }

  public TypeBinding resolveForCatch(BlockScope scope)
  {
    TypeBinding exceptionType = this.type.resolveType(scope, true);
    if (exceptionType == null) return null;
    boolean hasError = false;
    if (exceptionType.isBoundParameterizedType()) {
      scope.problemReporter().invalidParameterizedExceptionType(exceptionType, this);
      hasError = true;
    }

    if (exceptionType.isTypeVariable()) {
      scope.problemReporter().invalidTypeVariableAsException(exceptionType, this);
      hasError = true;
    }

    if ((exceptionType.isArrayType()) && (((ArrayBinding)exceptionType).leafComponentType == TypeBinding.VOID)) {
      scope.problemReporter().variableTypeCannotBeVoidArray(this);
      hasError = true;
    }

    if (exceptionType.findSuperTypeErasingTo(21, true) == null) {
      scope.problemReporter().cannotThrowType(this.type, exceptionType);
      hasError = true;
    }

    Binding existingVariable = scope.getBinding(this.name, 3, this, false);
    if ((existingVariable != null) && (existingVariable.isValidBinding())) {
      if (((existingVariable instanceof LocalVariableBinding)) && (this.hiddenVariableDepth == 0))
        scope.problemReporter().redefineArgument(this);
      else {
        scope.problemReporter().localVariableHiding(this, existingVariable, false);
      }
    }

    this.binding = new LocalVariableBinding(this, exceptionType, this.modifiers, false);
    resolveAnnotations(scope, this.annotations, this.binding);

    scope.addLocalVariable(this.binding);
    this.binding.setConstant(Constant.NotAConstant);
    if (hasError) return null;
    return exceptionType;
  }

  public void traverse(ASTVisitor visitor, BlockScope scope)
  {
    if (visitor.visit(this, scope)) {
      if (this.annotations != null) {
        int annotationsLength = this.annotations.length;
        for (int i = 0; i < annotationsLength; i++)
          this.annotations[i].traverse(visitor, scope);
      }
      if (this.type != null)
        this.type.traverse(visitor, scope);
    }
    visitor.endVisit(this, scope);
  }

  public void traverse(ASTVisitor visitor, ClassScope scope) {
    if (visitor.visit(this, scope)) {
      if (this.annotations != null) {
        int annotationsLength = this.annotations.length;
        for (int i = 0; i < annotationsLength; i++)
          this.annotations[i].traverse(visitor, scope);
      }
      if (this.type != null)
        this.type.traverse(visitor, scope);
    }
    visitor.endVisit(this, scope);
  }
}