package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.NullInfoRegistry;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.BaseTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class Assignment extends Expression
{
  public Expression lhs;
  public Expression expression;

  public Assignment(Expression lhs, Expression expression, int sourceEnd)
  {
    this.lhs = lhs;
    lhs.bits |= 8192;
    this.expression = expression;
    this.sourceStart = lhs.sourceStart;
    this.sourceEnd = sourceEnd;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    LocalVariableBinding local = this.lhs.localVariableBinding();
    int nullStatus = this.expression.nullStatus(flowInfo);
    if ((local != null) && ((local.type.tagBits & 0x2) == 0L) && 
      (nullStatus == 1)) {
      flowContext.recordUsingNullReference(currentScope, local, this.lhs, 
        769, flowInfo);
    }

    flowInfo = ((Reference)this.lhs)
      .analyseAssignment(currentScope, flowContext, flowInfo, this, false)
      .unconditionalInits();
    if ((local != null) && ((local.type.tagBits & 0x2) == 0L)) {
      switch (nullStatus) {
      case 1:
        flowInfo.markAsDefinitelyNull(local);
        break;
      case -1:
        flowInfo.markAsDefinitelyNonNull(local);
        break;
      case 0:
      default:
        flowInfo.markAsDefinitelyUnknown(local);
      }
      if (flowContext.initsOnFinally != null) {
        switch (nullStatus) {
        case 1:
          flowContext.initsOnFinally.markAsDefinitelyNull(local);
          break;
        case -1:
          flowContext.initsOnFinally.markAsDefinitelyNonNull(local);
          break;
        case 0:
        default:
          flowContext.initsOnFinally.markAsDefinitelyUnknown(local);
        }
      }
    }
    return flowInfo;
  }

  void checkAssignment(BlockScope scope, TypeBinding lhsType, TypeBinding rhsType) {
    FieldBinding leftField = getLastField(this.lhs);
    if ((leftField != null) && (rhsType != TypeBinding.NULL) && (lhsType.isWildcard()) && (((WildcardBinding)lhsType).boundKind != 2))
      scope.problemReporter().wildcardAssignment(lhsType, rhsType, this.expression);
    else if ((leftField != null) && (!leftField.isStatic()) && (leftField.declaringClass != null) && (leftField.declaringClass.isRawType()))
      scope.problemReporter().unsafeRawFieldAssignment(leftField, rhsType, this.lhs);
    else if (rhsType.needsUncheckedConversion(lhsType))
      scope.problemReporter().unsafeTypeConversion(this.expression, rhsType, lhsType);
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired)
  {
    int pc = codeStream.position;
    ((Reference)this.lhs).generateAssignment(currentScope, codeStream, this, valueRequired);

    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public static Binding getDirectBinding(Expression someExpression) {
    if ((someExpression.bits & 0x20000000) != 0) {
      return null;
    }
    if ((someExpression instanceof SingleNameReference))
      return ((SingleNameReference)someExpression).binding;
    if ((someExpression instanceof FieldReference)) {
      FieldReference fieldRef = (FieldReference)someExpression;
      if ((fieldRef.receiver.isThis()) && (!(fieldRef.receiver instanceof QualifiedThisReference)))
        return fieldRef.binding;
    }
    else if ((someExpression instanceof Assignment)) {
      Expression lhs = ((Assignment)someExpression).lhs;
      if ((lhs.bits & 0x2000) != 0)
      {
        return getDirectBinding(((Assignment)someExpression).lhs);
      }if ((someExpression instanceof PrefixExpression))
      {
        return getDirectBinding(((Assignment)someExpression).lhs);
      }

    }

    return null;
  }

  FieldBinding getLastField(Expression someExpression) {
    if ((someExpression instanceof SingleNameReference)) {
      if ((someExpression.bits & 0x7) == 1)
        return (FieldBinding)((SingleNameReference)someExpression).binding;
    } else {
      if ((someExpression instanceof FieldReference))
        return ((FieldReference)someExpression).binding;
      if ((someExpression instanceof QualifiedNameReference)) {
        QualifiedNameReference qName = (QualifiedNameReference)someExpression;
        if (qName.otherBindings == null) {
          if ((someExpression.bits & 0x7) == 1)
            return (FieldBinding)qName.binding;
        }
        else
          return qName.otherBindings[(qName.otherBindings.length - 1)];
      }
    }
    return null;
  }

  public int nullStatus(FlowInfo flowInfo) {
    return this.expression.nullStatus(flowInfo);
  }

  public StringBuffer print(int indent, StringBuffer output)
  {
    printIndent(indent, output);
    return printExpressionNoParenthesis(indent, output);
  }

  public StringBuffer printExpression(int indent, StringBuffer output) {
    output.append('(');
    return printExpressionNoParenthesis(0, output).append(')');
  }

  public StringBuffer printExpressionNoParenthesis(int indent, StringBuffer output) {
    this.lhs.printExpression(indent, output).append(" = ");
    return this.expression.printExpression(0, output);
  }

  public StringBuffer printStatement(int indent, StringBuffer output)
  {
    return print(indent, output).append(';');
  }

  public TypeBinding resolveType(BlockScope scope)
  {
    this.constant = Constant.NotAConstant;
    if ((!(this.lhs instanceof Reference)) || (this.lhs.isThis())) {
      scope.problemReporter().expressionShouldBeAVariable(this.lhs);
      return null;
    }
    TypeBinding lhsType = this.lhs.resolveType(scope);
    this.expression.setExpectedType(lhsType);
    if (lhsType != null)
      this.resolvedType = lhsType.capture(scope, this.sourceEnd);
    TypeBinding rhsType = this.expression.resolveType(scope);
    if ((lhsType == null) || (rhsType == null)) {
      return null;
    }

    Binding left = getDirectBinding(this.lhs);
    if ((left != null) && (left == getDirectBinding(this.expression))) {
      scope.problemReporter().assignmentHasNoEffect(this, left.shortReadableName());
    }

    if (lhsType != rhsType)
      scope.compilationUnitScope().recordTypeConversion(lhsType, rhsType);
    if ((this.expression.isConstantValueOfTypeAssignableToType(rhsType, lhsType)) || 
      ((lhsType.isBaseType()) && (BaseTypeBinding.isWidening(lhsType.id, rhsType.id))) || 
      (rhsType.isCompatibleWith(lhsType))) {
      this.expression.computeConversion(scope, lhsType, rhsType);
      checkAssignment(scope, lhsType, rhsType);
      if (((this.expression instanceof CastExpression)) && 
        ((this.expression.bits & 0x4000) == 0)) {
        CastExpression.checkNeedForAssignedCast(scope, lhsType, (CastExpression)this.expression);
      }
      return this.resolvedType;
    }if ((scope.isBoxingCompatibleWith(rhsType, lhsType)) || (
      (rhsType.isBaseType()) && 
      (scope.compilerOptions().sourceLevel >= 3211264L) && 
      (!lhsType.isBaseType()) && 
      (this.expression.isConstantValueOfTypeAssignableToType(rhsType, scope.environment().computeBoxingType(lhsType))))) {
      this.expression.computeConversion(scope, lhsType, rhsType);
      if (((this.expression instanceof CastExpression)) && 
        ((this.expression.bits & 0x4000) == 0)) {
        CastExpression.checkNeedForAssignedCast(scope, lhsType, (CastExpression)this.expression);
      }
      return this.resolvedType;
    }
    scope.problemReporter().typeMismatchError(rhsType, lhsType, this.expression);
    return lhsType;
  }

  public TypeBinding resolveTypeExpecting(BlockScope scope, TypeBinding expectedType)
  {
    TypeBinding type = super.resolveTypeExpecting(scope, expectedType);
    if (type == null) return null;
    TypeBinding lhsType = this.resolvedType;
    TypeBinding rhsType = this.expression.resolvedType;

    if ((expectedType == TypeBinding.BOOLEAN) && 
      (lhsType == TypeBinding.BOOLEAN) && 
      ((this.lhs.bits & 0x2000) != 0)) {
      scope.problemReporter().possibleAccidentalBooleanAssignment(this);
    }
    checkAssignment(scope, lhsType, rhsType);
    return type;
  }

  public void traverse(ASTVisitor visitor, BlockScope scope) {
    if (visitor.visit(this, scope)) {
      this.lhs.traverse(visitor, scope);
      this.expression.traverse(visitor, scope);
    }
    visitor.endVisit(this, scope);
  }
  public LocalVariableBinding localVariableBinding() {
    return this.lhs.localVariableBinding();
  }
}