package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.codegen.ConstantPool;
import org.eclipse.jdt.internal.compiler.flow.ExceptionHandlingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.InitializationFlowContext;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.problem.AbortMethod;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class Clinit extends AbstractMethodDeclaration
{
  private FieldBinding assertionSyntheticFieldBinding = null;
  private FieldBinding classLiteralSyntheticField = null;

  public Clinit(CompilationResult compilationResult) {
    super(compilationResult);
    this.modifiers = 0;
    this.selector = TypeConstants.CLINIT;
  }

  public void analyseCode(ClassScope classScope, InitializationFlowContext staticInitializerFlowContext, FlowInfo flowInfo)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    try {
      ExceptionHandlingFlowContext clinitContext = 
        new ExceptionHandlingFlowContext(
        staticInitializerFlowContext.parent, 
        this, 
        Binding.NO_EXCEPTIONS, 
        this.scope, 
        FlowInfo.DEAD_END);

      if ((flowInfo.tagBits & 0x1) == 0) {
        this.bits |= 64;
      }

      flowInfo = flowInfo.mergedWith(staticInitializerFlowContext.initsOnReturn);
      FieldBinding[] fields = this.scope.enclosingSourceType().fields();
      int i = 0; for (int count = fields.length; i < count; i++)
      {
        FieldBinding field;
        if (((field = fields[i]).isStatic()) && 
          (field.isFinal()) && 
          (!flowInfo.isDefinitelyAssigned(fields[i]))) {
          this.scope.problemReporter().uninitializedBlankFinalField(
            field, 
            this.scope.referenceType().declarationOf(field.original()));
        }

      }

      staticInitializerFlowContext.checkInitializerExceptions(
        this.scope, 
        clinitContext, 
        flowInfo);
    } catch (AbortMethod localAbortMethod) {
      this.ignoreFurtherInvestigation = true;
    }
  }

  public void generateCode(ClassScope classScope, ClassFile classFile)
  {
    int clinitOffset = 0;
    if (this.ignoreFurtherInvestigation)
    {
      return;
    }
    try {
      clinitOffset = classFile.contentsOffset;
      generateCode(classScope, classFile, clinitOffset);
    }
    catch (AbortMethod e)
    {
      if (e.compilationResult == CodeStream.RESTART_IN_WIDE_MODE)
      {
        try {
          classFile.contentsOffset = clinitOffset;
          classFile.methodCount -= 1;
          classFile.codeStream.wideMode = true;
          generateCode(classScope, classFile, clinitOffset);
        }
        catch (AbortMethod localAbortMethod1) {
          classFile.contentsOffset = clinitOffset;
          classFile.methodCount -= 1;
        }
      }
      else {
        classFile.contentsOffset = clinitOffset;
        classFile.methodCount -= 1;
      }
    }
  }

  private void generateCode(ClassScope classScope, ClassFile classFile, int clinitOffset)
  {
    ConstantPool constantPool = classFile.constantPool;
    int constantPoolOffset = constantPool.currentOffset;
    int constantPoolIndex = constantPool.currentIndex;
    classFile.generateMethodInfoHeaderForClinit();
    int codeAttributeOffset = classFile.contentsOffset;
    classFile.generateCodeAttributeHeader();
    CodeStream codeStream = classFile.codeStream;
    resolve(classScope);

    codeStream.reset(this, classFile);
    TypeDeclaration declaringType = classScope.referenceContext;

    MethodScope staticInitializerScope = declaringType.staticInitializerScope;
    staticInitializerScope.computeLocalVariablePositions(0, codeStream);

    if (this.assertionSyntheticFieldBinding != null)
    {
      codeStream.generateClassLiteralAccessForType(
        classScope.outerMostClassScope().enclosingSourceType(), 
        this.classLiteralSyntheticField);
      codeStream.invokeJavaLangClassDesiredAssertionStatus();
      BranchLabel falseLabel = new BranchLabel(codeStream);
      codeStream.ifne(falseLabel);
      codeStream.iconst_1();
      BranchLabel jumpLabel = new BranchLabel(codeStream);
      codeStream.decrStackSize(1);
      codeStream.goto_(jumpLabel);
      falseLabel.place();
      codeStream.iconst_0();
      jumpLabel.place();
      codeStream.putstatic(this.assertionSyntheticFieldBinding);
    }

    FieldDeclaration[] fieldDeclarations = declaringType.fields;
    BlockScope lastInitializerScope = null;
    if (TypeDeclaration.kind(declaringType.modifiers) == 3) {
      int enumCount = 0;
      int remainingFieldCount = 0;
      if (fieldDeclarations != null) {
        int i = 0; for (int max = fieldDeclarations.length; i < max; i++) {
          FieldDeclaration fieldDecl = fieldDeclarations[i];
          if (fieldDecl.isStatic()) {
            if (fieldDecl.getKind() == 3) {
              fieldDecl.generateCode(staticInitializerScope, codeStream);
              enumCount++;
            } else {
              remainingFieldCount++;
            }
          }
        }

      }

      codeStream.generateInlinedValue(enumCount);
      codeStream.anewarray(declaringType.binding);
      if ((enumCount > 0) && 
        (fieldDeclarations != null)) {
        int i = 0; for (int max = fieldDeclarations.length; i < max; i++) {
          FieldDeclaration fieldDecl = fieldDeclarations[i];

          if (fieldDecl.getKind() == 3) {
            codeStream.dup();
            codeStream.generateInlinedValue(fieldDecl.binding.id);
            codeStream.getstatic(fieldDecl.binding);
            codeStream.aastore();
          }
        }
      }

      codeStream.putstatic(declaringType.enumValuesSyntheticfield);
      if (remainingFieldCount != 0)
      {
        int i = 0; for (int max = fieldDeclarations.length; i < max; i++) {
          FieldDeclaration fieldDecl = fieldDeclarations[i];
          switch (fieldDecl.getKind()) {
          case 3:
            break;
          case 2:
            if (fieldDecl.isStatic())
            {
              lastInitializerScope = ((Initializer)fieldDecl).block.scope;
              fieldDecl.generateCode(staticInitializerScope, codeStream);
            }break;
          case 1:
            if (fieldDecl.binding.isStatic())
            {
              lastInitializerScope = null;
              fieldDecl.generateCode(staticInitializerScope, codeStream);
            }
            break;
          }
        }
      }
    } else if (fieldDeclarations != null) {
      int i = 0; for (int max = fieldDeclarations.length; i < max; i++) {
        FieldDeclaration fieldDecl = fieldDeclarations[i];
        switch (fieldDecl.getKind()) {
        case 2:
          if (fieldDecl.isStatic())
          {
            lastInitializerScope = ((Initializer)fieldDecl).block.scope;
            fieldDecl.generateCode(staticInitializerScope, codeStream);
          }break;
        case 1:
          if (fieldDecl.binding.isStatic())
          {
            lastInitializerScope = null;
            fieldDecl.generateCode(staticInitializerScope, codeStream);
          }
          break;
        }
      }
    }

    if (codeStream.position == 0)
    {
      classFile.contentsOffset = clinitOffset;

      classFile.methodCount -= 1;

      constantPool.resetForClinit(constantPoolIndex, constantPoolOffset);
    } else {
      if ((this.bits & 0x40) != 0) {
        int before = codeStream.position;
        codeStream.return_();
        if (lastInitializerScope != null)
        {
          codeStream.updateLastRecordedEndPC(lastInitializerScope, before);
        }
      }

      codeStream.recordPositionsFrom(0, declaringType.sourceStart);
      classFile.completeCodeAttributeForClinit(codeAttributeOffset);
    }
  }

  public boolean isClinit()
  {
    return true;
  }

  public boolean isInitializationMethod()
  {
    return true;
  }

  public boolean isStatic()
  {
    return true;
  }

  public void parseStatements(Parser parser, CompilationUnitDeclaration unit)
  {
  }

  public StringBuffer print(int tab, StringBuffer output)
  {
    printIndent(tab, output).append("<clinit>()");
    printBody(tab + 1, output);
    return output;
  }

  public void resolve(ClassScope classScope)
  {
    this.scope = new MethodScope(classScope, classScope.referenceContext, true);
  }

  public void traverse(ASTVisitor visitor, ClassScope classScope)
  {
    visitor.visit(this, classScope);
    visitor.endVisit(this, classScope);
  }

  public void setAssertionSupport(FieldBinding assertionSyntheticFieldBinding, boolean needClassLiteralField)
  {
    this.assertionSyntheticFieldBinding = assertionSyntheticFieldBinding;

    if (needClassLiteralField) {
      SourceTypeBinding sourceType = 
        this.scope.outerMostClassScope().enclosingSourceType();

      if ((!sourceType.isInterface()) && (!sourceType.isBaseType()))
        this.classLiteralSyntheticField = sourceType.addSyntheticFieldForClassLiteral(sourceType, this.scope);
    }
  }
}