package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LocalTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ProblemMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class ExplicitConstructorCall extends Statement
  implements InvocationSite
{
  public Expression[] arguments;
  public Expression qualification;
  public MethodBinding binding;
  protected MethodBinding codegenBinding;
  MethodBinding syntheticAccessor;
  public int accessMode;
  public TypeReference[] typeArguments;
  public TypeBinding[] genericTypeArguments;
  public static final int ImplicitSuper = 1;
  public static final int Super = 2;
  public static final int This = 3;
  public VariableBinding[][] implicitArguments;
  public int typeArgumentsSourceStart;

  public ExplicitConstructorCall(int accessMode)
  {
    this.accessMode = accessMode;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    try
    {
      ((MethodScope)currentScope).isConstructorCall = true;

      if (this.qualification != null) {
        flowInfo = 
          this.qualification
          .analyseCode(currentScope, flowContext, flowInfo)
          .unconditionalInits();
      }

      if (this.arguments != null) {
        int i = 0; for (int max = this.arguments.length; i < max; i++)
          flowInfo = 
            this.arguments[i]
            .analyseCode(currentScope, flowContext, flowInfo)
            .unconditionalInits();
      }
      ReferenceBinding[] thrownExceptions;
      if ((thrownExceptions = this.binding.thrownExceptions) != Binding.NO_EXCEPTIONS)
      {
        flowContext.checkExceptionHandlers(
          thrownExceptions, 
          this.accessMode == 1 ? 
          (ASTNode)currentScope.methodScope().referenceContext : 
          this, 
          flowInfo, 
          currentScope);
      }
      manageEnclosingInstanceAccessIfNecessary(currentScope, flowInfo);
      manageSyntheticAccessIfNecessary(currentScope, flowInfo);
      return flowInfo;
    } finally {
      ((MethodScope)currentScope).isConstructorCall = false;
    }
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream)
  {
    if ((this.bits & 0x80000000) == 0)
      return;
    try
    {
      ((MethodScope)currentScope).isConstructorCall = true;

      int pc = codeStream.position;
      codeStream.aload_0();

      ReferenceBinding targetType = this.codegenBinding.declaringClass;

      if ((targetType.erasure().id == 41) || (targetType.isEnum())) {
        codeStream.aload_1();
        codeStream.iload_2();
      }

      if (targetType.isNestedType()) {
        codeStream.generateSyntheticEnclosingInstanceValues(
          currentScope, 
          targetType, 
          (this.bits & 0x2000) != 0 ? null : this.qualification, 
          this);
      }

      generateArguments(this.binding, this.arguments, currentScope, codeStream);

      if (targetType.isNestedType()) {
        codeStream.generateSyntheticOuterArgumentValues(
          currentScope, 
          targetType, 
          this);
      }
      if (this.syntheticAccessor != null)
      {
        int i = 0;
        int max = this.syntheticAccessor.parameters.length - this.codegenBinding.parameters.length;
        for (; i < max; 
          i++) {
          codeStream.aconst_null();
        }
        codeStream.invokespecial(this.syntheticAccessor);
      } else {
        codeStream.invokespecial(this.codegenBinding);
      }
      codeStream.recordPositionsFrom(pc, this.sourceStart);
    } finally {
      ((MethodScope)currentScope).isConstructorCall = false;
    }
  }

  public TypeBinding[] genericTypeArguments()
  {
    return this.genericTypeArguments;
  }

  public boolean isImplicitSuper()
  {
    return this.accessMode == 1;
  }

  public boolean isSuperAccess()
  {
    return this.accessMode != 3;
  }

  public boolean isTypeAccess()
  {
    return true;
  }

  void manageEnclosingInstanceAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo)
  {
    ReferenceBinding superTypeErasure = (ReferenceBinding)this.binding.declaringClass.erasure();

    if ((flowInfo.tagBits & 0x1) == 0)
    {
      if ((superTypeErasure.isNestedType()) && 
        (currentScope.enclosingSourceType().isLocalType()))
      {
        if (superTypeErasure.isLocalType()) {
          ((LocalTypeBinding)superTypeErasure).addInnerEmulationDependent(currentScope, this.qualification != null);
        }
        else
          currentScope.propagateInnerEmulation(superTypeErasure, this.qualification != null);
      }
    }
  }

  public void manageSyntheticAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo)
  {
    if ((flowInfo.tagBits & 0x1) == 0)
    {
      this.codegenBinding = this.binding.original();

      if ((this.binding.isPrivate()) && (this.accessMode != 3)) {
        ReferenceBinding declaringClass = this.codegenBinding.declaringClass;

        if (((declaringClass.tagBits & 0x10) != 0L) && (currentScope.compilerOptions().complianceLevel >= 3145728L))
        {
          this.codegenBinding.tagBits |= 1024L;
        } else {
          this.syntheticAccessor = ((SourceTypeBinding)declaringClass).addSyntheticMethod(this.codegenBinding, isSuperAccess());
          currentScope.problemReporter().needToEmulateMethodAccess(this.codegenBinding, this);
        }
      }
    }
  }

  public StringBuffer printStatement(int indent, StringBuffer output)
  {
    printIndent(indent, output);
    if (this.qualification != null) this.qualification.printExpression(0, output).append('.');
    if (this.typeArguments != null) {
      output.append('<');
      int max = this.typeArguments.length - 1;
      for (int j = 0; j < max; j++) {
        this.typeArguments[j].print(0, output);
        output.append(", ");
      }
      this.typeArguments[max].print(0, output);
      output.append('>');
    }
    if (this.accessMode == 3)
      output.append("this(");
    else {
      output.append("super(");
    }
    if (this.arguments != null) {
      for (int i = 0; i < this.arguments.length; i++) {
        if (i > 0) output.append(", ");
        this.arguments[i].printExpression(0, output);
      }
    }
    return output.append(");");
  }

  public void resolve(BlockScope scope)
  {
    MethodScope methodScope = scope.methodScope();
    try {
      AbstractMethodDeclaration methodDeclaration = methodScope.referenceMethod();
      if ((methodDeclaration == null) || 
        (!methodDeclaration.isConstructor()) || 
        (((ConstructorDeclaration)methodDeclaration).constructorCall != this)) {
        scope.problemReporter().invalidExplicitConstructorCall(this);

        if (this.qualification != null) {
          this.qualification.resolveType(scope);
        }
        if (this.typeArguments != null) {
          int i = 0; for (int max = this.typeArguments.length; i < max; i++) {
            this.typeArguments[i].resolveType(scope, true);
          }
        }
        if (this.arguments != null) {
          int i = 0; for (int max = this.arguments.length; i < max; i++)
            this.arguments[i].resolveType(scope);  }  } ReferenceBinding receiverType;
      TypeBinding[] argumentTypes;
      boolean argsContainCast;
      while (true) { return;

        methodScope.isConstructorCall = true;
        receiverType = scope.enclosingReceiverType();
        if (this.accessMode != 3) {
          receiverType = receiverType.superclass();
        }
        if (receiverType != null)
        {
          if ((this.accessMode == 2) && (receiverType.erasure().id == 41)) {
            scope.problemReporter().cannotInvokeSuperConstructorInEnum(this, methodScope.referenceMethod().binding);
          }

          if (this.qualification != null) {
            if (this.accessMode != 2) {
              scope.problemReporter().unnecessaryEnclosingInstanceSpecification(
                this.qualification, 
                receiverType);
            }
            ReferenceBinding enclosingType = receiverType.enclosingType();
            if (enclosingType == null) {
              scope.problemReporter().unnecessaryEnclosingInstanceSpecification(
                this.qualification, 
                receiverType);
              this.bits |= 8192;
            } else {
              TypeBinding qTb = this.qualification.resolveTypeExpecting(scope, enclosingType);
              this.qualification.computeConversion(scope, qTb, qTb);
            }
          }

          if (this.typeArguments != null) {
            int length = this.typeArguments.length;
            boolean argHasError = false;
            this.genericTypeArguments = new TypeBinding[length];
            for (int i = 0; i < length; i++) {
              TypeReference typeReference = this.typeArguments[i];
              if ((this.genericTypeArguments[i] =  = typeReference.resolveType(scope, true)) == null) {
                argHasError = true;
              }
              if ((argHasError) && ((typeReference instanceof Wildcard))) {
                scope.problemReporter().illegalUsageOfWildcard(typeReference);
              }

            }

            if (argHasError);
          }
          else
          {
            argumentTypes = Binding.NO_PARAMETERS;
            argsContainCast = false;
            if (this.arguments == null) break;
            boolean argHasError = false;
            int length = this.arguments.length;
            argumentTypes = new TypeBinding[length];
            for (int i = 0; i < length; i++) {
              Expression argument = this.arguments[i];
              if ((argument instanceof CastExpression)) {
                argument.bits |= 32;
                argsContainCast = true;
              }
              if ((argumentTypes[i] =  = argument.resolveType(scope)) == null) {
                argHasError = true;
              }
            }
            if (!argHasError)
              break label728;
            TypeBinding[] pseudoArgs = new TypeBinding[length];
            int i = length;
            do { pseudoArgs[i] = (argumentTypes[i] == null ? TypeBinding.NULL : argumentTypes[i]);

              i--; } while (i >= 0);

            this.binding = scope.findMethod(receiverType, TypeConstants.INIT, pseudoArgs, this);
            if ((this.binding != null) && (!this.binding.isValidBinding())) {
              MethodBinding closestMatch = ((ProblemMethodBinding)this.binding).closestMatch;

              if (closestMatch != null) {
                if (closestMatch.original().typeVariables != Binding.NO_TYPE_VARIABLES)
                {
                  closestMatch = scope.environment().createParameterizedGenericMethod(closestMatch.original(), null);
                }
                this.binding = closestMatch;
                MethodBinding closestMatchOriginal = closestMatch.original();
                if (((closestMatchOriginal.isPrivate()) || (closestMatchOriginal.declaringClass.isLocalType())) && (!scope.isDefinedInMethod(closestMatchOriginal)))
                {
                  closestMatchOriginal.modifiers |= 134217728;
                }
              }
            }
          }
        } }
      if (receiverType.erasure().id == 41)
      {
        argumentTypes = new TypeBinding[] { scope.getJavaLangString(), TypeBinding.INT };
      }
      label728: if ((this.binding = scope.getConstructor(receiverType, argumentTypes, this)).isValidBinding()) {
        if (isMethodUseDeprecated(this.binding, scope, this.accessMode != 1))
          scope.problemReporter().deprecatedMethod(this.binding, this);
        checkInvocationArguments(scope, null, receiverType, this.binding, this.arguments, argumentTypes, argsContainCast, this);
        if ((this.binding.isPrivate()) || (receiverType.isLocalType()))
          this.binding.original().modifiers |= 134217728;
      }
      else {
        if (this.binding.declaringClass == null)
          this.binding.declaringClass = receiverType;
        scope.problemReporter().invalidConstructor(this, this.binding);
      }
    } finally {
      methodScope.isConstructorCall = false; } methodScope.isConstructorCall = false;
  }

  public void setActualReceiverType(ReferenceBinding receiverType)
  {
  }

  public void setDepth(int depth)
  {
  }

  public void setFieldIndex(int depth)
  {
  }

  public void traverse(ASTVisitor visitor, BlockScope scope)
  {
    if (visitor.visit(this, scope)) {
      if (this.qualification != null) {
        this.qualification.traverse(visitor, scope);
      }
      if (this.typeArguments != null) {
        int i = 0; for (int typeArgumentsLength = this.typeArguments.length; i < typeArgumentsLength; i++) {
          this.typeArguments[i].traverse(visitor, scope);
        }
      }
      if (this.arguments != null) {
        int i = 0; for (int argumentLength = this.arguments.length; i < argumentLength; i++)
          this.arguments[i].traverse(visitor, scope);
      }
    }
    visitor.endVisit(this, scope);
  }
}