package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.BaseTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class FieldDeclaration extends AbstractVariableDeclaration
{
  public FieldBinding binding;
  public Javadoc javadoc;
  public int endPart1Position;
  public int endPart2Position;

  public FieldDeclaration()
  {
  }

  public FieldDeclaration(char[] name, int sourceStart, int sourceEnd)
  {
    this.name = name;

    this.sourceStart = sourceStart;
    this.sourceEnd = sourceEnd;
  }

  public FlowInfo analyseCode(MethodScope initializationScope, FlowContext flowContext, FlowInfo flowInfo) {
    if ((this.binding != null) && (!this.binding.isUsed()) && 
      ((this.binding.isPrivate()) || ((this.binding.declaringClass != null) && (this.binding.declaringClass.isLocalType()))) && 
      (!initializationScope.referenceCompilationUnit().compilationResult.hasSyntaxError)) {
      initializationScope.problemReporter().unusedPrivateField(this);
    }

    if ((this.binding != null) && 
      (this.binding.isValidBinding()) && 
      (this.binding.isStatic()) && 
      (this.binding.constant() == Constant.NotAConstant) && 
      (this.binding.declaringClass.isNestedType()) && 
      (!this.binding.declaringClass.isStatic())) {
      initializationScope.problemReporter().unexpectedStaticModifierForField(
        (SourceTypeBinding)this.binding.declaringClass, 
        this);
    }

    if (this.initialization != null) {
      flowInfo = 
        this.initialization
        .analyseCode(initializationScope, flowContext, flowInfo)
        .unconditionalInits();
      flowInfo.markAsDefinitelyAssigned(this.binding);
    }
    return flowInfo;
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream)
  {
    if ((this.bits & 0x80000000) == 0) {
      return;
    }

    int pc = codeStream.position;
    boolean isStatic;
    if ((this.initialization != null) && (
      (!(isStatic = this.binding.isStatic())) || (this.binding.constant() == Constant.NotAConstant)))
    {
      if (!isStatic) {
        codeStream.aload_0();
      }
      this.initialization.generateCode(currentScope, codeStream, true);

      if (isStatic)
        codeStream.putstatic(this.binding);
      else {
        codeStream.putfield(this.binding);
      }
    }
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public int getKind()
  {
    return this.type == null ? 3 : 1;
  }

  public boolean isStatic() {
    if (this.binding != null)
      return this.binding.isStatic();
    return (this.modifiers & 0x8) != 0;
  }

  public StringBuffer printStatement(int indent, StringBuffer output) {
    if (this.javadoc != null) {
      this.javadoc.print(indent, output);
    }
    return super.printStatement(indent, output);
  }

  public void resolve(MethodScope initializationScope)
  {
    if ((this.bits & 0x10) != 0) return;
    if ((this.binding == null) || (!this.binding.isValidBinding())) return;

    this.bits |= 16;

    ClassScope classScope = initializationScope.enclosingClassScope();

    if (classScope != null)
    {
      SourceTypeBinding declaringType = classScope.enclosingSourceType();

      if (declaringType.superclass != null) {
        Binding existingVariable = classScope.findField(declaringType.superclass, this.name, this, false);
        if ((existingVariable != null) && 
          (existingVariable.isValidBinding())) {
          if ((existingVariable instanceof FieldBinding)) { FieldBinding existingField = (FieldBinding)existingVariable;
            if (existingField.original() == this.binding);
          }
          else {
            initializationScope.problemReporter().fieldHiding(this, existingVariable);
            break label263;
          }
        }
      }
      Scope outerScope = classScope.parent;
      if (outerScope.kind != 4) {
        Binding existingVariable = outerScope.getBinding(this.name, 3, this, false);
        if ((existingVariable != null) && 
          (existingVariable.isValidBinding()) && 
          (existingVariable != this.binding))
          if ((existingVariable instanceof FieldBinding)) { FieldBinding existingField = (FieldBinding)existingVariable;
            if ((existingField.original() == this.binding) || (
              (!existingField.isStatic()) && (declaringType.isStatic())));
          }
          else {
            initializationScope.problemReporter().fieldHiding(this, existingVariable);
          }
      }
    }
    label263: if (this.type != null) {
      this.type.resolvedType = this.binding.type;
    }

    FieldBinding previousField = initializationScope.initializedField;
    int previousFieldID = initializationScope.lastVisibleFieldID;
    try {
      initializationScope.initializedField = this.binding;
      initializationScope.lastVisibleFieldID = this.binding.id;

      resolveAnnotations(initializationScope, this.annotations, this.binding);

      if (((this.binding.getAnnotationTagBits() & 0x0) == 0L) && 
        ((this.binding.modifiers & 0x100000) != 0) && 
        (initializationScope.compilerOptions().sourceLevel >= 3211264L)) {
        initializationScope.problemReporter().missingDeprecatedAnnotationForField(this);
      }

      if (this.initialization == null) {
        this.binding.setConstant(Constant.NotAConstant);
      }
      else {
        this.binding.setConstant(Constant.NotAConstant);

        TypeBinding fieldType = this.binding.type;

        this.initialization.setExpectedType(fieldType);
        if ((this.initialization instanceof ArrayInitializer))
        {
          TypeBinding initializationType;
          if ((initializationType = this.initialization.resolveTypeExpecting(initializationScope, fieldType)) != null) {
            ((ArrayInitializer)this.initialization).binding = ((ArrayBinding)initializationType);
            this.initialization.computeConversion(initializationScope, fieldType, initializationType);
          }
        }
        else
        {
          TypeBinding initializationType;
          if ((initializationType = this.initialization.resolveType(initializationScope)) != null)
          {
            if (fieldType != initializationType)
              initializationScope.compilationUnitScope().recordTypeConversion(fieldType, initializationType);
            if ((this.initialization.isConstantValueOfTypeAssignableToType(initializationType, fieldType)) || 
              ((fieldType.isBaseType()) && (BaseTypeBinding.isWidening(fieldType.id, initializationType.id))) || 
              (initializationType.isCompatibleWith(fieldType))) {
              this.initialization.computeConversion(initializationScope, fieldType, initializationType);
              if (initializationType.needsUncheckedConversion(fieldType)) {
                initializationScope.problemReporter().unsafeTypeConversion(this.initialization, initializationType, fieldType);
              }
              if (((this.initialization instanceof CastExpression)) && 
                ((this.initialization.bits & 0x4000) == 0))
                CastExpression.checkNeedForAssignedCast(initializationScope, fieldType, (CastExpression)this.initialization);
            }
            else if ((initializationScope.isBoxingCompatibleWith(initializationType, fieldType)) || (
              (initializationType.isBaseType()) && 
              (initializationScope.compilerOptions().sourceLevel >= 3211264L) && 
              (!fieldType.isBaseType()) && 
              (this.initialization.isConstantValueOfTypeAssignableToType(initializationType, initializationScope.environment().computeBoxingType(fieldType))))) {
              this.initialization.computeConversion(initializationScope, fieldType, initializationType);
              if (((this.initialization instanceof CastExpression)) && 
                ((this.initialization.bits & 0x4000) == 0))
                CastExpression.checkNeedForAssignedCast(initializationScope, fieldType, (CastExpression)this.initialization);
            }
            else {
              initializationScope.problemReporter().typeMismatchError(initializationType, fieldType, this);
            }
            if (this.binding.isFinal())
              this.binding.setConstant(this.initialization.constant.castTo((this.binding.type.id << 4) + this.initialization.constant.typeID()));
          }
          else {
            this.binding.setConstant(Constant.NotAConstant);
          }
        }
        if (this.binding == Assignment.getDirectBinding(this.initialization)) {
          initializationScope.problemReporter().assignmentHasNoEffect(this, this.name);
        }
      }

      if (this.javadoc != null)
      {
        this.javadoc.resolve(initializationScope);
      } else if ((this.binding.declaringClass != null) && (!this.binding.declaringClass.isLocalType()))
        initializationScope.problemReporter().javadocMissing(this.sourceStart, this.sourceEnd, this.binding.modifiers);
    }
    finally {
      initializationScope.initializedField = previousField;
      initializationScope.lastVisibleFieldID = previousFieldID;
      if (this.binding.constant() == null)
        this.binding.setConstant(Constant.NotAConstant);
    }
  }

  public void traverse(ASTVisitor visitor, MethodScope scope) {
    if (visitor.visit(this, scope)) {
      if (this.javadoc != null) {
        this.javadoc.traverse(visitor, scope);
      }
      if (this.annotations != null) {
        int annotationsLength = this.annotations.length;
        for (int i = 0; i < annotationsLength; i++)
          this.annotations[i].traverse(visitor, scope);
      }
      if (this.type != null) {
        this.type.traverse(visitor, scope);
      }
      if (this.initialization != null)
        this.initialization.traverse(visitor, scope);
    }
    visitor.endVisit(this, scope);
  }
}