package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class IfStatement extends Statement
{
  public Expression condition;
  public Statement thenStatement;
  public Statement elseStatement;
  int thenInitStateIndex = -1;
  int elseInitStateIndex = -1;
  int mergedInitStateIndex = -1;

  public IfStatement(Expression condition, Statement thenStatement, int sourceStart, int sourceEnd)
  {
    this.condition = condition;
    this.thenStatement = thenStatement;

    if ((thenStatement instanceof EmptyStatement)) thenStatement.bits |= 1;
    this.sourceStart = sourceStart;
    this.sourceEnd = sourceEnd;
  }

  public IfStatement(Expression condition, Statement thenStatement, Statement elseStatement, int sourceStart, int sourceEnd)
  {
    this.condition = condition;
    this.thenStatement = thenStatement;

    if ((thenStatement instanceof EmptyStatement)) thenStatement.bits |= 1;
    this.elseStatement = elseStatement;
    if ((elseStatement instanceof IfStatement)) elseStatement.bits |= 536870912;
    if ((elseStatement instanceof EmptyStatement)) elseStatement.bits |= 1;
    this.sourceStart = sourceStart;
    this.sourceEnd = sourceEnd;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    FlowInfo conditionFlowInfo = 
      this.condition.analyseCode(currentScope, flowContext, flowInfo);

    Constant cst = this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedTrue = (cst != Constant.NotAConstant) && (cst.booleanValue());
    boolean isConditionOptimizedFalse = (cst != Constant.NotAConstant) && (!cst.booleanValue());

    FlowInfo thenFlowInfo = conditionFlowInfo.safeInitsWhenTrue();
    if (isConditionOptimizedFalse) {
      thenFlowInfo.setReachMode(1);
    }
    FlowInfo elseFlowInfo = conditionFlowInfo.initsWhenFalse();
    if (isConditionOptimizedTrue) {
      elseFlowInfo.setReachMode(1);
    }
    if (this.thenStatement != null)
    {
      this.thenInitStateIndex = 
        currentScope.methodScope().recordInitializationStates(thenFlowInfo);
      if (!this.thenStatement.complainIfUnreachable(thenFlowInfo, currentScope, false)) {
        thenFlowInfo = 
          this.thenStatement.analyseCode(currentScope, flowContext, thenFlowInfo);
      }
    }

    if ((thenFlowInfo.tagBits & 0x1) != 0) {
      this.bits |= 1073741824;
    }

    if (this.elseStatement != null)
    {
      if ((thenFlowInfo == FlowInfo.DEAD_END) && 
        ((this.bits & 0x20000000) == 0) && 
        (!(this.elseStatement instanceof IfStatement))) {
        currentScope.problemReporter().unnecessaryElse(this.elseStatement);
      }

      this.elseInitStateIndex = 
        currentScope.methodScope().recordInitializationStates(elseFlowInfo);
      if (!this.elseStatement.complainIfUnreachable(elseFlowInfo, currentScope, false)) {
        elseFlowInfo = 
          this.elseStatement.analyseCode(currentScope, flowContext, elseFlowInfo);
      }

    }

    FlowInfo mergedInfo = FlowInfo.mergedOptimizedBranches(
      thenFlowInfo, 
      isConditionOptimizedTrue, 
      elseFlowInfo, 
      isConditionOptimizedFalse, 
      true);
    this.mergedInitStateIndex = currentScope.methodScope().recordInitializationStates(mergedInfo);
    return mergedInfo;
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream)
  {
    if ((this.bits & 0x80000000) == 0) {
      return;
    }
    int pc = codeStream.position;
    BranchLabel endifLabel = new BranchLabel(codeStream);
    Constant cst;
    boolean hasThenPart = 
      (((cst = this.condition.optimizedBooleanConstant()) == Constant.NotAConstant) || 
      (cst.booleanValue())) && 
      (this.thenStatement != null) && (!
      this.thenStatement.isEmptyBlock());
    boolean hasElsePart = 
      ((cst == Constant.NotAConstant) || (!cst.booleanValue())) && 
      (this.elseStatement != null) && (!
      this.elseStatement.isEmptyBlock());
    if (hasThenPart) {
      BranchLabel falseLabel = null;

      this.condition.generateOptimizedBoolean(
        currentScope, 
        codeStream, 
        null, 
        hasElsePart ? (falseLabel = new BranchLabel(codeStream)) : endifLabel, 
        true);

      if (this.thenInitStateIndex != -1) {
        codeStream.removeNotDefinitelyAssignedVariables(currentScope, this.thenInitStateIndex);
        codeStream.addDefinitelyAssignedVariables(currentScope, this.thenInitStateIndex);
      }

      this.thenStatement.generateCode(currentScope, codeStream);

      if (hasElsePart) {
        if ((this.bits & 0x40000000) == 0) {
          this.thenStatement.branchChainTo(endifLabel);
          int position = codeStream.position;
          codeStream.goto_(endifLabel);

          codeStream.updateLastRecordedEndPC((this.thenStatement instanceof Block) ? ((Block)this.thenStatement).scope : currentScope, position);
        }

        if (this.elseInitStateIndex != -1) {
          codeStream.removeNotDefinitelyAssignedVariables(
            currentScope, 
            this.elseInitStateIndex);
          codeStream.addDefinitelyAssignedVariables(currentScope, this.elseInitStateIndex);
        }
        if (falseLabel != null) falseLabel.place();
        this.elseStatement.generateCode(currentScope, codeStream);
      }
    } else if (hasElsePart)
    {
      this.condition.generateOptimizedBoolean(
        currentScope, 
        codeStream, 
        endifLabel, 
        null, 
        true);

      if (this.elseInitStateIndex != -1) {
        codeStream.removeNotDefinitelyAssignedVariables(
          currentScope, 
          this.elseInitStateIndex);
        codeStream.addDefinitelyAssignedVariables(currentScope, this.elseInitStateIndex);
      }
      this.elseStatement.generateCode(currentScope, codeStream);
    }
    else {
      this.condition.generateCode(currentScope, codeStream, false);
      codeStream.recordPositionsFrom(pc, this.sourceStart);
    }

    if (this.mergedInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables(
        currentScope, 
        this.mergedInitStateIndex);
      codeStream.addDefinitelyAssignedVariables(currentScope, this.mergedInitStateIndex);
    }
    endifLabel.place();
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public StringBuffer printStatement(int indent, StringBuffer output)
  {
    printIndent(indent, output).append("if (");
    this.condition.printExpression(0, output).append(")\n");
    this.thenStatement.printStatement(indent + 2, output);
    if (this.elseStatement != null) {
      output.append('\n');
      printIndent(indent, output);
      output.append("else\n");
      this.elseStatement.printStatement(indent + 2, output);
    }
    return output;
  }

  public void resolve(BlockScope scope)
  {
    TypeBinding type = this.condition.resolveTypeExpecting(scope, TypeBinding.BOOLEAN);
    this.condition.computeConversion(scope, type, type);
    if (this.thenStatement != null)
      this.thenStatement.resolve(scope);
    if (this.elseStatement != null)
      this.elseStatement.resolve(scope);
  }

  public void traverse(ASTVisitor visitor, BlockScope blockScope)
  {
    if (visitor.visit(this, blockScope)) {
      this.condition.traverse(visitor, blockScope);
      if (this.thenStatement != null)
        this.thenStatement.traverse(visitor, blockScope);
      if (this.elseStatement != null)
        this.elseStatement.traverse(visitor, blockScope);
    }
    visitor.endVisit(this, blockScope);
  }
}