package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.flow.ExceptionHandlingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.InitializationFlowContext;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.problem.AbortMethod;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class MethodDeclaration extends AbstractMethodDeclaration
{
  public TypeReference returnType;
  public TypeParameter[] typeParameters;

  public MethodDeclaration(CompilationResult compilationResult)
  {
    super(compilationResult);
  }

  public void analyseCode(ClassScope classScope, InitializationFlowContext initializationContext, FlowInfo flowInfo)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    try {
      if (this.binding == null) {
        return;
      }
      if ((!this.binding.isUsed()) && 
        ((this.binding.isPrivate()) || (
        ((this.binding.modifiers & 0x30000000) == 0) && (this.binding.declaringClass.isLocalType()))) && 
        (!classScope.referenceCompilationUnit().compilationResult.hasSyntaxError)) {
        this.scope.problemReporter().unusedPrivateMethod(this);
      }

      if ((this.binding.declaringClass.isEnum()) && ((this.selector == TypeConstants.VALUES) || (this.selector == TypeConstants.VALUEOF))) {
        return;
      }

      if ((this.binding.isAbstract()) || (this.binding.isNative())) {
        return;
      }
      ExceptionHandlingFlowContext methodContext = 
        new ExceptionHandlingFlowContext(
        initializationContext, 
        this, 
        this.binding.thrownExceptions, 
        this.scope, 
        FlowInfo.DEAD_END);

      if (this.arguments != null) {
        int i = 0; for (int count = this.arguments.length; i < count; i++) {
          flowInfo.markAsDefinitelyAssigned(this.arguments[i].binding);
        }
      }

      if (this.statements != null) {
        boolean didAlreadyComplain = false;
        int i = 0; for (int count = this.statements.length; i < count; i++) {
          Statement stat = this.statements[i];
          if (!stat.complainIfUnreachable(flowInfo, this.scope, didAlreadyComplain))
            flowInfo = stat.analyseCode(this.scope, methodContext, flowInfo);
          else {
            didAlreadyComplain = true;
          }
        }
      }

      TypeBinding returnTypeBinding = this.binding.returnType;
      if ((returnTypeBinding == TypeBinding.VOID) || (isAbstract())) {
        if ((flowInfo.tagBits & 0x1) == 0) {
          this.bits |= 64;
        }
      }
      else if (flowInfo != FlowInfo.DEAD_END) {
        this.scope.problemReporter().shouldReturn(returnTypeBinding, this);
      }

      methodContext.complainIfUnusedExceptionHandlers(this);
    } catch (AbortMethod localAbortMethod) {
      this.ignoreFurtherInvestigation = true;
    }
  }

  public boolean isMethod()
  {
    return true;
  }

  public void parseStatements(Parser parser, CompilationUnitDeclaration unit)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    parser.parse(this, unit);
  }

  public StringBuffer printReturnType(int indent, StringBuffer output)
  {
    if (this.returnType == null) return output;
    return this.returnType.printExpression(0, output).append(' ');
  }

  public void resolveStatements()
  {
    if ((this.returnType != null) && (this.binding != null)) {
      this.returnType.resolvedType = this.binding.returnType;
    }

    if (CharOperation.equals(this.scope.enclosingSourceType().sourceName, this.selector)) {
      this.scope.problemReporter().methodWithConstructorName(this);
    }

    if (this.typeParameters != null) {
      int i = 0; for (int length = this.typeParameters.length; i < length; i++) {
        this.typeParameters[i].resolve(this.scope);
      }

    }

    CompilerOptions compilerOptions = this.scope.compilerOptions();

    if (this.binding != null) {
      long sourceLevel = compilerOptions.sourceLevel;
      if (sourceLevel >= 3211264L) {
        int bindingModifiers = this.binding.modifiers;
        boolean hasOverrideAnnotation = (this.binding.tagBits & 0x0) != 0L;
        boolean isInterfaceMethod = this.binding.declaringClass.isInterface();
        if (hasOverrideAnnotation)
        {
          if ((isInterfaceMethod) || ((bindingModifiers & 0x10000008) != 268435456))
          {
            if ((sourceLevel < 3276800L) || 
              ((bindingModifiers & 0x20000008) != 536870912))
            {
              this.scope.problemReporter().methodMustOverride(this); } 
          } } else if ((!isInterfaceMethod) && 
          ((bindingModifiers & 0x10000008) == 268435456))
        {
          this.scope.problemReporter().missingOverrideAnnotation(this);
        }
      }
    }

    switch (TypeDeclaration.kind(this.scope.referenceType().modifiers)) {
    case 3:
      if ((this.selector == TypeConstants.VALUES) || 
        (this.selector == TypeConstants.VALUEOF)) {
        break;
      }
    case 1:
      if ((this.modifiers & 0x1000000) != 0) {
        if (((this.modifiers & 0x100) == 0) && 
          ((this.modifiers & 0x400) == 0)) {
          this.scope.problemReporter().methodNeedBody(this);
        }
      }
      else if (((this.modifiers & 0x100) != 0) || ((this.modifiers & 0x400) != 0))
        this.scope.problemReporter().methodNeedingNoBody(this); break;
    case 2:
    }
    super.resolveStatements();

    if ((compilerOptions.getSeverity(562949953421312L) != -1) && 
      (this.binding != null)) {
      int bindingModifiers = this.binding.modifiers;
      if (((bindingModifiers & 0x30000000) == 268435456) && 
        ((this.bits & 0x10) == 0))
        this.scope.problemReporter().overridesMethodWithoutSuperInvocation(this.binding);
    }
  }

  public void traverse(ASTVisitor visitor, ClassScope classScope)
  {
    if (visitor.visit(this, classScope)) {
      if (this.javadoc != null) {
        this.javadoc.traverse(visitor, this.scope);
      }
      if (this.annotations != null) {
        int annotationsLength = this.annotations.length;
        for (int i = 0; i < annotationsLength; i++)
          this.annotations[i].traverse(visitor, this.scope);
      }
      if (this.typeParameters != null) {
        int typeParametersLength = this.typeParameters.length;
        for (int i = 0; i < typeParametersLength; i++) {
          this.typeParameters[i].traverse(visitor, this.scope);
        }
      }
      if (this.returnType != null)
        this.returnType.traverse(visitor, this.scope);
      if (this.arguments != null) {
        int argumentLength = this.arguments.length;
        for (int i = 0; i < argumentLength; i++)
          this.arguments[i].traverse(visitor, this.scope);
      }
      if (this.thrownExceptions != null) {
        int thrownExceptionsLength = this.thrownExceptions.length;
        for (int i = 0; i < thrownExceptionsLength; i++)
          this.thrownExceptions[i].traverse(visitor, this.scope);
      }
      if (this.statements != null) {
        int statementsLength = this.statements.length;
        for (int i = 0; i < statementsLength; i++)
          this.statements[i].traverse(visitor, this.scope);
      }
    }
    visitor.endVisit(this, classScope);
  }
  public TypeParameter[] typeParameters() {
    return this.typeParameters;
  }
}