package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class ParameterizedSingleTypeReference extends ArrayTypeReference
{
  public TypeReference[] typeArguments;

  public ParameterizedSingleTypeReference(char[] name, TypeReference[] typeArguments, int dim, long pos)
  {
    super(name, dim, pos);
    this.originalSourceEnd = this.sourceEnd;
    this.typeArguments = typeArguments;
  }
  public void checkBounds(Scope scope) {
    if (this.resolvedType == null) return;

    if ((this.resolvedType.leafComponentType() instanceof ParameterizedTypeBinding)) {
      ParameterizedTypeBinding parameterizedType = (ParameterizedTypeBinding)this.resolvedType.leafComponentType();
      ReferenceBinding currentType = parameterizedType.genericType();
      TypeVariableBinding[] typeVariables = currentType.typeVariables();
      TypeBinding[] argTypes = parameterizedType.arguments;
      if ((argTypes != null) && (typeVariables != null))
        parameterizedType.boundCheck(scope, this.typeArguments);
    }
  }

  public TypeReference copyDims(int dim)
  {
    return new ParameterizedSingleTypeReference(this.token, this.typeArguments, dim, (this.sourceStart << 32) + this.sourceEnd);
  }

  public char[][] getParameterizedTypeName()
  {
    StringBuffer buffer = new StringBuffer(5);
    buffer.append(this.token).append('<');
    int i = 0; for (int length = this.typeArguments.length; i < length; i++) {
      if (i > 0) buffer.append(',');
      buffer.append(CharOperation.concatWith(this.typeArguments[i].getParameterizedTypeName(), '.'));
    }
    buffer.append('>');
    int nameLength = buffer.length();
    char[] name = new char[nameLength];
    buffer.getChars(0, nameLength, name, 0);
    int dim = this.dimensions;
    if (dim > 0) {
      char[] dimChars = new char[dim * 2];
      for (int i = 0; i < dim; i++) {
        int index = i * 2;
        dimChars[index] = '[';
        dimChars[(index + 1)] = ']';
      }
      name = CharOperation.concat(name, dimChars);
    }
    return new char[][] { name };
  }

  protected TypeBinding getTypeBinding(Scope scope)
  {
    return null;
  }

  private TypeBinding internalResolveType(Scope scope, ReferenceBinding enclosingType, boolean checkBounds)
  {
    this.constant = Constant.NotAConstant;
    if ((this.bits & 0x40000) != 0) {
      if ((this.resolvedType != null) && (!this.resolvedType.isValidBinding()))
        return null;
      return this.resolvedType;
    }
    this.bits |= 262144;
    if (enclosingType == null) {
      this.resolvedType = scope.getType(this.token);
      if (!this.resolvedType.isValidBinding()) {
        reportInvalidType(scope);

        boolean isClassScope = scope.kind == 3;
        int argLength = this.typeArguments.length;
        for (int i = 0; i < argLength; i++) {
          TypeReference typeArgument = this.typeArguments[i];
          if (isClassScope)
            typeArgument.resolveType((ClassScope)scope);
          else {
            typeArgument.resolveType((BlockScope)scope, checkBounds);
          }
        }
        return null;
      }
      enclosingType = this.resolvedType.enclosingType();
      if ((enclosingType != null) && ((enclosingType.isGenericType()) || (enclosingType.isParameterizedType()))) {
        ReferenceBinding currentType = (ReferenceBinding)this.resolvedType;
        enclosingType = currentType.isStatic() ? 
          (ReferenceBinding)scope.environment().convertToRawType(enclosingType) : 
          scope.environment().convertToParameterizedType(enclosingType);
      }
    } else {
      this.resolvedType = scope.getMemberType(this.token, enclosingType);
      if (!this.resolvedType.isValidBinding()) {
        scope.problemReporter().invalidEnclosingType(this, this.resolvedType, enclosingType);
        return null;
      }
      if (isTypeUseDeprecated(this.resolvedType, scope)) {
        scope.problemReporter().deprecatedType(this.resolvedType, this);
      }
    }

    boolean isClassScope = scope.kind == 3;
    TypeReference keep = null;
    if (isClassScope) {
      keep = ((ClassScope)scope).superTypeReference;
      ((ClassScope)scope).superTypeReference = null;
    }
    ReferenceBinding currentType = (ReferenceBinding)this.resolvedType;
    int argLength = this.typeArguments.length;
    TypeBinding[] argTypes = new TypeBinding[argLength];
    boolean argHasError = false;
    for (int i = 0; i < argLength; i++) {
      TypeReference typeArgument = this.typeArguments[i];
      TypeBinding argType = isClassScope ? 
        typeArgument.resolveTypeArgument((ClassScope)scope, currentType, i) : 
        typeArgument.resolveTypeArgument((BlockScope)scope, currentType, i);
      if (argType == null)
        argHasError = true;
      else {
        argTypes[i] = argType;
      }
    }
    if (argHasError) return null;
    if (isClassScope) {
      ((ClassScope)scope).superTypeReference = keep;
      if (((ClassScope)scope).detectHierarchyCycle(currentType, this)) {
        return null;
      }
    }
    TypeVariableBinding[] typeVariables = currentType.typeVariables();
    if (typeVariables == Binding.NO_TYPE_VARIABLES) {
      scope.problemReporter().nonGenericTypeCannotBeParameterized(this, currentType, argTypes);
      return null;
    }if (argLength != typeVariables.length) {
      scope.problemReporter().incorrectArityForParameterizedType(this, currentType, argTypes);
      return null;
    }if (!currentType.isStatic()) {
      ReferenceBinding actualEnclosing = currentType.enclosingType();
      if ((actualEnclosing != null) && (actualEnclosing.isRawType())) {
        scope.problemReporter().rawMemberTypeCannotBeParameterized(
          this, scope.environment().createRawType((ReferenceBinding)currentType.erasure(), actualEnclosing), argTypes);
        return null;
      }
    }

    ParameterizedTypeBinding parameterizedType = scope.environment().createParameterizedType((ReferenceBinding)currentType.erasure(), argTypes, enclosingType);

    if (checkBounds) {
      parameterizedType.boundCheck(scope, this.typeArguments);
    }
    this.resolvedType = parameterizedType;
    if (isTypeUseDeprecated(this.resolvedType, scope)) {
      reportDeprecatedType(this.resolvedType, scope);
    }

    if (this.dimensions > 0) {
      if (this.dimensions > 255)
        scope.problemReporter().tooManyDimensions(this);
      this.resolvedType = scope.createArrayType(this.resolvedType, this.dimensions);
    }
    return this.resolvedType;
  }

  public StringBuffer printExpression(int indent, StringBuffer output) {
    output.append(this.token);
    output.append("<");
    int max = this.typeArguments.length - 1;
    for (int i = 0; i < max; i++) {
      this.typeArguments[i].print(0, output);
      output.append(", ");
    }
    this.typeArguments[max].print(0, output);
    output.append(">");
    if ((this.bits & 0x4000) != 0) {
      for (int i = 0; i < this.dimensions - 1; i++) {
        output.append("[]");
      }
      output.append("...");
    } else {
      for (int i = 0; i < this.dimensions; i++) {
        output.append("[]");
      }
    }
    return output;
  }

  public TypeBinding resolveType(BlockScope scope, boolean checkBounds) {
    return internalResolveType(scope, null, checkBounds);
  }

  public TypeBinding resolveType(ClassScope scope) {
    return internalResolveType(scope, null, false);
  }

  public TypeBinding resolveTypeEnclosing(BlockScope scope, ReferenceBinding enclosingType) {
    return internalResolveType(scope, enclosingType, true);
  }

  public void traverse(ASTVisitor visitor, BlockScope scope) {
    if (visitor.visit(this, scope)) {
      int i = 0; for (int max = this.typeArguments.length; i < max; i++) {
        this.typeArguments[i].traverse(visitor, scope);
      }
    }
    visitor.endVisit(this, scope);
  }

  public void traverse(ASTVisitor visitor, ClassScope scope) {
    if (visitor.visit(this, scope)) {
      int i = 0; for (int max = this.typeArguments.length; i < max; i++) {
        this.typeArguments[i].traverse(visitor, scope);
      }
    }
    visitor.endVisit(this, scope);
  }
}