package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.NullInfoRegistry;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ProblemFieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class QualifiedNameReference extends NameReference
{
  public char[][] tokens;
  public long[] sourcePositions;
  public FieldBinding[] otherBindings;
  public FieldBinding[] otherCodegenBindings;
  int[] otherDepths;
  public int indexOfFirstFieldBinding;
  SyntheticMethodBinding syntheticWriteAccessor;
  SyntheticMethodBinding[] syntheticReadAccessors;
  public TypeBinding genericCast;
  public TypeBinding[] otherGenericCasts;

  public QualifiedNameReference(char[][] tokens, long[] positions, int sourceStart, int sourceEnd)
  {
    this.tokens = tokens;
    this.sourcePositions = positions;
    this.sourceStart = sourceStart;
    this.sourceEnd = sourceEnd;
  }

  public FlowInfo analyseAssignment(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo, Assignment assignment, boolean isCompound)
  {
    int otherBindingsCount = this.otherBindings == null ? 0 : this.otherBindings.length;
    boolean needValue = (otherBindingsCount == 0) || (!this.otherBindings[0].isStatic());
    boolean complyTo14 = currentScope.compilerOptions().complianceLevel >= 3145728L;
    FieldBinding lastFieldBinding = null;
    switch (this.bits & 0x7) {
    case 1:
      lastFieldBinding = (FieldBinding)this.binding;
      if ((needValue) || (complyTo14)) {
        manageSyntheticAccessIfNecessary(currentScope, lastFieldBinding, this.actualReceiverType, 0, flowInfo);
      }
      if (this.indexOfFirstFieldBinding == 1) {
        ReferenceBinding declaringClass = lastFieldBinding.declaringClass;

        if (declaringClass.isEnum()) {
          MethodScope methodScope = currentScope.methodScope();
          SourceTypeBinding sourceType = methodScope.enclosingSourceType();
          if ((lastFieldBinding.isStatic()) && 
            ((sourceType == declaringClass) || (sourceType.superclass == declaringClass)) && 
            (lastFieldBinding.constant() == Constant.NotAConstant) && 
            (!methodScope.isStatic) && 
            (methodScope.isInsideInitializerOrConstructor())) {
            currentScope.problemReporter().enumStaticFieldUsedDuringInitialization(lastFieldBinding, this);
          }
        }
      }

      if ((lastFieldBinding.isBlankFinal()) && 
        (this.otherBindings != null) && 
        (currentScope.allowBlankFinalFieldAssignment(lastFieldBinding)) && 
        (!flowInfo.isDefinitelyAssigned(lastFieldBinding))) {
        currentScope.problemReporter().uninitializedBlankFinalField(
          lastFieldBinding, 
          this);
      }

      break;
    case 2:
      LocalVariableBinding localBinding;
      if (!flowInfo
        .isDefinitelyAssigned(localBinding = (LocalVariableBinding)this.binding)) {
        currentScope.problemReporter().uninitializedLocalVariable(localBinding, this);
      }
      if ((flowInfo.tagBits & 0x1) == 0)
        localBinding.useFlag = 1;
      else if (localBinding.useFlag == 0) {
        localBinding.useFlag = 2;
      }
      checkNPE(currentScope, flowContext, flowInfo, true);
    }

    if (needValue) {
      manageEnclosingInstanceAccessIfNecessary(currentScope, flowInfo);
    }

    if (this.otherBindings != null) {
      for (int i = 0; i < otherBindingsCount - 1; i++) {
        lastFieldBinding = this.otherBindings[i];
        needValue = !this.otherBindings[(i + 1)].isStatic();
        if ((needValue) || (complyTo14)) {
          manageSyntheticAccessIfNecessary(
            currentScope, 
            lastFieldBinding, 
            i == 0 ? 
            ((VariableBinding)this.binding).type : 
            this.otherBindings[(i - 1)].type, 
            i + 1, 
            flowInfo);
        }
      }
      lastFieldBinding = this.otherBindings[(otherBindingsCount - 1)];
    }

    if (isCompound) {
      if ((otherBindingsCount == 0) && 
        (lastFieldBinding.isBlankFinal()) && 
        (currentScope.allowBlankFinalFieldAssignment(lastFieldBinding)) && 
        (!flowInfo.isDefinitelyAssigned(lastFieldBinding)))
        currentScope.problemReporter().uninitializedBlankFinalField(lastFieldBinding, this);
      TypeBinding lastReceiverType;
      TypeBinding lastReceiverType;
      TypeBinding lastReceiverType;
      switch (otherBindingsCount) {
      case 0:
        lastReceiverType = this.actualReceiverType;
        break;
      case 1:
        lastReceiverType = ((VariableBinding)this.binding).type;
        break;
      default:
        lastReceiverType = this.otherBindings[(otherBindingsCount - 2)].type;
      }

      manageSyntheticAccessIfNecessary(
        currentScope, 
        lastFieldBinding, 
        lastReceiverType, 
        otherBindingsCount, 
        flowInfo);
    }

    if (assignment.expression != null) {
      flowInfo = 
        assignment.expression
        .analyseCode(currentScope, flowContext, flowInfo)
        .unconditionalInits();
    }

    if (lastFieldBinding.isFinal())
    {
      if ((otherBindingsCount == 0) && 
        (this.indexOfFirstFieldBinding == 1) && 
        (lastFieldBinding.isBlankFinal()) && 
        (!isCompound) && 
        (currentScope.allowBlankFinalFieldAssignment(lastFieldBinding))) {
        if (flowInfo.isPotentiallyAssigned(lastFieldBinding))
          currentScope.problemReporter().duplicateInitializationOfBlankFinalField(lastFieldBinding, this);
        else {
          flowContext.recordSettingFinal(lastFieldBinding, this, flowInfo);
        }
        flowInfo.markAsDefinitelyAssigned(lastFieldBinding);
      } else {
        currentScope.problemReporter().cannotAssignToFinalField(lastFieldBinding, this);
        if ((otherBindingsCount == 0) && (currentScope.allowBlankFinalFieldAssignment(lastFieldBinding)))
          flowInfo.markAsDefinitelyAssigned(lastFieldBinding);
      }
    }
    TypeBinding lastReceiverType;
    TypeBinding lastReceiverType;
    TypeBinding lastReceiverType;
    switch (otherBindingsCount) {
    case 0:
      lastReceiverType = this.actualReceiverType;
      break;
    case 1:
      lastReceiverType = ((VariableBinding)this.binding).type;
      break;
    default:
      lastReceiverType = this.otherBindings[(otherBindingsCount - 2)].type;
    }

    manageSyntheticAccessIfNecessary(currentScope, lastFieldBinding, lastReceiverType, -1, flowInfo);

    return flowInfo;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo) {
    return analyseCode(currentScope, flowContext, flowInfo, true);
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo, boolean valueRequired)
  {
    int otherBindingsCount = this.otherBindings == null ? 0 : this.otherBindings.length;

    boolean needValue = this.otherBindings[0].isStatic() ? false : otherBindingsCount == 0 ? valueRequired : true;
    boolean complyTo14 = currentScope.compilerOptions().complianceLevel >= 3145728L;
    switch (this.bits & 0x7) {
    case 1:
      if ((needValue) || (complyTo14)) {
        manageSyntheticAccessIfNecessary(currentScope, (FieldBinding)this.binding, this.actualReceiverType, 0, flowInfo);
      }
      if (this.indexOfFirstFieldBinding == 1) {
        FieldBinding fieldBinding = (FieldBinding)this.binding;
        ReferenceBinding declaringClass = fieldBinding.declaringClass;

        if (declaringClass.isEnum()) {
          MethodScope methodScope = currentScope.methodScope();
          SourceTypeBinding sourceType = methodScope.enclosingSourceType();
          if ((fieldBinding.isStatic()) && 
            ((sourceType == declaringClass) || (sourceType.superclass == declaringClass)) && 
            (fieldBinding.constant() == Constant.NotAConstant) && 
            (!methodScope.isStatic) && 
            (methodScope.isInsideInitializerOrConstructor())) {
            currentScope.problemReporter().enumStaticFieldUsedDuringInitialization(fieldBinding, this);
          }
        }

        if ((fieldBinding.isBlankFinal()) && 
          (currentScope.allowBlankFinalFieldAssignment(fieldBinding)) && 
          (!flowInfo.isDefinitelyAssigned(fieldBinding))) {
          currentScope.problemReporter().uninitializedBlankFinalField(fieldBinding, this);
        }
      }
      break;
    case 2:
      LocalVariableBinding localBinding;
      if (!flowInfo
        .isDefinitelyAssigned(localBinding = (LocalVariableBinding)this.binding)) {
        currentScope.problemReporter().uninitializedLocalVariable(localBinding, this);
      }
      if ((flowInfo.tagBits & 0x1) == 0)
        localBinding.useFlag = 1;
      else if (localBinding.useFlag == 0) {
        localBinding.useFlag = 2;
      }
      checkNPE(currentScope, flowContext, flowInfo, true);
    }
    if (needValue) {
      manageEnclosingInstanceAccessIfNecessary(currentScope, flowInfo);
    }

    if (this.otherBindings != null) {
      for (int i = 0; i < otherBindingsCount; i++) {
        needValue = i < otherBindingsCount - 1 ? true : this.otherBindings[(i + 1)].isStatic() ? false : valueRequired;
        if ((needValue) || (complyTo14)) {
          TypeBinding lastReceiverType = getGenericCast(i);
          if (lastReceiverType == null) {
            if (i == 0)
              lastReceiverType = ((VariableBinding)this.binding).type;
            else {
              lastReceiverType = this.otherBindings[(i - 1)].type;
            }
          }
          manageSyntheticAccessIfNecessary(
            currentScope, 
            this.otherBindings[i], 
            lastReceiverType, 
            i + 1, 
            flowInfo);
        }
      }
    }
    return flowInfo;
  }

  public TypeBinding checkFieldAccess(BlockScope scope)
  {
    FieldBinding fieldBinding = (FieldBinding)this.binding;
    MethodScope methodScope = scope.methodScope();

    if ((this.indexOfFirstFieldBinding == 1) && 
      (methodScope.enclosingSourceType() == fieldBinding.original().declaringClass) && 
      (methodScope.lastVisibleFieldID >= 0) && 
      (fieldBinding.id >= methodScope.lastVisibleFieldID) && (
      (!fieldBinding.isStatic()) || (methodScope.isStatic))) {
      scope.problemReporter().forwardReference(this, 0, methodScope.enclosingSourceType());
    }
    this.bits &= -8;
    this.bits |= 1;
    return getOtherFieldBindings(scope);
  }

  public void checkNPE(BlockScope scope, FlowContext flowContext, FlowInfo flowInfo, boolean checkString)
  {
    if ((this.bits & 0x7) == 2) {
      LocalVariableBinding local = (LocalVariableBinding)this.binding;
      if ((local != null) && 
        ((local.type.tagBits & 0x2) == 0L) && (
        (checkString) || (local.type.id != 11))) {
        if ((this.bits & 0x20000) == 0) {
          flowContext.recordUsingNullReference(scope, local, this, 
            3, flowInfo);
        }
        flowInfo.markAsComparedEqualToNonNull(local);

        if (flowContext.initsOnFinally != null)
          flowContext.initsOnFinally.markAsComparedEqualToNonNull(local);
      }
    }
  }

  public void computeConversion(Scope scope, TypeBinding runtimeTimeType, TypeBinding compileTimeType)
  {
    if ((runtimeTimeType == null) || (compileTimeType == null)) {
      return;
    }
    FieldBinding field = null;
    int length = this.otherBindings == null ? 0 : this.otherBindings.length;
    if (length == 0) {
      if (((this.bits & 0x1) != 0) && (this.binding != null) && (this.binding.isValidBinding()))
        field = (FieldBinding)this.binding;
    }
    else {
      field = this.otherBindings[(length - 1)];
    }
    if (field != null) {
      FieldBinding originalBinding = field.original();
      TypeBinding originalType = originalBinding.type;

      if ((originalBinding != field) && 
        (originalType != field.type) && 
        (runtimeTimeType.id != 1) && 
        ((originalType.tagBits & 0x20000000) != 0L)) {
        TypeBinding targetType = (!compileTimeType.isBaseType()) && (runtimeTimeType.isBaseType()) ? 
          compileTimeType : 
          runtimeTimeType;
        setGenericCast(length, originalType.genericCast(targetType));
      }
    }
    super.computeConversion(scope, runtimeTimeType, compileTimeType);
  }

  public void generateAssignment(BlockScope currentScope, CodeStream codeStream, Assignment assignment, boolean valueRequired) {
    int pc = codeStream.position;
    FieldBinding lastFieldBinding = generateReadSequence(currentScope, codeStream);
    codeStream.recordPositionsFrom(pc, this.sourceStart);
    assignment.expression.generateCode(currentScope, codeStream, true);
    fieldStore(codeStream, lastFieldBinding, this.syntheticWriteAccessor, valueRequired);

    if (valueRequired)
      codeStream.generateImplicitConversion(assignment.implicitConversion);
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired)
  {
    int pc = codeStream.position;
    if (this.constant != Constant.NotAConstant) {
      if (valueRequired)
        codeStream.generateConstant(this.constant, this.implicitConversion);
    }
    else {
      FieldBinding lastFieldBinding = generateReadSequence(currentScope, codeStream);
      if (lastFieldBinding != null) {
        boolean isStatic = lastFieldBinding.isStatic();
        Constant fieldConstant = lastFieldBinding.constant();
        if (fieldConstant != Constant.NotAConstant) {
          if (!isStatic) {
            codeStream.invokeObjectGetClass();
            codeStream.pop();
          }
          if (valueRequired)
            codeStream.generateConstant(fieldConstant, this.implicitConversion);
        }
        else {
          boolean isFirst = (lastFieldBinding == this.binding) && 
            ((this.indexOfFirstFieldBinding == 1) || (lastFieldBinding.declaringClass == currentScope.enclosingReceiverType())) && 
            (this.otherBindings == null);
          TypeBinding requiredGenericCast = getGenericCast(this.otherCodegenBindings == null ? 0 : this.otherCodegenBindings.length);
          if ((valueRequired) || 
            ((!isFirst) && (currentScope.compilerOptions().complianceLevel >= 3145728L)) || 
            ((this.implicitConversion & 0x400) != 0) || 
            (requiredGenericCast != null)) {
            int lastFieldPc = codeStream.position;
            if (lastFieldBinding.declaringClass == null) {
              codeStream.arraylength();
              if (valueRequired) {
                codeStream.generateImplicitConversion(this.implicitConversion);
              }
              else
                codeStream.pop();
            }
            else {
              SyntheticMethodBinding accessor = 
                this.syntheticReadAccessors == null ? 
                null : 
                this.syntheticReadAccessors[(this.syntheticReadAccessors.length - 1)];
              if (accessor == null) {
                if (isStatic)
                  codeStream.getstatic(lastFieldBinding);
                else
                  codeStream.getfield(lastFieldBinding);
              }
              else {
                codeStream.invokestatic(accessor);
              }
              if (requiredGenericCast != null) codeStream.checkcast(requiredGenericCast);
              if (valueRequired) {
                codeStream.generateImplicitConversion(this.implicitConversion);
              } else {
                boolean isUnboxing = (this.implicitConversion & 0x400) != 0;

                if (isUnboxing) codeStream.generateImplicitConversion(this.implicitConversion);
                switch (isUnboxing ? postConversionType(currentScope).id : lastFieldBinding.type.id) {
                case 7:
                case 8:
                  codeStream.pop2();
                  break;
                default:
                  codeStream.pop();
                }
              }
            }

            int fieldPosition = (int)(this.sourcePositions[(this.sourcePositions.length - 1)] >>> 32);
            codeStream.recordPositionsFrom(lastFieldPc, fieldPosition);
          }
          else if (!isStatic) {
            codeStream.invokeObjectGetClass();
            codeStream.pop();
          }
        }
      }
    }

    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public void generateCompoundAssignment(BlockScope currentScope, CodeStream codeStream, Expression expression, int operator, int assignmentImplicitConversion, boolean valueRequired)
  {
    FieldBinding lastFieldBinding = generateReadSequence(currentScope, codeStream);
    SyntheticMethodBinding accessor = 
      this.syntheticReadAccessors == null ? 
      null : 
      this.syntheticReadAccessors[(this.syntheticReadAccessors.length - 1)];
    if (lastFieldBinding.isStatic()) {
      if (accessor == null)
        codeStream.getstatic(lastFieldBinding);
      else
        codeStream.invokestatic(accessor);
    }
    else {
      codeStream.dup();
      if (accessor == null)
        codeStream.getfield(lastFieldBinding);
      else
        codeStream.invokestatic(accessor);
    }
    int operationTypeID;
    switch (operationTypeID = (this.implicitConversion & 0xFF) >> 4) {
    case 0:
    case 1:
    case 11:
      codeStream.generateStringConcatenationAppend(currentScope, null, expression);
      break;
    default:
      TypeBinding requiredGenericCast = getGenericCast(this.otherCodegenBindings == null ? 0 : this.otherCodegenBindings.length);
      if (requiredGenericCast != null) codeStream.checkcast(requiredGenericCast);

      codeStream.generateImplicitConversion(this.implicitConversion);

      if (expression == IntLiteral.One)
        codeStream.generateConstant(expression.constant, this.implicitConversion);
      else {
        expression.generateCode(currentScope, codeStream, true);
      }

      codeStream.sendOperator(operator, operationTypeID);

      codeStream.generateImplicitConversion(assignmentImplicitConversion);
    }

    fieldStore(codeStream, lastFieldBinding, this.syntheticWriteAccessor, valueRequired);
  }

  public void generatePostIncrement(BlockScope currentScope, CodeStream codeStream, CompoundAssignment postIncrement, boolean valueRequired)
  {
    FieldBinding lastFieldBinding = generateReadSequence(currentScope, codeStream);
    SyntheticMethodBinding accessor = 
      this.syntheticReadAccessors == null ? 
      null : 
      this.syntheticReadAccessors[(this.syntheticReadAccessors.length - 1)];
    if (lastFieldBinding.isStatic()) {
      if (accessor == null)
        codeStream.getstatic(lastFieldBinding);
      else
        codeStream.invokestatic(accessor);
    }
    else {
      codeStream.dup();
      if (accessor == null)
        codeStream.getfield(lastFieldBinding);
      else {
        codeStream.invokestatic(accessor);
      }
    }

    if (valueRequired) {
      if (lastFieldBinding.isStatic()) {
        if ((lastFieldBinding.type == TypeBinding.LONG) || 
          (lastFieldBinding.type == TypeBinding.DOUBLE))
          codeStream.dup2();
        else {
          codeStream.dup();
        }
      }
      else if ((lastFieldBinding.type == TypeBinding.LONG) || 
        (lastFieldBinding.type == TypeBinding.DOUBLE))
        codeStream.dup2_x1();
      else {
        codeStream.dup_x1();
      }
    }

    TypeBinding requiredGenericCast = getGenericCast(this.otherCodegenBindings == null ? 0 : this.otherCodegenBindings.length);
    if (requiredGenericCast != null) codeStream.checkcast(requiredGenericCast);

    codeStream.generateImplicitConversion(this.implicitConversion);
    codeStream.generateConstant(
      postIncrement.expression.constant, 
      this.implicitConversion);
    codeStream.sendOperator(postIncrement.operator, this.implicitConversion & 0xF);
    codeStream.generateImplicitConversion(
      postIncrement.preAssignImplicitConversion);
    fieldStore(codeStream, lastFieldBinding, this.syntheticWriteAccessor, false);
  }

  public FieldBinding generateReadSequence(BlockScope currentScope, CodeStream codeStream)
  {
    int otherBindingsCount = this.otherCodegenBindings == null ? 0 : this.otherCodegenBindings.length;
    boolean needValue = (otherBindingsCount == 0) || (!this.otherBindings[0].isStatic());
    FieldBinding lastFieldBinding = null;
    TypeBinding lastGenericCast = null;
    boolean complyTo14 = currentScope.compilerOptions().complianceLevel >= 3145728L;

    switch (this.bits & 0x7) {
    case 1:
      lastFieldBinding = (FieldBinding)this.codegenBinding;
      lastGenericCast = this.genericCast;

      if (lastFieldBinding.constant() == Constant.NotAConstant)
      {
        if (((needValue) && (!lastFieldBinding.isStatic())) || (lastGenericCast != null)) {
          int pc = codeStream.position;
          if ((this.bits & 0x1FE0) != 0) {
            ReferenceBinding targetType = currentScope.enclosingSourceType().enclosingTypeAt((this.bits & 0x1FE0) >> 5);
            Object[] emulationPath = currentScope.getEmulationPath(targetType, true, false);
            codeStream.generateOuterAccess(emulationPath, this, targetType, currentScope);
          } else {
            generateReceiver(codeStream);
          }
          codeStream.recordPositionsFrom(pc, this.sourceStart);
        }
      }
      break;
    case 2:
      if (needValue) {
        LocalVariableBinding localBinding = (LocalVariableBinding)this.codegenBinding;

        Constant localConstant = localBinding.constant();
        if (localConstant != Constant.NotAConstant) {
          codeStream.generateConstant(localConstant, 0);
        }
        else if ((this.bits & 0x1FE0) != 0)
        {
          VariableBinding[] path = currentScope.getEmulationPath(localBinding);
          codeStream.generateOuterAccess(path, this, localBinding, currentScope);
        } else {
          codeStream.load(localBinding);
        }
      }

      break;
    }

    int positionsLength = this.sourcePositions.length;
    if (this.otherCodegenBindings != null) {
      for (int i = 0; i < otherBindingsCount; i++) {
        int pc = codeStream.position;
        FieldBinding nextField = this.otherCodegenBindings[i];
        TypeBinding nextGenericCast = this.otherGenericCasts == null ? null : this.otherGenericCasts[i];
        if (lastFieldBinding != null) {
          needValue = !nextField.isStatic();
          Constant fieldConstant = lastFieldBinding.constant();
          if (fieldConstant != Constant.NotAConstant) {
            if ((i > 0) && (!lastFieldBinding.isStatic())) {
              codeStream.invokeObjectGetClass();
              codeStream.pop();
            }
            if (needValue)
              codeStream.generateConstant(fieldConstant, 0);
          }
          else {
            if ((needValue) || ((i > 0) && (complyTo14)) || (lastGenericCast != null)) {
              MethodBinding accessor = this.syntheticReadAccessors == null ? null : this.syntheticReadAccessors[i];
              if (accessor == null) {
                if (lastFieldBinding.isStatic())
                  codeStream.getstatic(lastFieldBinding);
                else
                  codeStream.getfield(lastFieldBinding);
              }
              else {
                codeStream.invokestatic(accessor);
              }
              if (lastGenericCast != null) codeStream.checkcast(lastGenericCast);
              if (!needValue) codeStream.pop();
            }
            else if (this.codegenBinding == lastFieldBinding) {
              if (lastFieldBinding.isStatic())
              {
                if (((FieldBinding)this.binding).original().declaringClass != this.actualReceiverType.erasure()) {
                  MethodBinding accessor = this.syntheticReadAccessors == null ? null : this.syntheticReadAccessors[i];
                  if (accessor == null)
                    codeStream.getstatic(lastFieldBinding);
                  else {
                    codeStream.invokestatic(accessor);
                  }
                  codeStream.pop();
                }
              }
            } else if (!lastFieldBinding.isStatic()) {
              codeStream.invokeObjectGetClass();
              codeStream.pop();
            }

            if (positionsLength - otherBindingsCount + i - 1 >= 0) {
              int fieldPosition = (int)(this.sourcePositions[(positionsLength - otherBindingsCount + i - 1)] >>> 32);
              codeStream.recordPositionsFrom(pc, fieldPosition);
            }
          }
        }
        lastFieldBinding = nextField;
        lastGenericCast = nextGenericCast;
      }
    }
    return lastFieldBinding;
  }

  public void generateReceiver(CodeStream codeStream) {
    codeStream.aload_0();
  }

  public TypeBinding[] genericTypeArguments()
  {
    return null;
  }

  protected FieldBinding getCodegenBinding(int index)
  {
    if (index == 0) {
      return (FieldBinding)this.codegenBinding;
    }
    return this.otherCodegenBindings[(index - 1)];
  }

  protected TypeBinding getGenericCast(int index)
  {
    if (index == 0) {
      return this.genericCast;
    }
    if (this.otherGenericCasts == null) return null;
    return this.otherGenericCasts[(index - 1)];
  }

  public TypeBinding getOtherFieldBindings(BlockScope scope)
  {
    int length = this.tokens.length;
    FieldBinding field;
    if ((this.bits & 0x1) != 0) {
      FieldBinding field = (FieldBinding)this.binding;
      if (!field.isStatic())
      {
        if ((this.indexOfFirstFieldBinding > 1) || 
          (scope.methodScope().isStatic)) {
          scope.problemReporter().staticFieldAccessToNonStaticVariable(this, field);
          return null;
        }

      }
      else if ((this.indexOfFirstFieldBinding > 1) && 
        (field.declaringClass != this.actualReceiverType) && 
        (field.declaringClass.canBeSeenBy(scope))) {
        scope.problemReporter().indirectAccessToStaticField(this, field);
      }

      if (isFieldUseDeprecated(field, scope, ((this.bits & 0x2000) != 0) && (this.indexOfFirstFieldBinding == length)))
        scope.problemReporter().deprecatedField(field, this);
    } else {
      field = null;
    }
    TypeBinding type = ((VariableBinding)this.binding).type;
    int index = this.indexOfFirstFieldBinding;
    if (index == length) {
      this.constant = ((FieldBinding)this.binding).constant();

      return (type != null) && ((this.bits & 0x2000) == 0) ? 
        type.capture(scope, this.sourceEnd) : 
        type;
    }

    int otherBindingsLength = length - index;
    this.otherCodegenBindings = (this.otherBindings = new FieldBinding[otherBindingsLength]);
    this.otherDepths = new int[otherBindingsLength];

    this.constant = ((VariableBinding)this.binding).constant();

    int firstDepth = (this.bits & 0x1FE0) >> 5;

    while (index < length) {
      char[] token = this.tokens[index];
      if (type == null) {
        return null;
      }
      this.bits &= -8161;
      FieldBinding previousField = field;
      field = scope.getField(type.capture(scope, (int)this.sourcePositions[index]), token, this);
      int place = index - this.indexOfFirstFieldBinding;
      this.otherBindings[place] = field;
      this.otherDepths[place] = ((this.bits & 0x1FE0) >> 5);
      if (field.isValidBinding())
      {
        if (previousField != null) {
          TypeBinding fieldReceiverType = type;
          TypeBinding receiverErasure = type.erasure();
          if (((receiverErasure instanceof ReferenceBinding)) && 
            (receiverErasure.findSuperTypeWithSameErasure(field.declaringClass) == null)) {
            fieldReceiverType = field.declaringClass;
          }

          FieldBinding originalBinding = previousField.original();
          if (((originalBinding.type.tagBits & 0x20000000) != 0L) && (fieldReceiverType.id != 1)) {
            setGenericCast(index - 1, originalBinding.type.genericCast(fieldReceiverType));
          }
        }

        if (isFieldUseDeprecated(field, scope, ((this.bits & 0x2000) != 0) && (index + 1 == length))) {
          scope.problemReporter().deprecatedField(field, this);
        }

        if (this.constant != Constant.NotAConstant) {
          this.constant = field.constant();
        }

        if (field.isStatic())
        {
          scope.problemReporter().nonStaticAccessToStaticField(this, field);

          if (field.declaringClass != type) {
            scope.problemReporter().indirectAccessToStaticField(this, field);
          }
        }
        type = field.type;
        index++;
      } else {
        this.constant = Constant.NotAConstant;
        scope.problemReporter().invalidField(this, field, index, type);
        setDepth(firstDepth);
        return null;
      }
    }
    setDepth(firstDepth);
    type = this.otherBindings[(otherBindingsLength - 1)].type;

    return (type != null) && ((this.bits & 0x2000) == 0) ? 
      type.capture(scope, this.sourceEnd) : 
      type;
  }

  public void manageEnclosingInstanceAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo) {
    if ((flowInfo.tagBits & 0x1) == 0)
    {
      if (((this.bits & 0x1FE0) == 0) || (this.constant != Constant.NotAConstant)) {
        return;
      }
      if ((this.bits & 0x7) == 2)
        currentScope.emulateOuterAccess((LocalVariableBinding)this.binding);
    }
  }

  public void manageSyntheticAccessIfNecessary(BlockScope currentScope, FieldBinding fieldBinding, TypeBinding lastReceiverType, int index, FlowInfo flowInfo)
  {
    if ((flowInfo.tagBits & 0x1) != 0) return;

    if (fieldBinding.constant() != Constant.NotAConstant) {
      return;
    }

    FieldBinding originalField = fieldBinding.original();
    if (originalField != fieldBinding) {
      setCodegenBinding(index < 0 ? this.otherBindings.length : this.otherBindings == null ? 0 : index, originalField);
    }

    if (fieldBinding.isPrivate()) {
      FieldBinding someCodegenBinding = getCodegenBinding(index < 0 ? this.otherBindings.length : this.otherBindings == null ? 0 : index);
      if (someCodegenBinding.declaringClass != currentScope.enclosingSourceType()) {
        setSyntheticAccessor(fieldBinding, index, 
          ((SourceTypeBinding)someCodegenBinding.declaringClass).addSyntheticMethod(someCodegenBinding, index >= 0));
        currentScope.problemReporter().needToEmulateFieldAccess(someCodegenBinding, this, index >= 0);
      }
    }
    else if (fieldBinding.isProtected()) {
      int depth = (index == 0) || ((index < 0) && (this.otherDepths == null)) ? 
        (this.bits & 0x1FE0) >> 5 : 
        this.otherDepths[(index - 1)];

      if ((depth > 0) && (fieldBinding.declaringClass.getPackage() != currentScope.enclosingSourceType().getPackage())) {
        FieldBinding someCodegenBinding = getCodegenBinding(index < 0 ? this.otherBindings.length : this.otherBindings == null ? 0 : index);
        setSyntheticAccessor(fieldBinding, index, 
          ((SourceTypeBinding)currentScope.enclosingSourceType().enclosingTypeAt(depth)).addSyntheticMethod(someCodegenBinding, index >= 0));
        currentScope.problemReporter().needToEmulateFieldAccess(someCodegenBinding, this, index >= 0);
        return;
      }

    }

    if ((fieldBinding.declaringClass != lastReceiverType) && 
      (!lastReceiverType.isArrayType()) && 
      (fieldBinding.declaringClass != null) && 
      (fieldBinding.constant() == Constant.NotAConstant)) {
      CompilerOptions options = currentScope.compilerOptions();
      if (((options.targetJDK >= 3014656L) && 
        ((options.complianceLevel >= 3145728L) || (index > 1) || (this.indexOfFirstFieldBinding != 1) || (!fieldBinding.isStatic())) && 
        (fieldBinding.declaringClass.id != 1)) || 
        (!fieldBinding.declaringClass.canBeSeenBy(currentScope)))
      {
        setCodegenBinding(
          index < 0 ? this.otherBindings.length : this.otherBindings == null ? 0 : index, 
          currentScope.enclosingSourceType().getUpdatedFieldBinding(
          getCodegenBinding(index < 0 ? this.otherBindings.length : this.otherBindings == null ? 0 : index), 
          (ReferenceBinding)lastReceiverType.erasure()));
      }
    }
  }

  public int nullStatus(FlowInfo flowInfo) {
    return 0;
  }

  public Constant optimizedBooleanConstant() {
    switch (this.resolvedType.id) {
    case 5:
    case 33:
      if (this.constant != Constant.NotAConstant) return this.constant;
      switch (this.bits & 0x7) {
      case 1:
        if (this.otherBindings == null) {
          return ((FieldBinding)this.binding).constant();
        }
      case 2:
        return this.otherBindings[(this.otherBindings.length - 1)].constant();
      }
    }
    return Constant.NotAConstant;
  }

  public TypeBinding postConversionType(Scope scope)
  {
    TypeBinding convertedType = this.resolvedType;
    TypeBinding requiredGenericCast = getGenericCast(this.otherCodegenBindings == null ? 0 : this.otherCodegenBindings.length);
    if (requiredGenericCast != null)
      convertedType = requiredGenericCast;
    int runtimeType = (this.implicitConversion & 0xFF) >> 4;
    switch (runtimeType) {
    case 5:
      convertedType = TypeBinding.BOOLEAN;
      break;
    case 3:
      convertedType = TypeBinding.BYTE;
      break;
    case 4:
      convertedType = TypeBinding.SHORT;
      break;
    case 2:
      convertedType = TypeBinding.CHAR;
      break;
    case 10:
      convertedType = TypeBinding.INT;
      break;
    case 9:
      convertedType = TypeBinding.FLOAT;
      break;
    case 7:
      convertedType = TypeBinding.LONG;
      break;
    case 8:
      convertedType = TypeBinding.DOUBLE;
    case 6:
    }

    if ((this.implicitConversion & 0x200) != 0) {
      convertedType = scope.environment().computeBoxingType(convertedType);
    }
    return convertedType;
  }

  public StringBuffer printExpression(int indent, StringBuffer output) {
    for (int i = 0; i < this.tokens.length; i++) {
      if (i > 0) output.append('.');
      output.append(this.tokens[i]);
    }
    return output;
  }

  public TypeBinding reportError(BlockScope scope)
  {
    if ((this.binding instanceof ProblemFieldBinding))
      scope.problemReporter().invalidField(this, (FieldBinding)this.binding);
    else if ((this.binding instanceof ProblemReferenceBinding))
      scope.problemReporter().invalidType(this, (TypeBinding)this.binding);
    else {
      scope.problemReporter().unresolvableReference(this, this.binding);
    }
    return null;
  }

  public TypeBinding resolveType(BlockScope scope)
  {
    this.actualReceiverType = scope.enclosingReceiverType();
    this.constant = Constant.NotAConstant;
    if ((this.codegenBinding = this.binding = scope.getBinding(this.tokens, this.bits & 0x7, this, true)).isValidBinding())
      switch (this.bits & 0x7) {
      case 3:
      case 7:
        if ((this.binding instanceof LocalVariableBinding)) {
          if ((!((LocalVariableBinding)this.binding).isFinal()) && ((this.bits & 0x1FE0) != 0))
            scope.problemReporter().cannotReferToNonFinalOuterLocal(
              (LocalVariableBinding)this.binding, 
              this);
          this.bits &= -8;
          this.bits |= 2;
          return this.resolvedType = getOtherFieldBindings(scope);
        }
        if ((this.binding instanceof FieldBinding)) {
          FieldBinding fieldBinding = (FieldBinding)this.binding;
          MethodScope methodScope = scope.methodScope();

          if ((this.indexOfFirstFieldBinding == 1) && 
            (methodScope.enclosingSourceType() == fieldBinding.original().declaringClass) && 
            (methodScope.lastVisibleFieldID >= 0) && 
            (fieldBinding.id >= methodScope.lastVisibleFieldID) && (
            (!fieldBinding.isStatic()) || (methodScope.isStatic))) {
            scope.problemReporter().forwardReference(this, 0, methodScope.enclosingSourceType());
          }
          if ((!fieldBinding.isStatic()) && 
            (this.indexOfFirstFieldBinding == 1) && 
            (scope.compilerOptions().getSeverity(4194304L) != -1)) {
            scope.problemReporter().unqualifiedFieldAccess(this, fieldBinding);
          }
          this.bits &= -8;
          this.bits |= 1;

          return this.resolvedType = getOtherFieldBindings(scope);
        }

        this.bits &= -8;
        this.bits |= 4;
      case 4:
        TypeBinding type = (TypeBinding)this.binding;

        type = scope.environment().convertToRawType(type);
        return this.resolvedType = type;
      case 5:
      case 6:
      }
    return this.resolvedType = reportError(scope);
  }

  protected void setCodegenBinding(int index, FieldBinding someCodegenBinding)
  {
    if (index == 0) {
      this.codegenBinding = someCodegenBinding;
    } else {
      int length = this.otherBindings.length;
      if (this.otherCodegenBindings == this.otherBindings) {
        System.arraycopy(this.otherBindings, 0, this.otherCodegenBindings = new FieldBinding[length], 0, length);
      }
      this.otherCodegenBindings[(index - 1)] = someCodegenBinding;
    }
  }

  public void setFieldIndex(int index) {
    this.indexOfFirstFieldBinding = index;
  }

  protected void setGenericCast(int index, TypeBinding someGenericCast)
  {
    if (index == 0) {
      this.genericCast = someGenericCast;
    } else {
      if (this.otherGenericCasts == null) {
        this.otherGenericCasts = new TypeBinding[this.otherBindings.length];
      }
      this.otherGenericCasts[(index - 1)] = someGenericCast;
    }
  }

  protected void setSyntheticAccessor(FieldBinding fieldBinding, int index, SyntheticMethodBinding syntheticAccessor)
  {
    if (index < 0) {
      this.syntheticWriteAccessor = syntheticAccessor;
    } else {
      if (this.syntheticReadAccessors == null) {
        this.syntheticReadAccessors = new SyntheticMethodBinding[this.otherBindings == null ? 1 : this.otherBindings.length + 1];
      }
      this.syntheticReadAccessors[index] = syntheticAccessor;
    }
  }

  public void traverse(ASTVisitor visitor, BlockScope scope) {
    visitor.visit(this, scope);
    visitor.endVisit(this, scope);
  }

  public void traverse(ASTVisitor visitor, ClassScope scope) {
    visitor.visit(this, scope);
    visitor.endVisit(this, scope);
  }

  public String unboundReferenceErrorName() {
    return new String(this.tokens[0]);
  }
}