package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ProblemFieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class SingleNameReference extends NameReference
  implements OperatorIds
{
  public static final int READ = 0;
  public static final int WRITE = 1;
  public char[] token;
  public MethodBinding[] syntheticAccessors;
  public TypeBinding genericCast;

  public SingleNameReference(char[] source, long pos)
  {
    this.token = source;
    this.sourceStart = ((int)(pos >>> 32));
    this.sourceEnd = ((int)pos);
  }

  public FlowInfo analyseAssignment(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo, Assignment assignment, boolean isCompound) {
    boolean isReachable = (flowInfo.tagBits & 0x1) == 0;

    if (isCompound) {
      switch (this.bits & 0x7)
      {
      case 1:
        FieldBinding fieldBinding;
        if (((fieldBinding = (FieldBinding)this.binding).isBlankFinal()) && 
          (currentScope.allowBlankFinalFieldAssignment(fieldBinding)) && 
          (!flowInfo.isDefinitelyAssigned(fieldBinding))) {
          currentScope.problemReporter().uninitializedBlankFinalField(fieldBinding, this);
        }

        manageSyntheticAccessIfNecessary(currentScope, flowInfo, true);
        break;
      case 2:
        LocalVariableBinding localBinding;
        if (!flowInfo.isDefinitelyAssigned(localBinding = (LocalVariableBinding)this.binding)) {
          currentScope.problemReporter().uninitializedLocalVariable(localBinding, this);
        }

        if (isReachable)
          localBinding.useFlag = 1;
        else if (localBinding.useFlag == 0)
          localBinding.useFlag = 2;
        break;
      }
    }
    if (assignment.expression != null)
    {
      flowInfo = assignment.expression.analyseCode(currentScope, flowContext, flowInfo).unconditionalInits();
    }
    switch (this.bits & 0x7) {
    case 1:
      manageSyntheticAccessIfNecessary(currentScope, flowInfo, false);

      FieldBinding fieldBinding = (FieldBinding)this.binding;
      ReferenceBinding declaringClass = fieldBinding.declaringClass;

      if (declaringClass.isEnum()) {
        MethodScope methodScope = currentScope.methodScope();
        SourceTypeBinding sourceType = currentScope.enclosingSourceType();
        if ((fieldBinding.isStatic()) && 
          (this.constant == Constant.NotAConstant) && 
          (!methodScope.isStatic) && 
          ((sourceType == declaringClass) || (sourceType.superclass == declaringClass)) && 
          (methodScope.isInsideInitializerOrConstructor())) {
          currentScope.problemReporter().enumStaticFieldUsedDuringInitialization(fieldBinding, this);
        }
      }

      if (fieldBinding.isFinal())
      {
        if ((!isCompound) && (fieldBinding.isBlankFinal()) && (currentScope.allowBlankFinalFieldAssignment(fieldBinding))) {
          if (flowInfo.isPotentiallyAssigned(fieldBinding))
            currentScope.problemReporter().duplicateInitializationOfBlankFinalField(fieldBinding, this);
          else {
            flowContext.recordSettingFinal(fieldBinding, this, flowInfo);
          }
          flowInfo.markAsDefinitelyAssigned(fieldBinding);
        } else {
          currentScope.problemReporter().cannotAssignToFinalField(fieldBinding, this);
        }
      }
      break;
    case 2:
      LocalVariableBinding localBinding = (LocalVariableBinding)this.binding;
      if (!flowInfo.isDefinitelyAssigned(localBinding))
        this.bits |= 8;
      else {
        this.bits &= -9;
      }
      if (localBinding.isFinal()) {
        if ((this.bits & 0x1FE0) == 0)
        {
          if (((isReachable) && (isCompound)) || (!localBinding.isBlankFinal()))
            currentScope.problemReporter().cannotAssignToFinalLocal(localBinding, this);
          else if (flowInfo.isPotentiallyAssigned(localBinding))
            currentScope.problemReporter().duplicateInitializationOfFinalLocal(localBinding, this);
          else
            flowContext.recordSettingFinal(localBinding, this, flowInfo);
        }
        else {
          currentScope.problemReporter().cannotAssignToFinalOuterLocal(localBinding, this);
        }
      }
      else if ((localBinding.tagBits & 0x400) != 0L) {
        currentScope.problemReporter().parameterAssignment(localBinding, this);
      }
      flowInfo.markAsDefinitelyAssigned(localBinding);
    }
    manageEnclosingInstanceAccessIfNecessary(currentScope, flowInfo);
    return flowInfo;
  }
  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo) {
    return analyseCode(currentScope, flowContext, flowInfo, true);
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo, boolean valueRequired) {
    switch (this.bits & 0x7) {
    case 1:
      if ((valueRequired) || (currentScope.compilerOptions().complianceLevel >= 3145728L)) {
        manageSyntheticAccessIfNecessary(currentScope, flowInfo, true);
      }
      FieldBinding fieldBinding = (FieldBinding)this.binding;
      ReferenceBinding declaringClass = fieldBinding.declaringClass;

      if (declaringClass.isEnum()) {
        MethodScope methodScope = currentScope.methodScope();
        SourceTypeBinding sourceType = currentScope.enclosingSourceType();
        if ((fieldBinding.isStatic()) && 
          (this.constant == Constant.NotAConstant) && 
          (!methodScope.isStatic) && 
          ((sourceType == declaringClass) || (sourceType.superclass == declaringClass)) && 
          (methodScope.isInsideInitializerOrConstructor())) {
          currentScope.problemReporter().enumStaticFieldUsedDuringInitialization(fieldBinding, this);
        }
      }

      if ((fieldBinding.isBlankFinal()) && (currentScope.allowBlankFinalFieldAssignment(fieldBinding)) && 
        (!flowInfo.isDefinitelyAssigned(fieldBinding))) {
        currentScope.problemReporter().uninitializedBlankFinalField(fieldBinding, this);
      }

      break;
    case 2:
      LocalVariableBinding localBinding;
      if (!flowInfo.isDefinitelyAssigned(localBinding = (LocalVariableBinding)this.binding)) {
        currentScope.problemReporter().uninitializedLocalVariable(localBinding, this);
      }
      if ((flowInfo.tagBits & 0x1) == 0)
        localBinding.useFlag = 1;
      else if (localBinding.useFlag == 0)
        localBinding.useFlag = 2;
      break;
    }
    if (valueRequired) {
      manageEnclosingInstanceAccessIfNecessary(currentScope, flowInfo);
    }
    return flowInfo;
  }

  public TypeBinding checkFieldAccess(BlockScope scope)
  {
    FieldBinding fieldBinding = (FieldBinding)this.binding;

    this.bits &= -8;
    this.bits |= 1;
    MethodScope methodScope = scope.methodScope();
    boolean isStatic = fieldBinding.isStatic();
    if (!isStatic)
    {
      if (methodScope.isStatic) {
        scope.problemReporter().staticFieldAccessToNonStaticVariable(this, fieldBinding);
        this.constant = Constant.NotAConstant;
        return fieldBinding.type;
      }
    }
    this.constant = fieldBinding.constant();

    if (isFieldUseDeprecated(fieldBinding, scope, (this.bits & 0x2000) != 0)) {
      scope.problemReporter().deprecatedField(fieldBinding, this);
    }
    if (((this.bits & 0x2000) == 0) && 
      (methodScope.enclosingSourceType() == fieldBinding.original().declaringClass) && 
      (methodScope.lastVisibleFieldID >= 0) && 
      (fieldBinding.id >= methodScope.lastVisibleFieldID) && (
      (!fieldBinding.isStatic()) || (methodScope.isStatic))) {
      scope.problemReporter().forwardReference(this, 0, methodScope.enclosingSourceType());
      this.bits |= 536870912;
    }
    return fieldBinding.type;
  }

  public void computeConversion(Scope scope, TypeBinding runtimeTimeType, TypeBinding compileTimeType)
  {
    if ((runtimeTimeType == null) || (compileTimeType == null))
      return;
    if (((this.bits & 0x1) != 0) && (this.binding != null) && (this.binding.isValidBinding()))
    {
      FieldBinding field = (FieldBinding)this.binding;
      FieldBinding originalBinding = field.original();
      TypeBinding originalType = originalBinding.type;

      if ((originalBinding != field) && 
        (originalType != field.type) && 
        (runtimeTimeType.id != 1) && 
        ((originalType.tagBits & 0x20000000) != 0L)) {
        TypeBinding targetType = (!compileTimeType.isBaseType()) && (runtimeTimeType.isBaseType()) ? 
          compileTimeType : 
          runtimeTimeType;
        this.genericCast = originalType.genericCast(scope.boxing(targetType));
      }
    }
    super.computeConversion(scope, runtimeTimeType, compileTimeType);
  }

  public void generateAssignment(BlockScope currentScope, CodeStream codeStream, Assignment assignment, boolean valueRequired)
  {
    if (assignment.expression.isCompactableOperation()) {
      BinaryExpression operation = (BinaryExpression)assignment.expression;
      int operator = (operation.bits & 0xFC0) >> 6;
      SingleNameReference variableReference;
      if (((operation.left instanceof SingleNameReference)) && ((variableReference = (SingleNameReference)operation.left).binding == this.binding))
      {
        variableReference.generateCompoundAssignment(currentScope, codeStream, this.syntheticAccessors == null ? null : this.syntheticAccessors[1], operation.right, operator, operation.implicitConversion, valueRequired);
        if (valueRequired)
          codeStream.generateImplicitConversion(assignment.implicitConversion);
        return;
      }
      SingleNameReference variableReference;
      if (((operation.right instanceof SingleNameReference)) && 
        ((operator == 14) || (operator == 15)) && 
        ((variableReference = (SingleNameReference)operation.right).binding == this.binding) && 
        (operation.left.constant != Constant.NotAConstant) && 
        ((operation.left.implicitConversion & 0xFF) >> 4 != 11) && 
        ((operation.right.implicitConversion & 0xFF) >> 4 != 11))
      {
        variableReference.generateCompoundAssignment(currentScope, codeStream, this.syntheticAccessors == null ? null : this.syntheticAccessors[1], operation.left, operator, operation.implicitConversion, valueRequired);
        if (valueRequired) {
          codeStream.generateImplicitConversion(assignment.implicitConversion);
        }
        return;
      }
    }
    switch (this.bits & 0x7)
    {
    case 1:
      int pc = codeStream.position;
      FieldBinding fieldBinding;
      if (!(fieldBinding = (FieldBinding)this.codegenBinding).isStatic()) {
        if ((this.bits & 0x1FE0) != 0) {
          ReferenceBinding targetType = currentScope.enclosingSourceType().enclosingTypeAt((this.bits & 0x1FE0) >> 5);
          Object[] emulationPath = currentScope.getEmulationPath(targetType, true, false);
          codeStream.generateOuterAccess(emulationPath, this, targetType, currentScope);
        } else {
          generateReceiver(codeStream);
        }
      }
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      assignment.expression.generateCode(currentScope, codeStream, true);
      fieldStore(codeStream, fieldBinding, this.syntheticAccessors == null ? null : this.syntheticAccessors[1], valueRequired);
      if (valueRequired) {
        codeStream.generateImplicitConversion(assignment.implicitConversion);
      }

      return;
    case 2:
      LocalVariableBinding localBinding = (LocalVariableBinding)this.codegenBinding;
      if (localBinding.resolvedPosition != -1) {
        assignment.expression.generateCode(currentScope, codeStream, true);
      } else {
        if (assignment.expression.constant != Constant.NotAConstant)
        {
          if (valueRequired)
            codeStream.generateConstant(assignment.expression.constant, assignment.implicitConversion);
        }
        else {
          assignment.expression.generateCode(currentScope, codeStream, true);

          if (valueRequired) {
            codeStream.generateImplicitConversion(assignment.implicitConversion);
          }
          else if ((localBinding.type == TypeBinding.LONG) || (localBinding.type == TypeBinding.DOUBLE))
            codeStream.pop2();
          else {
            codeStream.pop();
          }
        }

        return;
      }

      if ((localBinding.type.isArrayType()) && (
        (assignment.expression.resolvedType == TypeBinding.NULL) || (
        ((assignment.expression instanceof CastExpression)) && 
        (((CastExpression)assignment.expression).innermostCastedExpression().resolvedType == TypeBinding.NULL)))) {
        codeStream.checkcast(localBinding.type);
      }

      codeStream.store(localBinding, valueRequired);
      if ((this.bits & 0x8) != 0) {
        localBinding.recordInitializationStartPC(codeStream.position);
      }

      if (valueRequired)
        codeStream.generateImplicitConversion(assignment.implicitConversion); break;
    }
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired) {
    int pc = codeStream.position;
    if (this.constant != Constant.NotAConstant) {
      if (valueRequired) {
        codeStream.generateConstant(this.constant, this.implicitConversion);
      }
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      return;
    }
    switch (this.bits & 0x7) {
    case 1:
      FieldBinding fieldBinding = (FieldBinding)this.codegenBinding;
      Constant fieldConstant = fieldBinding.constant();
      if (fieldConstant != Constant.NotAConstant)
      {
        if (valueRequired) {
          codeStream.generateConstant(fieldConstant, this.implicitConversion);
        }
        codeStream.recordPositionsFrom(pc, this.sourceStart);
        return;
      }
      if (fieldBinding.isStatic()) {
        if (!valueRequired)
        {
          if ((((FieldBinding)this.binding).original().declaringClass == this.actualReceiverType.erasure()) && 
            ((this.implicitConversion & 0x400) == 0) && 
            (this.genericCast == null))
          {
            codeStream.recordPositionsFrom(pc, this.sourceStart);
            return;
          }
        }
        if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
          codeStream.getstatic(fieldBinding);
        else
          codeStream.invokestatic(this.syntheticAccessors[0]);
      }
      else {
        if ((!valueRequired) && 
          ((this.implicitConversion & 0x400) == 0) && 
          (this.genericCast == null))
        {
          codeStream.recordPositionsFrom(pc, this.sourceStart);
          return;
        }

        if ((this.bits & 0x1FE0) != 0) {
          ReferenceBinding targetType = currentScope.enclosingSourceType().enclosingTypeAt((this.bits & 0x1FE0) >> 5);
          Object[] emulationPath = currentScope.getEmulationPath(targetType, true, false);
          codeStream.generateOuterAccess(emulationPath, this, targetType, currentScope);
        } else {
          generateReceiver(codeStream);
        }

        if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
          codeStream.getfield(fieldBinding);
        else {
          codeStream.invokestatic(this.syntheticAccessors[0]);
        }
      }
      break;
    case 2:
      LocalVariableBinding localBinding = (LocalVariableBinding)this.codegenBinding;
      if ((!valueRequired) && ((this.implicitConversion & 0x400) == 0))
      {
        codeStream.recordPositionsFrom(pc, this.sourceStart);
        return;
      }

      if ((this.bits & 0x1FE0) != 0)
      {
        VariableBinding[] path = currentScope.getEmulationPath(localBinding);
        codeStream.generateOuterAccess(path, this, localBinding, currentScope);
      }
      else {
        codeStream.load(localBinding);
      }
      break;
    default:
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      return;
    }

    if (this.genericCast != null) codeStream.checkcast(this.genericCast);
    if (valueRequired) {
      codeStream.generateImplicitConversion(this.implicitConversion);
    } else {
      boolean isUnboxing = (this.implicitConversion & 0x400) != 0;

      if (isUnboxing) codeStream.generateImplicitConversion(this.implicitConversion);
      switch (isUnboxing ? postConversionType(currentScope).id : this.resolvedType.id) {
      case 7:
      case 8:
        codeStream.pop2();
        break;
      default:
        codeStream.pop();
      }
    }
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public void generateCompoundAssignment(BlockScope currentScope, CodeStream codeStream, Expression expression, int operator, int assignmentImplicitConversion, boolean valueRequired)
  {
    generateCompoundAssignment(
      currentScope, 
      codeStream, 
      this.syntheticAccessors == null ? null : this.syntheticAccessors[1], 
      expression, 
      operator, 
      assignmentImplicitConversion, 
      valueRequired);
  }

  public void generateCompoundAssignment(BlockScope currentScope, CodeStream codeStream, MethodBinding writeAccessor, Expression expression, int operator, int assignmentImplicitConversion, boolean valueRequired)
  {
    switch (this.bits & 0x7)
    {
    case 1:
      FieldBinding fieldBinding;
      if ((fieldBinding = (FieldBinding)this.codegenBinding).isStatic()) {
        if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
          codeStream.getstatic(fieldBinding);
        else
          codeStream.invokestatic(this.syntheticAccessors[0]);
      }
      else {
        if ((this.bits & 0x1FE0) != 0) {
          ReferenceBinding targetType = currentScope.enclosingSourceType().enclosingTypeAt((this.bits & 0x1FE0) >> 5);
          Object[] emulationPath = currentScope.getEmulationPath(targetType, true, false);
          codeStream.generateOuterAccess(emulationPath, this, targetType, currentScope);
        } else {
          codeStream.aload_0();
        }
        codeStream.dup();
        if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
          codeStream.getfield(fieldBinding);
        else {
          codeStream.invokestatic(this.syntheticAccessors[0]);
        }
      }
      break;
    case 2:
      LocalVariableBinding localBinding = (LocalVariableBinding)this.codegenBinding;

      switch (localBinding.type.id) {
      case 11:
        codeStream.generateStringConcatenationAppend(currentScope, this, expression);
        if (valueRequired) {
          codeStream.dup();
        }
        codeStream.store(localBinding, false);
        return;
      case 10:
        Constant assignConstant;
        int increment;
        if (((assignConstant = expression.constant) != Constant.NotAConstant) && 
          (assignConstant.typeID() != 9) && 
          (assignConstant.typeID() != 8) && 
          ((increment = assignConstant.intValue()) == (short)increment)) {
          switch (operator) {
          case 14:
            codeStream.iinc(localBinding.resolvedPosition, increment);
            if (valueRequired) {
              codeStream.load(localBinding);
            }
            return;
          case 13:
            codeStream.iinc(localBinding.resolvedPosition, -increment);
            if (valueRequired)
              codeStream.load(localBinding);
            return;
          }
        }
        break;
      }
      codeStream.load(localBinding);
    }
    int operationTypeID;
    switch (operationTypeID = (this.implicitConversion & 0xFF) >> 4)
    {
    case 0:
    case 1:
    case 11:
      codeStream.generateStringConcatenationAppend(currentScope, null, expression);

      break;
    default:
      if (this.genericCast != null)
        codeStream.checkcast(this.genericCast);
      codeStream.generateImplicitConversion(this.implicitConversion);

      if (expression == IntLiteral.One)
        codeStream.generateConstant(expression.constant, this.implicitConversion);
      else {
        expression.generateCode(currentScope, codeStream, true);
      }

      codeStream.sendOperator(operator, operationTypeID);

      codeStream.generateImplicitConversion(assignmentImplicitConversion);
    }

    switch (this.bits & 0x7) {
    case 1:
      fieldStore(codeStream, (FieldBinding)this.codegenBinding, writeAccessor, valueRequired);

      return;
    case 2:
      LocalVariableBinding localBinding = (LocalVariableBinding)this.codegenBinding;
      if (valueRequired) {
        if ((localBinding.type == TypeBinding.LONG) || (localBinding.type == TypeBinding.DOUBLE))
          codeStream.dup2();
        else {
          codeStream.dup();
        }
      }
      codeStream.store(localBinding, false);
    }
  }

  public void generatePostIncrement(BlockScope currentScope, CodeStream codeStream, CompoundAssignment postIncrement, boolean valueRequired) {
    switch (this.bits & 0x7)
    {
    case 1:
      FieldBinding fieldBinding;
      if ((fieldBinding = (FieldBinding)this.codegenBinding).isStatic()) {
        if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
          codeStream.getstatic(fieldBinding);
        else
          codeStream.invokestatic(this.syntheticAccessors[0]);
      }
      else {
        if ((this.bits & 0x1FE0) != 0) {
          ReferenceBinding targetType = currentScope.enclosingSourceType().enclosingTypeAt((this.bits & 0x1FE0) >> 5);
          Object[] emulationPath = currentScope.getEmulationPath(targetType, true, false);
          codeStream.generateOuterAccess(emulationPath, this, targetType, currentScope);
        } else {
          codeStream.aload_0();
        }
        codeStream.dup();
        if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
          codeStream.getfield(fieldBinding);
        else {
          codeStream.invokestatic(this.syntheticAccessors[0]);
        }
      }
      if (valueRequired) {
        if (fieldBinding.isStatic()) {
          if ((fieldBinding.type == TypeBinding.LONG) || (fieldBinding.type == TypeBinding.DOUBLE))
            codeStream.dup2();
          else {
            codeStream.dup();
          }
        }
        else if ((fieldBinding.type == TypeBinding.LONG) || (fieldBinding.type == TypeBinding.DOUBLE))
          codeStream.dup2_x1();
        else {
          codeStream.dup_x1();
        }
      }

      if (this.genericCast != null)
        codeStream.checkcast(this.genericCast);
      codeStream.generateImplicitConversion(this.implicitConversion);
      codeStream.generateConstant(postIncrement.expression.constant, this.implicitConversion);
      codeStream.sendOperator(postIncrement.operator, this.implicitConversion & 0xF);
      codeStream.generateImplicitConversion(postIncrement.preAssignImplicitConversion);
      fieldStore(codeStream, fieldBinding, this.syntheticAccessors == null ? null : this.syntheticAccessors[1], false);

      return;
    case 2:
      LocalVariableBinding localBinding = (LocalVariableBinding)this.codegenBinding;

      if (localBinding.type == TypeBinding.INT) {
        if (valueRequired) {
          codeStream.load(localBinding);
        }
        if (postIncrement.operator == 14)
          codeStream.iinc(localBinding.resolvedPosition, 1);
        else
          codeStream.iinc(localBinding.resolvedPosition, -1);
      }
      else {
        codeStream.load(localBinding);
        if (valueRequired) {
          if ((localBinding.type == TypeBinding.LONG) || (localBinding.type == TypeBinding.DOUBLE))
            codeStream.dup2();
          else {
            codeStream.dup();
          }
        }
        codeStream.generateImplicitConversion(this.implicitConversion);
        codeStream.generateConstant(postIncrement.expression.constant, this.implicitConversion);
        codeStream.sendOperator(postIncrement.operator, this.implicitConversion & 0xF);
        codeStream.generateImplicitConversion(postIncrement.preAssignImplicitConversion);

        codeStream.store(localBinding, false);
      }break;
    }
  }

  public void generateReceiver(CodeStream codeStream) {
    codeStream.aload_0();
  }

  public TypeBinding[] genericTypeArguments()
  {
    return null;
  }

  public LocalVariableBinding localVariableBinding()
  {
    switch (this.bits & 0x7) {
    case 1:
      break;
    case 2:
      return (LocalVariableBinding)this.binding;
    }
    return null;
  }

  public void manageEnclosingInstanceAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo)
  {
    if ((flowInfo.tagBits & 0x1) == 0)
    {
      if (((this.bits & 0x1FE0) == 0) || (this.constant != Constant.NotAConstant)) return;

      if ((this.bits & 0x7) == 2)
        currentScope.emulateOuterAccess((LocalVariableBinding)this.binding);
    }
  }

  public void manageSyntheticAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo, boolean isReadAccess)
  {
    if ((flowInfo.tagBits & 0x1) != 0) return;

    if (this.constant != Constant.NotAConstant) {
      return;
    }
    if ((this.bits & 0x1) != 0) {
      FieldBinding fieldBinding = (FieldBinding)this.binding;
      FieldBinding codegenField = fieldBinding.original();
      this.codegenBinding = codegenField;
      if (((this.bits & 0x1FE0) != 0) && (
        (codegenField.isPrivate()) || (
        (codegenField.isProtected()) && 
        (codegenField.declaringClass.getPackage() != currentScope.enclosingSourceType().getPackage())))) {
        if (this.syntheticAccessors == null)
          this.syntheticAccessors = new MethodBinding[2];
        this.syntheticAccessors[(isReadAccess ? 0 : 1)] = 
          ((SourceTypeBinding)currentScope.enclosingSourceType()
          .enclosingTypeAt((this.bits & 0x1FE0) >> 5)).addSyntheticMethod(codegenField, isReadAccess);
        currentScope.problemReporter().needToEmulateFieldAccess(codegenField, this, isReadAccess);
        return;
      }

      if ((fieldBinding.declaringClass != this.actualReceiverType) && 
        (!this.actualReceiverType.isArrayType()) && 
        (fieldBinding.declaringClass != null) && 
        (fieldBinding.constant() == Constant.NotAConstant)) {
        CompilerOptions options = currentScope.compilerOptions();
        if (((options.targetJDK >= 3014656L) && 
          ((options.complianceLevel >= 3145728L) || (!fieldBinding.isStatic())) && 
          (fieldBinding.declaringClass.id != 1)) || 
          (!fieldBinding.declaringClass.canBeSeenBy(currentScope)))
        {
          this.codegenBinding = 
            currentScope.enclosingSourceType().getUpdatedFieldBinding(
            codegenField, 
            (ReferenceBinding)this.actualReceiverType.erasure());
        }
      }
    }
  }

  public int nullStatus(FlowInfo flowInfo) {
    if ((this.constant != null) && (this.constant != Constant.NotAConstant)) {
      return -1;
    }
    switch (this.bits & 0x7) {
    case 1:
      return 0;
    case 2:
      LocalVariableBinding local = (LocalVariableBinding)this.binding;
      if (local != null) {
        if (flowInfo.isDefinitelyNull(local))
          return 1;
        if (flowInfo.isDefinitelyNonNull(local))
          return -1;
        return 0;
      }break;
    }
    return -1;
  }

  public TypeBinding postConversionType(Scope scope)
  {
    TypeBinding convertedType = this.resolvedType;
    if (this.genericCast != null)
      convertedType = this.genericCast;
    int runtimeType = (this.implicitConversion & 0xFF) >> 4;
    switch (runtimeType) {
    case 5:
      convertedType = TypeBinding.BOOLEAN;
      break;
    case 3:
      convertedType = TypeBinding.BYTE;
      break;
    case 4:
      convertedType = TypeBinding.SHORT;
      break;
    case 2:
      convertedType = TypeBinding.CHAR;
      break;
    case 10:
      convertedType = TypeBinding.INT;
      break;
    case 9:
      convertedType = TypeBinding.FLOAT;
      break;
    case 7:
      convertedType = TypeBinding.LONG;
      break;
    case 8:
      convertedType = TypeBinding.DOUBLE;
    case 6:
    }

    if ((this.implicitConversion & 0x200) != 0) {
      convertedType = scope.environment().computeBoxingType(convertedType);
    }
    return convertedType;
  }

  public StringBuffer printExpression(int indent, StringBuffer output)
  {
    return output.append(this.token);
  }

  public TypeBinding reportError(BlockScope scope)
  {
    this.constant = Constant.NotAConstant;
    if ((this.binding instanceof ProblemFieldBinding))
      scope.problemReporter().invalidField(this, (FieldBinding)this.binding);
    else if ((this.binding instanceof ProblemReferenceBinding))
      scope.problemReporter().invalidType(this, (TypeBinding)this.binding);
    else {
      scope.problemReporter().unresolvableReference(this, this.binding);
    }
    return null;
  }

  public TypeBinding resolveType(BlockScope scope)
  {
    if (this.actualReceiverType != null) {
      this.binding = scope.getField(this.actualReceiverType, this.token, this);
    } else {
      this.actualReceiverType = scope.enclosingSourceType();
      this.binding = scope.getBinding(this.token, this.bits & 0x7, this, true);
    }
    this.codegenBinding = this.binding;
    if (this.binding.isValidBinding()) {
      switch (this.bits & 0x7) {
      case 3:
      case 7:
        if ((this.binding instanceof VariableBinding)) {
          VariableBinding variable = (VariableBinding)this.binding;
          if ((this.binding instanceof LocalVariableBinding)) {
            this.bits &= -8;
            this.bits |= 2;
            if ((!variable.isFinal()) && ((this.bits & 0x1FE0) != 0)) {
              scope.problemReporter().cannotReferToNonFinalOuterLocal((LocalVariableBinding)variable, this);
            }
            TypeBinding fieldType = variable.type;
            if ((this.bits & 0x2000) == 0) {
              this.constant = variable.constant();
              if (fieldType != null)
                fieldType = fieldType.capture(scope, this.sourceEnd);
            } else {
              this.constant = Constant.NotAConstant;
            }
            return this.resolvedType = fieldType;
          }

          FieldBinding field = (FieldBinding)this.binding;
          if ((!field.isStatic()) && (scope.compilerOptions().getSeverity(4194304L) != -1)) {
            scope.problemReporter().unqualifiedFieldAccess(this, field);
          }

          TypeBinding fieldType = checkFieldAccess(scope);
          return this.resolvedType = 
            (this.bits & 0x2000) == 0 ? 
            fieldType.capture(scope, this.sourceEnd) : 
            fieldType;
        }

        this.bits &= -8;
        this.bits |= 4;
      case 4:
        this.constant = Constant.NotAConstant;

        TypeBinding type = (TypeBinding)this.binding;
        if (isTypeUseDeprecated(type, scope))
          scope.problemReporter().deprecatedType(type, this);
        type = scope.environment().convertToRawType(type);
        return this.resolvedType = type;
      case 5:
      case 6:
      }
    }
    return this.resolvedType = reportError(scope);
  }

  public void traverse(ASTVisitor visitor, BlockScope scope) {
    visitor.visit(this, scope);
    visitor.endVisit(this, scope);
  }

  public void traverse(ASTVisitor visitor, ClassScope scope) {
    visitor.visit(this, scope);
    visitor.endVisit(this, scope);
  }

  public String unboundReferenceErrorName()
  {
    return new String(this.token);
  }
}