package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.core.compiler.CategorizedProblem;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.InitializationFlowContext;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.MemberTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.NestedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticArgumentBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilationUnit;
import org.eclipse.jdt.internal.compiler.problem.AbortMethod;
import org.eclipse.jdt.internal.compiler.problem.AbortType;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.problem.ProblemSeverities;

public class TypeDeclaration extends Statement
  implements ProblemSeverities, ReferenceContext
{
  public static final int CLASS_DECL = 1;
  public static final int INTERFACE_DECL = 2;
  public static final int ENUM_DECL = 3;
  public static final int ANNOTATION_TYPE_DECL = 4;
  public int modifiers = 0;
  public int modifiersSourceStart;
  public Annotation[] annotations;
  public char[] name;
  public TypeReference superclass;
  public TypeReference[] superInterfaces;
  public FieldDeclaration[] fields;
  public AbstractMethodDeclaration[] methods;
  public TypeDeclaration[] memberTypes;
  public SourceTypeBinding binding;
  public ClassScope scope;
  public MethodScope initializerScope;
  public MethodScope staticInitializerScope;
  public boolean ignoreFurtherInvestigation = false;
  public int maxFieldCount;
  public int declarationSourceStart;
  public int declarationSourceEnd;
  public int bodyStart;
  public int bodyEnd;
  public CompilationResult compilationResult;
  public MethodDeclaration[] missingAbstractMethods;
  public Javadoc javadoc;
  public QualifiedAllocationExpression allocation;
  public TypeDeclaration enclosingType;
  public FieldBinding enumValuesSyntheticfield;
  public TypeParameter[] typeParameters;

  public TypeDeclaration(CompilationResult compilationResult)
  {
    this.compilationResult = compilationResult;
  }

  public void abort(int abortLevel, CategorizedProblem problem)
  {
    switch (abortLevel) {
    case 2:
      throw new AbortCompilation(this.compilationResult, problem);
    case 4:
      throw new AbortCompilationUnit(this.compilationResult, problem);
    case 16:
      throw new AbortMethod(this.compilationResult, problem);
    }
    throw new AbortType(this.compilationResult, problem);
  }

  public final void addClinit()
  {
    if (needClassInitMethod())
    {
      AbstractMethodDeclaration[] methodDeclarations;
      if ((methodDeclarations = this.methods) == null) {
        int length = 0;
        methodDeclarations = new AbstractMethodDeclaration[1];
      } else {
        int length = methodDeclarations.length;
        System.arraycopy(
          methodDeclarations, 
          0, 
          methodDeclarations = new AbstractMethodDeclaration[length + 1], 
          1, 
          length);
      }
      Clinit clinit = new Clinit(this.compilationResult);
      methodDeclarations[0] = clinit;

      clinit.declarationSourceStart = (clinit.sourceStart = this.sourceStart);
      clinit.declarationSourceEnd = (clinit.sourceEnd = this.sourceEnd);
      clinit.bodyEnd = this.sourceEnd;
      this.methods = methodDeclarations;
    }
  }

  public MethodDeclaration addMissingAbstractMethodFor(MethodBinding methodBinding)
  {
    TypeBinding[] argumentTypes = methodBinding.parameters;
    int argumentsLength = argumentTypes.length;

    MethodDeclaration methodDeclaration = new MethodDeclaration(this.compilationResult);
    methodDeclaration.selector = methodBinding.selector;
    methodDeclaration.sourceStart = this.sourceStart;
    methodDeclaration.sourceEnd = this.sourceEnd;
    methodDeclaration.modifiers = (methodBinding.getAccessFlags() & 0xFFFFFBFF);

    if (argumentsLength > 0) {
      String baseName = "arg";
      Argument[] arguments = methodDeclaration.arguments = new Argument[argumentsLength];
      int i = argumentsLength;
      do { arguments[i] = new Argument((baseName + i).toCharArray(), 0L, null, 0);

        i--; } while (i >= 0);
    }

    if (this.missingAbstractMethods == null) {
      this.missingAbstractMethods = new MethodDeclaration[] { methodDeclaration };
    }
    else
    {
      MethodDeclaration[] newMethods;
      System.arraycopy(
        this.missingAbstractMethods, 
        0, 
        newMethods = new MethodDeclaration[this.missingAbstractMethods.length + 1], 
        1, 
        this.missingAbstractMethods.length);
      newMethods[0] = methodDeclaration;
      this.missingAbstractMethods = newMethods;
    }

    methodDeclaration.binding = new MethodBinding(
      methodDeclaration.modifiers, 
      methodBinding.selector, 
      methodBinding.returnType, 
      argumentsLength == 0 ? Binding.NO_PARAMETERS : argumentTypes, 
      methodBinding.thrownExceptions, 
      this.binding);

    methodDeclaration.scope = new MethodScope(this.scope, methodDeclaration, true);
    methodDeclaration.bindArguments();

    return methodDeclaration;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    if (this.ignoreFurtherInvestigation)
      return flowInfo;
    try {
      if ((flowInfo.tagBits & 0x1) == 0) {
        this.bits |= -2147483648;
        LocalTypeBinding localType = (LocalTypeBinding)this.binding;
        localType.setConstantPoolName(currentScope.compilationUnitScope().computeConstantPoolName(localType));
      }
      manageEnclosingInstanceAccessIfNecessary(currentScope, flowInfo);
      updateMaxFieldCount();
      internalAnalyseCode(flowContext, flowInfo);
    } catch (AbortType localAbortType) {
      this.ignoreFurtherInvestigation = true;
    }
    return flowInfo;
  }

  public void analyseCode(ClassScope enclosingClassScope)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    try
    {
      updateMaxFieldCount();
      internalAnalyseCode(null, FlowInfo.initial(this.maxFieldCount));
    } catch (AbortType localAbortType) {
      this.ignoreFurtherInvestigation = true;
    }
  }

  public void analyseCode(ClassScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    try {
      if ((flowInfo.tagBits & 0x1) == 0) {
        this.bits |= -2147483648;
        LocalTypeBinding localType = (LocalTypeBinding)this.binding;
        localType.setConstantPoolName(currentScope.compilationUnitScope().computeConstantPoolName(localType));
      }
      manageEnclosingInstanceAccessIfNecessary(currentScope, flowInfo);
      updateMaxFieldCount();
      internalAnalyseCode(flowContext, flowInfo);
    } catch (AbortType localAbortType) {
      this.ignoreFurtherInvestigation = true;
    }
  }

  public void analyseCode(CompilationUnitScope unitScope)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    try {
      internalAnalyseCode(null, FlowInfo.initial(this.maxFieldCount));
    } catch (AbortType localAbortType) {
      this.ignoreFurtherInvestigation = true;
    }
  }

  public boolean checkConstructors(Parser parser)
  {
    boolean hasConstructor = false;
    if (this.methods != null) {
      int i = this.methods.length;
      do
      {
        AbstractMethodDeclaration am;
        if ((am = this.methods[i]).isConstructor())
          if (!CharOperation.equals(am.selector, this.name))
          {
            ConstructorDeclaration c = (ConstructorDeclaration)am;
            if ((c.constructorCall == null) || (c.constructorCall.isImplicitSuper())) {
              MethodDeclaration m = parser.convertToMethodDeclaration(c, this.compilationResult);
              this.methods[i] = m;
            }
          } else {
            switch (kind(this.modifiers))
            {
            case 2:
              parser.problemReporter().interfaceCannotHaveConstructors((ConstructorDeclaration)am);
              break;
            case 4:
              parser.problemReporter().annotationTypeDeclarationCannotHaveConstructor((ConstructorDeclaration)am);
            case 3:
            }

            hasConstructor = true;
          }
        i--; } while (i >= 0);
    }

    return hasConstructor;
  }

  public CompilationResult compilationResult() {
    return this.compilationResult;
  }

  public ConstructorDeclaration createDefaultConstructor(boolean needExplicitConstructorCall, boolean needToInsert)
  {
    ConstructorDeclaration constructor = new ConstructorDeclaration(this.compilationResult);
    constructor.bits |= 128;
    constructor.selector = this.name;
    constructor.modifiers = (this.modifiers & 0x7);

    constructor.declarationSourceStart = (constructor.sourceStart = this.sourceStart);
    constructor.declarationSourceEnd = 
      (constructor.sourceEnd = constructor.bodyEnd = this.sourceEnd);

    if (needExplicitConstructorCall) {
      constructor.constructorCall = SuperReference.implicitSuperConstructorCall();
      constructor.constructorCall.sourceStart = this.sourceStart;
      constructor.constructorCall.sourceEnd = this.sourceEnd;
    }

    if (needToInsert) {
      if (this.methods == null) {
        this.methods = new AbstractMethodDeclaration[] { constructor };
      }
      else
      {
        AbstractMethodDeclaration[] newMethods;
        System.arraycopy(
          this.methods, 
          0, 
          newMethods = new AbstractMethodDeclaration[this.methods.length + 1], 
          1, 
          this.methods.length);
        newMethods[0] = constructor;
        this.methods = newMethods;
      }
    }
    return constructor;
  }

  public MethodBinding createDefaultConstructorWithBinding(MethodBinding inheritedConstructorBinding)
  {
    String baseName = "$anonymous";
    TypeBinding[] argumentTypes = inheritedConstructorBinding.parameters;
    int argumentsLength = argumentTypes.length;

    ConstructorDeclaration constructor = new ConstructorDeclaration(this.compilationResult);
    constructor.selector = new char[] { 'x' };
    constructor.sourceStart = this.sourceStart;
    constructor.sourceEnd = this.sourceEnd;
    int newModifiers = this.modifiers & 0x7;
    if (inheritedConstructorBinding.isVarargs()) {
      newModifiers |= 128;
    }
    constructor.modifiers = newModifiers;
    constructor.bits |= 128;

    if (argumentsLength > 0) {
      Argument[] arguments = constructor.arguments = new Argument[argumentsLength];
      int i = argumentsLength;
      do { arguments[i] = new Argument((baseName + i).toCharArray(), 0L, null, 0);

        i--; } while (i >= 0);
    }

    constructor.constructorCall = SuperReference.implicitSuperConstructorCall();
    constructor.constructorCall.sourceStart = this.sourceStart;
    constructor.constructorCall.sourceEnd = this.sourceEnd;

    if (argumentsLength > 0)
    {
      Expression[] args = constructor.constructorCall.arguments = new Expression[argumentsLength];
      int i = argumentsLength;
      do { args[i] = new SingleNameReference((baseName + i).toCharArray(), 0L);

        i--; } while (i >= 0);
    }

    if (this.methods == null) {
      this.methods = new AbstractMethodDeclaration[] { constructor };
    }
    else
    {
      AbstractMethodDeclaration[] newMethods;
      System.arraycopy(this.methods, 0, newMethods = new AbstractMethodDeclaration[this.methods.length + 1], 1, this.methods.length);
      newMethods[0] = constructor;
      this.methods = newMethods;
    }

    SourceTypeBinding sourceType = this.binding;
    constructor.binding = new MethodBinding(
      constructor.modifiers, 
      argumentsLength == 0 ? Binding.NO_PARAMETERS : argumentTypes, 
      inheritedConstructorBinding.thrownExceptions, 
      sourceType);

    constructor.binding.modifiers |= 67108864;

    constructor.scope = new MethodScope(this.scope, constructor, true);
    constructor.bindArguments();
    constructor.constructorCall.resolve(constructor.scope);

    MethodBinding[] methodBindings = sourceType.methods();
    int length;
    System.arraycopy(methodBindings, 0, methodBindings = new MethodBinding[(length = methodBindings.length) + 1], 1, length);
    methodBindings[0] = constructor.binding;
    length++; if (length > 1)
      ReferenceBinding.sortMethods(methodBindings, 0, length);
    sourceType.setMethods(methodBindings);

    return constructor.binding;
  }

  public FieldDeclaration declarationOf(FieldBinding fieldBinding)
  {
    if ((fieldBinding != null) && (this.fields != null)) {
      int i = 0; for (int max = this.fields.length; i < max; i++)
      {
        FieldDeclaration fieldDecl;
        if ((fieldDecl = this.fields[i]).binding == fieldBinding)
          return fieldDecl;
      }
    }
    return null;
  }

  public TypeDeclaration declarationOf(MemberTypeBinding memberTypeBinding)
  {
    if ((memberTypeBinding != null) && (this.memberTypes != null)) {
      int i = 0; for (int max = this.memberTypes.length; i < max; i++)
      {
        TypeDeclaration memberTypeDecl;
        if ((memberTypeDecl = this.memberTypes[i]).binding == memberTypeBinding)
          return memberTypeDecl;
      }
    }
    return null;
  }

  public AbstractMethodDeclaration declarationOf(MethodBinding methodBinding)
  {
    if ((methodBinding != null) && (this.methods != null)) {
      int i = 0; for (int max = this.methods.length; i < max; i++)
      {
        AbstractMethodDeclaration methodDecl;
        if ((methodDecl = this.methods[i]).binding == methodBinding)
          return methodDecl;
      }
    }
    return null;
  }

  public TypeDeclaration declarationOfType(char[][] typeName)
  {
    int typeNameLength = typeName.length;
    if ((typeNameLength < 1) || (!CharOperation.equals(typeName[0], this.name))) {
      return null;
    }
    if (typeNameLength == 1) {
      return this;
    }
    char[][] subTypeName = new char[typeNameLength - 1][];
    System.arraycopy(typeName, 1, subTypeName, 0, typeNameLength - 1);
    for (int i = 0; i < this.memberTypes.length; i++) {
      TypeDeclaration typeDecl = this.memberTypes[i].declarationOfType(subTypeName);
      if (typeDecl != null) {
        return typeDecl;
      }
    }
    return null;
  }

  public void generateCode(ClassFile enclosingClassFile)
  {
    if ((this.bits & 0x2000) != 0)
      return;
    this.bits |= 8192;
    if (this.ignoreFurtherInvestigation) {
      if (this.binding == null)
        return;
      ClassFile.createProblemType(
        this, 
        this.scope.referenceCompilationUnit().compilationResult);
      return;
    }
    try
    {
      ClassFile classFile = ClassFile.getNewInstance(this.binding);
      classFile.initialize(this.binding, enclosingClassFile, false);
      if (this.binding.isMemberType()) {
        classFile.recordInnerClasses(this.binding);
      } else if (this.binding.isLocalType()) {
        enclosingClassFile.recordInnerClasses(this.binding);
        classFile.recordInnerClasses(this.binding);
      }

      classFile.addFieldInfos();

      if (this.memberTypes != null) {
        int i = 0; for (int max = this.memberTypes.length; i < max; i++) {
          TypeDeclaration memberType = this.memberTypes[i];
          classFile.recordInnerClasses(memberType.binding);
          memberType.generateCode(this.scope, classFile);
        }
      }

      classFile.setForMethodInfos();
      if (this.methods != null) {
        int i = 0; for (int max = this.methods.length; i < max; i++) {
          this.methods[i].generateCode(this.scope, classFile);
        }
      }

      classFile.addSpecialMethods();

      if (this.ignoreFurtherInvestigation) {
        throw new AbortType(this.scope.referenceCompilationUnit().compilationResult, null);
      }

      classFile.addAttributes();
      this.scope.referenceCompilationUnit().compilationResult.record(
        this.binding.constantPoolName(), 
        classFile);
    } catch (AbortType localAbortType) {
      if (this.binding == null)
        return;
      ClassFile.createProblemType(
        this, 
        this.scope.referenceCompilationUnit().compilationResult);
    }
  }

  public void generateCode(BlockScope blockScope, CodeStream codeStream)
  {
    if ((this.bits & 0x80000000) == 0) {
      return;
    }
    if ((this.bits & 0x2000) != 0) return;
    int pc = codeStream.position;
    if (this.binding != null) ((NestedTypeBinding)this.binding).computeSyntheticArgumentSlotSizes();
    generateCode(codeStream.classFile);
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public void generateCode(ClassScope classScope, ClassFile enclosingClassFile)
  {
    if ((this.bits & 0x2000) != 0) return;
    if (this.binding != null) ((NestedTypeBinding)this.binding).computeSyntheticArgumentSlotSizes();
    generateCode(enclosingClassFile);
  }

  public void generateCode(CompilationUnitScope unitScope)
  {
    generateCode(null);
  }

  public boolean hasErrors() {
    return this.ignoreFurtherInvestigation;
  }

  private void internalAnalyseCode(FlowContext flowContext, FlowInfo flowInfo)
  {
    if (((this.binding.isPrivate()) || ((this.binding.tagBits & 0x30) == 16L)) && (!this.binding.isUsed()) && 
      (!this.scope.referenceCompilationUnit().compilationResult.hasSyntaxError)) {
      this.scope.problemReporter().unusedPrivateType(this);
    }

    InitializationFlowContext initializerContext = new InitializationFlowContext(null, this, this.initializerScope);
    InitializationFlowContext staticInitializerContext = new InitializationFlowContext(null, this, this.staticInitializerScope);
    FlowInfo nonStaticFieldInfo = flowInfo.unconditionalFieldLessCopy();
    FlowInfo staticFieldInfo = flowInfo.unconditionalFieldLessCopy();
    if (this.fields != null) {
      int i = 0; for (int count = this.fields.length; i < count; i++) {
        FieldDeclaration field = this.fields[i];
        if (field.isStatic()) {
          if ((staticFieldInfo.tagBits & 0x1) != 0) {
            field.bits &= 2147483647;
          }

          staticInitializerContext.handledExceptions = Binding.ANY_EXCEPTION;

          staticFieldInfo = 
            field.analyseCode(
            this.staticInitializerScope, 
            staticInitializerContext, 
            staticFieldInfo);

          if (staticFieldInfo == FlowInfo.DEAD_END) {
            this.staticInitializerScope.problemReporter().initializerMustCompleteNormally(field);
            staticFieldInfo = FlowInfo.initial(this.maxFieldCount).setReachMode(1);
          }
        } else {
          if ((nonStaticFieldInfo.tagBits & 0x1) != 0) {
            field.bits &= 2147483647;
          }

          initializerContext.handledExceptions = Binding.ANY_EXCEPTION;

          nonStaticFieldInfo = 
            field.analyseCode(this.initializerScope, initializerContext, nonStaticFieldInfo);

          if (nonStaticFieldInfo == FlowInfo.DEAD_END) {
            this.initializerScope.problemReporter().initializerMustCompleteNormally(field);
            nonStaticFieldInfo = FlowInfo.initial(this.maxFieldCount).setReachMode(1);
          }
        }
      }
    }
    if (this.memberTypes != null) {
      int i = 0; for (int count = this.memberTypes.length; i < count; i++) {
        if (flowContext != null)
          this.memberTypes[i].analyseCode(this.scope, flowContext, nonStaticFieldInfo.copy().setReachMode(flowInfo.reachMode()));
        else {
          this.memberTypes[i].analyseCode(this.scope);
        }
      }
    }
    if (this.methods != null) {
      UnconditionalFlowInfo outerInfo = flowInfo.unconditionalFieldLessCopy();
      FlowInfo constructorInfo = nonStaticFieldInfo.unconditionalInits().discardNonFieldInitializations().addInitializationsFrom(outerInfo);
      int i = 0; for (int count = this.methods.length; i < count; i++) {
        AbstractMethodDeclaration method = this.methods[i];
        if (!method.ignoreFurtherInvestigation)
        {
          if (method.isInitializationMethod()) {
            if (method.isStatic())
              method.analyseCode(
                this.scope, 
                staticInitializerContext, 
                staticFieldInfo.unconditionalInits().discardNonFieldInitializations().addInitializationsFrom(outerInfo));
            else
              ((ConstructorDeclaration)method).analyseCode(this.scope, initializerContext, constructorInfo.copy(), flowInfo.reachMode());
          }
          else {
            method.analyseCode(this.scope, null, flowInfo.copy());
          }
        }
      }
    }
    if ((this.binding.isEnum()) && (!this.binding.isAnonymousType()))
      this.enumValuesSyntheticfield = this.binding.addSyntheticFieldForEnumValues();
  }

  public static final int kind(int flags)
  {
    switch (flags & 0x6200) {
    case 512:
      return 2;
    case 8704:
      return 4;
    case 16384:
      return 3;
    }
    return 1;
  }

  public void manageEnclosingInstanceAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo)
  {
    if ((flowInfo.tagBits & 0x1) != 0) return;
    NestedTypeBinding nestedType = (NestedTypeBinding)this.binding;

    MethodScope methodScope = currentScope.methodScope();
    if ((!methodScope.isStatic) && (!methodScope.isConstructorCall)) {
      nestedType.addSyntheticArgumentAndField(nestedType.enclosingType());
    }

    if (nestedType.isAnonymousType()) {
      ReferenceBinding superclassBinding = (ReferenceBinding)nestedType.superclass.erasure();
      if ((superclassBinding.enclosingType() != null) && (!superclassBinding.isStatic()) && (
        (!superclassBinding.isLocalType()) || 
        (((NestedTypeBinding)superclassBinding).getSyntheticField(superclassBinding.enclosingType(), true) != null)))
      {
        nestedType.addSyntheticArgument(superclassBinding.enclosingType());
      }

      if ((!methodScope.isStatic) && (methodScope.isConstructorCall) && (currentScope.compilerOptions().complianceLevel >= 3211264L)) {
        ReferenceBinding enclosing = nestedType.enclosingType();
        if (enclosing.isNestedType()) {
          NestedTypeBinding nestedEnclosing = (NestedTypeBinding)enclosing;

          SyntheticArgumentBinding syntheticEnclosingInstanceArgument = nestedEnclosing.getSyntheticArgument(nestedEnclosing.enclosingType(), true);
          if (syntheticEnclosingInstanceArgument != null)
            nestedType.addSyntheticArgumentAndField(syntheticEnclosingInstanceArgument);
        }
      }
    }
  }

  public void manageEnclosingInstanceAccessIfNecessary(ClassScope currentScope, FlowInfo flowInfo)
  {
    if ((flowInfo.tagBits & 0x1) == 0) {
      NestedTypeBinding nestedType = (NestedTypeBinding)this.binding;
      nestedType.addSyntheticArgumentAndField(this.binding.enclosingType());
    }
  }

  public final boolean needClassInitMethod()
  {
    if ((this.bits & 0x1) != 0) {
      return true;
    }
    switch (kind(this.modifiers)) {
    case 2:
    case 4:
      return this.fields != null;
    case 3:
      return true;
    }
    if (this.fields != null) {
      int i = this.fields.length;
      do { FieldDeclaration field = this.fields[i];

        if ((field.modifiers & 0x8) != 0)
          return true;
        i--; } while (i >= 0);
    }

    return false;
  }

  public void parseMethod(Parser parser, CompilationUnitDeclaration unit)
  {
    if (unit.ignoreMethodBodies) {
      return;
    }

    if (this.memberTypes != null) {
      int length = this.memberTypes.length;
      for (int i = 0; i < length; i++) {
        this.memberTypes[i].parseMethod(parser, unit);
      }
    }

    if (this.methods != null) {
      int length = this.methods.length;
      for (int i = 0; i < length; i++) {
        this.methods[i].parseStatements(parser, unit);
      }

    }

    if (this.fields != null) {
      int length = this.fields.length;
      for (int i = 0; i < length; i++) {
        FieldDeclaration fieldDeclaration = this.fields[i];
        switch (fieldDeclaration.getKind()) {
        case 2:
          ((Initializer)fieldDeclaration).parseStatements(parser, this, unit);
        }
      }
    }
  }

  public StringBuffer print(int indent, StringBuffer output)
  {
    if (this.javadoc != null) {
      this.javadoc.print(indent, output);
    }
    if ((this.bits & 0x200) == 0) {
      printIndent(indent, output);
      printHeader(0, output);
    }
    return printBody(indent, output);
  }

  public StringBuffer printBody(int indent, StringBuffer output) {
    output.append(" {");
    if (this.memberTypes != null) {
      for (int i = 0; i < this.memberTypes.length; i++) {
        if (this.memberTypes[i] != null) {
          output.append('\n');
          this.memberTypes[i].print(indent + 1, output);
        }
      }
    }
    if (this.fields != null) {
      for (int fieldI = 0; fieldI < this.fields.length; fieldI++) {
        if (this.fields[fieldI] != null) {
          output.append('\n');
          this.fields[fieldI].print(indent + 1, output);
        }
      }
    }
    if (this.methods != null) {
      for (int i = 0; i < this.methods.length; i++) {
        if (this.methods[i] != null) {
          output.append('\n');
          this.methods[i].print(indent + 1, output);
        }
      }
    }
    output.append('\n');
    return printIndent(indent, output).append('}');
  }

  public StringBuffer printHeader(int indent, StringBuffer output) {
    printModifiers(this.modifiers, output);
    if (this.annotations != null) printAnnotations(this.annotations, output);

    switch (kind(this.modifiers)) {
    case 1:
      output.append("class ");
      break;
    case 2:
      output.append("interface ");
      break;
    case 3:
      output.append("enum ");
      break;
    case 4:
      output.append("@interface ");
    }

    output.append(this.name);
    if (this.typeParameters != null) {
      output.append("<");
      for (int i = 0; i < this.typeParameters.length; i++) {
        if (i > 0) output.append(", ");
        this.typeParameters[i].print(0, output);
      }
      output.append(">");
    }
    if (this.superclass != null) {
      output.append(" extends ");
      this.superclass.print(0, output);
    }
    if ((this.superInterfaces != null) && (this.superInterfaces.length > 0)) {
      switch (kind(this.modifiers)) {
      case 1:
      case 3:
        output.append(" implements ");
        break;
      case 2:
      case 4:
        output.append(" extends ");
      }

      for (int i = 0; i < this.superInterfaces.length; i++) {
        if (i > 0) output.append(", ");
        this.superInterfaces[i].print(0, output);
      }
    }
    return output;
  }

  public StringBuffer printStatement(int tab, StringBuffer output) {
    return print(tab, output);
  }

  public void resolve()
  {
    SourceTypeBinding sourceType = this.binding;
    if (sourceType == null) {
      this.ignoreFurtherInvestigation = true;
      return;
    }
    try {
      boolean old = this.staticInitializerScope.insideTypeAnnotation;
      try {
        this.staticInitializerScope.insideTypeAnnotation = true;
        resolveAnnotations(this.staticInitializerScope, this.annotations, sourceType);
      } finally {
        this.staticInitializerScope.insideTypeAnnotation = old;
      }

      if (((sourceType.getAnnotationTagBits() & 0x0) == 0L) && 
        ((sourceType.modifiers & 0x100000) != 0) && 
        (this.scope.compilerOptions().sourceLevel >= 3211264L)) {
        this.scope.problemReporter().missingDeprecatedAnnotationForType(this);
      }
      if ((this.bits & 0x8) != 0) {
        this.scope.problemReporter().undocumentedEmptyBlock(this.bodyStart - 1, this.bodyEnd);
      }
      boolean needSerialVersion = 
        (this.scope.compilerOptions().getSeverity(4294967296L) != -1) && 
        (sourceType.isClass()) && 
        (!sourceType.isAbstract()) && 
        (sourceType.findSuperTypeErasingTo(56, false) == null) && 
        (sourceType.findSuperTypeErasingTo(37, false) != null);

      if (needSerialVersion)
      {
        CompilationUnitScope compilationUnitScope = this.scope.compilationUnitScope();
        MethodBinding methodBinding = sourceType.getExactMethod(TypeConstants.WRITEREPLACE, new TypeBinding[0], compilationUnitScope);
        ReferenceBinding[] throwsExceptions;
        needSerialVersion = 
          (methodBinding == null) || 
          (!methodBinding.isValidBinding()) || 
          (methodBinding.returnType.id != 1) || 
          ((throwsExceptions = methodBinding.thrownExceptions).length != 1) || 
          (throwsExceptions[0].id != 57);
        if (needSerialVersion)
        {
          boolean hasWriteObjectMethod = false;
          boolean hasReadObjectMethod = false;
          TypeBinding argumentTypeBinding = this.scope.getType(TypeConstants.JAVA_IO_OBJECTOUTPUTSTREAM, 3);
          if (argumentTypeBinding.isValidBinding()) {
            methodBinding = sourceType.getExactMethod(TypeConstants.WRITEOBJECT, new TypeBinding[] { argumentTypeBinding }, compilationUnitScope);
            ReferenceBinding[] throwsExceptions;
            hasWriteObjectMethod = (methodBinding != null) && 
              (methodBinding.isValidBinding()) && 
              (methodBinding.modifiers == 2) && 
              (methodBinding.returnType == TypeBinding.VOID) && 
              ((throwsExceptions = methodBinding.thrownExceptions).length == 1) && 
              (throwsExceptions[0].id == 58);
          }
          argumentTypeBinding = this.scope.getType(TypeConstants.JAVA_IO_OBJECTINPUTSTREAM, 3);
          if (argumentTypeBinding.isValidBinding()) {
            methodBinding = sourceType.getExactMethod(TypeConstants.READOBJECT, new TypeBinding[] { argumentTypeBinding }, compilationUnitScope);
            ReferenceBinding[] throwsExceptions;
            hasReadObjectMethod = (methodBinding != null) && 
              (methodBinding.isValidBinding()) && 
              (methodBinding.modifiers == 2) && 
              (methodBinding.returnType == TypeBinding.VOID) && 
              ((throwsExceptions = methodBinding.thrownExceptions).length == 1) && 
              (throwsExceptions[0].id == 58);
          }
          needSerialVersion = (!hasWriteObjectMethod) || (!hasReadObjectMethod);
        }
      }

      if (sourceType.findSuperTypeErasingTo(21, true) != null) {
        ReferenceBinding current = sourceType;
        do
          if (current.isGenericType()) {
            this.scope.problemReporter().genericTypeCannotExtendThrowable(this);
          }
          else if (!current.isStatic())
            if (current.isLocalType()) {
              NestedTypeBinding nestedType = (NestedTypeBinding)current.erasure();
              if (nestedType.scope.methodScope().isStatic) break;
            }
        while ((current = current.enclosingType()) != null);
      }
      this.maxFieldCount = 0;
      int lastVisibleFieldID = -1;
      boolean hasEnumConstants = false;
      boolean hasEnumConstantsWithoutBody = false;

      if (this.typeParameters != null) {
        int i = 0; for (int count = this.typeParameters.length; i < count; i++) {
          this.typeParameters[i].resolve(this.scope);
        }
      }
      if (this.memberTypes != null) {
        int i = 0; for (int count = this.memberTypes.length; i < count; i++) {
          this.memberTypes[i].resolve(this.scope);
        }
      }
      if (this.fields != null) {
        int i = 0; for (int count = this.fields.length; i < count; i++) {
          FieldDeclaration field = this.fields[i];
          switch (field.getKind()) {
          case 3:
            hasEnumConstants = true;
            if (!(field.initialization instanceof QualifiedAllocationExpression))
              hasEnumConstantsWithoutBody = true;
          case 1:
            FieldBinding fieldBinding = field.binding;
            if (fieldBinding == null)
            {
              if (field.initialization != null) field.initialization.resolve(field.isStatic() ? this.staticInitializerScope : this.initializerScope);
              this.ignoreFurtherInvestigation = true;
              continue;
            }
            if ((needSerialVersion) && 
              ((fieldBinding.modifiers & 0x18) == 24) && 
              (CharOperation.equals(TypeConstants.SERIALVERSIONUID, fieldBinding.name)) && 
              (TypeBinding.LONG == fieldBinding.type)) {
              needSerialVersion = false;
            }
            this.maxFieldCount += 1;
            lastVisibleFieldID = field.binding.id;
            break;
          case 2:
            ((Initializer)field).lastVisibleFieldID = (lastVisibleFieldID + 1);
          }

          field.resolve(field.isStatic() ? this.staticInitializerScope : this.initializerScope);
        }
      }
      if (needSerialVersion) {
        this.scope.problemReporter().missingSerialVersion(this);
      }

      switch (kind(this.modifiers)) {
      case 4:
        if (this.superclass != null) {
          this.scope.problemReporter().annotationTypeDeclarationCannotHaveSuperclass(this);
        }
        if (this.superInterfaces != null) {
          this.scope.problemReporter().annotationTypeDeclarationCannotHaveSuperinterfaces(this);
        }
        break;
      case 3:
        if ((this.binding.isAbstract()) && (
          (!hasEnumConstants) || (hasEnumConstantsWithoutBody))) {
          int i = 0; for (int count = this.methods.length; i < count; i++) {
            AbstractMethodDeclaration methodDeclaration = this.methods[i];
            if ((methodDeclaration.isAbstract()) && (methodDeclaration.binding != null)) {
              this.scope.problemReporter().enumAbstractMethodMustBeImplemented(methodDeclaration);
            }
          }
        }

        break;
      }

      int missingAbstractMethodslength = this.missingAbstractMethods == null ? 0 : this.missingAbstractMethods.length;
      int methodsLength = this.methods == null ? 0 : this.methods.length;
      if (methodsLength + missingAbstractMethodslength > 65535) {
        this.scope.problemReporter().tooManyMethods(this);
      }
      if (this.methods != null) {
        int i = 0; for (int count = this.methods.length; i < count; i++) {
          this.methods[i].resolve(this.scope);
        }
      }

      if (this.javadoc != null) {
        if ((this.scope != null) && (this.name != TypeConstants.PACKAGE_INFO_NAME))
        {
          this.javadoc.resolve(this.scope);
        }
      } else if ((sourceType != null) && (!sourceType.isLocalType()))
        this.scope.problemReporter().javadocMissing(this.sourceStart, this.sourceEnd, sourceType.modifiers);
    }
    catch (AbortType localAbortType)
    {
      this.ignoreFurtherInvestigation = true;
      return;
    }
  }

  public void resolve(BlockScope blockScope)
  {
    if ((this.bits & 0x200) == 0)
    {
      Binding existing = blockScope.getType(this.name);
      if (((existing instanceof ReferenceBinding)) && 
        (existing != this.binding) && 
        (existing.isValidBinding())) {
        ReferenceBinding existingType = (ReferenceBinding)existing;
        if ((existingType instanceof TypeVariableBinding))
          blockScope.problemReporter().typeHiding(this, (TypeVariableBinding)existingType);
        else if (((existingType instanceof LocalTypeBinding)) && 
          (((LocalTypeBinding)existingType).scope.methodScope() == blockScope.methodScope()))
        {
          blockScope.problemReporter().duplicateNestedType(this);
        } else if (blockScope.isDefinedInType(existingType))
        {
          blockScope.problemReporter().typeCollidesWithEnclosingType(this);
        } else if (blockScope.isDefinedInSameUnit(existingType))
        {
          blockScope.problemReporter().typeHiding(this, existingType);
        }
      }
      blockScope.addLocalType(this);
    }

    if (this.binding != null)
    {
      blockScope.referenceCompilationUnit().record((LocalTypeBinding)this.binding);

      resolve();
      updateMaxFieldCount();
    }
  }

  public void resolve(ClassScope upperScope)
  {
    if ((this.binding != null) && ((this.binding instanceof LocalTypeBinding)))
    {
      upperScope.referenceCompilationUnit().record((LocalTypeBinding)this.binding);
    }
    resolve();
    updateMaxFieldCount();
  }

  public void resolve(CompilationUnitScope upperScope)
  {
    resolve();
    updateMaxFieldCount();
  }

  public void tagAsHavingErrors() {
    this.ignoreFurtherInvestigation = true;
  }

  public void traverse(ASTVisitor visitor, CompilationUnitScope unitScope)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    try {
      if (visitor.visit(this, unitScope)) {
        if (this.javadoc != null) {
          this.javadoc.traverse(visitor, this.scope);
        }
        if (this.annotations != null) {
          int annotationsLength = this.annotations.length;
          for (int i = 0; i < annotationsLength; i++)
            this.annotations[i].traverse(visitor, this.staticInitializerScope);
        }
        if (this.superclass != null)
          this.superclass.traverse(visitor, this.scope);
        if (this.superInterfaces != null) {
          int length = this.superInterfaces.length;
          for (int i = 0; i < length; i++)
            this.superInterfaces[i].traverse(visitor, this.scope);
        }
        if (this.typeParameters != null) {
          int length = this.typeParameters.length;
          for (int i = 0; i < length; i++) {
            this.typeParameters[i].traverse(visitor, this.scope);
          }
        }
        if (this.memberTypes != null) {
          int length = this.memberTypes.length;
          for (int i = 0; i < length; i++)
            this.memberTypes[i].traverse(visitor, this.scope);
        }
        if (this.fields != null) {
          int length = this.fields.length;
          for (int i = 0; i < length; i++)
          {
            FieldDeclaration field;
            if ((field = this.fields[i]).isStatic())
              field.traverse(visitor, this.staticInitializerScope);
            else {
              field.traverse(visitor, this.initializerScope);
            }
          }
        }
        if (this.methods != null) {
          int length = this.methods.length;
          for (int i = 0; i < length; i++)
            this.methods[i].traverse(visitor, this.scope);
        }
      }
      visitor.endVisit(this, unitScope);
    }
    catch (AbortType localAbortType)
    {
    }
  }

  public void traverse(ASTVisitor visitor, BlockScope blockScope)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    try {
      if (visitor.visit(this, blockScope)) {
        if (this.javadoc != null) {
          this.javadoc.traverse(visitor, this.scope);
        }
        if (this.annotations != null) {
          int annotationsLength = this.annotations.length;
          for (int i = 0; i < annotationsLength; i++)
            this.annotations[i].traverse(visitor, this.staticInitializerScope);
        }
        if (this.superclass != null)
          this.superclass.traverse(visitor, this.scope);
        if (this.superInterfaces != null) {
          int length = this.superInterfaces.length;
          for (int i = 0; i < length; i++)
            this.superInterfaces[i].traverse(visitor, this.scope);
        }
        if (this.typeParameters != null) {
          int length = this.typeParameters.length;
          for (int i = 0; i < length; i++) {
            this.typeParameters[i].traverse(visitor, this.scope);
          }
        }
        if (this.memberTypes != null) {
          int length = this.memberTypes.length;
          for (int i = 0; i < length; i++)
            this.memberTypes[i].traverse(visitor, this.scope);
        }
        if (this.fields != null) {
          int length = this.fields.length;
          for (int i = 0; i < length; i++)
          {
            FieldDeclaration field;
            if (!(field = this.fields[i]).isStatic())
            {
              field.traverse(visitor, this.initializerScope);
            }
          }
        }
        if (this.methods != null) {
          int length = this.methods.length;
          for (int i = 0; i < length; i++)
            this.methods[i].traverse(visitor, this.scope);
        }
      }
      visitor.endVisit(this, blockScope);
    }
    catch (AbortType localAbortType)
    {
    }
  }

  public void traverse(ASTVisitor visitor, ClassScope classScope)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    try {
      if (visitor.visit(this, classScope)) {
        if (this.javadoc != null) {
          this.javadoc.traverse(visitor, this.scope);
        }
        if (this.annotations != null) {
          int annotationsLength = this.annotations.length;
          for (int i = 0; i < annotationsLength; i++)
            this.annotations[i].traverse(visitor, this.staticInitializerScope);
        }
        if (this.superclass != null)
          this.superclass.traverse(visitor, this.scope);
        if (this.superInterfaces != null) {
          int length = this.superInterfaces.length;
          for (int i = 0; i < length; i++)
            this.superInterfaces[i].traverse(visitor, this.scope);
        }
        if (this.typeParameters != null) {
          int length = this.typeParameters.length;
          for (int i = 0; i < length; i++) {
            this.typeParameters[i].traverse(visitor, this.scope);
          }
        }
        if (this.memberTypes != null) {
          int length = this.memberTypes.length;
          for (int i = 0; i < length; i++)
            this.memberTypes[i].traverse(visitor, this.scope);
        }
        if (this.fields != null) {
          int length = this.fields.length;
          for (int i = 0; i < length; i++)
          {
            FieldDeclaration field;
            if ((field = this.fields[i]).isStatic())
              field.traverse(visitor, this.staticInitializerScope);
            else {
              field.traverse(visitor, this.initializerScope);
            }
          }
        }
        if (this.methods != null) {
          int length = this.methods.length;
          for (int i = 0; i < length; i++)
            this.methods[i].traverse(visitor, this.scope);
        }
      }
      visitor.endVisit(this, classScope);
    }
    catch (AbortType localAbortType)
    {
    }
  }

  void updateMaxFieldCount()
  {
    if (this.binding == null)
      return;
    TypeDeclaration outerMostType = this.scope.outerMostClassScope().referenceType();
    if (this.maxFieldCount > outerMostType.maxFieldCount)
      outerMostType.maxFieldCount = this.maxFieldCount;
    else
      this.maxFieldCount = outerMostType.maxFieldCount;
  }

  public boolean isSecondary()
  {
    return (this.bits & 0x1000) != 0;
  }
}