package org.eclipse.jdt.internal.compiler.flow;

import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.Reference;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class LoopingFlowContext extends SwitchFlowContext
{
  public BranchLabel continueLabel;
  public UnconditionalFlowInfo initsOnContinue = FlowInfo.DEAD_END;
  private UnconditionalFlowInfo upstreamNullFlowInfo;
  private LoopingFlowContext[] innerFlowContexts = null;
  private UnconditionalFlowInfo[] innerFlowInfos = null;
  private int innerFlowContextsCount = 0;
  private LabelFlowContext[] breakTargetContexts = null;
  private int breakTargetsCount = 0;
  Reference[] finalAssignments;
  VariableBinding[] finalVariables;
  int assignCount = 0;
  LocalVariableBinding[] nullLocals;
  Expression[] nullReferences;
  int[] nullCheckTypes;
  int nullCount;
  Scope associatedScope;

  public LoopingFlowContext(FlowContext parent, FlowInfo upstreamNullFlowInfo, ASTNode associatedNode, BranchLabel breakLabel, BranchLabel continueLabel, Scope associatedScope)
  {
    super(parent, associatedNode, breakLabel);
    this.preemptNullDiagnostic = true;

    this.continueLabel = continueLabel;
    this.associatedScope = associatedScope;
    this.upstreamNullFlowInfo = upstreamNullFlowInfo.unconditionalCopy();
  }

  public void complainOnDeferredFinalChecks(BlockScope scope, FlowInfo flowInfo)
  {
    for (int i = 0; i < this.assignCount; i++) {
      VariableBinding variable = this.finalVariables[i];
      if (variable != null) {
        boolean complained = false;
        if ((variable instanceof FieldBinding)) {
          if (flowInfo.isPotentiallyAssigned((FieldBinding)variable)) {
            complained = true;
            scope.problemReporter().duplicateInitializationOfBlankFinalField(
              (FieldBinding)variable, 
              this.finalAssignments[i]);
          }
        }
        else if (flowInfo.isPotentiallyAssigned((LocalVariableBinding)variable)) {
          complained = true;
          scope.problemReporter().duplicateInitializationOfFinalLocal(
            (LocalVariableBinding)variable, 
            this.finalAssignments[i]);
        }

        if (complained) {
          FlowContext context = this.parent;
          while (context != null) {
            context.removeFinalAssignmentIfAny(this.finalAssignments[i]);
            context = context.parent;
          }
        }
      }
    }
  }

  public void complainOnDeferredNullChecks(BlockScope scope, FlowInfo callerFlowInfo)
  {
    for (int i = 0; i < this.innerFlowContextsCount; i++) {
      this.upstreamNullFlowInfo
        .addPotentialNullInfoFrom(
        this.innerFlowContexts[i].upstreamNullFlowInfo)
        .addPotentialNullInfoFrom(this.innerFlowInfos[i]);
    }
    this.innerFlowContextsCount = 0;
    UnconditionalFlowInfo flowInfo = this.upstreamNullFlowInfo
      .addPotentialNullInfoFrom(callerFlowInfo.unconditionalInitsWithoutSideEffect());
    if (this.deferNullDiagnostic)
    {
      for (int i = 0; i < this.nullCount; i++) {
        LocalVariableBinding local = this.nullLocals[i];
        Expression expression = this.nullReferences[i];

        switch (this.nullCheckTypes[i]) {
        case 258:
        case 514:
          if (flowInfo.isDefinitelyNonNull(local)) {
            this.nullReferences[i] = null;
            if (this.nullCheckTypes[i] == 514)
              scope.problemReporter().localVariableRedundantCheckOnNonNull(local, expression);
            else
              scope.problemReporter().localVariableNonNullComparedToNull(local, expression);
          }
          break;
        case 256:
        case 512:
          if (flowInfo.isDefinitelyNonNull(local)) {
            this.nullReferences[i] = null;
            if (this.nullCheckTypes[i] == 512)
              scope.problemReporter().localVariableRedundantCheckOnNonNull(local, expression);
            else {
              scope.problemReporter().localVariableNonNullComparedToNull(local, expression);
            }

          }
          else if (flowInfo.isDefinitelyNull(local)) {
            this.nullReferences[i] = null;
            if (this.nullCheckTypes[i] == 256)
              scope.problemReporter().localVariableRedundantCheckOnNull(local, expression);
            else
              scope.problemReporter().localVariableNullComparedToNonNull(local, expression);
          }
          break;
        case 257:
        case 513:
        case 769:
        case 1025:
          if (flowInfo.isDefinitelyNull(local)) {
            this.nullReferences[i] = null;
            switch (this.nullCheckTypes[i] & 0xFFFFFF00) {
            case 256:
              scope.problemReporter().localVariableRedundantCheckOnNull(local, expression);
              break;
            case 512:
              scope.problemReporter().localVariableNullComparedToNonNull(local, expression);
              break;
            case 768:
              scope.problemReporter().localVariableRedundantNullAssignment(local, expression);
              break;
            case 1024:
              scope.problemReporter().localVariableNullInstanceof(local, expression);
              continue;
            }
          }
          break;
        case 3:
          if (flowInfo.isDefinitelyNull(local)) {
            this.nullReferences[i] = null;
            scope.problemReporter().localVariableNullReference(local, expression);
          }break;
        }

        this.parent.recordUsingNullReference(scope, local, expression, 
          this.nullCheckTypes[i], flowInfo);
      }
    }
    else
    {
      for (int i = 0; i < this.nullCount; i++) {
        Expression expression = this.nullReferences[i];

        LocalVariableBinding local = this.nullLocals[i];
        switch (this.nullCheckTypes[i]) {
        case 256:
        case 512:
          if (flowInfo.isDefinitelyNonNull(local)) {
            this.nullReferences[i] = null;
            if (this.nullCheckTypes[i] == 512)
              scope.problemReporter().localVariableRedundantCheckOnNonNull(local, expression);
            else
              scope.problemReporter().localVariableNonNullComparedToNull(local, expression);
          }
          break;
        case 257:
        case 513:
        case 769:
        case 1025:
          if (flowInfo.isDefinitelyNull(local)) {
            this.nullReferences[i] = null;
            switch (this.nullCheckTypes[i] & 0xFFFFFF00) {
            case 256:
              scope.problemReporter().localVariableRedundantCheckOnNull(local, expression);
              break;
            case 512:
              scope.problemReporter().localVariableNullComparedToNonNull(local, expression);
              break;
            case 768:
              scope.problemReporter().localVariableRedundantNullAssignment(local, expression);
              break;
            case 1024:
              scope.problemReporter().localVariableNullInstanceof(local, expression);
            }
          }

          break;
        case 3:
          if (flowInfo.isDefinitelyNull(local)) {
            this.nullReferences[i] = null;
            scope.problemReporter().localVariableNullReference(local, expression);
          }
          else if (flowInfo.isPotentiallyNull(local)) {
            this.nullReferences[i] = null;
            scope.problemReporter().localVariablePotentialNullReference(local, expression);
          }

          break;
        }

      }

    }

    for (int i = 0; i < this.breakTargetsCount; i++)
      this.breakTargetContexts[i].initsOnBreak.addPotentialNullInfoFrom(flowInfo);
  }

  public BranchLabel continueLabel()
  {
    return this.continueLabel;
  }

  public String individualToString() {
    StringBuffer buffer = new StringBuffer("Looping flow context");
    buffer.append("[initsOnBreak - ").append(this.initsOnBreak.toString()).append(']');
    buffer.append("[initsOnContinue - ").append(this.initsOnContinue.toString()).append(']');
    buffer.append("[finalAssignments count - ").append(this.assignCount).append(']');
    buffer.append("[nullReferences count - ").append(this.nullCount).append(']');
    return buffer.toString();
  }

  public boolean isContinuable() {
    return true;
  }

  public boolean isContinuedTo() {
    return this.initsOnContinue != FlowInfo.DEAD_END;
  }

  public void recordBreakTo(FlowContext targetContext) {
    if ((targetContext instanceof LabelFlowContext))
    {
      int current;
      if ((current = this.breakTargetsCount++) == 0)
        this.breakTargetContexts = new LabelFlowContext[2];
      else if (current == this.breakTargetContexts.length) {
        System.arraycopy(this.breakTargetContexts, 0, this.breakTargetContexts = new LabelFlowContext[current + 2], 0, current);
      }
      this.breakTargetContexts[current] = ((LabelFlowContext)targetContext);
    }
  }

  public void recordContinueFrom(FlowContext innerFlowContext, FlowInfo flowInfo) {
    if ((flowInfo.tagBits & 0x1) == 0) {
      if ((this.initsOnContinue.tagBits & 0x1) == 0) {
        this.initsOnContinue = 
          this.initsOnContinue.mergedWith(flowInfo.unconditionalInitsWithoutSideEffect());
      }
      else {
        this.initsOnContinue = flowInfo.unconditionalCopy();
      }
      FlowContext inner = innerFlowContext;
      while ((inner != this) && (!(inner instanceof LoopingFlowContext))) {
        inner = inner.parent;
      }
      if (inner == this) {
        this.upstreamNullFlowInfo
          .addPotentialNullInfoFrom(
          flowInfo.unconditionalInitsWithoutSideEffect());
      }
      else {
        int length = 0;
        if (this.innerFlowContexts == null) {
          this.innerFlowContexts = new LoopingFlowContext[5];
          this.innerFlowInfos = new UnconditionalFlowInfo[5];
        }
        else if (this.innerFlowContextsCount == 
          (length = this.innerFlowContexts.length) - 1) {
          System.arraycopy(this.innerFlowContexts, 0, 
            this.innerFlowContexts = new LoopingFlowContext[length + 5], 
            0, length);
          System.arraycopy(this.innerFlowInfos, 0, 
            this.innerFlowInfos = new UnconditionalFlowInfo[length + 5], 
            0, length);
        }
        this.innerFlowContexts[this.innerFlowContextsCount] = ((LoopingFlowContext)inner);
        this.innerFlowInfos[(this.innerFlowContextsCount++)] = 
          flowInfo.unconditionalInitsWithoutSideEffect();
      }
    }
  }

  protected boolean recordFinalAssignment(VariableBinding binding, Reference finalAssignment)
  {
    if ((binding instanceof LocalVariableBinding)) {
      Scope scope = ((LocalVariableBinding)binding).declaringScope;
      while ((scope = scope.parent) != null) {
        if (scope == this.associatedScope)
          return false;
      }
    }
    if (this.assignCount == 0) {
      this.finalAssignments = new Reference[5];
      this.finalVariables = new VariableBinding[5];
    } else {
      if (this.assignCount == this.finalAssignments.length)
        System.arraycopy(
          this.finalAssignments, 
          0, 
          this.finalAssignments = new Reference[this.assignCount * 2], 
          0, 
          this.assignCount);
      System.arraycopy(
        this.finalVariables, 
        0, 
        this.finalVariables = new VariableBinding[this.assignCount * 2], 
        0, 
        this.assignCount);
    }
    this.finalAssignments[this.assignCount] = finalAssignment;
    this.finalVariables[(this.assignCount++)] = binding;
    return true;
  }

  protected void recordNullReference(LocalVariableBinding local, Expression expression, int status)
  {
    if (this.nullCount == 0) {
      this.nullLocals = new LocalVariableBinding[5];
      this.nullReferences = new Expression[5];
      this.nullCheckTypes = new int[5];
    }
    else if (this.nullCount == this.nullLocals.length) {
      System.arraycopy(this.nullLocals, 0, 
        this.nullLocals = new LocalVariableBinding[this.nullCount * 2], 0, this.nullCount);
      System.arraycopy(this.nullReferences, 0, 
        this.nullReferences = new Expression[this.nullCount * 2], 0, this.nullCount);
      System.arraycopy(this.nullCheckTypes, 0, 
        this.nullCheckTypes = new int[this.nullCount * 2], 0, this.nullCount);
    }
    this.nullLocals[this.nullCount] = local;
    this.nullReferences[this.nullCount] = expression;
    this.nullCheckTypes[(this.nullCount++)] = status;
  }

  public void recordUsingNullReference(Scope scope, LocalVariableBinding local, Expression reference, int checkType, FlowInfo flowInfo)
  {
    if (((flowInfo.tagBits & 0x1) != 0) || 
      (flowInfo.isDefinitelyUnknown(local))) {
      return;
    }
    switch (checkType) {
    case 256:
    case 512:
      if (flowInfo.isDefinitelyNonNull(local)) {
        if (checkType == 512)
          scope.problemReporter().localVariableRedundantCheckOnNonNull(local, reference);
        else
          scope.problemReporter().localVariableNonNullComparedToNull(local, reference);
      }
      else if (flowInfo.isDefinitelyNull(local)) {
        if (checkType == 256)
          scope.problemReporter().localVariableRedundantCheckOnNull(local, reference);
        else
          scope.problemReporter().localVariableNullComparedToNonNull(local, reference);
      }
      else if (!flowInfo.cannotBeDefinitelyNullOrNonNull(local)) {
        if (flowInfo.isPotentiallyNonNull(local))
          recordNullReference(local, reference, 0x2 | checkType & 0xFFFFFF00);
        else {
          recordNullReference(local, reference, checkType);
        }
      }
      return;
    case 257:
    case 513:
    case 769:
    case 1025:
      if ((flowInfo.isPotentiallyNonNull(local)) || 
        (flowInfo.isPotentiallyUnknown(local))) {
        return;
      }
      if (flowInfo.isDefinitelyNull(local)) {
        switch (checkType & 0xFFFFFF00) {
        case 256:
          scope.problemReporter().localVariableRedundantCheckOnNull(local, reference);
          return;
        case 512:
          scope.problemReporter().localVariableNullComparedToNonNull(local, reference);
          return;
        case 768:
          scope.problemReporter().localVariableRedundantNullAssignment(local, reference);
          return;
        case 1024:
          scope.problemReporter().localVariableNullInstanceof(local, reference);
          return;
        }
      }
      recordNullReference(local, reference, checkType);
      return;
    case 3:
      if (flowInfo.isDefinitelyNonNull(local)) {
        return;
      }
      if (flowInfo.isDefinitelyNull(local)) {
        scope.problemReporter().localVariableNullReference(local, reference);
        return;
      }
      if (flowInfo.isPotentiallyNull(local)) {
        scope.problemReporter().localVariablePotentialNullReference(local, reference);
        return;
      }
      recordNullReference(local, reference, checkType);
      return;
    }
  }

  void removeFinalAssignmentIfAny(Reference reference)
  {
    for (int i = 0; i < this.assignCount; i++)
      if (this.finalAssignments[i] == reference) {
        this.finalAssignments[i] = null;
        this.finalVariables[i] = null;
        return;
      }
  }
}