package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ImportReference;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.env.AccessRestriction;
import org.eclipse.jdt.internal.compiler.env.ICompilationUnit;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.CompoundNameVector;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
import org.eclipse.jdt.internal.compiler.util.HashtableOfType;
import org.eclipse.jdt.internal.compiler.util.ObjectVector;
import org.eclipse.jdt.internal.compiler.util.SimpleNameVector;

public class CompilationUnitScope extends Scope
{
  public LookupEnvironment environment;
  public CompilationUnitDeclaration referenceContext;
  public char[][] currentPackageName;
  public PackageBinding fPackage;
  public ImportBinding[] imports;
  public HashtableOfObject typeOrPackageCache;
  public SourceTypeBinding[] topLevelTypes;
  private CompoundNameVector qualifiedReferences;
  private SimpleNameVector simpleNameReferences;
  private ObjectVector referencedTypes;
  private ObjectVector referencedSuperTypes;
  HashtableOfType constantPoolNameUsage;
  private int captureID = 1;

  public CompilationUnitScope(CompilationUnitDeclaration unit, LookupEnvironment environment) {
    super(4, null);
    this.environment = environment;
    this.referenceContext = unit;
    unit.scope = this;
    this.currentPackageName = (unit.currentPackage == null ? CharOperation.NO_CHAR_CHAR : unit.currentPackage.tokens);

    if (compilerOptions().produceReferenceInfo) {
      this.qualifiedReferences = new CompoundNameVector();
      this.simpleNameReferences = new SimpleNameVector();
      this.referencedTypes = new ObjectVector();
      this.referencedSuperTypes = new ObjectVector();
    } else {
      this.qualifiedReferences = null;
      this.simpleNameReferences = null;
      this.referencedTypes = null;
      this.referencedSuperTypes = null;
    }
  }

  void buildFieldsAndMethods() { int i = 0; for (int length = this.topLevelTypes.length; i < length; i++)
      this.topLevelTypes[i].scope.buildFieldsAndMethods(); }

  void buildTypeBindings(AccessRestriction accessRestriction) {
    this.topLevelTypes = new SourceTypeBinding[0];
    boolean firstIsSynthetic = false;
    if (this.referenceContext.compilationResult.compilationUnit != null) {
      char[][] expectedPackageName = this.referenceContext.compilationResult.compilationUnit.getPackageName();
      if ((expectedPackageName != null) && 
        (!CharOperation.equals(this.currentPackageName, expectedPackageName)))
      {
        if ((this.referenceContext.currentPackage != null) || 
          (this.referenceContext.types != null) || 
          (this.referenceContext.imports != null)) {
          problemReporter().packageIsNotExpectedPackage(this.referenceContext);
        }
        this.currentPackageName = (expectedPackageName.length == 0 ? CharOperation.NO_CHAR_CHAR : expectedPackageName);
      }
    }
    if (this.currentPackageName == CharOperation.NO_CHAR_CHAR) {
      if ((this.fPackage = this.environment.defaultPackage) == null)
        problemReporter().mustSpecifyPackage(this.referenceContext);
    }
    else
    {
      if ((this.fPackage = this.environment.createPackage(this.currentPackageName)) == null) {
        problemReporter().packageCollidesWithType(this.referenceContext);
        return;
      }if (this.referenceContext.isPackageInfo())
      {
        if ((this.referenceContext.types == null) || (this.referenceContext.types.length == 0)) {
          this.referenceContext.types = new TypeDeclaration[1];
          TypeDeclaration declaration = new TypeDeclaration(this.referenceContext.compilationResult);
          this.referenceContext.types[0] = declaration;
          declaration.name = TypeConstants.PACKAGE_INFO_NAME;
          declaration.modifiers = 512;
          firstIsSynthetic = true;
        }
      }
      recordQualifiedReference(this.currentPackageName);
    }

    TypeDeclaration[] types = this.referenceContext.types;
    int typeLength = types == null ? 0 : types.length;
    this.topLevelTypes = new SourceTypeBinding[typeLength];
    int count = 0;
    for (int i = 0; i < typeLength; i++) {
      TypeDeclaration typeDecl = types[i];
      ReferenceBinding typeBinding = this.fPackage.getType0(typeDecl.name);
      recordSimpleReference(typeDecl.name);
      if ((typeBinding != null) && (typeBinding.isValidBinding()) && (!(typeBinding instanceof UnresolvedReferenceBinding)))
      {
        problemReporter().duplicateTypes(this.referenceContext, typeDecl);
      }
      else {
        if ((this.fPackage != this.environment.defaultPackage) && (this.fPackage.getPackage(typeDecl.name) != null))
        {
          problemReporter().typeCollidesWithPackage(this.referenceContext, typeDecl);
        }

        if ((typeDecl.modifiers & 0x1) != 0)
        {
          char[] mainTypeName;
          if (((mainTypeName = this.referenceContext.getMainTypeName()) != null) && 
            (!CharOperation.equals(mainTypeName, typeDecl.name))) {
            problemReporter().publicClassMustMatchFileName(this.referenceContext, typeDecl);
          }

        }

        ClassScope child = new ClassScope(this, typeDecl);
        SourceTypeBinding type = child.buildType(null, this.fPackage, accessRestriction);
        if ((firstIsSynthetic) && (i == 0))
          type.modifiers |= 4096;
        if (type != null) {
          this.topLevelTypes[(count++)] = type;
        }
      }
    }
    if (count != this.topLevelTypes.length)
      System.arraycopy(this.topLevelTypes, 0, this.topLevelTypes = new SourceTypeBinding[count], 0, count); 
  }

  void checkAndSetImports() { if (this.referenceContext.imports == null) {
      this.imports = getDefaultImports();
      return;
    }

    int numberOfStatements = this.referenceContext.imports.length;
    int numberOfImports = numberOfStatements + 1;
    for (int i = 0; i < numberOfStatements; i++) {
      ImportReference importReference = this.referenceContext.imports[i];
      if (((importReference.bits & 0x20000) != 0) && (CharOperation.equals(JAVA_LANG, importReference.tokens)) && (!importReference.isStatic())) {
        numberOfImports--;
        break;
      }
    }
    ImportBinding[] resolvedImports = new ImportBinding[numberOfImports];
    resolvedImports[0] = getDefaultImports()[0];
    int index = 1;

    for (int i = 0; i < numberOfStatements; i++) {
      ImportReference importReference = this.referenceContext.imports[i];
      char[][] compoundName = importReference.tokens;

      for (int j = 0; j < index; j++) {
        ImportBinding resolved = resolvedImports[j];
        if (resolved.onDemand == ((importReference.bits & 0x20000) != 0)) if ((resolved.isStatic() == importReference.isStatic()) && 
            (CharOperation.equals(compoundName, resolvedImports[j].compoundName))) {
            break;
          }
      }
      if ((importReference.bits & 0x20000) != 0) {
        if (!CharOperation.equals(compoundName, this.currentPackageName))
        {
          Binding importBinding = findImport(compoundName, compoundName.length);
          if ((importBinding.isValidBinding()) && ((!importReference.isStatic()) || (!(importBinding instanceof PackageBinding))))
          {
            resolvedImports[(index++)] = new ImportBinding(compoundName, true, importBinding, importReference);
          }
        }
      } else resolvedImports[(index++)] = new ImportBinding(compoundName, false, null, importReference);

    }

    if (resolvedImports.length > index)
      System.arraycopy(resolvedImports, 0, resolvedImports = new ImportBinding[index], 0, index);
    this.imports = resolvedImports;
  }

  void checkParameterizedTypes()
  {
    if (compilerOptions().sourceLevel < 3211264L) return;

    int i = 0; for (int length = this.topLevelTypes.length; i < length; i++) {
      ClassScope scope = this.topLevelTypes[i].scope;
      scope.checkParameterizedTypeBounds();
      scope.checkParameterizedSuperTypeCollisions();
    }
  }

  public char[] computeConstantPoolName(LocalTypeBinding localType)
  {
    if (localType.constantPoolName() != null) {
      return localType.constantPoolName();
    }

    if (this.constantPoolNameUsage == null) {
      this.constantPoolNameUsage = new HashtableOfType();
    }
    ReferenceBinding outerMostEnclosingType = localType.scope.outerMostClassScope().enclosingSourceType();

    int index = 0;

    boolean isCompliant15 = compilerOptions().complianceLevel >= 3211264L;
    char[] candidateName;
    while (true)
    {
      char[] candidateName;
      if (localType.isMemberType())
      {
        char[] candidateName;
        if (index == 0) {
          candidateName = CharOperation.concat(
            localType.enclosingType().constantPoolName(), 
            localType.sourceName, 
            '$');
        }
        else
        {
          candidateName = CharOperation.concat(
            localType.enclosingType().constantPoolName(), 
            '$', 
            String.valueOf(index).toCharArray(), 
            '$', 
            localType.sourceName);
        }
      }
      else
      {
        char[] candidateName;
        if (localType.isAnonymousType())
        {
          char[] candidateName;
          if (isCompliant15)
          {
            candidateName = CharOperation.concat(
              localType.enclosingType.constantPoolName(), 
              String.valueOf(index + 1).toCharArray(), 
              '$');
          }
          else candidateName = CharOperation.concat(
              outerMostEnclosingType.constantPoolName(), 
              String.valueOf(index + 1).toCharArray(), 
              '$');
        }
        else
        {
          char[] candidateName;
          if (isCompliant15)
            candidateName = CharOperation.concat(
              CharOperation.concat(
              localType.enclosingType().constantPoolName(), 
              String.valueOf(index + 1).toCharArray(), 
              '$'), 
              localType.sourceName);
          else
            candidateName = CharOperation.concat(
              outerMostEnclosingType.constantPoolName(), 
              '$', 
              String.valueOf(index + 1).toCharArray(), 
              '$', 
              localType.sourceName);
        }
      }
      if (this.constantPoolNameUsage.get(candidateName) == null) break;
      index++;
    }
    this.constantPoolNameUsage.put(candidateName, localType);

    return candidateName;
  }

  void connectTypeHierarchy() {
    int i = 0; for (int length = this.topLevelTypes.length; i < length; i++)
      this.topLevelTypes[i].scope.connectTypeHierarchy(); 
  }

  void faultInImports() { if (this.typeOrPackageCache != null)
      return;
    if (this.referenceContext.imports == null) {
      this.typeOrPackageCache = new HashtableOfObject(1);
      return;
    }

    int numberOfStatements = this.referenceContext.imports.length;
    HashtableOfType typesBySimpleNames = null;
    for (int i = 0; i < numberOfStatements; i++) {
      if ((this.referenceContext.imports[i].bits & 0x20000) == 0) {
        typesBySimpleNames = new HashtableOfType(this.topLevelTypes.length + numberOfStatements);
        int j = 0; for (int length = this.topLevelTypes.length; j < length; j++)
          typesBySimpleNames.put(this.topLevelTypes[j].sourceName, this.topLevelTypes[j]);
        break;
      }

    }

    int numberOfImports = numberOfStatements + 1;
    for (int i = 0; i < numberOfStatements; i++) {
      ImportReference importReference = this.referenceContext.imports[i];
      if (((importReference.bits & 0x20000) != 0) && (CharOperation.equals(JAVA_LANG, importReference.tokens)) && (!importReference.isStatic())) {
        numberOfImports--;
        break;
      }
    }
    ImportBinding[] resolvedImports = new ImportBinding[numberOfImports];
    resolvedImports[0] = getDefaultImports()[0];
    int index = 1;

    for (int i = 0; i < numberOfStatements; i++) {
      ImportReference importReference = this.referenceContext.imports[i];
      char[][] compoundName = importReference.tokens;

      for (int j = 0; j < index; j++) {
        ImportBinding resolved = resolvedImports[j];
        if (resolved.onDemand == ((importReference.bits & 0x20000) != 0)) if ((resolved.isStatic() == importReference.isStatic()) && 
            (CharOperation.equals(compoundName, resolved.compoundName))) {
            problemReporter().unusedImport(importReference);
            break;
          }
      }

      if ((importReference.bits & 0x20000) != 0) {
        if (CharOperation.equals(compoundName, this.currentPackageName)) {
          problemReporter().unusedImport(importReference);
        }
        else
        {
          Binding importBinding = findImport(compoundName, compoundName.length);
          if (!importBinding.isValidBinding()) {
            problemReporter().importProblem(importReference, importBinding);
          }
          else if ((importReference.isStatic()) && ((importBinding instanceof PackageBinding))) {
            problemReporter().cannotImportPackage(importReference);
          }
          else
            resolvedImports[(index++)] = new ImportBinding(compoundName, true, importBinding, importReference); 
        }
      } else { Binding importBinding = findSingleImport(compoundName, importReference.isStatic());
        if (!importBinding.isValidBinding()) {
          problemReporter().importProblem(importReference, importBinding);
        }
        else if ((importBinding instanceof PackageBinding)) {
          problemReporter().cannotImportPackage(importReference);
        }
        else {
          ReferenceBinding conflictingType = null;
          if ((importBinding instanceof MethodBinding)) {
            conflictingType = (ReferenceBinding)getType(compoundName, compoundName.length);
            if (!conflictingType.isValidBinding()) {
              conflictingType = null;
            }
          }
          if (((importBinding instanceof ReferenceBinding)) || (conflictingType != null)) {
            ReferenceBinding referenceBinding = conflictingType == null ? (ReferenceBinding)importBinding : conflictingType;
            if (importReference.isTypeUseDeprecated(referenceBinding, this)) {
              problemReporter().deprecatedType(referenceBinding, importReference);
            }
            ReferenceBinding existingType = typesBySimpleNames.get(compoundName[(compoundName.length - 1)]);
            if (existingType != null)
            {
              if (existingType == referenceBinding) {
                continue;
              }
              int j = 0; for (int length = this.topLevelTypes.length; j < length; j++) {
                if (CharOperation.equals(this.topLevelTypes[j].sourceName, existingType.sourceName)) {
                  problemReporter().conflictingImport(importReference);
                  break;
                }
              }
              problemReporter().duplicateImport(importReference);
              continue;
            }
            typesBySimpleNames.put(compoundName[(compoundName.length - 1)], referenceBinding);
          } else if ((importBinding instanceof FieldBinding)) {
            for (int j = 0; j < index; j++) {
              ImportBinding resolved = resolvedImports[j];

              if ((resolved.isStatic()) && ((resolved.resolvedImport instanceof FieldBinding)) && (importBinding != resolved.resolvedImport) && 
                (CharOperation.equals(compoundName[(compoundName.length - 1)], resolved.compoundName[(resolved.compoundName.length - 1)]))) {
                problemReporter().duplicateImport(importReference);
                break;
              }
            }
          }

          resolvedImports[(index++)] = (conflictingType == null ? 
            new ImportBinding(compoundName, false, importBinding, importReference) : 
            new ImportConflictBinding(compoundName, importBinding, conflictingType, importReference));
        }
      }
    }

    if (resolvedImports.length > index)
      System.arraycopy(resolvedImports, 0, resolvedImports = new ImportBinding[index], 0, index);
    this.imports = resolvedImports;

    int length = this.imports.length;
    this.typeOrPackageCache = new HashtableOfObject(length);
    for (int i = 0; i < length; i++) {
      ImportBinding binding = this.imports[i];
      if (((!binding.onDemand) && ((binding.resolvedImport instanceof ReferenceBinding))) || ((binding instanceof ImportConflictBinding)))
        this.typeOrPackageCache.put(binding.compoundName[(binding.compoundName.length - 1)], binding);
    } }

  public void faultInTypes() {
    faultInImports();

    int i = 0; for (int length = this.topLevelTypes.length; i < length; i++)
      this.topLevelTypes[i].faultInTypesForFieldsAndMethods();
  }

  public Binding findImport(char[][] compoundName, boolean findStaticImports, boolean onDemand) {
    if (onDemand) {
      return findImport(compoundName, compoundName.length);
    }
    return findSingleImport(compoundName, findStaticImports);
  }

  private Binding findImport(char[][] compoundName, int length) {
    recordQualifiedReference(compoundName);

    Binding binding = this.environment.getTopLevelPackage(compoundName[0]);
    int i = 1;
    if (binding != null) {
      PackageBinding packageBinding = (PackageBinding)binding;
      while (i < length) {
        binding = packageBinding.getTypeOrPackage(compoundName[(i++)]);
        if ((binding == null) || (!binding.isValidBinding())) {
          binding = null;
          break;
        }
        if (!(binding instanceof PackageBinding)) {
          break;
        }
        packageBinding = (PackageBinding)binding;
      }
      return packageBinding;
    }
    ReferenceBinding type;
    if (binding == null) {
      if ((this.environment.defaultPackage == null) || (compilerOptions().complianceLevel >= 3145728L))
        return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, i), null, 1);
      ReferenceBinding type = findType(compoundName[0], this.environment.defaultPackage, this.environment.defaultPackage);
      if ((type == null) || (!type.isValidBinding()))
        return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, i), null, 1);
      i = 1;
    } else {
      type = (ReferenceBinding)binding;
    }

    while (i < length) {
      type = (ReferenceBinding)this.environment.convertToRawType(type);
      if (!type.canBeSeenBy(this.fPackage)) {
        return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, i), type, 2);
      }
      char[] name = compoundName[(i++)];

      type = type.getMemberType(name);
      if (type == null)
        return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, i), null, 1);
    }
    if (!type.canBeSeenBy(this.fPackage))
      return new ProblemReferenceBinding(compoundName, type, 2);
    return type;
  }
  private Binding findSingleImport(char[][] compoundName, boolean findStaticImports) {
    if (compoundName.length == 1)
    {
      if ((this.environment.defaultPackage == null) || (compilerOptions().complianceLevel >= 3145728L))
        return new ProblemReferenceBinding(compoundName, null, 1);
      ReferenceBinding typeBinding = findType(compoundName[0], this.environment.defaultPackage, this.fPackage);
      if (typeBinding == null)
        return new ProblemReferenceBinding(compoundName, null, 1);
      return typeBinding;
    }

    if (findStaticImports)
      return findSingleStaticImport(compoundName);
    return findImport(compoundName, compoundName.length);
  }
  private Binding findSingleStaticImport(char[][] compoundName) {
    Binding binding = findImport(compoundName, compoundName.length - 1);
    if (!binding.isValidBinding()) return binding;

    char[] name = compoundName[(compoundName.length - 1)];
    if ((binding instanceof PackageBinding)) {
      Binding temp = ((PackageBinding)binding).getTypeOrPackage(name);
      if ((temp != null) && ((temp instanceof ReferenceBinding)))
        return new ProblemReferenceBinding(compoundName, (ReferenceBinding)temp, 14);
      return binding;
    }

    ReferenceBinding type = (ReferenceBinding)binding;
    FieldBinding field = findField(type, name, null, true);
    if ((field != null) && (field.isValidBinding()) && (field.isStatic()) && (field.canBeSeenBy(type, null, this))) {
      return field;
    }

    MethodBinding method = findStaticMethod(type, name);
    if (method != null) return method;

    type = findMemberType(name, type);
    if ((type == null) || (!type.isStatic())) {
      if ((field != null) && (!field.isValidBinding()) && (field.problemId() != 1))
        return field;
      return new ProblemReferenceBinding(compoundName, type, 1);
    }
    if (!type.canBeSeenBy(this.fPackage))
      return new ProblemReferenceBinding(compoundName, type, 2);
    return type;
  }
  MethodBinding findStaticMethod(ReferenceBinding currentType, char[] selector) {
    if (!currentType.canBeSeenBy(this))
      return null;
    do
    {
      MethodBinding[] methods = currentType.getMethods(selector);
      if (methods != Binding.NO_METHODS) {
        int i = methods.length;
        do { MethodBinding method = methods[i];
          if ((method.isStatic()) && (method.canBeSeenBy(this.fPackage)))
            return method;
          i--; } while (i >= 0);
      }

      if (currentType.superInterfaces() == null)
        ((SourceTypeBinding)currentType).scope.connectTypeHierarchy(); 
    }
    while ((currentType = currentType.superclass()) != null);
    return null;
  }

  ImportBinding[] getDefaultImports() {
    if (this.environment.defaultImports != null) return this.environment.defaultImports;

    Binding importBinding = this.environment.getTopLevelPackage(JAVA);
    if (importBinding != null) {
      importBinding = ((PackageBinding)importBinding).getTypeOrPackage(JAVA_LANG[1]);
    }
    if ((importBinding == null) || (!importBinding.isValidBinding()))
    {
      BinaryTypeBinding missingObject = this.environment.cacheMissingBinaryType(JAVA_LANG_OBJECT, this.referenceContext);
      importBinding = missingObject.fPackage;
    }

    return this.environment.defaultImports = new ImportBinding[] { new ImportBinding(JAVA_LANG, true, importBinding, null) };
  }

  public final Binding getImport(char[][] compoundName, boolean onDemand, boolean isStaticImport) {
    if (onDemand)
      return findImport(compoundName, compoundName.length);
    return findSingleImport(compoundName, isStaticImport);
  }

  public int nextCaptureID() {
    return this.captureID++;
  }

  public ProblemReporter problemReporter()
  {
    ProblemReporter problemReporter = this.referenceContext.problemReporter;
    problemReporter.referenceContext = this.referenceContext;
    return problemReporter;
  }

  void recordQualifiedReference(char[][] qualifiedName)
  {
    if (this.qualifiedReferences == null) return;

    int length = qualifiedName.length;
    if (length > 1)
      while (!this.qualifiedReferences.contains(qualifiedName)) {
        this.qualifiedReferences.add(qualifiedName);
        if (length == 2) {
          recordSimpleReference(qualifiedName[0]);
          recordSimpleReference(qualifiedName[1]);
          return;
        }
        length--;
        recordSimpleReference(qualifiedName[length]);
        System.arraycopy(qualifiedName, 0, qualifiedName = new char[length][], 0, length);
      }
    else if (length == 1)
      recordSimpleReference(qualifiedName[0]);
  }

  void recordReference(char[][] qualifiedEnclosingName, char[] simpleName) {
    recordQualifiedReference(qualifiedEnclosingName);
    recordSimpleReference(simpleName);
  }
  void recordReference(ReferenceBinding type, char[] simpleName) {
    ReferenceBinding actualType = typeToRecord(type);
    if (actualType != null)
      recordReference(actualType.compoundName, simpleName); 
  }

  void recordSimpleReference(char[] simpleName) { if (this.simpleNameReferences == null) return;

    if (!this.simpleNameReferences.contains(simpleName))
      this.simpleNameReferences.add(simpleName); }

  void recordSuperTypeReference(TypeBinding type) {
    if (this.referencedSuperTypes == null) return;

    ReferenceBinding actualType = typeToRecord(type);
    if ((actualType != null) && (!this.referencedSuperTypes.containsIdentical(actualType)))
      this.referencedSuperTypes.add(actualType); 
  }

  public void recordTypeConversion(TypeBinding superType, TypeBinding subType) { recordSuperTypeReference(subType); }

  void recordTypeReference(TypeBinding type) {
    if (this.referencedTypes == null) return;

    ReferenceBinding actualType = typeToRecord(type);
    if ((actualType != null) && (!this.referencedTypes.containsIdentical(actualType)))
      this.referencedTypes.add(actualType); 
  }

  void recordTypeReferences(TypeBinding[] types) { if (this.referencedTypes == null) return;
    if ((types == null) || (types.length == 0)) return;

    int i = 0; for (int max = types.length; i < max; i++)
    {
      ReferenceBinding actualType = typeToRecord(types[i]);
      if ((actualType != null) && (!this.referencedTypes.containsIdentical(actualType)))
        this.referencedTypes.add(actualType);
    } }

  Binding resolveSingleImport(ImportBinding importBinding) {
    if (importBinding.resolvedImport == null) {
      importBinding.resolvedImport = findSingleImport(importBinding.compoundName, importBinding.isStatic());
      if ((!importBinding.resolvedImport.isValidBinding()) || ((importBinding.resolvedImport instanceof PackageBinding))) {
        if (this.imports != null) {
          ImportBinding[] newImports = new ImportBinding[this.imports.length - 1];
          int i = 0; int n = 0; for (int max = this.imports.length; i < max; i++)
            if (this.imports[i] != importBinding)
              newImports[(n++)] = this.imports[i];
          this.imports = newImports;
        }
        return null;
      }
    }
    return importBinding.resolvedImport;
  }

  public void storeDependencyInfo()
  {
    for (int i = 0; i < this.referencedSuperTypes.size; i++) {
      ReferenceBinding type = (ReferenceBinding)this.referencedSuperTypes.elementAt(i);
      if (!this.referencedTypes.containsIdentical(type)) {
        this.referencedTypes.add(type);
      }
      if (!type.isLocalType()) {
        ReferenceBinding enclosing = type.enclosingType();
        if (enclosing != null)
          recordSuperTypeReference(enclosing);
      }
      ReferenceBinding superclass = type.superclass();
      if (superclass != null)
        recordSuperTypeReference(superclass);
      ReferenceBinding[] interfaces = type.superInterfaces();
      if (interfaces != null) {
        int j = 0; for (int length = interfaces.length; j < length; j++)
          recordSuperTypeReference(interfaces[j]);
      }
    }
    int i = 0; for (int l = this.referencedTypes.size; i < l; i++) {
      ReferenceBinding type = (ReferenceBinding)this.referencedTypes.elementAt(i);
      if (!type.isLocalType()) {
        recordQualifiedReference(type.isMemberType() ? 
          CharOperation.splitOn('.', type.readableName()) : 
          type.compoundName);
      }
    }
    int size = this.qualifiedReferences.size;
    char[][][] qualifiedRefs = new char[size][][];
    for (int i = 0; i < size; i++)
      qualifiedRefs[i] = this.qualifiedReferences.elementAt(i);
    this.referenceContext.compilationResult.qualifiedReferences = qualifiedRefs;

    size = this.simpleNameReferences.size;
    char[][] simpleRefs = new char[size][];
    for (int i = 0; i < size; i++)
      simpleRefs[i] = this.simpleNameReferences.elementAt(i);
    this.referenceContext.compilationResult.simpleNameReferences = simpleRefs;
  }
  public String toString() {
    return "--- CompilationUnit Scope : " + new String(this.referenceContext.getFileName());
  }
  private ReferenceBinding typeToRecord(TypeBinding type) {
    if (type.isArrayType()) {
      type = ((ArrayBinding)type).leafComponentType;
    }
    switch (type.kind()) {
    case 132:
    case 516:
    case 4100:
      return null;
    case 260:
    case 1028:
      type = type.erasure();
    }
    ReferenceBinding refType = (ReferenceBinding)type;
    if (refType.isLocalType()) return null;
    return refType;
  }
  public void verifyMethods(MethodVerifier verifier) {
    int i = 0; for (int length = this.topLevelTypes.length; i < length; i++)
      this.topLevelTypes[i].verifyMethods(verifier);
  }
}