package org.eclipse.jdt.internal.compiler.lookup;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ClassFilePool;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.env.AccessRestriction;
import org.eclipse.jdt.internal.compiler.env.IBinaryType;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;
import org.eclipse.jdt.internal.compiler.env.NameEnvironmentAnswer;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.ITypeRequestor;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.HashtableOfPackage;
import org.eclipse.jdt.internal.compiler.util.SimpleLookupTable;

public class LookupEnvironment
  implements ProblemReasons, TypeConstants
{
  static final int BUILD_FIELDS_AND_METHODS = 4;
  static final int BUILD_TYPE_HIERARCHY = 1;
  static final int CHECK_AND_SET_IMPORTS = 2;
  static final int CONNECT_TYPE_HIERARCHY = 3;
  static final ProblemPackageBinding TheNotFoundPackage = new ProblemPackageBinding(CharOperation.NO_CHAR, 1);
  static final ProblemReferenceBinding TheNotFoundType = new ProblemReferenceBinding(CharOperation.NO_CHAR, null, 1);
  private Map accessRestrictions;
  ImportBinding[] defaultImports;
  public PackageBinding defaultPackage;
  HashtableOfPackage knownPackages;
  private int lastCompletedUnitIndex = -1;
  private int lastUnitIndex = -1;
  public INameEnvironment nameEnvironment;
  public CompilerOptions globalOptions;
  public ProblemReporter problemReporter;
  public ClassFilePool classFilePool;
  private int stepCompleted;
  public ITypeRequestor typeRequestor;
  private ArrayBinding[][] uniqueArrayBindings;
  private SimpleLookupTable uniqueParameterizedTypeBindings;
  private SimpleLookupTable uniqueRawTypeBindings;
  private SimpleLookupTable uniqueWildcardBindings;
  private SimpleLookupTable uniqueParameterizedGenericMethodBindings;
  public CompilationUnitDeclaration unitBeingCompleted = null;
  public Object missingClassFileLocation = null;

  private CompilationUnitDeclaration[] units = new CompilationUnitDeclaration[4];
  private MethodVerifier verifier;

  public LookupEnvironment(ITypeRequestor typeRequestor, CompilerOptions globalOptions, ProblemReporter problemReporter, INameEnvironment nameEnvironment)
  {
    this.typeRequestor = typeRequestor;
    this.globalOptions = globalOptions;
    this.problemReporter = problemReporter;
    this.defaultPackage = new PackageBinding(this);
    this.defaultImports = null;
    this.nameEnvironment = nameEnvironment;
    this.knownPackages = new HashtableOfPackage();
    this.uniqueArrayBindings = new ArrayBinding[5][];
    this.uniqueArrayBindings[0] = new ArrayBinding[50];
    this.uniqueParameterizedTypeBindings = new SimpleLookupTable(3);
    this.uniqueRawTypeBindings = new SimpleLookupTable(3);
    this.uniqueWildcardBindings = new SimpleLookupTable(3);
    this.uniqueParameterizedGenericMethodBindings = new SimpleLookupTable(3);
    this.accessRestrictions = new HashMap(3);
    this.classFilePool = ClassFilePool.newInstance();
  }

  public ReferenceBinding askForType(char[][] compoundName)
  {
    NameEnvironmentAnswer answer = this.nameEnvironment.findType(compoundName);
    if (answer == null) {
      return null;
    }
    if (answer.isBinaryType())
    {
      this.typeRequestor.accept(answer.getBinaryType(), computePackageFrom(compoundName), answer.getAccessRestriction());
    } else if (answer.isCompilationUnit())
    {
      this.typeRequestor.accept(answer.getCompilationUnit(), answer.getAccessRestriction());
    } else if (answer.isSourceType())
    {
      this.typeRequestor.accept(answer.getSourceTypes(), computePackageFrom(compoundName), answer.getAccessRestriction());
    }
    return getCachedType(compoundName);
  }

  ReferenceBinding askForType(PackageBinding packageBinding, char[] name)
  {
    if (packageBinding == null) {
      if (this.defaultPackage == null)
        return null;
      packageBinding = this.defaultPackage;
    }
    NameEnvironmentAnswer answer = this.nameEnvironment.findType(name, packageBinding.compoundName);
    if (answer == null) {
      return null;
    }
    if (answer.isBinaryType())
    {
      this.typeRequestor.accept(answer.getBinaryType(), packageBinding, answer.getAccessRestriction());
    } else if (answer.isCompilationUnit())
    {
      this.typeRequestor.accept(answer.getCompilationUnit(), answer.getAccessRestriction());
    } else if (answer.isSourceType())
    {
      this.typeRequestor.accept(answer.getSourceTypes(), packageBinding, answer.getAccessRestriction());
    }
    return packageBinding.getType0(name);
  }

  public void buildTypeBindings(CompilationUnitDeclaration unit, AccessRestriction accessRestriction)
  {
    CompilationUnitScope scope = new CompilationUnitScope(unit, this);
    scope.buildTypeBindings(accessRestriction);

    int unitsLength = this.units.length;
    if (++this.lastUnitIndex >= unitsLength)
      System.arraycopy(this.units, 0, this.units = new CompilationUnitDeclaration[2 * unitsLength], 0, unitsLength);
    this.units[this.lastUnitIndex] = unit;
  }

  public BinaryTypeBinding cacheBinaryType(IBinaryType binaryType, AccessRestriction accessRestriction)
  {
    return cacheBinaryType(binaryType, true, accessRestriction);
  }

  public BinaryTypeBinding cacheBinaryType(IBinaryType binaryType, boolean needFieldsAndMethods, AccessRestriction accessRestriction)
  {
    char[][] compoundName = CharOperation.splitOn('/', binaryType.getName());
    ReferenceBinding existingType = getCachedType(compoundName);

    if ((existingType == null) || ((existingType instanceof UnresolvedReferenceBinding)))
    {
      return createBinaryTypeFrom(binaryType, computePackageFrom(compoundName), needFieldsAndMethods, accessRestriction);
    }return null;
  }

  public BinaryTypeBinding cacheMissingBinaryType(char[][] compoundName, CompilationUnitDeclaration unit) {
    this.problemReporter.isClassPathCorrect(
      compoundName, 
      unit == null ? this.unitBeingCompleted : unit, 
      this.missingClassFileLocation);

    PackageBinding packageBinding = computePackageFrom(compoundName);

    MissingBinaryTypeBinding type = new MissingBinaryTypeBinding(packageBinding, compoundName, this);
    if (type.id != 1)
    {
      ReferenceBinding objectType = getType(TypeConstants.JAVA_LANG_OBJECT);
      if (objectType == null)
        objectType = cacheMissingBinaryType(TypeConstants.JAVA_LANG_OBJECT, unit);
      type.setMissingSuperclass(objectType);
    }
    packageBinding.addType(type);
    return type;
  }

  public void completeTypeBindings()
  {
    this.stepCompleted = 1;

    for (int i = this.lastCompletedUnitIndex + 1; i <= this.lastUnitIndex; i++) {
      (this.unitBeingCompleted = this.units[i]).scope.checkAndSetImports();
    }
    this.stepCompleted = 2;

    for (int i = this.lastCompletedUnitIndex + 1; i <= this.lastUnitIndex; i++) {
      (this.unitBeingCompleted = this.units[i]).scope.connectTypeHierarchy();
    }
    this.stepCompleted = 3;

    for (int i = this.lastCompletedUnitIndex + 1; i <= this.lastUnitIndex; i++) {
      CompilationUnitScope unitScope = (this.unitBeingCompleted = this.units[i]).scope;
      unitScope.checkParameterizedTypes();
      unitScope.buildFieldsAndMethods();
      this.units[i] = null;
    }
    this.stepCompleted = 4;
    this.lastCompletedUnitIndex = this.lastUnitIndex;
    this.unitBeingCompleted = null;
  }

  public void completeTypeBindings(CompilationUnitDeclaration parsedUnit)
  {
    if (this.stepCompleted == 4)
    {
      completeTypeBindings();
    } else {
      if (parsedUnit.scope == null) return;

      if (this.stepCompleted >= 2) {
        (this.unitBeingCompleted = parsedUnit).scope.checkAndSetImports();
      }
      if (this.stepCompleted >= 3) {
        (this.unitBeingCompleted = parsedUnit).scope.connectTypeHierarchy();
      }
      this.unitBeingCompleted = null;
    }
  }

  public void completeTypeBindings(CompilationUnitDeclaration parsedUnit, boolean buildFieldsAndMethods)
  {
    if (parsedUnit.scope == null) return;

    (this.unitBeingCompleted = parsedUnit).scope.checkAndSetImports();
    parsedUnit.scope.connectTypeHierarchy();
    parsedUnit.scope.checkParameterizedTypes();
    if (buildFieldsAndMethods)
      parsedUnit.scope.buildFieldsAndMethods();
    this.unitBeingCompleted = null;
  }

  public TypeBinding computeBoxingType(TypeBinding type) {
    switch (type.id) {
    case 33:
      return TypeBinding.BOOLEAN;
    case 26:
      return TypeBinding.BYTE;
    case 28:
      return TypeBinding.CHAR;
    case 27:
      return TypeBinding.SHORT;
    case 32:
      return TypeBinding.DOUBLE;
    case 31:
      return TypeBinding.FLOAT;
    case 29:
      return TypeBinding.INT;
    case 30:
      return TypeBinding.LONG;
    case 10:
      TypeBinding boxedType = getType(JAVA_LANG_INTEGER);
      if (boxedType != null) return boxedType;
      return new ProblemReferenceBinding(JAVA_LANG_INTEGER, null, 1);
    case 3:
      TypeBinding boxedType = getType(JAVA_LANG_BYTE);
      if (boxedType != null) return boxedType;
      return new ProblemReferenceBinding(JAVA_LANG_BYTE, null, 1);
    case 4:
      TypeBinding boxedType = getType(JAVA_LANG_SHORT);
      if (boxedType != null) return boxedType;
      return new ProblemReferenceBinding(JAVA_LANG_SHORT, null, 1);
    case 2:
      TypeBinding boxedType = getType(JAVA_LANG_CHARACTER);
      if (boxedType != null) return boxedType;
      return new ProblemReferenceBinding(JAVA_LANG_CHARACTER, null, 1);
    case 7:
      TypeBinding boxedType = getType(JAVA_LANG_LONG);
      if (boxedType != null) return boxedType;
      return new ProblemReferenceBinding(JAVA_LANG_LONG, null, 1);
    case 9:
      TypeBinding boxedType = getType(JAVA_LANG_FLOAT);
      if (boxedType != null) return boxedType;
      return new ProblemReferenceBinding(JAVA_LANG_FLOAT, null, 1);
    case 8:
      TypeBinding boxedType = getType(JAVA_LANG_DOUBLE);
      if (boxedType != null) return boxedType;
      return new ProblemReferenceBinding(JAVA_LANG_DOUBLE, null, 1);
    case 5:
      TypeBinding boxedType = getType(JAVA_LANG_BOOLEAN);
      if (boxedType != null) return boxedType;
      return new ProblemReferenceBinding(JAVA_LANG_BOOLEAN, null, 1);
    case 6:
    case 11:
    case 12:
    case 13:
    case 14:
    case 15:
    case 16:
    case 17:
    case 18:
    case 19:
    case 20:
    case 21:
    case 22:
    case 23:
    case 24:
    case 25:
    }

    switch (type.kind()) {
    case 516:
    case 4100:
      switch (type.erasure().id) {
      case 33:
        return TypeBinding.BOOLEAN;
      case 26:
        return TypeBinding.BYTE;
      case 28:
        return TypeBinding.CHAR;
      case 27:
        return TypeBinding.SHORT;
      case 32:
        return TypeBinding.DOUBLE;
      case 31:
        return TypeBinding.FLOAT;
      case 29:
        return TypeBinding.INT;
      case 30:
        return TypeBinding.LONG;
      }break;
    }
    return type;
  }
  private PackageBinding computePackageFrom(char[][] constantPoolName) {
    if (constantPoolName.length == 1) {
      return this.defaultPackage;
    }
    PackageBinding packageBinding = getPackage0(constantPoolName[0]);
    if ((packageBinding == null) || (packageBinding == TheNotFoundPackage)) {
      packageBinding = new PackageBinding(constantPoolName[0], this);
      this.knownPackages.put(constantPoolName[0], packageBinding);
    }

    int i = 1; for (int length = constantPoolName.length - 1; i < length; i++) {
      PackageBinding parent = packageBinding;
      if (((packageBinding = parent.getPackage0(constantPoolName[i])) == null) || (packageBinding == TheNotFoundPackage)) {
        packageBinding = new PackageBinding(CharOperation.subarray(constantPoolName, 0, i + 1), parent, this);
        parent.addPackage(packageBinding);
      }
    }
    return packageBinding;
  }

  public ReferenceBinding convertToParameterizedType(ReferenceBinding originalType)
  {
    if (originalType != null) {
      boolean isGeneric = originalType.isGenericType();
      ReferenceBinding originalEnclosingType = originalType.enclosingType();
      ReferenceBinding convertedEnclosingType = originalEnclosingType;
      boolean needToConvert = isGeneric;
      if (originalEnclosingType != null) {
        convertedEnclosingType = originalType.isStatic() ? 
          (ReferenceBinding)convertToRawType(originalEnclosingType) : 
          convertToParameterizedType(originalEnclosingType);
        needToConvert |= originalEnclosingType != convertedEnclosingType;
      }
      if (needToConvert) {
        return createParameterizedType(originalType, isGeneric ? originalType.typeVariables() : null, convertedEnclosingType);
      }
    }
    return originalType;
  }

  public TypeBinding convertToRawType(TypeBinding type)
  {
    TypeBinding originalType;
    int dimension;
    TypeBinding originalType;
    switch (type.kind()) {
    case 132:
    case 516:
    case 1028:
    case 4100:
      return type;
    case 68:
      int dimension = type.dimensions();
      originalType = type.leafComponentType();
      break;
    default:
      if (type.id == 1)
        return type;
      dimension = 0;
      originalType = type;
    }
    boolean needToConvert;
    boolean needToConvert;
    boolean needToConvert;
    switch (originalType.kind()) {
    case 132:
      return type;
    case 2052:
      needToConvert = true;
      break;
    case 260:
      ParameterizedTypeBinding paramType = (ParameterizedTypeBinding)originalType;
      needToConvert = paramType.genericType().isGenericType();
      break;
    default:
      needToConvert = false;
    }

    ReferenceBinding originalEnclosing = originalType.enclosingType();
    TypeBinding convertedType;
    TypeBinding convertedType;
    if (originalEnclosing == null) {
      convertedType = needToConvert ? createRawType((ReferenceBinding)originalType.erasure(), null) : originalType;
    }
    else
    {
      ReferenceBinding convertedEnclosing;
      ReferenceBinding convertedEnclosing;
      if (originalEnclosing.kind() == 1028) {
        needToConvert |= !((ReferenceBinding)originalType).isStatic();
        convertedEnclosing = originalEnclosing;
      }
      else
      {
        ReferenceBinding convertedEnclosing;
        if ((needToConvert) || (((ReferenceBinding)originalType).isStatic()))
          convertedEnclosing = (ReferenceBinding)convertToRawType(originalEnclosing);
        else
          convertedEnclosing = convertToParameterizedType(originalEnclosing);
      }
      TypeBinding convertedType;
      if (needToConvert) {
        convertedType = createRawType((ReferenceBinding)originalType.erasure(), convertedEnclosing);
      }
      else
      {
        TypeBinding convertedType;
        if (originalEnclosing != convertedEnclosing)
          convertedType = createParameterizedType((ReferenceBinding)originalType.erasure(), null, convertedEnclosing);
        else
          convertedType = originalType;
      }
    }
    if (originalType != convertedType) {
      return dimension > 0 ? createArrayType(convertedType, dimension) : convertedType;
    }
    return type;
  }

  public TypeBinding convertUnresolvedBinaryToRawType(TypeBinding type)
  {
    TypeBinding originalType;
    int dimension;
    TypeBinding originalType;
    switch (type.kind()) {
    case 132:
    case 516:
    case 1028:
    case 4100:
      return type;
    case 68:
      int dimension = type.dimensions();
      originalType = type.leafComponentType();
      break;
    default:
      if (type.id == 1)
        return type;
      dimension = 0;
      originalType = type;
    }
    boolean needToConvert;
    boolean needToConvert;
    boolean needToConvert;
    switch (originalType.kind()) {
    case 132:
      return type;
    case 2052:
      needToConvert = true;
      break;
    case 260:
      ParameterizedTypeBinding paramType = (ParameterizedTypeBinding)originalType;
      needToConvert = paramType.genericType().isGenericType();
      break;
    default:
      needToConvert = false;
    }

    ReferenceBinding originalEnclosing = originalType.enclosingType();
    TypeBinding convertedType;
    TypeBinding convertedType;
    if (originalEnclosing == null) {
      convertedType = needToConvert ? createRawType((ReferenceBinding)originalType.erasure(), null) : originalType;
    } else {
      ReferenceBinding convertedEnclosing = (ReferenceBinding)convertUnresolvedBinaryToRawType(originalEnclosing);
      if (convertedEnclosing != originalEnclosing)
        needToConvert |= !((ReferenceBinding)originalType).isStatic();
      TypeBinding convertedType;
      if (needToConvert) {
        convertedType = createRawType((ReferenceBinding)originalType.erasure(), convertedEnclosing);
      }
      else
      {
        TypeBinding convertedType;
        if (originalEnclosing != convertedEnclosing)
          convertedType = createParameterizedType((ReferenceBinding)originalType.erasure(), null, convertedEnclosing);
        else
          convertedType = originalType;
      }
    }
    if (originalType != convertedType) {
      return dimension > 0 ? createArrayType(convertedType, dimension) : convertedType;
    }
    return type;
  }

  public AnnotationBinding createAnnotation(ReferenceBinding annotationType, ElementValuePair[] pairs)
  {
    if (pairs.length != 0) {
      AnnotationBinding.setMethodBindings(annotationType, pairs);
    }
    return new AnnotationBinding(annotationType, pairs);
  }

  public ArrayBinding createArrayType(TypeBinding leafComponentType, int dimensionCount)
  {
    if ((leafComponentType instanceof LocalTypeBinding)) {
      return ((LocalTypeBinding)leafComponentType).createArrayType(dimensionCount, this);
    }

    int dimIndex = dimensionCount - 1;
    int length = this.uniqueArrayBindings.length;
    ArrayBinding[] arrayBindings;
    if (dimIndex < length)
    {
      ArrayBinding[] arrayBindings;
      if ((arrayBindings = this.uniqueArrayBindings[dimIndex]) == null)
      {
        ArrayBinding[] tmp56_53 = new ArrayBinding[10]; arrayBindings = tmp56_53; this.uniqueArrayBindings[dimIndex] = tmp56_53;
      }
    } else { System.arraycopy(
        this.uniqueArrayBindings, 0, 
        this.uniqueArrayBindings = new ArrayBinding[dimensionCount][], 0, 
        length);
      ArrayBinding[] tmp93_90 = new ArrayBinding[10]; arrayBindings = tmp93_90; this.uniqueArrayBindings[dimIndex] = tmp93_90;
    }

    int index = -1;
    length = arrayBindings.length;
    do {
      ArrayBinding currentBinding = arrayBindings[index];
      if (currentBinding == null)
        return arrayBindings[index] =  = new ArrayBinding(leafComponentType, dimensionCount, this);
      if (currentBinding.leafComponentType == leafComponentType)
        return currentBinding;
      index++; } while (index < length);

    System.arraycopy(
      arrayBindings, 0, 
      arrayBindings = new ArrayBinding[length * 2], 0, 
      length);
    this.uniqueArrayBindings[dimIndex] = arrayBindings;
    return arrayBindings[length] =  = new ArrayBinding(leafComponentType, dimensionCount, this);
  }
  public BinaryTypeBinding createBinaryTypeFrom(IBinaryType binaryType, PackageBinding packageBinding, AccessRestriction accessRestriction) {
    return createBinaryTypeFrom(binaryType, packageBinding, true, accessRestriction);
  }
  public BinaryTypeBinding createBinaryTypeFrom(IBinaryType binaryType, PackageBinding packageBinding, boolean needFieldsAndMethods, AccessRestriction accessRestriction) {
    BinaryTypeBinding binaryBinding = new BinaryTypeBinding(packageBinding, binaryType, this);

    ReferenceBinding cachedType = packageBinding.getType0(binaryBinding.compoundName[(binaryBinding.compoundName.length - 1)]);
    if (cachedType != null) {
      if ((cachedType instanceof UnresolvedReferenceBinding)) {
        ((UnresolvedReferenceBinding)cachedType).setResolvedType(binaryBinding, this);
      } else {
        if (cachedType.isBinaryBinding()) {
          return (BinaryTypeBinding)cachedType;
        }

        return null;
      }
    }

    packageBinding.addType(binaryBinding);
    setAccessRestriction(binaryBinding, accessRestriction);
    binaryBinding.cachePartsFrom(binaryType, needFieldsAndMethods);
    return binaryBinding;
  }

  public PackageBinding createPackage(char[][] compoundName)
  {
    PackageBinding packageBinding = getPackage0(compoundName[0]);
    if ((packageBinding == null) || (packageBinding == TheNotFoundPackage)) {
      packageBinding = new PackageBinding(compoundName[0], this);
      this.knownPackages.put(compoundName[0], packageBinding);
    }

    int i = 1; for (int length = compoundName.length; i < length; i++)
    {
      ReferenceBinding type = packageBinding.getType0(compoundName[i]);
      if ((type != null) && (type != TheNotFoundType) && (!(type instanceof UnresolvedReferenceBinding))) {
        return null;
      }
      PackageBinding parent = packageBinding;
      if (((packageBinding = parent.getPackage0(compoundName[i])) == null) || (packageBinding == TheNotFoundPackage))
      {
        if (this.nameEnvironment.findType(compoundName[i], parent.compoundName) != null) {
          return null;
        }
        packageBinding = new PackageBinding(CharOperation.subarray(compoundName, 0, i + 1), parent, this);
        parent.addPackage(packageBinding);
      }
    }
    return packageBinding;
  }

  public ParameterizedGenericMethodBinding createParameterizedGenericMethod(MethodBinding genericMethod, RawTypeBinding rawType)
  {
    ParameterizedGenericMethodBinding[] cachedInfo = (ParameterizedGenericMethodBinding[])this.uniqueParameterizedGenericMethodBindings.get(genericMethod);
    boolean needToGrow = false;
    int index = 0;
    if (cachedInfo != null)
    {
      for (int max = cachedInfo.length; index < max; index++) {
        ParameterizedGenericMethodBinding cachedMethod = cachedInfo[index];
        if (cachedMethod == null) break;
        if (cachedMethod.isRaw)
          if (cachedMethod.declaringClass == (rawType == null ? genericMethod.declaringClass : rawType))
            return cachedMethod;
      }
      needToGrow = true;
    } else {
      cachedInfo = new ParameterizedGenericMethodBinding[5];
      this.uniqueParameterizedGenericMethodBindings.put(genericMethod, cachedInfo);
    }

    int length = cachedInfo.length;
    if ((needToGrow) && (index == length)) {
      System.arraycopy(cachedInfo, 0, cachedInfo = new ParameterizedGenericMethodBinding[length * 2], 0, length);
      this.uniqueParameterizedGenericMethodBindings.put(genericMethod, cachedInfo);
    }

    ParameterizedGenericMethodBinding parameterizedGenericMethod = new ParameterizedGenericMethodBinding(genericMethod, rawType, this);
    cachedInfo[index] = parameterizedGenericMethod;
    return parameterizedGenericMethod;
  }

  public ParameterizedGenericMethodBinding createParameterizedGenericMethod(MethodBinding genericMethod, TypeBinding[] typeArguments)
  {
    ParameterizedGenericMethodBinding[] cachedInfo = (ParameterizedGenericMethodBinding[])this.uniqueParameterizedGenericMethodBindings.get(genericMethod);
    int argLength = typeArguments == null ? 0 : typeArguments.length;
    boolean needToGrow = false;
    int index = 0;
    if (cachedInfo != null)
    {
      for (int max = cachedInfo.length; index < max; index++) {
        ParameterizedGenericMethodBinding cachedMethod = cachedInfo[index];
        if (cachedMethod == null) break;
        if (!cachedMethod.isRaw) {
          TypeBinding[] cachedArguments = cachedMethod.typeArguments;
          int cachedArgLength = cachedArguments == null ? 0 : cachedArguments.length;
          if (argLength == cachedArgLength) {
            int j = 0;
            while (typeArguments[j] == cachedArguments[j])
            {
              j++; if (j >= cachedArgLength)
              {
                return cachedMethod;
              }
            }
          }
        }
      }
      needToGrow = true;
    } else {
      cachedInfo = new ParameterizedGenericMethodBinding[5];
      this.uniqueParameterizedGenericMethodBindings.put(genericMethod, cachedInfo);
    }

    int length = cachedInfo.length;
    if ((needToGrow) && (index == length)) {
      System.arraycopy(cachedInfo, 0, cachedInfo = new ParameterizedGenericMethodBinding[length * 2], 0, length);
      this.uniqueParameterizedGenericMethodBindings.put(genericMethod, cachedInfo);
    }

    ParameterizedGenericMethodBinding parameterizedGenericMethod = new ParameterizedGenericMethodBinding(genericMethod, typeArguments, this);
    cachedInfo[index] = parameterizedGenericMethod;
    return parameterizedGenericMethod;
  }

  public ParameterizedTypeBinding createParameterizedType(ReferenceBinding genericType, TypeBinding[] typeArguments, ReferenceBinding enclosingType)
  {
    ParameterizedTypeBinding[] cachedInfo = (ParameterizedTypeBinding[])this.uniqueParameterizedTypeBindings.get(genericType);
    int argLength = typeArguments == null ? 0 : typeArguments.length;
    boolean needToGrow = false;
    int index = 0;
    if (cachedInfo != null)
    {
      for (int max = cachedInfo.length; index < max; index++) {
        ParameterizedTypeBinding cachedType = cachedInfo[index];
        if (cachedType == null) break;
        if ((cachedType.actualType() == genericType) && 
          (cachedType.enclosingType() == enclosingType)) {
          TypeBinding[] cachedArguments = cachedType.arguments;
          int cachedArgLength = cachedArguments == null ? 0 : cachedArguments.length;
          if (argLength == cachedArgLength) {
            int j = 0;
            while (typeArguments[j] == cachedArguments[j])
            {
              j++; if (j >= cachedArgLength)
              {
                return cachedType;
              }
            }
          }
        }
      }
      needToGrow = true;
    } else {
      cachedInfo = new ParameterizedTypeBinding[5];
      this.uniqueParameterizedTypeBindings.put(genericType, cachedInfo);
    }

    int length = cachedInfo.length;
    if ((needToGrow) && (index == length)) {
      System.arraycopy(cachedInfo, 0, cachedInfo = new ParameterizedTypeBinding[length * 2], 0, length);
      this.uniqueParameterizedTypeBindings.put(genericType, cachedInfo);
    }

    ParameterizedTypeBinding parameterizedType = new ParameterizedTypeBinding(genericType, typeArguments, enclosingType, this);
    cachedInfo[index] = parameterizedType;
    return parameterizedType;
  }

  public RawTypeBinding createRawType(ReferenceBinding genericType, ReferenceBinding enclosingType)
  {
    RawTypeBinding[] cachedInfo = (RawTypeBinding[])this.uniqueRawTypeBindings.get(genericType);
    boolean needToGrow = false;
    int index = 0;
    if (cachedInfo != null)
    {
      for (int max = cachedInfo.length; index < max; index++) {
        RawTypeBinding cachedType = cachedInfo[index];
        if (cachedType == null) break;
        if ((cachedType.actualType() == genericType) && 
          (cachedType.enclosingType() == enclosingType))
        {
          return cachedType;
        }
      }
      needToGrow = true;
    } else {
      cachedInfo = new RawTypeBinding[1];
      this.uniqueRawTypeBindings.put(genericType, cachedInfo);
    }

    int length = cachedInfo.length;
    if ((needToGrow) && (index == length)) {
      System.arraycopy(cachedInfo, 0, cachedInfo = new RawTypeBinding[length * 2], 0, length);
      this.uniqueRawTypeBindings.put(genericType, cachedInfo);
    }

    RawTypeBinding rawType = new RawTypeBinding(genericType, enclosingType, this);
    cachedInfo[index] = rawType;
    return rawType;
  }

  public WildcardBinding createWildcard(ReferenceBinding genericType, int rank, TypeBinding bound, TypeBinding[] otherBounds, int boundKind)
  {
    if (genericType == null)
      genericType = ReferenceBinding.LUB_GENERIC;
    WildcardBinding[] cachedInfo = (WildcardBinding[])this.uniqueWildcardBindings.get(genericType);
    boolean needToGrow = false;
    int index = 0;
    if (cachedInfo != null)
    {
      for (int max = cachedInfo.length; index < max; index++) {
        WildcardBinding cachedType = cachedInfo[index];
        if (cachedType == null) break;
        if ((cachedType.genericType == genericType) && 
          (cachedType.rank == rank) && 
          (cachedType.boundKind == boundKind) && 
          (cachedType.bound == bound))
          if (cachedType.otherBounds != otherBounds) {
            int cachedLength = cachedType.otherBounds == null ? 0 : cachedType.otherBounds.length;
            int length = otherBounds == null ? 0 : otherBounds.length;
            if (cachedLength == length)
              for (int j = 0; j < length; j++)
                if (cachedType.otherBounds[j] != otherBounds[j])
                  break;
          }
          else {
            return cachedType;
          }
      }
      needToGrow = true;
    } else {
      cachedInfo = new WildcardBinding[10];
      this.uniqueWildcardBindings.put(genericType, cachedInfo);
    }

    int length = cachedInfo.length;
    if ((needToGrow) && (index == length)) {
      System.arraycopy(cachedInfo, 0, cachedInfo = new WildcardBinding[length * 2], 0, length);
      this.uniqueWildcardBindings.put(genericType, cachedInfo);
    }

    WildcardBinding wildcard = new WildcardBinding(genericType, rank, bound, otherBounds, boundKind, this);
    cachedInfo[index] = wildcard;
    return wildcard;
  }

  public AccessRestriction getAccessRestriction(TypeBinding type)
  {
    return (AccessRestriction)this.accessRestrictions.get(type);
  }

  public ReferenceBinding getCachedType(char[][] compoundName)
  {
    if (compoundName.length == 1) {
      if (this.defaultPackage == null)
        return null;
      return this.defaultPackage.getType0(compoundName[0]);
    }

    PackageBinding packageBinding = getPackage0(compoundName[0]);
    if ((packageBinding == null) || (packageBinding == TheNotFoundPackage)) {
      return null;
    }
    int i = 1; for (int packageLength = compoundName.length - 1; i < packageLength; i++)
      if (((packageBinding = packageBinding.getPackage0(compoundName[i])) == null) || (packageBinding == TheNotFoundPackage))
        return null;
    return packageBinding.getType0(compoundName[(compoundName.length - 1)]);
  }

  PackageBinding getPackage0(char[] name)
  {
    return this.knownPackages.get(name);
  }

  public ReferenceBinding getResolvedType(char[][] compoundName, Scope scope)
  {
    ReferenceBinding type = getType(compoundName);
    if (type != null) return type;

    return cacheMissingBinaryType(
      compoundName, 
      scope == null ? this.unitBeingCompleted : scope.referenceCompilationUnit());
  }

  PackageBinding getTopLevelPackage(char[] name)
  {
    PackageBinding packageBinding = getPackage0(name);
    if (packageBinding != null) {
      if (packageBinding == TheNotFoundPackage)
        return null;
      return packageBinding;
    }

    if (this.nameEnvironment.isPackage(null, name)) {
      this.knownPackages.put(name, packageBinding = new PackageBinding(name, this));
      return packageBinding;
    }

    this.knownPackages.put(name, TheNotFoundPackage);
    return null;
  }

  public ReferenceBinding getType(char[][] compoundName)
  {
    if (compoundName.length == 1) {
      if (this.defaultPackage == null)
        return null;
      ReferenceBinding referenceBinding;
      if ((referenceBinding = this.defaultPackage.getType0(compoundName[0])) == null) {
        PackageBinding packageBinding = getPackage0(compoundName[0]);
        if ((packageBinding != null) && (packageBinding != TheNotFoundPackage))
          return null;
        referenceBinding = askForType(this.defaultPackage, compoundName[0]);
      }
    } else {
      PackageBinding packageBinding = getPackage0(compoundName[0]);
      if (packageBinding == TheNotFoundPackage) {
        return null;
      }
      if (packageBinding != null) {
        int i = 1; for (int packageLength = compoundName.length - 1; i < packageLength; i++) {
          if ((packageBinding = packageBinding.getPackage0(compoundName[i])) == null)
            break;
          if (packageBinding == TheNotFoundPackage)
            return null;
        }
      }
      ReferenceBinding referenceBinding;
      if (packageBinding == null)
        referenceBinding = askForType(compoundName);
      else if ((referenceBinding = packageBinding.getType0(compoundName[(compoundName.length - 1)])) == null) {
        referenceBinding = askForType(packageBinding, compoundName[(compoundName.length - 1)]);
      }
    }
    if ((referenceBinding == null) || (referenceBinding == TheNotFoundType))
      return null;
    ReferenceBinding referenceBinding = BinaryTypeBinding.resolveType(referenceBinding, this, false);

    if (referenceBinding.isNestedType())
      return new ProblemReferenceBinding(compoundName, referenceBinding, 4);
    return referenceBinding;
  }
  private TypeBinding[] getTypeArgumentsFromSignature(SignatureWrapper wrapper, TypeVariableBinding[] staticVariables, ReferenceBinding enclosingType, ReferenceBinding genericType) {
    ArrayList args = new ArrayList(2);
    int rank = 0;
    do
      args.add(getTypeFromVariantTypeSignature(wrapper, staticVariables, enclosingType, genericType, rank++));
    while (wrapper.signature[wrapper.start] != '>');
    wrapper.start += 1;
    TypeBinding[] typeArguments = new TypeBinding[args.size()];
    args.toArray(typeArguments);
    return typeArguments;
  }

  ReferenceBinding getTypeFromCompoundName(char[][] compoundName, boolean isParameterized)
  {
    ReferenceBinding binding = getCachedType(compoundName);
    if (binding == null) {
      PackageBinding packageBinding = computePackageFrom(compoundName);
      binding = new UnresolvedReferenceBinding(compoundName, packageBinding);
      packageBinding.addType(binding);
    } else if (binding == TheNotFoundType)
    {
      binding = cacheMissingBinaryType(compoundName, this.unitBeingCompleted);
    } else if (!isParameterized)
    {
      binding = (ReferenceBinding)convertUnresolvedBinaryToRawType(binding);
    }
    return binding;
  }

  ReferenceBinding getTypeFromConstantPoolName(char[] signature, int start, int end, boolean isParameterized)
  {
    if (end == -1) {
      end = signature.length;
    }
    char[][] compoundName = CharOperation.splitOn('/', signature, start, end);
    return getTypeFromCompoundName(compoundName, isParameterized);
  }

  TypeBinding getTypeFromSignature(char[] signature, int start, int end, boolean isParameterized, TypeBinding enclosingType)
  {
    int dimension = 0;
    while (signature[start] == '[') {
      start++;
      dimension++;
    }
    if (end == -1) {
      end = signature.length - 1;
    }

    TypeBinding binding = null;
    if (start == end) {
      switch (signature[start]) {
      case 'I':
        binding = TypeBinding.INT;
        break;
      case 'Z':
        binding = TypeBinding.BOOLEAN;
        break;
      case 'V':
        binding = TypeBinding.VOID;
        break;
      case 'C':
        binding = TypeBinding.CHAR;
        break;
      case 'D':
        binding = TypeBinding.DOUBLE;
        break;
      case 'B':
        binding = TypeBinding.BYTE;
        break;
      case 'F':
        binding = TypeBinding.FLOAT;
        break;
      case 'J':
        binding = TypeBinding.LONG;
        break;
      case 'S':
        binding = TypeBinding.SHORT;
        break;
      default:
        this.problemReporter.corruptedSignature(enclosingType, signature, start); break;
      }
    }
    else {
      binding = getTypeFromConstantPoolName(signature, start + 1, end, isParameterized);
    }

    if (dimension == 0)
      return binding;
    return createArrayType(binding, dimension);
  }

  TypeBinding getTypeFromTypeSignature(SignatureWrapper wrapper, TypeVariableBinding[] staticVariables, ReferenceBinding enclosingType)
  {
    int dimension = 0;
    while (wrapper.signature[wrapper.start] == '[') {
      wrapper.start += 1;
      dimension++;
    }

    if (wrapper.signature[wrapper.start] == 'T') {
      int varStart = wrapper.start + 1;
      int varEnd = wrapper.computeEnd();
      int i = staticVariables.length;
      do { if (CharOperation.equals(staticVariables[i].sourceName, wrapper.signature, varStart, varEnd))
          return dimension == 0 ? staticVariables[i] : createArrayType(staticVariables[i], dimension);
        i--; } while (i >= 0);

      ReferenceBinding initialType = enclosingType;
      do {
        if ((enclosingType instanceof BinaryTypeBinding)) {
          TypeVariableBinding[] enclosingVariables = ((BinaryTypeBinding)enclosingType).typeVariables;
          int i = enclosingVariables.length;
          do { if (CharOperation.equals(enclosingVariables[i].sourceName, wrapper.signature, varStart, varEnd))
              return dimension == 0 ? enclosingVariables[i] : createArrayType(enclosingVariables[i], dimension);
            i--; } while (i >= 0);
        }
      }

      while ((enclosingType = enclosingType.enclosingType()) != null);
      this.problemReporter.undefinedTypeVariableSignature(CharOperation.subarray(wrapper.signature, varStart, varEnd), initialType);
      return null;
    }
    boolean isParameterized;
    TypeBinding type = getTypeFromSignature(wrapper.signature, wrapper.start, wrapper.computeEnd(), isParameterized = wrapper.end == wrapper.bracket ? 1 : 0, enclosingType);
    if (!isParameterized) {
      return dimension == 0 ? type : createArrayType(type, dimension);
    }

    ReferenceBinding actualType = (ReferenceBinding)type;
    TypeBinding[] typeArguments = getTypeArgumentsFromSignature(wrapper, staticVariables, enclosingType, actualType);
    ReferenceBinding actualEnclosing = actualType.enclosingType();
    if (actualEnclosing != null) {
      actualEnclosing = (ReferenceBinding)convertToRawType(actualEnclosing);
    }
    ParameterizedTypeBinding parameterizedType = createParameterizedType(actualType, typeArguments, actualEnclosing);

    while (wrapper.signature[wrapper.start] == '.') {
      wrapper.start += 1;
      char[] memberName = wrapper.nextWord();
      BinaryTypeBinding.resolveType(parameterizedType, this, false);
      ReferenceBinding memberType = parameterizedType.genericType().getMemberType(memberName);
      if (wrapper.signature[wrapper.start] == '<') {
        wrapper.start += 1;
        typeArguments = getTypeArgumentsFromSignature(wrapper, staticVariables, enclosingType, memberType);
      } else {
        typeArguments = (TypeBinding[])null;
      }
      parameterizedType = createParameterizedType(memberType, typeArguments, parameterizedType);
    }
    wrapper.start += 1;
    return dimension == 0 ? parameterizedType : createArrayType(parameterizedType, dimension);
  }

  TypeBinding getTypeFromVariantTypeSignature(SignatureWrapper wrapper, TypeVariableBinding[] staticVariables, ReferenceBinding enclosingType, ReferenceBinding genericType, int rank)
  {
    switch (wrapper.signature[wrapper.start])
    {
    case '-':
      wrapper.start += 1;
      TypeBinding bound = getTypeFromTypeSignature(wrapper, staticVariables, enclosingType);
      return createWildcard(genericType, rank, bound, null, 2);
    case '+':
      wrapper.start += 1;
      TypeBinding bound = getTypeFromTypeSignature(wrapper, staticVariables, enclosingType);
      return createWildcard(genericType, rank, bound, null, 1);
    case '*':
      wrapper.start += 1;
      return createWildcard(genericType, rank, null, null, 0);
    case ',':
    }return getTypeFromTypeSignature(wrapper, staticVariables, enclosingType);
  }

  boolean isPackage(char[][] compoundName, char[] name)
  {
    if ((compoundName == null) || (compoundName.length == 0))
      return this.nameEnvironment.isPackage(null, name);
    return this.nameEnvironment.isPackage(compoundName, name);
  }

  public MethodVerifier methodVerifier()
  {
    if (this.verifier == null)
      this.verifier = (this.globalOptions.sourceLevel < 3211264L ? 
        new MethodVerifier(this) : 
        new MethodVerifier15(this));
    return this.verifier;
  }
  public void reset() {
    this.defaultPackage = new PackageBinding(this);
    this.defaultImports = null;
    this.knownPackages = new HashtableOfPackage();
    this.accessRestrictions = new HashMap(3);

    this.verifier = null;
    int i = this.uniqueArrayBindings.length;
    do { ArrayBinding[] arrayBindings = this.uniqueArrayBindings[i];
      if (arrayBindings != null) {
        int j = arrayBindings.length;
        do { arrayBindings[j] = null;

          j--; } while (j >= 0);
      }
      i--; } while (i >= 0);

    this.uniqueParameterizedTypeBindings = new SimpleLookupTable(3);
    this.uniqueRawTypeBindings = new SimpleLookupTable(3);
    this.uniqueWildcardBindings = new SimpleLookupTable(3);
    this.uniqueParameterizedGenericMethodBindings = new SimpleLookupTable(3);

    int i = this.units.length;
    do { this.units[i] = null;

      i--; } while (i >= 0);

    this.lastUnitIndex = -1;
    this.lastCompletedUnitIndex = -1;
    this.unitBeingCompleted = null;

    this.classFilePool.reset();
  }

  public void setAccessRestriction(ReferenceBinding type, AccessRestriction accessRestriction)
  {
    if (accessRestriction == null) return;
    type.modifiers |= 262144;
    this.accessRestrictions.put(type, accessRestriction);
  }

  void updateCaches(UnresolvedReferenceBinding unresolvedType, ReferenceBinding resolvedType)
  {
    if (this.uniqueParameterizedTypeBindings.get(unresolvedType) != null) {
      Object[] keys = this.uniqueParameterizedTypeBindings.keyTable;
      int i = 0; for (int l = keys.length; i < l; i++) {
        if (keys[i] == unresolvedType) {
          keys[i] = resolvedType;
          break;
        }
      }
    }

    if (this.uniqueWildcardBindings.get(unresolvedType) != null) {
      Object[] keys = this.uniqueWildcardBindings.keyTable;
      int i = 0; for (int l = keys.length; i < l; i++)
        if (keys[i] == unresolvedType) {
          keys[i] = resolvedType;
          break;
        }
    }
  }
}