package org.quartz.core;

import java.io.InputStream;
import java.lang.management.ManagementFactory;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Random;
import java.util.Set;
import java.util.Timer;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import org.quartz.Calendar;
import org.quartz.InterruptableJob;
import org.quartz.Job;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.quartz.JobKey;
import org.quartz.JobListener;
import org.quartz.JobPersistenceException;
import org.quartz.ListenerManager;
import org.quartz.Matcher;
import org.quartz.ObjectAlreadyExistsException;
import org.quartz.SchedulerContext;
import org.quartz.SchedulerException;
import org.quartz.SchedulerListener;
import org.quartz.SchedulerMetaData;
import org.quartz.Trigger;
import org.quartz.Trigger.CompletedExecutionInstruction;
import org.quartz.Trigger.TriggerState;
import org.quartz.TriggerKey;
import org.quartz.TriggerListener;
import org.quartz.UnableToInterruptJobException;
import org.quartz.core.jmx.QuartzSchedulerMBean;
import org.quartz.impl.SchedulerRepository;
import org.quartz.impl.matchers.GroupMatcher;
import org.quartz.impl.triggers.SimpleTriggerImpl;
import org.quartz.simpl.PropertySettingJobFactory;
import org.quartz.spi.JobFactory;
import org.quartz.spi.JobStore;
import org.quartz.spi.OperableTrigger;
import org.quartz.spi.SchedulerPlugin;
import org.quartz.spi.SchedulerSignaler;
import org.quartz.spi.ThreadExecutor;
import org.quartz.spi.ThreadPool;
import org.quartz.utils.UpdateChecker;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class QuartzScheduler
  implements RemotableQuartzScheduler
{
  private static String VERSION_MAJOR = "UNKNOWN";
  private static String VERSION_MINOR = "UNKNOWN";
  private static String VERSION_ITERATION = "UNKNOWN";
  private QuartzSchedulerResources resources;
  private QuartzSchedulerThread schedThread;
  private ThreadGroup threadGroup;
  private SchedulerContext context = new SchedulerContext();

  private ListenerManager listenerManager = new ListenerManagerImpl();

  private HashMap<String, JobListener> internalJobListeners = new HashMap(10);

  private HashMap<String, TriggerListener> internalTriggerListeners = new HashMap(10);

  private ArrayList<SchedulerListener> internalSchedulerListeners = new ArrayList(10);

  private JobFactory jobFactory = new PropertySettingJobFactory();

  ExecutingJobsManager jobMgr = null;

  ErrorLogger errLogger = null;
  private SchedulerSignaler signaler;
  private Random random = new Random();

  private ArrayList<Object> holdToPreventGC = new ArrayList(5);

  private boolean signalOnSchedulingChange = true;

  private volatile boolean closed = false;
  private volatile boolean shuttingDown = false;
  private boolean boundRemotely = false;

  private QuartzSchedulerMBean jmxBean = null;

  private Date initialStart = null;
  private final Timer updateTimer;
  private final Logger log = LoggerFactory.getLogger(getClass());
  private long dbRetryInterval;

  public QuartzScheduler(QuartzSchedulerResources resources, long idleWaitTime, long dbRetryInterval)
    throws SchedulerException
  {
    this.resources = resources;
    if ((resources.getJobStore() instanceof JobListener)) {
      addInternalJobListener((JobListener)resources.getJobStore());
    }

    this.schedThread = new QuartzSchedulerThread(this, resources);
    ThreadExecutor schedThreadExecutor = resources.getThreadExecutor();
    schedThreadExecutor.execute(this.schedThread);
    if (idleWaitTime > 0L) {
      this.schedThread.setIdleWaitTime(idleWaitTime);
    }
    if (dbRetryInterval > 0L) {
      this.schedThread.setDbFailureRetryInterval(dbRetryInterval);
    }

    this.jobMgr = new ExecutingJobsManager();
    addInternalJobListener(this.jobMgr);
    this.errLogger = new ErrorLogger();
    addInternalSchedulerListener(this.errLogger);

    this.signaler = new SchedulerSignalerImpl(this, this.schedThread);

    if (shouldRunUpdateCheck())
      this.updateTimer = scheduleUpdateCheck();
    else {
      this.updateTimer = null;
    }
    this.dbRetryInterval = dbRetryInterval;

    getLog().info("Quartz Scheduler v." + getVersion() + " created.");
  }

  public long getDbRetryInterval() {
    return this.dbRetryInterval;
  }

  public void initialize() throws SchedulerException
  {
    try {
      bind();
    } catch (Exception re) {
      throw new SchedulerException("Unable to bind scheduler to RMI Registry.", re);
    }

    if (this.resources.getJMXExport()) {
      try {
        registerJMX();
      } catch (Exception e) {
        throw new SchedulerException("Unable to register scheduler with MBeanServer.", e);
      }

    }

    getLog().info("Scheduler meta-data: " + new SchedulerMetaData(getSchedulerName(), getSchedulerInstanceId(), getClass(), this.boundRemotely, runningSince() != null, isInStandbyMode(), isShutdown(), runningSince(), numJobsExecuted(), getJobStoreClass(), supportsPersistence(), isClustered(), getThreadPoolClass(), getThreadPoolSize(), getVersion()).toString());
  }

  public String getVersion()
  {
    return getVersionMajor() + "." + getVersionMinor() + "." + getVersionIteration();
  }

  public static String getVersionMajor()
  {
    return VERSION_MAJOR;
  }

  private boolean shouldRunUpdateCheck() {
    if ((this.resources.isRunUpdateCheck()) && (!Boolean.getBoolean("org.quartz.scheduler.skipUpdateCheck")) && (!Boolean.getBoolean("org.terracotta.quartz.skipUpdateCheck")))
    {
      return true;
    }
    return false;
  }

  public static String getVersionMinor() {
    return VERSION_MINOR;
  }

  public static String getVersionIteration() {
    return VERSION_ITERATION;
  }

  public SchedulerSignaler getSchedulerSignaler() {
    return this.signaler;
  }

  public Logger getLog() {
    return this.log;
  }

  private Timer scheduleUpdateCheck()
  {
    Timer rval = new Timer(true);
    rval.scheduleAtFixedRate(new UpdateChecker(), 1000L, 604800000L);
    return rval;
  }

  private void registerJMX()
    throws Exception
  {
    String jmxObjectName = this.resources.getJMXObjectName();
    MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();
    this.jmxBean = new QuartzSchedulerMBeanImpl(this);
    mbs.registerMBean(this.jmxBean, new ObjectName(jmxObjectName));
  }

  private void unregisterJMX()
    throws Exception
  {
    String jmxObjectName = this.resources.getJMXObjectName();
    MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();
    mbs.unregisterMBean(new ObjectName(jmxObjectName));
    this.jmxBean.setSampledStatisticsEnabled(false);
    getLog().info("Scheduler unregistered from name '" + jmxObjectName + "' in the local MBeanServer.");
  }

  private void bind()
    throws RemoteException
  {
    String host = this.resources.getRMIRegistryHost();

    if ((host == null) || (host.length() == 0)) {
      return;
    }

    RemotableQuartzScheduler exportable = null;

    if (this.resources.getRMIServerPort() > 0) {
      exportable = (RemotableQuartzScheduler)UnicastRemoteObject.exportObject(this, this.resources.getRMIServerPort());
    }
    else {
      exportable = (RemotableQuartzScheduler)UnicastRemoteObject.exportObject(this);
    }

    Registry registry = null;

    if (this.resources.getRMICreateRegistryStrategy().equals("as_needed"))
    {
      try
      {
        registry = LocateRegistry.getRegistry(this.resources.getRMIRegistryPort());

        registry.list();
      } catch (Exception e) {
        registry = LocateRegistry.createRegistry(this.resources.getRMIRegistryPort());
      }
    }
    else if (this.resources.getRMICreateRegistryStrategy().equals("always")) {
      try
      {
        registry = LocateRegistry.createRegistry(this.resources.getRMIRegistryPort());
      }
      catch (Exception e)
      {
        registry = LocateRegistry.getRegistry(this.resources.getRMIRegistryPort());
      }
    }
    else {
      registry = LocateRegistry.getRegistry(this.resources.getRMIRegistryHost(), this.resources.getRMIRegistryPort());
    }

    String bindName = this.resources.getRMIBindName();

    registry.rebind(bindName, exportable);

    this.boundRemotely = true;

    getLog().info("Scheduler bound to RMI registry under name '" + bindName + "'");
  }

  private void unBind()
    throws RemoteException
  {
    String host = this.resources.getRMIRegistryHost();

    if ((host == null) || (host.length() == 0)) {
      return;
    }

    Registry registry = LocateRegistry.getRegistry(this.resources.getRMIRegistryHost(), this.resources.getRMIRegistryPort());

    String bindName = this.resources.getRMIBindName();
    try
    {
      registry.unbind(bindName);
      UnicastRemoteObject.unexportObject(this, true);
    }
    catch (NotBoundException nbe) {
    }
    getLog().info("Scheduler un-bound from name '" + bindName + "' in RMI registry");
  }

  public String getSchedulerName()
  {
    return this.resources.getName();
  }

  public String getSchedulerInstanceId()
  {
    return this.resources.getInstanceId();
  }

  public ThreadGroup getSchedulerThreadGroup()
  {
    if (this.threadGroup == null) {
      this.threadGroup = new ThreadGroup("QuartzScheduler:" + getSchedulerName());

      if (this.resources.getMakeSchedulerThreadDaemon()) {
        this.threadGroup.setDaemon(true);
      }
    }

    return this.threadGroup;
  }

  public void addNoGCObject(Object obj) {
    this.holdToPreventGC.add(obj);
  }

  public boolean removeNoGCObject(Object obj) {
    return this.holdToPreventGC.remove(obj);
  }

  public SchedulerContext getSchedulerContext()
    throws SchedulerException
  {
    return this.context;
  }

  public boolean isSignalOnSchedulingChange() {
    return this.signalOnSchedulingChange;
  }

  public void setSignalOnSchedulingChange(boolean signalOnSchedulingChange) {
    this.signalOnSchedulingChange = signalOnSchedulingChange;
  }

  public void start()
    throws SchedulerException
  {
    if ((this.shuttingDown) || (this.closed)) {
      throw new SchedulerException("The Scheduler cannot be restarted after shutdown() has been called.");
    }

    if (this.initialStart == null) {
      this.initialStart = new Date();
      this.resources.getJobStore().schedulerStarted();
      startPlugins();
    } else {
      this.resources.getJobStore().schedulerResumed();
    }

    this.schedThread.togglePause(false);

    getLog().info("Scheduler " + this.resources.getUniqueIdentifier() + " started.");

    notifySchedulerListenersStarted();
  }

  public void startDelayed(final int seconds) throws SchedulerException
  {
    if ((this.shuttingDown) || (this.closed)) {
      throw new SchedulerException("The Scheduler cannot be restarted after shutdown() has been called.");
    }

    Thread t = new Thread(new Runnable() {
      public void run() {
        try { Thread.sleep(seconds * 1000L); } catch (InterruptedException ignore) {
        }try {
          QuartzScheduler.this.start();
        } catch (SchedulerException se) {
          QuartzScheduler.this.getLog().error("Unable to start secheduler after startup delay.", se);
        }
      }
    });
    t.start();
  }

  public void standby()
  {
    this.resources.getJobStore().schedulerPaused();
    this.schedThread.togglePause(true);
    getLog().info("Scheduler " + this.resources.getUniqueIdentifier() + " paused.");

    notifySchedulerListenersInStandbyMode();
  }

  public boolean isInStandbyMode()
  {
    return this.schedThread.isPaused();
  }

  public Date runningSince() {
    if (this.initialStart == null)
      return null;
    return new Date(this.initialStart.getTime());
  }

  public int numJobsExecuted() {
    return this.jobMgr.getNumJobsFired();
  }

  public Class<?> getJobStoreClass() {
    return this.resources.getJobStore().getClass();
  }

  public boolean supportsPersistence() {
    return this.resources.getJobStore().supportsPersistence();
  }

  public boolean isClustered() {
    return this.resources.getJobStore().isClustered();
  }

  public Class<?> getThreadPoolClass() {
    return this.resources.getThreadPool().getClass();
  }

  public int getThreadPoolSize() {
    return this.resources.getThreadPool().getPoolSize();
  }

  public void shutdown()
  {
    shutdown(false);
  }

  public void shutdown(boolean waitForJobsToComplete)
  {
    if ((this.shuttingDown) || (this.closed)) {
      return;
    }

    this.shuttingDown = true;

    getLog().info("Scheduler " + this.resources.getUniqueIdentifier() + " shutting down.");

    standby();

    this.schedThread.halt();

    notifySchedulerListenersShuttingdown();

    if (((this.resources.isInterruptJobsOnShutdown()) && (!waitForJobsToComplete)) || ((this.resources.isInterruptJobsOnShutdownWithWait()) && (waitForJobsToComplete)))
    {
      List jobs = getCurrentlyExecutingJobs();
      for (JobExecutionContext job : jobs) {
        if ((job.getJobInstance() instanceof InterruptableJob)) {
          try {
            ((InterruptableJob)job.getJobInstance()).interrupt();
          }
          catch (Throwable e) {
            getLog().warn("Encountered error when interrupting job {} during shutdown: {}", job.getJobDetail().getKey(), e);
          }
        }
      }
    }
    this.resources.getThreadPool().shutdown(waitForJobsToComplete);

    if (waitForJobsToComplete) {
      while (this.jobMgr.getNumJobsCurrentlyExecuting() > 0) {
        try {
          Thread.sleep(100L);
        }
        catch (Exception ignore)
        {
        }
      }
    }

    try
    {
      this.schedThread.join();
    }
    catch (InterruptedException ignore) {
    }
    this.closed = true;

    if (this.resources.getJMXExport())
      try {
        unregisterJMX();
      }
      catch (Exception e)
      {
      }
    if (this.boundRemotely)
      try {
        unBind();
      }
      catch (RemoteException re)
      {
      }
    shutdownPlugins();

    this.resources.getJobStore().shutdown();

    notifySchedulerListenersShutdown();

    SchedulerRepository.getInstance().remove(this.resources.getName());

    this.holdToPreventGC.clear();

    if (this.updateTimer != null) {
      this.updateTimer.cancel();
    }
    getLog().info("Scheduler " + this.resources.getUniqueIdentifier() + " shutdown complete.");
  }

  public boolean isShutdown()
  {
    return this.closed;
  }

  public boolean isShuttingDown() {
    return this.shuttingDown;
  }

  public boolean isStarted() {
    return (!this.shuttingDown) && (!this.closed) && (!isInStandbyMode()) && (this.initialStart != null);
  }

  public void validateState() throws SchedulerException {
    if (isShutdown())
      throw new SchedulerException("The Scheduler has been shutdown.");
  }

  public List<JobExecutionContext> getCurrentlyExecutingJobs()
  {
    return this.jobMgr.getExecutingJobs();
  }

  public Date scheduleJob(JobDetail jobDetail, Trigger trigger)
    throws SchedulerException
  {
    validateState();

    if (jobDetail == null) {
      throw new SchedulerException("JobDetail cannot be null");
    }

    if (trigger == null) {
      throw new SchedulerException("Trigger cannot be null");
    }

    if (jobDetail.getKey() == null) {
      throw new SchedulerException("Job's key cannot be null");
    }

    if (jobDetail.getJobClass() == null) {
      throw new SchedulerException("Job's class cannot be null");
    }

    OperableTrigger trig = (OperableTrigger)trigger;

    if (trigger.getJobKey() == null)
      trig.setJobKey(jobDetail.getKey());
    else if (!trigger.getJobKey().equals(jobDetail.getKey())) {
      throw new SchedulerException("Trigger does not reference given job!");
    }

    trig.validate();

    Calendar cal = null;
    if (trigger.getCalendarName() != null) {
      cal = this.resources.getJobStore().retrieveCalendar(trigger.getCalendarName());
    }
    Date ft = trig.computeFirstFireTime(cal);

    if (ft == null) {
      throw new SchedulerException("Based on configured schedule, the given trigger '" + trigger.getKey() + "' will never fire.");
    }

    this.resources.getJobStore().storeJobAndTrigger(jobDetail, trig);
    notifySchedulerListenersJobAdded(jobDetail);
    notifySchedulerThread(trigger.getNextFireTime().getTime());
    notifySchedulerListenersSchduled(trigger);

    return ft;
  }

  public Date scheduleJob(Trigger trigger)
    throws SchedulerException
  {
    validateState();

    if (trigger == null) {
      throw new SchedulerException("Trigger cannot be null");
    }

    OperableTrigger trig = (OperableTrigger)trigger;

    trig.validate();

    Calendar cal = null;
    if (trigger.getCalendarName() != null) {
      cal = this.resources.getJobStore().retrieveCalendar(trigger.getCalendarName());
      if (cal == null) {
        throw new SchedulerException("Calendar not found: " + trigger.getCalendarName());
      }
    }

    Date ft = trig.computeFirstFireTime(cal);

    if (ft == null) {
      throw new SchedulerException("Based on configured schedule, the given trigger '" + trigger.getKey() + "' will never fire.");
    }

    this.resources.getJobStore().storeTrigger(trig, false);
    notifySchedulerThread(trigger.getNextFireTime().getTime());
    notifySchedulerListenersSchduled(trigger);

    return ft;
  }

  public void addJob(JobDetail jobDetail, boolean replace)
    throws SchedulerException
  {
    validateState();

    if ((!jobDetail.isDurable()) && (!replace)) {
      throw new SchedulerException("Jobs added with no trigger must be durable.");
    }

    this.resources.getJobStore().storeJob(jobDetail, replace);
    notifySchedulerThread(0L);
    notifySchedulerListenersJobAdded(jobDetail);
  }

  public boolean deleteJob(JobKey jobKey)
    throws SchedulerException
  {
    validateState();

    boolean result = false;

    List triggers = getTriggersOfJob(jobKey);
    for (Trigger trigger : triggers) {
      if (!unscheduleJob(trigger.getKey())) {
        StringBuilder sb = new StringBuilder().append("Unable to unschedule trigger [").append(trigger.getKey()).append("] while deleting job [").append(jobKey).append("]");

        throw new SchedulerException(sb.toString());
      }
      result = true;
    }

    result = (this.resources.getJobStore().removeJob(jobKey)) || (result);
    if (result) {
      notifySchedulerThread(0L);
      notifySchedulerListenersJobDeleted(jobKey);
    }
    return result;
  }

  public boolean deleteJobs(List<JobKey> jobKeys) throws SchedulerException {
    validateState();

    boolean result = false;

    result = this.resources.getJobStore().removeJobs(jobKeys);
    notifySchedulerThread(0L);
    for (JobKey key : jobKeys)
      notifySchedulerListenersJobDeleted(key);
    return result;
  }

  public void scheduleJobs(Map<JobDetail, List<Trigger>> triggersAndJobs, boolean replace) throws SchedulerException {
    validateState();

    for (Map.Entry e : triggersAndJobs.entrySet()) {
      job = (JobDetail)e.getKey();
      if (job != null)
      {
        List triggers = (List)e.getValue();
        if (triggers != null)
        {
          for (Trigger trigger : triggers) {
            OperableTrigger opt = (OperableTrigger)trigger;
            opt.setJobKey(job.getKey());

            opt.validate();

            Calendar cal = null;
            if (trigger.getCalendarName() != null) {
              cal = this.resources.getJobStore().retrieveCalendar(trigger.getCalendarName());
              if (cal == null) {
                throw new SchedulerException("Calendar '" + trigger.getCalendarName() + "' not found for trigger: " + trigger.getKey());
              }
            }

            Date ft = opt.computeFirstFireTime(cal);

            if (ft == null)
              throw new SchedulerException("Based on configured schedule, the given trigger will never fire.");
          }
        }
      }
    }
    JobDetail job;
    this.resources.getJobStore().storeJobsAndTriggers(triggersAndJobs, replace);
    notifySchedulerThread(0L);
    for (JobDetail job : triggersAndJobs.keySet())
      notifySchedulerListenersJobAdded(job);
  }

  public boolean unscheduleJobs(List<TriggerKey> triggerKeys) throws SchedulerException {
    validateState();

    boolean result = false;

    result = this.resources.getJobStore().removeTriggers(triggerKeys);
    notifySchedulerThread(0L);
    for (TriggerKey key : triggerKeys)
      notifySchedulerListenersUnscheduled(key);
    return result;
  }

  public boolean unscheduleJob(TriggerKey triggerKey)
    throws SchedulerException
  {
    validateState();

    if (this.resources.getJobStore().removeTrigger(triggerKey)) {
      notifySchedulerThread(0L);
      notifySchedulerListenersUnscheduled(triggerKey);
    } else {
      return false;
    }

    return true;
  }

  public Date rescheduleJob(TriggerKey triggerKey, Trigger newTrigger)
    throws SchedulerException
  {
    validateState();

    if (triggerKey == null) {
      throw new IllegalArgumentException("triggerKey cannot be null");
    }
    if (newTrigger == null) {
      throw new IllegalArgumentException("newTrigger cannot be null");
    }

    OperableTrigger trig = (OperableTrigger)newTrigger;
    Trigger oldTrigger = getTrigger(triggerKey);
    if (oldTrigger == null) {
      return null;
    }
    trig.setJobKey(oldTrigger.getJobKey());

    trig.validate();

    Calendar cal = null;
    if (newTrigger.getCalendarName() != null) {
      cal = this.resources.getJobStore().retrieveCalendar(newTrigger.getCalendarName());
    }

    Date ft = trig.computeFirstFireTime(cal);

    if (ft == null) {
      throw new SchedulerException("Based on configured schedule, the given trigger will never fire.");
    }

    if (this.resources.getJobStore().replaceTrigger(triggerKey, trig)) {
      notifySchedulerThread(newTrigger.getNextFireTime().getTime());
      notifySchedulerListenersUnscheduled(triggerKey);
      notifySchedulerListenersSchduled(newTrigger);
    } else {
      return null;
    }

    return ft;
  }

  private String newTriggerId()
  {
    long r = this.random.nextLong();
    if (r < 0L) {
      r = -r;
    }
    return "MT_" + Long.toString(r, 30 + (int)(System.currentTimeMillis() % 7L));
  }

  public void triggerJob(JobKey jobKey, JobDataMap data)
    throws SchedulerException
  {
    validateState();

    OperableTrigger trig = new SimpleTriggerImpl(newTriggerId(), "DEFAULT", jobKey.getName(), jobKey.getGroup(), new Date(), null, 0, 0L);

    trig.computeFirstFireTime(null);
    if (data != null) {
      trig.setJobDataMap(data);
    }

    boolean collision = true;
    while (collision) {
      try {
        this.resources.getJobStore().storeTrigger(trig, false);
        collision = false;
      } catch (ObjectAlreadyExistsException oaee) {
        trig.setKey(new TriggerKey(newTriggerId(), "DEFAULT"));
      }
    }

    notifySchedulerThread(trig.getNextFireTime().getTime());
    notifySchedulerListenersSchduled(trig);
  }

  public void triggerJob(OperableTrigger trig)
    throws SchedulerException
  {
    validateState();

    trig.computeFirstFireTime(null);

    boolean collision = true;
    while (collision) {
      try {
        this.resources.getJobStore().storeTrigger(trig, false);
        collision = false;
      } catch (ObjectAlreadyExistsException oaee) {
        trig.setKey(new TriggerKey(newTriggerId(), "DEFAULT"));
      }
    }

    notifySchedulerThread(trig.getNextFireTime().getTime());
    notifySchedulerListenersSchduled(trig);
  }

  public void pauseTrigger(TriggerKey triggerKey)
    throws SchedulerException
  {
    validateState();

    this.resources.getJobStore().pauseTrigger(triggerKey);
    notifySchedulerThread(0L);
    notifySchedulerListenersPausedTrigger(triggerKey);
  }

  public void pauseTriggers(GroupMatcher<TriggerKey> matcher)
    throws SchedulerException
  {
    validateState();

    if (matcher == null) {
      matcher = GroupMatcher.groupEquals("DEFAULT");
    }

    Collection pausedGroups = this.resources.getJobStore().pauseTriggers(matcher);
    notifySchedulerThread(0L);
    for (String pausedGroup : pausedGroups)
      notifySchedulerListenersPausedTriggers(pausedGroup);
  }

  public void pauseJob(JobKey jobKey)
    throws SchedulerException
  {
    validateState();

    this.resources.getJobStore().pauseJob(jobKey);
    notifySchedulerThread(0L);
    notifySchedulerListenersPausedJob(jobKey);
  }

  public void pauseJobs(GroupMatcher<JobKey> groupMatcher)
    throws SchedulerException
  {
    validateState();

    if (groupMatcher == null) {
      groupMatcher = GroupMatcher.groupEquals("DEFAULT");
    }

    Collection pausedGroups = this.resources.getJobStore().pauseJobs(groupMatcher);
    notifySchedulerThread(0L);
    for (String pausedGroup : pausedGroups)
      notifySchedulerListenersPausedJobs(pausedGroup);
  }

  public void resumeTrigger(TriggerKey triggerKey)
    throws SchedulerException
  {
    validateState();

    this.resources.getJobStore().resumeTrigger(triggerKey);
    notifySchedulerThread(0L);
    notifySchedulerListenersResumedTrigger(triggerKey);
  }

  public void resumeTriggers(GroupMatcher<TriggerKey> matcher)
    throws SchedulerException
  {
    validateState();

    if (matcher == null) {
      matcher = GroupMatcher.groupEquals("DEFAULT");
    }

    Collection pausedGroups = this.resources.getJobStore().resumeTriggers(matcher);
    notifySchedulerThread(0L);
    for (String pausedGroup : pausedGroups)
      notifySchedulerListenersResumedTriggers(pausedGroup);
  }

  public Set<String> getPausedTriggerGroups() throws SchedulerException
  {
    return this.resources.getJobStore().getPausedTriggerGroups();
  }

  public void resumeJob(JobKey jobKey)
    throws SchedulerException
  {
    validateState();

    this.resources.getJobStore().resumeJob(jobKey);
    notifySchedulerThread(0L);
    notifySchedulerListenersResumedJob(jobKey);
  }

  public void resumeJobs(GroupMatcher<JobKey> matcher)
    throws SchedulerException
  {
    validateState();

    if (matcher == null) {
      matcher = GroupMatcher.groupEquals("DEFAULT");
    }

    Collection resumedGroups = this.resources.getJobStore().resumeJobs(matcher);
    notifySchedulerThread(0L);
    for (String pausedGroup : resumedGroups)
      notifySchedulerListenersResumedJobs(pausedGroup);
  }

  public void pauseAll()
    throws SchedulerException
  {
    validateState();

    this.resources.getJobStore().pauseAll();
    notifySchedulerThread(0L);
    notifySchedulerListenersPausedTriggers(null);
  }

  public void resumeAll()
    throws SchedulerException
  {
    validateState();

    this.resources.getJobStore().resumeAll();
    notifySchedulerThread(0L);
    notifySchedulerListenersResumedTrigger(null);
  }

  public List<String> getJobGroupNames()
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().getJobGroupNames();
  }

  public Set<JobKey> getJobKeys(GroupMatcher<JobKey> matcher)
    throws SchedulerException
  {
    validateState();

    if (matcher == null) {
      matcher = GroupMatcher.groupEquals("DEFAULT");
    }

    return this.resources.getJobStore().getJobKeys(matcher);
  }

  public List<? extends Trigger> getTriggersOfJob(JobKey jobKey)
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().getTriggersForJob(jobKey);
  }

  public List<String> getTriggerGroupNames()
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().getTriggerGroupNames();
  }

  public Set<TriggerKey> getTriggerKeys(GroupMatcher<TriggerKey> matcher)
    throws SchedulerException
  {
    validateState();

    if (matcher == null) {
      matcher = GroupMatcher.groupEquals("DEFAULT");
    }

    return this.resources.getJobStore().getTriggerKeys(matcher);
  }

  public JobDetail getJobDetail(JobKey jobKey)
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().retrieveJob(jobKey);
  }

  public Trigger getTrigger(TriggerKey triggerKey)
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().retrieveTrigger(triggerKey);
  }

  public boolean checkExists(JobKey jobKey)
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().checkExists(jobKey);
  }

  public boolean checkExists(TriggerKey triggerKey)
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().checkExists(triggerKey);
  }

  public void clear()
    throws SchedulerException
  {
    validateState();

    this.resources.getJobStore().clearAllSchedulingData();
    notifySchedulerListenersUnscheduled(null);
  }

  public Trigger.TriggerState getTriggerState(TriggerKey triggerKey)
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().getTriggerState(triggerKey);
  }

  public void addCalendar(String calName, Calendar calendar, boolean replace, boolean updateTriggers)
    throws SchedulerException
  {
    validateState();

    this.resources.getJobStore().storeCalendar(calName, calendar, replace, updateTriggers);
  }

  public boolean deleteCalendar(String calName)
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().removeCalendar(calName);
  }

  public Calendar getCalendar(String calName)
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().retrieveCalendar(calName);
  }

  public List<String> getCalendarNames()
    throws SchedulerException
  {
    validateState();

    return this.resources.getJobStore().getCalendarNames();
  }

  public ListenerManager getListenerManager() {
    return this.listenerManager;
  }

  public void addInternalJobListener(JobListener jobListener)
  {
    if ((jobListener.getName() == null) || (jobListener.getName().length() == 0))
    {
      throw new IllegalArgumentException("JobListener name cannot be empty.");
    }

    synchronized (this.internalJobListeners) {
      this.internalJobListeners.put(jobListener.getName(), jobListener);
    }
  }

  public boolean removeInternalJobListener(String name)
  {
    synchronized (this.internalJobListeners) {
      return this.internalJobListeners.remove(name) != null;
    }
  }

  public List<JobListener> getInternalJobListeners()
  {
    synchronized (this.internalJobListeners) {
      return Collections.unmodifiableList(new LinkedList(this.internalJobListeners.values()));
    }
  }

  public JobListener getInternalJobListener(String name)
  {
    synchronized (this.internalJobListeners) {
      return (JobListener)this.internalJobListeners.get(name);
    }
  }

  public void addInternalTriggerListener(TriggerListener triggerListener)
  {
    if ((triggerListener.getName() == null) || (triggerListener.getName().length() == 0))
    {
      throw new IllegalArgumentException("TriggerListener name cannot be empty.");
    }

    synchronized (this.internalTriggerListeners) {
      this.internalTriggerListeners.put(triggerListener.getName(), triggerListener);
    }
  }

  public boolean removeinternalTriggerListener(String name)
  {
    synchronized (this.internalTriggerListeners) {
      return this.internalTriggerListeners.remove(name) != null;
    }
  }

  public List<TriggerListener> getInternalTriggerListeners()
  {
    synchronized (this.internalTriggerListeners) {
      return Collections.unmodifiableList(new LinkedList(this.internalTriggerListeners.values()));
    }
  }

  public TriggerListener getInternalTriggerListener(String name)
  {
    synchronized (this.internalTriggerListeners) {
      return (TriggerListener)this.internalTriggerListeners.get(name);
    }
  }

  public void addInternalSchedulerListener(SchedulerListener schedulerListener)
  {
    synchronized (this.internalSchedulerListeners) {
      this.internalSchedulerListeners.add(schedulerListener);
    }
  }

  public boolean removeInternalSchedulerListener(SchedulerListener schedulerListener)
  {
    synchronized (this.internalSchedulerListeners) {
      return this.internalSchedulerListeners.remove(schedulerListener);
    }
  }

  public List<SchedulerListener> getInternalSchedulerListeners()
  {
    synchronized (this.internalSchedulerListeners) {
      return Collections.unmodifiableList(new ArrayList(this.internalSchedulerListeners));
    }
  }

  protected void notifyJobStoreJobComplete(OperableTrigger trigger, JobDetail detail, Trigger.CompletedExecutionInstruction instCode)
    throws JobPersistenceException
  {
    this.resources.getJobStore().triggeredJobComplete(trigger, detail, instCode);
  }

  protected void notifyJobStoreJobVetoed(OperableTrigger trigger, JobDetail detail, Trigger.CompletedExecutionInstruction instCode)
    throws JobPersistenceException
  {
    this.resources.getJobStore().triggeredJobComplete(trigger, detail, instCode);
  }

  protected void notifySchedulerThread(long candidateNewNextFireTime) {
    if (isSignalOnSchedulingChange())
      this.signaler.signalSchedulingChange(candidateNewNextFireTime);
  }

  private List<TriggerListener> buildTriggerListenerList()
    throws SchedulerException
  {
    List allListeners = new LinkedList();
    allListeners.addAll(getListenerManager().getTriggerListeners());
    allListeners.addAll(getInternalTriggerListeners());

    return allListeners;
  }

  private List<JobListener> buildJobListenerList() throws SchedulerException
  {
    List allListeners = new LinkedList();
    allListeners.addAll(getListenerManager().getJobListeners());
    allListeners.addAll(getInternalJobListeners());

    return allListeners;
  }

  private List<SchedulerListener> buildSchedulerListenerList() {
    List allListeners = new LinkedList();
    allListeners.addAll(getListenerManager().getSchedulerListeners());
    allListeners.addAll(getInternalSchedulerListeners());

    return allListeners;
  }

  private boolean matchJobListener(JobListener listener, JobKey key) {
    List matchers = getListenerManager().getJobListenerMatchers(listener.getName());
    if (matchers == null)
      return true;
    for (Matcher matcher : matchers) {
      if (matcher.isMatch(key))
        return true;
    }
    return false;
  }

  private boolean matchTriggerListener(TriggerListener listener, TriggerKey key) {
    List matchers = getListenerManager().getTriggerListenerMatchers(listener.getName());
    if (matchers == null)
      return true;
    for (Matcher matcher : matchers) {
      if (matcher.isMatch(key))
        return true;
    }
    return false;
  }

  public boolean notifyTriggerListenersFired(JobExecutionContext jec)
    throws SchedulerException
  {
    boolean vetoedExecution = false;

    List triggerListeners = buildTriggerListenerList();

    for (TriggerListener tl : triggerListeners) {
      try {
        if (matchTriggerListener(tl, jec.getTrigger().getKey()))
        {
          tl.triggerFired(jec.getTrigger(), jec);

          if (tl.vetoJobExecution(jec.getTrigger(), jec))
            vetoedExecution = true;
        }
      } catch (Exception e) {
        SchedulerException se = new SchedulerException("TriggerListener '" + tl.getName() + "' threw exception: " + e.getMessage(), e);

        throw se;
      }
    }

    return vetoedExecution;
  }

  public void notifyTriggerListenersMisfired(Trigger trigger)
    throws SchedulerException
  {
    List triggerListeners = buildTriggerListenerList();

    for (TriggerListener tl : triggerListeners)
      try {
        if (matchTriggerListener(tl, trigger.getKey()))
        {
          tl.triggerMisfired(trigger);
        }
      } catch (Exception e) { SchedulerException se = new SchedulerException("TriggerListener '" + tl.getName() + "' threw exception: " + e.getMessage(), e);

        throw se;
      }
  }

  public void notifyTriggerListenersComplete(JobExecutionContext jec, Trigger.CompletedExecutionInstruction instCode)
    throws SchedulerException
  {
    List triggerListeners = buildTriggerListenerList();

    for (TriggerListener tl : triggerListeners)
      try {
        if (matchTriggerListener(tl, jec.getTrigger().getKey()))
        {
          tl.triggerComplete(jec.getTrigger(), jec, instCode);
        }
      } catch (Exception e) { SchedulerException se = new SchedulerException("TriggerListener '" + tl.getName() + "' threw exception: " + e.getMessage(), e);

        throw se;
      }
  }

  public void notifyJobListenersToBeExecuted(JobExecutionContext jec)
    throws SchedulerException
  {
    List jobListeners = buildJobListenerList();

    for (JobListener jl : jobListeners)
      try {
        if (matchJobListener(jl, jec.getJobDetail().getKey()))
        {
          jl.jobToBeExecuted(jec);
        }
      } catch (Exception e) { SchedulerException se = new SchedulerException("JobListener '" + jl.getName() + "' threw exception: " + e.getMessage(), e);

        throw se;
      }
  }

  public void notifyJobListenersWasVetoed(JobExecutionContext jec)
    throws SchedulerException
  {
    List jobListeners = buildJobListenerList();

    for (JobListener jl : jobListeners)
      try {
        if (matchJobListener(jl, jec.getJobDetail().getKey()))
        {
          jl.jobExecutionVetoed(jec);
        }
      } catch (Exception e) { SchedulerException se = new SchedulerException("JobListener '" + jl.getName() + "' threw exception: " + e.getMessage(), e);

        throw se;
      }
  }

  public void notifyJobListenersWasExecuted(JobExecutionContext jec, JobExecutionException je)
    throws SchedulerException
  {
    List jobListeners = buildJobListenerList();

    for (JobListener jl : jobListeners)
      try {
        if (matchJobListener(jl, jec.getJobDetail().getKey()))
        {
          jl.jobWasExecuted(jec, je);
        }
      } catch (Exception e) { SchedulerException se = new SchedulerException("JobListener '" + jl.getName() + "' threw exception: " + e.getMessage(), e);

        throw se;
      }
  }

  public void notifySchedulerListenersError(String msg, SchedulerException se)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.schedulerError(msg, se);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of error: ", e);

        getLog().error("  Original error (for notification) was: " + msg, se);
      }
  }

  public void notifySchedulerListenersSchduled(Trigger trigger)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.jobScheduled(trigger);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of scheduled job.  Triger=" + trigger.getKey(), e);
      }
  }

  public void notifySchedulerListenersUnscheduled(TriggerKey triggerKey)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        if (triggerKey == null)
          sl.schedulingDataCleared();
        else
          sl.jobUnscheduled(triggerKey);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of unscheduled job.  Triger=" + (triggerKey == null ? "ALL DATA" : triggerKey), e);
      }
  }

  public void notifySchedulerListenersFinalized(Trigger trigger)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.triggerFinalized(trigger);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of finalized trigger.  Triger=" + trigger.getKey(), e);
      }
  }

  public void notifySchedulerListenersPausedTrigger(TriggerKey triggerKey)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.triggerPaused(triggerKey);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of paused trigger: " + triggerKey, e);
      }
  }

  public void notifySchedulerListenersPausedTriggers(String group)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.triggersPaused(group);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of paused trigger group." + group, e);
      }
  }

  public void notifySchedulerListenersResumedTrigger(TriggerKey key)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.triggerResumed(key);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of resumed trigger: " + key, e);
      }
  }

  public void notifySchedulerListenersResumedTriggers(String group)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.triggersResumed(group);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of resumed group: " + group, e);
      }
  }

  public void notifySchedulerListenersPausedJob(JobKey key)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.jobPaused(key);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of paused job: " + key, e);
      }
  }

  public void notifySchedulerListenersPausedJobs(String group)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.jobsPaused(group);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of paused job group: " + group, e);
      }
  }

  public void notifySchedulerListenersResumedJob(JobKey key)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.jobResumed(key);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of resumed job: " + key, e);
      }
  }

  public void notifySchedulerListenersResumedJobs(String group)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.jobsResumed(group);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of resumed job group: " + group, e);
      }
  }

  public void notifySchedulerListenersInStandbyMode()
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.schedulerInStandbyMode();
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of inStandByMode.", e);
      }
  }

  public void notifySchedulerListenersStarted()
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.schedulerStarted();
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of startup.", e);
      }
  }

  public void notifySchedulerListenersShutdown()
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.schedulerShutdown();
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of shutdown.", e);
      }
  }

  public void notifySchedulerListenersShuttingdown()
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.schedulerShuttingdown();
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of shutdown.", e);
      }
  }

  public void notifySchedulerListenersJobAdded(JobDetail jobDetail)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.jobAdded(jobDetail);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of JobAdded.", e);
      }
  }

  public void notifySchedulerListenersJobDeleted(JobKey jobKey)
  {
    List schedListeners = buildSchedulerListenerList();

    for (SchedulerListener sl : schedListeners)
      try {
        sl.jobDeleted(jobKey);
      } catch (Exception e) {
        getLog().error("Error while notifying SchedulerListener of JobAdded.", e);
      }
  }

  public void setJobFactory(JobFactory factory)
    throws SchedulerException
  {
    if (factory == null) {
      throw new IllegalArgumentException("JobFactory cannot be set to null!");
    }

    getLog().info("JobFactory set to: " + factory);

    this.jobFactory = factory;
  }

  public JobFactory getJobFactory() {
    return this.jobFactory;
  }

  public boolean interrupt(JobKey jobKey)
    throws UnableToInterruptJobException
  {
    List jobs = getCurrentlyExecutingJobs();

    JobDetail jobDetail = null;
    Job job = null;

    boolean interrupted = false;

    for (JobExecutionContext jec : jobs) {
      jobDetail = jec.getJobDetail();
      if (jobKey.equals(jobDetail.getKey())) {
        job = jec.getJobInstance();
        if ((job instanceof InterruptableJob)) {
          ((InterruptableJob)job).interrupt();
          interrupted = true;
        } else {
          throw new UnableToInterruptJobException("Job " + jobDetail.getKey() + " can not be interrupted, since it does not implement " + InterruptableJob.class.getName());
        }

      }

    }

    return interrupted;
  }

  public boolean interrupt(String fireInstanceId)
    throws UnableToInterruptJobException
  {
    List jobs = getCurrentlyExecutingJobs();

    Job job = null;

    for (JobExecutionContext jec : jobs) {
      if (jec.getFireInstanceId().equals(fireInstanceId)) {
        job = jec.getJobInstance();
        if ((job instanceof InterruptableJob)) {
          ((InterruptableJob)job).interrupt();
          return true;
        }
        throw new UnableToInterruptJobException("Job " + jec.getJobDetail().getKey() + " can not be interrupted, since it does not implement " + InterruptableJob.class.getName());
      }

    }

    return false;
  }

  private void shutdownPlugins() {
    Iterator itr = this.resources.getSchedulerPlugins().iterator();
    while (itr.hasNext()) {
      SchedulerPlugin plugin = (SchedulerPlugin)itr.next();
      plugin.shutdown();
    }
  }

  private void startPlugins() {
    Iterator itr = this.resources.getSchedulerPlugins().iterator();
    while (itr.hasNext()) {
      SchedulerPlugin plugin = (SchedulerPlugin)itr.next();
      plugin.start();
    }
  }

  static
  {
    Properties props = new Properties();
    InputStream is = null;
    try {
      is = QuartzScheduler.class.getResourceAsStream("quartz-build.properties");
      if (is != null) {
        props.load(is);
        String version = props.getProperty("version");
        if (version != null) {
          String[] versionComponents = version.split("\\.");
          VERSION_MAJOR = versionComponents[0];
          VERSION_MINOR = versionComponents[1];
          if (versionComponents.length > 2)
            VERSION_ITERATION = versionComponents[2];
          else
            VERSION_ITERATION = "0";
        } else {
          LoggerFactory.getLogger(QuartzScheduler.class).error("Can't parse Quartz version from quartz-build.properties");
        }
      }
    }
    catch (Exception e) {
      LoggerFactory.getLogger(QuartzScheduler.class).error("Error loading version info from quartz-build.properties.", e);
    }
    finally {
      if (is != null) try {
          is.close();
        }
        catch (Exception ignore)
        {
        }
    }
  }
}