package org.quartz.core;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Random;
import java.util.concurrent.atomic.AtomicBoolean;
import org.quartz.JobPersistenceException;
import org.quartz.SchedulerException;
import org.quartz.Trigger.CompletedExecutionInstruction;
import org.quartz.spi.JobStore;
import org.quartz.spi.OperableTrigger;
import org.quartz.spi.ThreadPool;
import org.quartz.spi.TriggerFiredBundle;
import org.quartz.spi.TriggerFiredResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class QuartzSchedulerThread extends Thread
{
  private QuartzScheduler qs;
  private QuartzSchedulerResources qsRsrcs;
  private final Object sigLock = new Object();
  private boolean signaled;
  private long signaledNextFireTime;
  private boolean paused;
  private AtomicBoolean halted;
  private Random random = new Random(System.currentTimeMillis());

  private static long DEFAULT_IDLE_WAIT_TIME = 30000L;

  private long idleWaitTime = DEFAULT_IDLE_WAIT_TIME;

  private int idleWaitVariablness = 7000;

  private long dbFailureRetryInterval = 15000L;

  private final Logger log = LoggerFactory.getLogger(getClass());

  QuartzSchedulerThread(QuartzScheduler qs, QuartzSchedulerResources qsRsrcs)
  {
    this(qs, qsRsrcs, qsRsrcs.getMakeSchedulerThreadDaemon(), 5);
  }

  QuartzSchedulerThread(QuartzScheduler qs, QuartzSchedulerResources qsRsrcs, boolean setDaemon, int threadPrio)
  {
    super(qs.getSchedulerThreadGroup(), qsRsrcs.getThreadName());
    this.qs = qs;
    this.qsRsrcs = qsRsrcs;
    setDaemon(setDaemon);
    if (qsRsrcs.isThreadsInheritInitializersClassLoadContext()) {
      this.log.info("QuartzSchedulerThread Inheriting ContextClassLoader of thread: " + Thread.currentThread().getName());
      setContextClassLoader(Thread.currentThread().getContextClassLoader());
    }

    setPriority(threadPrio);

    this.paused = true;
    this.halted = new AtomicBoolean(false);
  }

  void setIdleWaitTime(long waitTime)
  {
    this.idleWaitTime = waitTime;
    this.idleWaitVariablness = ((int)(waitTime * 0.2D));
  }

  private long getDbFailureRetryInterval() {
    return this.dbFailureRetryInterval;
  }

  public void setDbFailureRetryInterval(long dbFailureRetryInterval) {
    this.dbFailureRetryInterval = dbFailureRetryInterval;
  }

  private long getRandomizedIdleWaitTime() {
    return this.idleWaitTime - this.random.nextInt(this.idleWaitVariablness);
  }

  void togglePause(boolean pause)
  {
    synchronized (this.sigLock) {
      this.paused = pause;

      if (this.paused)
        signalSchedulingChange(0L);
      else
        this.sigLock.notifyAll();
    }
  }

  void halt()
  {
    synchronized (this.sigLock) {
      this.halted.set(true);

      if (this.paused)
        this.sigLock.notifyAll();
      else
        signalSchedulingChange(0L);
    }
  }

  boolean isPaused()
  {
    return this.paused;
  }

  public void signalSchedulingChange(long candidateNewNextFireTime)
  {
    synchronized (this.sigLock) {
      this.signaled = true;
      this.signaledNextFireTime = candidateNewNextFireTime;
      this.sigLock.notifyAll();
    }
  }

  public void clearSignaledSchedulingChange() {
    synchronized (this.sigLock) {
      this.signaled = false;
      this.signaledNextFireTime = 0L;
    }
  }

  public boolean isScheduleChanged() {
    synchronized (this.sigLock) {
      return this.signaled;
    }
  }

  public long getSignaledNextFireTime() {
    synchronized (this.sigLock) {
      return this.signaledNextFireTime;
    }
  }

  public void run()
  {
    boolean lastAcquireFailed = false;

    while (!this.halted.get()) {
      try
      {
        synchronized (this.sigLock) {
          while ((this.paused) && (!this.halted.get()))
            try
            {
              this.sigLock.wait(1000L);
            }
            catch (InterruptedException ignore)
            {
            }
          if (this.halted.get()) {
            break;
          }
        }

        int availThreadCount = this.qsRsrcs.getThreadPool().blockForAvailableThreads();
        if (availThreadCount > 0)
        {
          List triggers = null;

          long now = System.currentTimeMillis();

          clearSignaledSchedulingChange();
          try {
            triggers = this.qsRsrcs.getJobStore().acquireNextTriggers(now + this.idleWaitTime, Math.min(availThreadCount, this.qsRsrcs.getMaxBatchSize()), this.qsRsrcs.getBatchTimeWindow());

            lastAcquireFailed = false;
            if (this.log.isDebugEnabled())
              this.log.debug("batch acquisition of " + (triggers == null ? 0 : triggers.size()) + " triggers");
          } catch (JobPersistenceException jpe) {
            if (!lastAcquireFailed) {
              this.qs.notifySchedulerListenersError("An error occurred while scanning for the next triggers to fire.", jpe);
            }

            lastAcquireFailed = true;
          } catch (RuntimeException e) {
            if (!lastAcquireFailed) {
              getLog().error("quartzSchedulerThreadLoop: RuntimeException " + e.getMessage(), e);
            }

            lastAcquireFailed = true;
          }

          if ((triggers != null) && (!triggers.isEmpty()))
          {
            now = System.currentTimeMillis();
            long triggerTime = ((OperableTrigger)triggers.get(0)).getNextFireTime().getTime();
            long timeUntilTrigger = triggerTime - now;
            while (timeUntilTrigger > 2L) {
              synchronized (this.sigLock) {
                if (this.halted.get()) {
                  break;
                }
                if (!isCandidateNewTimeEarlierWithinReason(triggerTime, false))
                {
                  try
                  {
                    now = System.currentTimeMillis();
                    timeUntilTrigger = triggerTime - now;
                    if (timeUntilTrigger >= 1L)
                      this.sigLock.wait(timeUntilTrigger);
                  } catch (InterruptedException ignore) {
                  }
                }
              }
              if (releaseIfScheduleChangedSignificantly(triggers, triggerTime)) {
                break;
              }
              now = System.currentTimeMillis();
              timeUntilTrigger = triggerTime - now;
            }

            if (!triggers.isEmpty())
            {
              List bndles = new ArrayList();

              boolean goAhead = true;
              synchronized (this.sigLock) {
                goAhead = !this.halted.get();
              }
              if (goAhead) {
                try {
                  List res = this.qsRsrcs.getJobStore().triggersFired(triggers);
                  if (res != null)
                    bndles = res;
                } catch (SchedulerException se) {
                  this.qs.notifySchedulerListenersError("An error occurred while firing triggers '" + triggers + "'", se);
                }

              }

              for (int i = 0; i < bndles.size(); i++) {
                TriggerFiredResult result = (TriggerFiredResult)bndles.get(i);
                TriggerFiredBundle bndle = result.getTriggerFiredBundle();
                Exception exception = result.getException();

                if ((exception instanceof RuntimeException)) {
                  getLog().error("RuntimeException while firing trigger " + triggers.get(i), exception);

                  releaseTriggerRetryLoop((OperableTrigger)triggers.get(i));
                }
                else if (bndle == null) {
                  try {
                    this.qsRsrcs.getJobStore().releaseAcquiredTrigger((OperableTrigger)triggers.get(i));
                  } catch (SchedulerException se) {
                    this.qs.notifySchedulerListenersError("An error occurred while releasing triggers '" + ((OperableTrigger)triggers.get(i)).getKey() + "'", se);

                    releaseTriggerRetryLoop((OperableTrigger)triggers.get(i));
                  }

                }
                else
                {
                  JobRunShell shell = null;
                  try {
                    shell = this.qsRsrcs.getJobRunShellFactory().createJobRunShell(bndle);
                    shell.initialize(this.qs);
                  } catch (SchedulerException se) {
                    try {
                      this.qsRsrcs.getJobStore().triggeredJobComplete((OperableTrigger)triggers.get(i), bndle.getJobDetail(), Trigger.CompletedExecutionInstruction.SET_ALL_JOB_TRIGGERS_ERROR);
                    }
                    catch (SchedulerException se2) {
                      this.qs.notifySchedulerListenersError("An error occurred while placing job's triggers in error state '" + ((OperableTrigger)triggers.get(i)).getKey() + "'", se2);

                      errorTriggerRetryLoop(bndle);
                    }
                    continue;
                  }

                  if (!this.qsRsrcs.getThreadPool().runInThread(shell))
                  {
                    try
                    {
                      getLog().error("ThreadPool.runInThread() return false!");
                      this.qsRsrcs.getJobStore().triggeredJobComplete((OperableTrigger)triggers.get(i), bndle.getJobDetail(), Trigger.CompletedExecutionInstruction.SET_ALL_JOB_TRIGGERS_ERROR);
                    }
                    catch (SchedulerException se2) {
                      this.qs.notifySchedulerListenersError("An error occurred while placing job's triggers in error state '" + ((OperableTrigger)triggers.get(i)).getKey() + "'", se2);

                      releaseTriggerRetryLoop((OperableTrigger)triggers.get(i));
                    }
                  }
                }
              }
            }
          }
          else;
        }
        else {
          continue;

          long now = System.currentTimeMillis();
          long waitTime = now + getRandomizedIdleWaitTime();
          long timeUntilContinue = waitTime - now;
          synchronized (this.sigLock) {
            try {
              if (!this.halted.get())
                this.sigLock.wait(timeUntilContinue);
            } catch (InterruptedException ignore) {
            }
          }
        }
      } catch (RuntimeException re) {
        getLog().error("Runtime error occurred in main trigger firing loop.", re);
      }

    }

    this.qs = null;
    this.qsRsrcs = null;
  }

  private boolean releaseIfScheduleChangedSignificantly(List<OperableTrigger> triggers, long triggerTime)
  {
    if (isCandidateNewTimeEarlierWithinReason(triggerTime, true)) {
      for (OperableTrigger trigger : triggers) {
        try
        {
          this.qsRsrcs.getJobStore().releaseAcquiredTrigger(trigger);
        } catch (JobPersistenceException jpe) {
          this.qs.notifySchedulerListenersError("An error occurred while releasing trigger '" + trigger.getKey() + "'", jpe);

          releaseTriggerRetryLoop(trigger);
        } catch (RuntimeException e) {
          getLog().error("releaseTriggerRetryLoop: RuntimeException " + e.getMessage(), e);

          releaseTriggerRetryLoop(trigger);
        }
      }
      triggers.clear();
      return true;
    }
    return false;
  }

  private boolean isCandidateNewTimeEarlierWithinReason(long oldTime, boolean clearSignal)
  {
    synchronized (this.sigLock)
    {
      if (!isScheduleChanged()) {
        return false;
      }
      boolean earlier = false;

      if (getSignaledNextFireTime() == 0L)
        earlier = true;
      else if (getSignaledNextFireTime() < oldTime) {
        earlier = true;
      }
      if (earlier)
      {
        long diff = oldTime - System.currentTimeMillis();
        if (diff < (this.qsRsrcs.getJobStore().supportsPersistence() ? 70L : 7L)) {
          earlier = false;
        }
      }
      if (clearSignal) {
        clearSignaledSchedulingChange();
      }

      return earlier;
    }
  }

  public void errorTriggerRetryLoop(TriggerFiredBundle bndle) {
    int retryCount = 0;
    try {
      while (!this.halted.get())
        try {
          Thread.sleep(getDbFailureRetryInterval());

          retryCount++;
          this.qsRsrcs.getJobStore().triggeredJobComplete(bndle.getTrigger(), bndle.getJobDetail(), Trigger.CompletedExecutionInstruction.SET_ALL_JOB_TRIGGERS_ERROR);

          retryCount = 0;
        }
        catch (JobPersistenceException jpe) {
          if (retryCount % 4 == 0) {
            this.qs.notifySchedulerListenersError("An error occurred while releasing trigger '" + bndle.getTrigger().getKey() + "'", jpe);
          }
        }
        catch (RuntimeException e)
        {
          getLog().error("releaseTriggerRetryLoop: RuntimeException " + e.getMessage(), e);
        } catch (InterruptedException e) {
          getLog().error("releaseTriggerRetryLoop: InterruptedException " + e.getMessage(), e);
        }
    }
    finally {
      if (retryCount == 0)
        getLog().info("releaseTriggerRetryLoop: connection restored.");
    }
  }

  public void releaseTriggerRetryLoop(OperableTrigger trigger)
  {
    int retryCount = 0;
    try {
      while (!this.halted.get())
        try {
          Thread.sleep(getDbFailureRetryInterval());

          retryCount++;
          this.qsRsrcs.getJobStore().releaseAcquiredTrigger(trigger);
          retryCount = 0;
        }
        catch (JobPersistenceException jpe) {
          if (retryCount % 4 == 0) {
            this.qs.notifySchedulerListenersError("An error occurred while releasing trigger '" + trigger.getKey() + "'", jpe);
          }
        }
        catch (RuntimeException e)
        {
          getLog().error("releaseTriggerRetryLoop: RuntimeException " + e.getMessage(), e);
        } catch (InterruptedException e) {
          getLog().error("releaseTriggerRetryLoop: InterruptedException " + e.getMessage(), e);
        }
    }
    finally {
      if (retryCount == 0)
        getLog().info("releaseTriggerRetryLoop: connection restored.");
    }
  }

  public Logger getLog()
  {
    return this.log;
  }
}