package org.quartz.impl;

import java.io.Serializable;
import org.quartz.DisallowConcurrentExecution;
import org.quartz.Job;
import org.quartz.JobBuilder;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobKey;
import org.quartz.PersistJobDataAfterExecution;
import org.quartz.utils.ClassUtils;

public class JobDetailImpl
  implements Cloneable, Serializable, JobDetail
{
  private String name;
  private String group = "DEFAULT";
  private String description;
  private Class<? extends Job> jobClass;
  private JobDataMap jobDataMap;
  private boolean durability = false;

  private boolean shouldRecover = false;

  private transient JobKey key = null;

  public JobDetailImpl()
  {
  }

  /** @deprecated */
  public JobDetailImpl(String name, Class<? extends Job> jobClass)
  {
    this(name, null, jobClass);
  }

  /** @deprecated */
  public JobDetailImpl(String name, String group, Class<? extends Job> jobClass)
  {
    setName(name);
    setGroup(group);
    setJobClass(jobClass);
  }

  /** @deprecated */
  public JobDetailImpl(String name, String group, Class<? extends Job> jobClass, boolean durability, boolean recover)
  {
    setName(name);
    setGroup(group);
    setJobClass(jobClass);
    setDurability(durability);
    setRequestsRecovery(recover);
  }

  public String getName()
  {
    return this.name;
  }

  public void setName(String name)
  {
    if ((name == null) || (name.trim().length() == 0)) {
      throw new IllegalArgumentException("Job name cannot be empty.");
    }

    this.name = name;
    this.key = null;
  }

  public String getGroup()
  {
    return this.group;
  }

  public void setGroup(String group)
  {
    if ((group != null) && (group.trim().length() == 0)) {
      throw new IllegalArgumentException("Group name cannot be empty.");
    }

    if (group == null) {
      group = "DEFAULT";
    }

    this.group = group;
    this.key = null;
  }

  public String getFullName()
  {
    return this.group + "." + this.name;
  }

  public JobKey getKey()
  {
    if (this.key == null) {
      if (getName() == null)
        return null;
      this.key = new JobKey(getName(), getGroup());
    }

    return this.key;
  }

  public void setKey(JobKey key) {
    if (key == null) {
      throw new IllegalArgumentException("Key cannot be null!");
    }
    setName(key.getName());
    setGroup(key.getGroup());
    this.key = key;
  }

  public String getDescription()
  {
    return this.description;
  }

  public void setDescription(String description)
  {
    this.description = description;
  }

  public Class<? extends Job> getJobClass()
  {
    return this.jobClass;
  }

  public void setJobClass(Class<? extends Job> jobClass)
  {
    if (jobClass == null) {
      throw new IllegalArgumentException("Job class cannot be null.");
    }

    if (!Job.class.isAssignableFrom(jobClass)) {
      throw new IllegalArgumentException("Job class must implement the Job interface.");
    }

    this.jobClass = jobClass;
  }

  public JobDataMap getJobDataMap()
  {
    if (this.jobDataMap == null) {
      this.jobDataMap = new JobDataMap();
    }
    return this.jobDataMap;
  }

  public void setJobDataMap(JobDataMap jobDataMap)
  {
    this.jobDataMap = jobDataMap;
  }

  public void setDurability(boolean durability)
  {
    this.durability = durability;
  }

  public void setRequestsRecovery(boolean shouldRecover)
  {
    this.shouldRecover = shouldRecover;
  }

  public boolean isDurable()
  {
    return this.durability;
  }

  public boolean isPersistJobDataAfterExecution()
  {
    return ClassUtils.isAnnotationPresent(this.jobClass, PersistJobDataAfterExecution.class);
  }

  public boolean isConcurrentExectionDisallowed()
  {
    return ClassUtils.isAnnotationPresent(this.jobClass, DisallowConcurrentExecution.class);
  }

  public boolean requestsRecovery()
  {
    return this.shouldRecover;
  }

  public String toString()
  {
    return "JobDetail '" + getFullName() + "':  jobClass: '" + (getJobClass() == null ? null : getJobClass().getName()) + " concurrentExectionDisallowed: " + isConcurrentExectionDisallowed() + " persistJobDataAfterExecution: " + isPersistJobDataAfterExecution() + " isDurable: " + isDurable() + " requestsRecovers: " + requestsRecovery();
  }

  public boolean equals(Object obj)
  {
    if (!(obj instanceof JobDetail)) {
      return false;
    }

    JobDetail other = (JobDetail)obj;

    if ((other.getKey() == null) || (getKey() == null)) {
      return false;
    }
    if (!other.getKey().equals(getKey())) {
      return false;
    }

    return true;
  }

  public int hashCode()
  {
    return getKey().hashCode();
  }

  public Object clone()
  {
    JobDetailImpl copy;
    try {
      copy = (JobDetailImpl)super.clone();
      if (this.jobDataMap != null)
        copy.jobDataMap = ((JobDataMap)this.jobDataMap.clone());
    }
    catch (CloneNotSupportedException ex) {
      throw new IncompatibleClassChangeError("Not Cloneable.");
    }

    return copy;
  }

  public JobBuilder getJobBuilder() {
    JobBuilder b = JobBuilder.newJob().ofType(getJobClass()).requestRecovery(requestsRecovery()).storeDurably(isDurable()).usingJobData(getJobDataMap()).withDescription(getDescription()).withIdentity(getKey());

    return b;
  }
}