package org.quartz.impl.jdbcjobstore;

import java.sql.Connection;
import java.util.HashSet;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.transaction.Synchronization;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class JTANonClusteredSemaphore
  implements Semaphore
{
  public static final String DEFAULT_TRANSACTION_MANANGER_LOCATION = "java:TransactionManager";
  ThreadLocal<HashSet<String>> lockOwners;
  HashSet<String> locks;
  private final Logger log;
  private String transactionManagerJNDIName;

  public JTANonClusteredSemaphore()
  {
    this.lockOwners = new ThreadLocal();

    this.locks = new HashSet();

    this.log = LoggerFactory.getLogger(getClass());

    this.transactionManagerJNDIName = "java:TransactionManager";
  }

  protected Logger getLog()
  {
    return this.log;
  }

  public void setTransactionManagerJNDIName(String transactionManagerJNDIName) {
    this.transactionManagerJNDIName = transactionManagerJNDIName;
  }

  private HashSet<String> getThreadLocks() {
    HashSet threadLocks = (HashSet)this.lockOwners.get();
    if (threadLocks == null) {
      threadLocks = new HashSet();
      this.lockOwners.set(threadLocks);
    }
    return threadLocks;
  }

  public synchronized boolean obtainLock(Connection conn, String lockName)
    throws LockException
  {
    lockName = lockName.intern();

    if (this.log.isDebugEnabled()) {
      this.log.debug("Lock '" + lockName + "' is desired by: " + Thread.currentThread().getName());
    }

    if (!isLockOwner(conn, lockName)) {
      if (this.log.isDebugEnabled()) {
        this.log.debug("Lock '" + lockName + "' is being obtained: " + Thread.currentThread().getName());
      }

      while (this.locks.contains(lockName)) {
        try {
          wait();
        } catch (InterruptedException ie) {
          if (this.log.isDebugEnabled()) {
            this.log.debug("Lock '" + lockName + "' was not obtained by: " + Thread.currentThread().getName());
          }

        }

      }

      Transaction t = getTransaction();
      if (t != null) {
        try {
          t.registerSynchronization(new SemaphoreSynchronization(lockName));
        } catch (Exception e) {
          throw new LockException("Failed to register semaphore with Transaction.", e);
        }
      }

      if (this.log.isDebugEnabled()) {
        this.log.debug("Lock '" + lockName + "' given to: " + Thread.currentThread().getName());
      }

      getThreadLocks().add(lockName);
      this.locks.add(lockName);
    } else if (this.log.isDebugEnabled()) {
      this.log.debug("Lock '" + lockName + "' already owned by: " + Thread.currentThread().getName() + " -- but not owner!", new Exception("stack-trace of wrongful returner"));
    }

    return true;
  }

  protected Transaction getTransaction()
    throws LockException
  {
    InitialContext ic = null;
    try {
      ic = new InitialContext();
      TransactionManager tm = (TransactionManager)ic.lookup(this.transactionManagerJNDIName);

      return tm.getTransaction();
    } catch (SystemException e) {
      throw new LockException("Failed to get Transaction from TransactionManager", e);
    } catch (NamingException e) {
      throw new LockException("Failed to find TransactionManager in JNDI under name: " + this.transactionManagerJNDIName, e);
    } finally {
      if (ic != null)
        try {
          ic.close();
        }
        catch (NamingException ignored)
        {
        }
    }
  }

  public synchronized void releaseLock(Connection conn, String lockName)
    throws LockException
  {
    releaseLock(lockName, false);
  }

  protected synchronized void releaseLock(String lockName, boolean fromSynchronization)
    throws LockException
  {
    lockName = lockName.intern();

    if (isLockOwner(null, lockName))
    {
      if (!fromSynchronization) {
        Transaction t = getTransaction();
        if (t != null) {
          if (getLog().isDebugEnabled()) {
            getLog().debug("Lock '" + lockName + "' is in a JTA transaction.  " + "Return deferred by: " + Thread.currentThread().getName());
          }

          return;
        }
      }

      if (getLog().isDebugEnabled()) {
        getLog().debug("Lock '" + lockName + "' returned by: " + Thread.currentThread().getName());
      }

      getThreadLocks().remove(lockName);
      this.locks.remove(lockName);
      notify();
    } else if (getLog().isDebugEnabled()) {
      getLog().debug("Lock '" + lockName + "' attempt to return by: " + Thread.currentThread().getName() + " -- but not owner!", new Exception("stack-trace of wrongful returner"));
    }
  }

  public synchronized boolean isLockOwner(Connection conn, String lockName)
  {
    lockName = lockName.intern();

    return getThreadLocks().contains(lockName);
  }

  public boolean requiresConnection()
  {
    return false;
  }

  private class SemaphoreSynchronization
    implements Synchronization
  {
    private String lockName;

    public SemaphoreSynchronization(String lockName)
    {
      this.lockName = lockName;
    }

    public void beforeCompletion()
    {
    }

    public void afterCompletion(int status) {
      try {
        JTANonClusteredSemaphore.this.releaseLock(this.lockName, true);
      }
      catch (LockException e)
      {
      }
    }
  }
}