package org.quartz.impl.jdbcjobstore;

import java.sql.Connection;
import java.sql.SQLException;
import org.quartz.JobPersistenceException;
import org.quartz.SchedulerConfigException;
import org.quartz.spi.ClassLoadHelper;
import org.quartz.spi.SchedulerSignaler;
import org.quartz.utils.DBConnectionManager;
import org.slf4j.Logger;

public class JobStoreCMT extends JobStoreSupport
{
  protected String nonManagedTxDsName;
  protected boolean dontSetNonManagedTXConnectionAutoCommitFalse = false;

  protected boolean setTxIsolationLevelReadCommitted = false;

  public void setNonManagedTXDataSource(String nonManagedTxDsName)
  {
    this.nonManagedTxDsName = nonManagedTxDsName;
  }

  public String getNonManagedTXDataSource()
  {
    return this.nonManagedTxDsName;
  }

  public boolean isDontSetNonManagedTXConnectionAutoCommitFalse() {
    return this.dontSetNonManagedTXConnectionAutoCommitFalse;
  }

  public void setDontSetNonManagedTXConnectionAutoCommitFalse(boolean b)
  {
    this.dontSetNonManagedTXConnectionAutoCommitFalse = b;
  }

  public boolean isTxIsolationLevelReadCommitted()
  {
    return this.setTxIsolationLevelReadCommitted;
  }

  public void setTxIsolationLevelReadCommitted(boolean b)
  {
    this.setTxIsolationLevelReadCommitted = b;
  }

  public void initialize(ClassLoadHelper loadHelper, SchedulerSignaler signaler)
    throws SchedulerConfigException
  {
    if (this.nonManagedTxDsName == null) {
      throw new SchedulerConfigException("Non-ManagedTX DataSource name not set!  If your 'org.quartz.jobStore.dataSource' is XA, then set 'org.quartz.jobStore.nonManagedTXDataSource' to a non-XA datasource (for the same DB).  Otherwise, you can set them to be the same.");
    }

    if (getLockHandler() == null)
    {
      setUseDBLocks(true);
    }

    super.initialize(loadHelper, signaler);

    getLog().info("JobStoreCMT initialized.");
  }

  public void shutdown()
  {
    super.shutdown();
    try
    {
      DBConnectionManager.getInstance().shutdown(getNonManagedTXDataSource());
    } catch (SQLException sqle) {
      getLog().warn("Database connection shutdown unsuccessful.", sqle);
    }
  }

  protected Connection getNonManagedTXConnection()
    throws JobPersistenceException
  {
    Connection conn = null;
    try {
      conn = DBConnectionManager.getInstance().getConnection(getNonManagedTXDataSource());
    }
    catch (SQLException sqle) {
      throw new JobPersistenceException("Failed to obtain DB connection from data source '" + getNonManagedTXDataSource() + "': " + sqle.toString(), sqle);
    }
    catch (Throwable e)
    {
      throw new JobPersistenceException("Failed to obtain DB connection from data source '" + getNonManagedTXDataSource() + "': " + e.toString(), e);
    }

    if (conn == null) {
      throw new JobPersistenceException("Could not get connection from DataSource '" + getNonManagedTXDataSource() + "'");
    }

    conn = getAttributeRestoringConnection(conn);
    try
    {
      if (!isDontSetNonManagedTXConnectionAutoCommitFalse()) {
        conn.setAutoCommit(false);
      }

      if (isTxIsolationLevelReadCommitted())
        conn.setTransactionIsolation(2);
    }
    catch (SQLException sqle) {
      getLog().warn("Failed to override connection auto commit/transaction isolation.", sqle); } catch (Throwable e) {
      try {
        conn.close(); } catch (Throwable tt) {
      }
      throw new JobPersistenceException("Failure setting up connection.", e);
    }

    return conn;
  }

  protected Object executeInLock(String lockName, JobStoreSupport.TransactionCallback txCallback)
    throws JobPersistenceException
  {
    boolean transOwner = false;
    Connection conn = null;
    try {
      if (lockName != null)
      {
        if (getLockHandler().requiresConnection()) {
          conn = getConnection();
        }

        transOwner = getLockHandler().obtainLock(conn, lockName);
      }

      if (conn == null) {
        conn = getConnection();
      }

      return txCallback.execute(conn);
    } finally {
      try {
        releaseLock(conn, "TRIGGER_ACCESS", transOwner);
      } finally {
        cleanupConnection(conn);
      }
    }
  }
}