package org.quartz.impl.triggers;

import java.util.Date;
import org.quartz.Calendar;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.quartz.JobKey;
import org.quartz.ScheduleBuilder;
import org.quartz.SchedulerException;
import org.quartz.Trigger;
import org.quartz.Trigger.CompletedExecutionInstruction;
import org.quartz.TriggerBuilder;
import org.quartz.TriggerKey;
import org.quartz.spi.OperableTrigger;

public abstract class AbstractTrigger<T extends Trigger>
  implements OperableTrigger
{
  private static final long serialVersionUID = -3904243490805975570L;
  private String name;
  private String group = "DEFAULT";
  private String jobName;
  private String jobGroup = "DEFAULT";
  private String description;
  private JobDataMap jobDataMap;
  private boolean volatility = false;

  private String calendarName = null;

  private String fireInstanceId = null;

  private int misfireInstruction = 0;

  private int priority = 5;

  private transient TriggerKey key = null;

  public AbstractTrigger()
  {
  }

  public AbstractTrigger(String name)
  {
    setName(name);
    setGroup(null);
  }

  public AbstractTrigger(String name, String group)
  {
    setName(name);
    setGroup(group);
  }

  public AbstractTrigger(String name, String group, String jobName, String jobGroup)
  {
    setName(name);
    setGroup(group);
    setJobName(jobName);
    setJobGroup(jobGroup);
  }

  public String getName()
  {
    return this.name;
  }

  public void setName(String name)
  {
    if ((name == null) || (name.trim().length() == 0)) {
      throw new IllegalArgumentException("Trigger name cannot be null or empty.");
    }

    this.name = name;
    this.key = null;
  }

  public String getGroup()
  {
    return this.group;
  }

  public void setGroup(String group)
  {
    if ((group != null) && (group.trim().length() == 0)) {
      throw new IllegalArgumentException("Group name cannot be an empty string.");
    }

    if (group == null) {
      group = "DEFAULT";
    }

    this.group = group;
    this.key = null;
  }

  public void setKey(TriggerKey key) {
    setName(key.getName());
    setGroup(key.getGroup());
    this.key = key;
  }

  public String getJobName()
  {
    return this.jobName;
  }

  public void setJobName(String jobName)
  {
    if ((jobName == null) || (jobName.trim().length() == 0)) {
      throw new IllegalArgumentException("Job name cannot be null or empty.");
    }

    this.jobName = jobName;
  }

  public String getJobGroup()
  {
    return this.jobGroup;
  }

  public void setJobGroup(String jobGroup)
  {
    if ((jobGroup != null) && (jobGroup.trim().length() == 0)) {
      throw new IllegalArgumentException("Group name cannot be null or empty.");
    }

    if (jobGroup == null) {
      jobGroup = "DEFAULT";
    }

    this.jobGroup = jobGroup;
  }

  public void setJobKey(JobKey key) {
    setJobName(key.getName());
    setJobGroup(key.getGroup());
  }

  public String getFullName()
  {
    return this.group + "." + this.name;
  }

  public TriggerKey getKey() {
    if (this.key == null) {
      if (getName() == null)
        return null;
      this.key = new TriggerKey(getName(), getGroup());
    }

    return this.key;
  }

  public JobKey getJobKey() {
    if (getJobName() == null) {
      return null;
    }
    return new JobKey(getJobName(), getJobGroup());
  }

  public String getFullJobName()
  {
    return this.jobGroup + "." + this.jobName;
  }

  public String getDescription()
  {
    return this.description;
  }

  public void setDescription(String description)
  {
    this.description = description;
  }

  public void setCalendarName(String calendarName)
  {
    this.calendarName = calendarName;
  }

  public String getCalendarName()
  {
    return this.calendarName;
  }

  public JobDataMap getJobDataMap()
  {
    if (this.jobDataMap == null) {
      this.jobDataMap = new JobDataMap();
    }
    return this.jobDataMap;
  }

  public void setJobDataMap(JobDataMap jobDataMap)
  {
    this.jobDataMap = jobDataMap;
  }

  public int getPriority()
  {
    return this.priority;
  }

  public void setPriority(int priority)
  {
    this.priority = priority;
  }

  public abstract void triggered(Calendar paramCalendar);

  public abstract Date computeFirstFireTime(Calendar paramCalendar);

  public Trigger.CompletedExecutionInstruction executionComplete(JobExecutionContext context, JobExecutionException result)
  {
    if ((result != null) && (result.refireImmediately())) {
      return Trigger.CompletedExecutionInstruction.RE_EXECUTE_JOB;
    }

    if ((result != null) && (result.unscheduleFiringTrigger())) {
      return Trigger.CompletedExecutionInstruction.SET_TRIGGER_COMPLETE;
    }

    if ((result != null) && (result.unscheduleAllTriggers())) {
      return Trigger.CompletedExecutionInstruction.SET_ALL_JOB_TRIGGERS_COMPLETE;
    }

    if (!mayFireAgain()) {
      return Trigger.CompletedExecutionInstruction.DELETE_TRIGGER;
    }

    return Trigger.CompletedExecutionInstruction.NOOP;
  }

  public abstract boolean mayFireAgain();

  public abstract Date getStartTime();

  public abstract void setStartTime(Date paramDate);

  public abstract void setEndTime(Date paramDate);

  public abstract Date getEndTime();

  public abstract Date getNextFireTime();

  public abstract Date getPreviousFireTime();

  public abstract Date getFireTimeAfter(Date paramDate);

  public abstract Date getFinalFireTime();

  public void setMisfireInstruction(int misfireInstruction)
  {
    if (!validateMisfireInstruction(misfireInstruction)) {
      throw new IllegalArgumentException("The misfire instruction code is invalid for this type of trigger.");
    }

    this.misfireInstruction = misfireInstruction;
  }

  protected abstract boolean validateMisfireInstruction(int paramInt);

  public int getMisfireInstruction()
  {
    return this.misfireInstruction;
  }

  public abstract void updateAfterMisfire(Calendar paramCalendar);

  public abstract void updateWithNewCalendar(Calendar paramCalendar, long paramLong);

  public void validate()
    throws SchedulerException
  {
    if (this.name == null) {
      throw new SchedulerException("Trigger's name cannot be null");
    }

    if (this.group == null) {
      throw new SchedulerException("Trigger's group cannot be null");
    }

    if (this.jobName == null) {
      throw new SchedulerException("Trigger's related Job's name cannot be null");
    }

    if (this.jobGroup == null)
      throw new SchedulerException("Trigger's related Job's group cannot be null");
  }

  public void setFireInstanceId(String id)
  {
    this.fireInstanceId = id;
  }

  public String getFireInstanceId()
  {
    return this.fireInstanceId;
  }

  public String toString()
  {
    return "Trigger '" + getFullName() + "':  triggerClass: '" + getClass().getName() + " calendar: '" + getCalendarName() + "' misfireInstruction: " + getMisfireInstruction() + " nextFireTime: " + getNextFireTime();
  }

  public int compareTo(Trigger other)
  {
    if ((other.getKey() == null) && (getKey() == null))
      return 0;
    if (other.getKey() == null)
      return -1;
    if (getKey() == null) {
      return 1;
    }
    return getKey().compareTo(other.getKey());
  }

  public boolean equals(Object o)
  {
    if (!(o instanceof Trigger)) {
      return false;
    }
    Trigger other = (Trigger)o;

    if ((other.getKey() == null) || (getKey() == null)) {
      return false;
    }

    return getKey().equals(other.getKey());
  }

  public int hashCode()
  {
    if (getKey() == null) {
      return super.hashCode();
    }
    return getKey().hashCode();
  }

  public Object clone()
  {
    AbstractTrigger copy;
    try {
      copy = (AbstractTrigger)super.clone();

      if (this.jobDataMap != null)
        copy.jobDataMap = ((JobDataMap)this.jobDataMap.clone());
    }
    catch (CloneNotSupportedException ex)
    {
      throw new IncompatibleClassChangeError("Not Cloneable.");
    }
    return copy;
  }

  public TriggerBuilder<T> getTriggerBuilder() {
    TriggerBuilder b = TriggerBuilder.newTrigger().forJob(getJobKey()).modifiedByCalendar(getCalendarName()).usingJobData(getJobDataMap()).withDescription(getDescription()).endAt(getEndTime()).withIdentity(getKey()).withPriority(getPriority()).startAt(getStartTime()).withSchedule(getScheduleBuilder());

    return b;
  }

  public abstract ScheduleBuilder<T> getScheduleBuilder();
}