package org.quartz.jobs.ee.ejb;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.Hashtable;
import javax.ejb.EJBHome;
import javax.ejb.EJBMetaData;
import javax.ejb.EJBObject;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import org.quartz.Job;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;

public class EJBInvokerJob
  implements Job
{
  public static final String EJB_JNDI_NAME_KEY = "ejb";
  public static final String EJB_METHOD_KEY = "method";
  public static final String EJB_ARG_TYPES_KEY = "argTypes";
  public static final String EJB_ARGS_KEY = "args";
  public static final String INITIAL_CONTEXT_FACTORY = "java.naming.factory.initial";
  public static final String PROVIDER_URL = "java.naming.provider.url";
  public static final String PRINCIPAL = "java.naming.security.principal";
  public static final String CREDENTIALS = "java.naming.security.credentials";

  public void execute(JobExecutionContext context)
    throws JobExecutionException
  {
    JobDataMap dataMap = context.getMergedJobDataMap();

    String ejb = dataMap.getString("ejb");
    String method = dataMap.getString("method");
    Object[] arguments = (Object[])dataMap.get("args");
    if (arguments == null) {
      arguments = new Object[0];
    }

    if (ejb == null)
    {
      throw new JobExecutionException();
    }

    InitialContext jndiContext = null;
    try
    {
      jndiContext = getInitialContext(dataMap);
    } catch (NamingException ne) {
      throw new JobExecutionException(ne);
    }
    try
    {
      Object value = null;
      try
      {
        value = jndiContext.lookup(ejb);
      } catch (NamingException ne) {
        throw new JobExecutionException(ne);
      }

      EJBHome ejbHome = (EJBHome)PortableRemoteObject.narrow(value, EJBHome.class);

      EJBMetaData metaData = null;
      try
      {
        metaData = ejbHome.getEJBMetaData();
      } catch (RemoteException re) {
        throw new JobExecutionException(re);
      }

      Class homeClass = metaData.getHomeInterfaceClass();

      Class remoteClass = metaData.getRemoteInterfaceClass();

      ejbHome = (EJBHome)PortableRemoteObject.narrow(ejbHome, homeClass);

      Method methodCreate = null;
      try
      {
        methodCreate = homeClass.getMethod("create", (Class[])null);
      } catch (NoSuchMethodException nsme) {
        throw new JobExecutionException(nsme);
      }

      EJBObject remoteObj = null;
      try
      {
        remoteObj = (EJBObject)methodCreate.invoke(ejbHome, (Object[])null);
      } catch (IllegalAccessException iae) {
        throw new JobExecutionException(iae);
      } catch (InvocationTargetException ite) {
        throw new JobExecutionException(ite);
      }

      Method methodExecute = null;
      try
      {
        Class[] argTypes = (Class[])dataMap.get("argTypes");
        if (argTypes == null) {
          argTypes = new Class[arguments.length];
          for (int i = 0; i < arguments.length; i++) {
            argTypes[i] = arguments[i].getClass();
          }

        }

        methodExecute = remoteClass.getMethod(method, argTypes);
      } catch (NoSuchMethodException nsme) {
        throw new JobExecutionException(nsme);
      }

      try
      {
        Object returnObj = methodExecute.invoke(remoteObj, arguments);

        context.setResult(returnObj);
      } catch (IllegalAccessException iae) {
        throw new JobExecutionException(iae);
      } catch (InvocationTargetException ite) {
        throw new JobExecutionException(ite);
      }

    }
    finally
    {
      if (jndiContext != null)
        try {
          jndiContext.close();
        }
        catch (NamingException e)
        {
        }
    }
  }

  private InitialContext getInitialContext(JobDataMap jobDataMap) throws NamingException
  {
    Hashtable params = new Hashtable(2);

    String initialContextFactory = jobDataMap.getString("java.naming.factory.initial");

    if (initialContextFactory != null) {
      params.put("java.naming.factory.initial", initialContextFactory);
    }

    String providerUrl = jobDataMap.getString("java.naming.provider.url");
    if (providerUrl != null) {
      params.put("java.naming.provider.url", providerUrl);
    }

    String principal = jobDataMap.getString("java.naming.security.principal");
    if (principal != null) {
      params.put("java.naming.security.principal", principal);
    }

    String credentials = jobDataMap.getString("java.naming.security.credentials");
    if (credentials != null) {
      params.put("java.naming.security.credentials", credentials);
    }

    return params.size() == 0 ? new InitialContext() : new InitialContext(params);
  }
}