package org.quartz.jobs.ee.mail;

import java.util.Date;
import java.util.Properties;
import javax.mail.Address;
import javax.mail.Message.RecipientType;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;
import org.quartz.Job;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class SendMailJob
  implements Job
{
  private final Logger log;
  public static final String PROP_SMTP_HOST = "smtp_host";
  public static final String PROP_RECIPIENT = "recipient";
  public static final String PROP_CC_RECIPIENT = "cc_recipient";
  public static final String PROP_SENDER = "sender";
  public static final String PROP_REPLY_TO = "reply_to";
  public static final String PROP_SUBJECT = "subject";
  public static final String PROP_MESSAGE = "message";
  public static final String PROP_CONTENT_TYPE = "content_type";

  public SendMailJob()
  {
    this.log = LoggerFactory.getLogger(getClass());
  }

  public void execute(JobExecutionContext context)
    throws JobExecutionException
  {
    JobDataMap data = context.getMergedJobDataMap();

    MailInfo mailInfo = populateMailInfo(data, createMailInfo());

    getLog().info("Sending message " + mailInfo);
    try
    {
      MimeMessage mimeMessage = prepareMimeMessage(mailInfo);

      Transport.send(mimeMessage);
    } catch (MessagingException e) {
      throw new JobExecutionException("Unable to send mail: " + mailInfo, e, false);
    }
  }

  protected Logger getLog()
  {
    return this.log;
  }

  protected MimeMessage prepareMimeMessage(MailInfo mailInfo) throws MessagingException
  {
    Session session = getMailSession(mailInfo);

    MimeMessage mimeMessage = new MimeMessage(session);

    Address[] toAddresses = InternetAddress.parse(mailInfo.getTo());
    mimeMessage.setRecipients(Message.RecipientType.TO, toAddresses);

    if (mailInfo.getCc() != null) {
      Address[] ccAddresses = InternetAddress.parse(mailInfo.getCc());
      mimeMessage.setRecipients(Message.RecipientType.CC, ccAddresses);
    }

    mimeMessage.setFrom(new InternetAddress(mailInfo.getFrom()));

    if (mailInfo.getReplyTo() != null) {
      mimeMessage.setReplyTo(new InternetAddress[] { new InternetAddress(mailInfo.getReplyTo()) });
    }

    mimeMessage.setSubject(mailInfo.getSubject());

    mimeMessage.setSentDate(new Date());

    setMimeMessageContent(mimeMessage, mailInfo);

    return mimeMessage;
  }

  protected void setMimeMessageContent(MimeMessage mimeMessage, MailInfo mailInfo) throws MessagingException
  {
    if (mailInfo.getContentType() == null)
      mimeMessage.setText(mailInfo.getMessage());
    else
      mimeMessage.setContent(mailInfo.getMessage(), mailInfo.getContentType());
  }

  protected Session getMailSession(MailInfo mailInfo) throws MessagingException
  {
    Properties properties = new Properties();
    properties.put("mail.smtp.host", mailInfo.getSmtpHost());

    return Session.getDefaultInstance(properties, null);
  }

  protected MailInfo createMailInfo() {
    return new MailInfo();
  }

  protected MailInfo populateMailInfo(JobDataMap data, MailInfo mailInfo)
  {
    mailInfo.setSmtpHost(getRequiredParm(data, "smtp_host", "PROP_SMTP_HOST"));
    mailInfo.setTo(getRequiredParm(data, "recipient", "PROP_RECIPIENT"));
    mailInfo.setFrom(getRequiredParm(data, "sender", "PROP_SENDER"));
    mailInfo.setSubject(getRequiredParm(data, "subject", "PROP_SUBJECT"));
    mailInfo.setMessage(getRequiredParm(data, "message", "PROP_MESSAGE"));

    mailInfo.setReplyTo(getOptionalParm(data, "reply_to"));
    mailInfo.setCc(getOptionalParm(data, "cc_recipient"));
    mailInfo.setContentType(getOptionalParm(data, "content_type"));

    return mailInfo;
  }

  protected String getRequiredParm(JobDataMap data, String property, String constantName)
  {
    String value = getOptionalParm(data, property);

    if (value == null) {
      throw new IllegalArgumentException(constantName + " not specified.");
    }

    return value;
  }

  protected String getOptionalParm(JobDataMap data, String property) {
    String value = data.getString(property);

    if ((value != null) && (value.trim().length() == 0)) {
      return null;
    }

    return value; } 
  protected static class MailInfo { private String smtpHost;
    private String to;
    private String from;
    private String subject;
    private String message;
    private String replyTo;
    private String cc;
    private String contentType;

    public String toString() { return "'" + getSubject() + "' to: " + getTo(); }

    public String getCc()
    {
      return this.cc;
    }

    public void setCc(String cc) {
      this.cc = cc;
    }

    public String getContentType() {
      return this.contentType;
    }

    public void setContentType(String contentType) {
      this.contentType = contentType;
    }

    public String getFrom() {
      return this.from;
    }

    public void setFrom(String from) {
      this.from = from;
    }

    public String getMessage() {
      return this.message;
    }

    public void setMessage(String message) {
      this.message = message;
    }

    public String getReplyTo() {
      return this.replyTo;
    }

    public void setReplyTo(String replyTo) {
      this.replyTo = replyTo;
    }

    public String getSmtpHost() {
      return this.smtpHost;
    }

    public void setSmtpHost(String smtpHost) {
      this.smtpHost = smtpHost;
    }

    public String getSubject() {
      return this.subject;
    }

    public void setSubject(String subject) {
      this.subject = subject;
    }

    public String getTo() {
      return this.to;
    }

    public void setTo(String to) {
      this.to = to;
    }
  }
}