package org.quartz.simpl;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import org.quartz.Job;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.spi.OperableTrigger;
import org.quartz.spi.TriggerFiredBundle;
import org.slf4j.Logger;

public class PropertySettingJobFactory extends SimpleJobFactory
{
  private boolean warnIfNotFound = false;
  private boolean throwIfNotFound = false;

  public Job newJob(TriggerFiredBundle bundle, Scheduler scheduler)
    throws SchedulerException
  {
    Job job = super.newJob(bundle, scheduler);

    JobDataMap jobDataMap = new JobDataMap();
    jobDataMap.putAll(scheduler.getContext());
    jobDataMap.putAll(bundle.getJobDetail().getJobDataMap());
    jobDataMap.putAll(bundle.getTrigger().getJobDataMap());

    setBeanProps(job, jobDataMap);

    return job;
  }

  protected void setBeanProps(Object obj, JobDataMap data) throws SchedulerException
  {
    BeanInfo bi = null;
    try {
      bi = Introspector.getBeanInfo(obj.getClass());
    } catch (IntrospectionException e) {
      handleError("Unable to introspect Job class.", e);
    }

    PropertyDescriptor[] propDescs = bi.getPropertyDescriptors();

    for (Iterator entryIter = data.getWrappedMap().entrySet().iterator(); entryIter.hasNext(); ) {
      Map.Entry entry = (Map.Entry)entryIter.next();

      String name = (String)entry.getKey();
      String c = name.substring(0, 1).toUpperCase(Locale.US);
      String methName = "set" + c + name.substring(1);

      Method setMeth = getSetMethod(methName, propDescs);

      Class paramType = null;
      Object o = null;
      try
      {
        if (setMeth == null) {
          handleError("No setter on Job class " + obj.getClass().getName() + " for property '" + name + "'");
        }
        else
        {
          paramType = setMeth.getParameterTypes()[0];
          o = entry.getValue();

          Object parm = null;
          if (paramType.isPrimitive()) {
            if (o == null) {
              handleError("Cannot set primitive property '" + name + "' on Job class " + obj.getClass().getName() + " to null.");
            }
            else if (paramType.equals(Integer.TYPE)) {
              if ((o instanceof String))
                parm = Integer.valueOf((String)o);
              else if ((o instanceof Integer))
                parm = o;
            }
            else if (paramType.equals(Long.TYPE)) {
              if ((o instanceof String))
                parm = Long.valueOf((String)o);
              else if ((o instanceof Long))
                parm = o;
            }
            else if (paramType.equals(Float.TYPE)) {
              if ((o instanceof String))
                parm = Float.valueOf((String)o);
              else if ((o instanceof Float))
                parm = o;
            }
            else if (paramType.equals(Double.TYPE)) {
              if ((o instanceof String))
                parm = Double.valueOf((String)o);
              else if ((o instanceof Double))
                parm = o;
            }
            else if (paramType.equals(Boolean.TYPE)) {
              if ((o instanceof String))
                parm = Boolean.valueOf((String)o);
              else if ((o instanceof Boolean))
                parm = o;
            }
            else if (paramType.equals(Byte.TYPE)) {
              if ((o instanceof String))
                parm = Byte.valueOf((String)o);
              else if ((o instanceof Byte))
                parm = o;
            }
            else if (paramType.equals(Short.TYPE)) {
              if ((o instanceof String))
                parm = Short.valueOf((String)o);
              else if ((o instanceof Short))
                parm = o;
            }
            else if (paramType.equals(Character.TYPE))
              if ((o instanceof String)) {
                String str = (String)o;
                if (str.length() == 1)
                  parm = Character.valueOf(str.charAt(0));
              }
              else if ((o instanceof Character)) {
                parm = o;
              }
          } else {
            if ((o != null) && (paramType.isAssignableFrom(o.getClass()))) {
              parm = o;
            }

            if ((o != null) && (parm == null)) {
              handleError("The setter on Job class " + obj.getClass().getName() + " for property '" + name + "' expects a " + paramType + " but was given " + o.getClass().getName());
            }
            else
            {
              setMeth.invoke(obj, new Object[] { parm }); } 
          }
        } } catch (NumberFormatException nfe) { handleError("The setter on Job class " + obj.getClass().getName() + " for property '" + name + "' expects a " + paramType + " but was given " + o.getClass().getName(), nfe); }
      catch (IllegalArgumentException e)
      {
        handleError("The setter on Job class " + obj.getClass().getName() + " for property '" + name + "' expects a " + paramType + " but was given " + o.getClass().getName(), e);
      }
      catch (IllegalAccessException e)
      {
        handleError("The setter on Job class " + obj.getClass().getName() + " for property '" + name + "' could not be accessed.", e);
      }
      catch (InvocationTargetException e)
      {
        handleError("The setter on Job class " + obj.getClass().getName() + " for property '" + name + "' could not be invoked.", e);
      }
    }
  }

  private void handleError(String message)
    throws SchedulerException
  {
    handleError(message, null);
  }

  private void handleError(String message, Exception e) throws SchedulerException {
    if (isThrowIfPropertyNotFound()) {
      throw new SchedulerException(message, e);
    }

    if (isWarnIfPropertyNotFound())
      if (e == null)
        getLog().warn(message);
      else
        getLog().warn(message, e);
  }

  private Method getSetMethod(String name, PropertyDescriptor[] props)
  {
    for (int i = 0; i < props.length; i++) {
      Method wMeth = props[i].getWriteMethod();

      if (wMeth != null)
      {
        if (wMeth.getParameterTypes().length == 1)
        {
          if (wMeth.getName().equals(name))
            return wMeth;
        }
      }
    }
    return null;
  }

  public boolean isThrowIfPropertyNotFound()
  {
    return this.throwIfNotFound;
  }

  public void setThrowIfPropertyNotFound(boolean throwIfNotFound)
  {
    this.throwIfNotFound = throwIfNotFound;
  }

  public boolean isWarnIfPropertyNotFound()
  {
    return this.warnIfNotFound;
  }

  public void setWarnIfPropertyNotFound(boolean warnIfNotFound)
  {
    this.warnIfNotFound = warnIfNotFound;
  }
}