package org.quartz.simpl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.atomic.AtomicLong;
import org.quartz.Calendar;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobKey;
import org.quartz.JobPersistenceException;
import org.quartz.ObjectAlreadyExistsException;
import org.quartz.Trigger;
import org.quartz.Trigger.CompletedExecutionInstruction;
import org.quartz.Trigger.TriggerState;
import org.quartz.TriggerKey;
import org.quartz.impl.JobDetailImpl;
import org.quartz.impl.matchers.GroupMatcher;
import org.quartz.impl.matchers.StringMatcher.StringOperatorName;
import org.quartz.spi.ClassLoadHelper;
import org.quartz.spi.JobStore;
import org.quartz.spi.OperableTrigger;
import org.quartz.spi.SchedulerSignaler;
import org.quartz.spi.TriggerFiredBundle;
import org.quartz.spi.TriggerFiredResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class RAMJobStore
  implements JobStore
{
  protected HashMap<JobKey, JobWrapper> jobsByKey = new HashMap(1000);

  protected HashMap<TriggerKey, TriggerWrapper> triggersByKey = new HashMap(1000);

  protected HashMap<String, HashMap<JobKey, JobWrapper>> jobsByGroup = new HashMap(25);

  protected HashMap<String, HashMap<TriggerKey, TriggerWrapper>> triggersByGroup = new HashMap(25);

  protected TreeSet<TriggerWrapper> timeTriggers = new TreeSet(new TriggerWrapperComparator());

  protected HashMap<String, Calendar> calendarsByName = new HashMap(25);

  protected ArrayList<TriggerWrapper> triggers = new ArrayList(1000);

  protected final Object lock = new Object();

  protected HashSet<String> pausedTriggerGroups = new HashSet();

  protected HashSet<String> pausedJobGroups = new HashSet();

  protected HashSet<JobKey> blockedJobs = new HashSet();

  protected long misfireThreshold = 5000L;
  protected SchedulerSignaler signaler;
  private final Logger log = LoggerFactory.getLogger(getClass());

  private static final AtomicLong ftrCtr = new AtomicLong(System.currentTimeMillis());

  protected Logger getLog()
  {
    return this.log;
  }

  public void initialize(ClassLoadHelper loadHelper, SchedulerSignaler schedSignaler)
  {
    this.signaler = schedSignaler;

    getLog().info("RAMJobStore initialized.");
  }

  public void schedulerStarted()
  {
  }

  public void schedulerPaused()
  {
  }

  public void schedulerResumed()
  {
  }

  public long getMisfireThreshold() {
    return this.misfireThreshold;
  }

  public void setMisfireThreshold(long misfireThreshold)
  {
    if (misfireThreshold < 1L) {
      throw new IllegalArgumentException("Misfirethreashold must be larger than 0");
    }
    this.misfireThreshold = misfireThreshold;
  }

  public void shutdown()
  {
  }

  public boolean supportsPersistence()
  {
    return false;
  }

  public void clearAllSchedulingData()
    throws JobPersistenceException
  {
    synchronized (this.lock)
    {
      List lst = getTriggerGroupNames();
      for (String group : lst) {
        Set keys = getTriggerKeys(GroupMatcher.triggerGroupEquals(group));
        for (TriggerKey key : keys) {
          removeTrigger(key);
        }
      }

      lst = getJobGroupNames();
      for (String group : lst) {
        Set keys = getJobKeys(GroupMatcher.jobGroupEquals(group));
        for (JobKey key : keys) {
          removeJob(key);
        }
      }

      lst = getCalendarNames();
      for (String name : lst)
        removeCalendar(name);
    }
  }

  public void storeJobAndTrigger(JobDetail newJob, OperableTrigger newTrigger)
    throws JobPersistenceException
  {
    storeJob(newJob, false);
    storeTrigger(newTrigger, false);
  }

  public void storeJob(JobDetail newJob, boolean replaceExisting)
    throws ObjectAlreadyExistsException
  {
    JobWrapper jw = new JobWrapper((JobDetail)newJob.clone());

    boolean repl = false;

    synchronized (this.lock) {
      if (this.jobsByKey.get(jw.key) != null) {
        if (!replaceExisting) {
          throw new ObjectAlreadyExistsException(newJob);
        }
        repl = true;
      }

      if (!repl)
      {
        HashMap grpMap = (HashMap)this.jobsByGroup.get(newJob.getKey().getGroup());
        if (grpMap == null) {
          grpMap = new HashMap(100);
          this.jobsByGroup.put(newJob.getKey().getGroup(), grpMap);
        }

        grpMap.put(newJob.getKey(), jw);

        this.jobsByKey.put(jw.key, jw);
      }
      else {
        JobWrapper orig = (JobWrapper)this.jobsByKey.get(jw.key);
        orig.jobDetail = jw.jobDetail;
      }
    }
  }

  public boolean removeJob(JobKey jobKey)
  {
    boolean found = false;

    synchronized (this.lock) {
      List triggersOfJob = getTriggersForJob(jobKey);
      for (OperableTrigger trig : triggersOfJob) {
        removeTrigger(trig.getKey());
        found = true;
      }

      found = this.jobsByKey.remove(jobKey) != null | found;
      if (found)
      {
        HashMap grpMap = (HashMap)this.jobsByGroup.get(jobKey.getGroup());
        if (grpMap != null) {
          grpMap.remove(jobKey);
          if (grpMap.size() == 0) {
            this.jobsByGroup.remove(jobKey.getGroup());
          }
        }
      }
    }

    return found;
  }

  public boolean removeJobs(List<JobKey> jobKeys) throws JobPersistenceException
  {
    boolean allFound = true;

    synchronized (this.lock) {
      for (JobKey key : jobKeys) {
        allFound = (removeJob(key)) && (allFound);
      }
    }
    return allFound;
  }

  public boolean removeTriggers(List<TriggerKey> triggerKeys) throws JobPersistenceException
  {
    boolean allFound = true;

    synchronized (this.lock) {
      for (TriggerKey key : triggerKeys) {
        allFound = (removeTrigger(key)) && (allFound);
      }
    }
    return allFound;
  }

  public void storeJobsAndTriggers(Map<JobDetail, List<Trigger>> triggersAndJobs, boolean replace)
    throws ObjectAlreadyExistsException, JobPersistenceException
  {
    synchronized (this.lock)
    {
      if (!replace) {
        for (Map.Entry e : triggersAndJobs.entrySet()) {
          if (checkExists(((JobDetail)e.getKey()).getKey()))
            throw new ObjectAlreadyExistsException((JobDetail)e.getKey());
          for (Trigger trigger : (List)e.getValue()) {
            if (checkExists(trigger.getKey())) {
              throw new ObjectAlreadyExistsException(trigger);
            }
          }
        }
      }
      for (Map.Entry e : triggersAndJobs.entrySet()) {
        storeJob((JobDetail)e.getKey(), true);
        for (Trigger trigger : (List)e.getValue())
          storeTrigger((OperableTrigger)trigger, true);
      }
    }
  }

  public void storeTrigger(OperableTrigger newTrigger, boolean replaceExisting)
    throws JobPersistenceException
  {
    TriggerWrapper tw = new TriggerWrapper((OperableTrigger)newTrigger.clone());

    synchronized (this.lock) {
      if (this.triggersByKey.get(tw.key) != null) {
        if (!replaceExisting) {
          throw new ObjectAlreadyExistsException(newTrigger);
        }

        removeTrigger(newTrigger.getKey(), false);
      }

      if (retrieveJob(newTrigger.getJobKey()) == null) {
        throw new JobPersistenceException("The job (" + newTrigger.getJobKey() + ") referenced by the trigger does not exist.");
      }

      this.triggers.add(tw);

      HashMap grpMap = (HashMap)this.triggersByGroup.get(newTrigger.getKey().getGroup());
      if (grpMap == null) {
        grpMap = new HashMap(100);
        this.triggersByGroup.put(newTrigger.getKey().getGroup(), grpMap);
      }
      grpMap.put(newTrigger.getKey(), tw);

      this.triggersByKey.put(tw.key, tw);

      if ((this.pausedTriggerGroups.contains(newTrigger.getKey().getGroup())) || (this.pausedJobGroups.contains(newTrigger.getJobKey().getGroup())))
      {
        tw.state = 4;
        if (this.blockedJobs.contains(tw.jobKey))
          tw.state = 6;
      }
      else if (this.blockedJobs.contains(tw.jobKey)) {
        tw.state = 5;
      } else {
        this.timeTriggers.add(tw);
      }
    }
  }

  public boolean removeTrigger(TriggerKey triggerKey)
  {
    return removeTrigger(triggerKey, true);
  }

  private boolean removeTrigger(TriggerKey key, boolean removeOrphanedJob)
  {
    boolean found = false;

    synchronized (this.lock)
    {
      found = this.triggersByKey.remove(key) != null;
      if (found) {
        TriggerWrapper tw = null;

        HashMap grpMap = (HashMap)this.triggersByGroup.get(key.getGroup());
        if (grpMap != null) {
          grpMap.remove(key);
          if (grpMap.size() == 0) {
            this.triggersByGroup.remove(key.getGroup());
          }
        }

        Iterator tgs = this.triggers.iterator();
        while (tgs.hasNext()) {
          tw = (TriggerWrapper)tgs.next();
          if (key.equals(tw.key)) {
            tgs.remove();
          }
        }

        this.timeTriggers.remove(tw);

        if (removeOrphanedJob) {
          JobWrapper jw = (JobWrapper)this.jobsByKey.get(tw.jobKey);
          List trigs = getTriggersForJob(tw.jobKey);
          if (((trigs == null) || (trigs.size() == 0)) && (!jw.jobDetail.isDurable()) && 
            (removeJob(jw.key))) {
            this.signaler.notifySchedulerListenersJobDeleted(jw.key);
          }
        }
      }

    }

    return found;
  }

  public boolean replaceTrigger(TriggerKey triggerKey, OperableTrigger newTrigger)
    throws JobPersistenceException
  {
    boolean found = false;

    synchronized (this.lock)
    {
      TriggerWrapper tw = (TriggerWrapper)this.triggersByKey.remove(triggerKey);
      found = tw != null;

      if (found)
      {
        if (!tw.getTrigger().getJobKey().equals(newTrigger.getJobKey())) {
          throw new JobPersistenceException("New trigger is not related to the same job as the old trigger.");
        }

        tw = null;

        HashMap grpMap = (HashMap)this.triggersByGroup.get(triggerKey.getGroup());
        if (grpMap != null) {
          grpMap.remove(triggerKey);
          if (grpMap.size() == 0) {
            this.triggersByGroup.remove(triggerKey.getGroup());
          }
        }

        Iterator tgs = this.triggers.iterator();
        while (tgs.hasNext()) {
          tw = (TriggerWrapper)tgs.next();
          if (triggerKey.equals(tw.key)) {
            tgs.remove();
          }
        }

        this.timeTriggers.remove(tw);
        try
        {
          storeTrigger(newTrigger, false);
        } catch (JobPersistenceException jpe) {
          storeTrigger(tw.getTrigger(), false);
          throw jpe;
        }
      }
    }

    return found;
  }

  public JobDetail retrieveJob(JobKey jobKey)
  {
    synchronized (this.lock) {
      JobWrapper jw = (JobWrapper)this.jobsByKey.get(jobKey);
      return jw != null ? (JobDetail)jw.jobDetail.clone() : null;
    }
  }

  public OperableTrigger retrieveTrigger(TriggerKey triggerKey)
  {
    synchronized (this.lock) {
      TriggerWrapper tw = (TriggerWrapper)this.triggersByKey.get(triggerKey);

      return tw != null ? (OperableTrigger)tw.getTrigger().clone() : null;
    }
  }

  public boolean checkExists(JobKey jobKey)
  {
    synchronized (this.lock) {
      JobWrapper jw = (JobWrapper)this.jobsByKey.get(jobKey);
      return jw != null;
    }
  }

  public boolean checkExists(TriggerKey triggerKey)
  {
    synchronized (this.lock) {
      TriggerWrapper tw = (TriggerWrapper)this.triggersByKey.get(triggerKey);

      return tw != null;
    }
  }

  public Trigger.TriggerState getTriggerState(TriggerKey triggerKey)
    throws JobPersistenceException
  {
    synchronized (this.lock) {
      TriggerWrapper tw = (TriggerWrapper)this.triggersByKey.get(triggerKey);

      if (tw == null) {
        return Trigger.TriggerState.NONE;
      }

      if (tw.state == 3) {
        return Trigger.TriggerState.COMPLETE;
      }

      if (tw.state == 4) {
        return Trigger.TriggerState.PAUSED;
      }

      if (tw.state == 6) {
        return Trigger.TriggerState.PAUSED;
      }

      if (tw.state == 5) {
        return Trigger.TriggerState.BLOCKED;
      }

      if (tw.state == 7) {
        return Trigger.TriggerState.ERROR;
      }

      return Trigger.TriggerState.NORMAL;
    }
  }

  public void storeCalendar(String name, Calendar calendar, boolean replaceExisting, boolean updateTriggers)
    throws ObjectAlreadyExistsException
  {
    calendar = (Calendar)calendar.clone();

    synchronized (this.lock)
    {
      Object obj = this.calendarsByName.get(name);

      if ((obj != null) && (!replaceExisting)) {
        throw new ObjectAlreadyExistsException("Calendar with name '" + name + "' already exists.");
      }
      if (obj != null) {
        this.calendarsByName.remove(name);
      }

      this.calendarsByName.put(name, calendar);

      if ((obj != null) && (updateTriggers)) {
        Iterator trigs = getTriggerWrappersForCalendar(name).iterator();
        while (trigs.hasNext()) {
          TriggerWrapper tw = (TriggerWrapper)trigs.next();
          OperableTrigger trig = tw.getTrigger();
          boolean removed = this.timeTriggers.remove(tw);

          trig.updateWithNewCalendar(calendar, getMisfireThreshold());

          if (removed)
            this.timeTriggers.add(tw);
        }
      }
    }
  }

  public boolean removeCalendar(String calName)
    throws JobPersistenceException
  {
    int numRefs = 0;

    synchronized (this.lock) {
      Iterator itr = this.triggers.iterator();
      while (itr.hasNext()) {
        OperableTrigger trigg = ((TriggerWrapper)itr.next()).trigger;
        if ((trigg.getCalendarName() != null) && (trigg.getCalendarName().equals(calName)))
        {
          numRefs++;
        }
      }
    }

    if (numRefs > 0) {
      throw new JobPersistenceException("Calender cannot be removed if it referenced by a Trigger!");
    }

    return this.calendarsByName.remove(calName) != null;
  }

  public Calendar retrieveCalendar(String calName)
  {
    synchronized (this.lock) {
      Calendar cal = (Calendar)this.calendarsByName.get(calName);
      if (cal != null)
        return (Calendar)cal.clone();
      return null;
    }
  }

  public int getNumberOfJobs()
  {
    synchronized (this.lock) {
      return this.jobsByKey.size();
    }
  }

  public int getNumberOfTriggers()
  {
    synchronized (this.lock) {
      return this.triggers.size();
    }
  }

  public int getNumberOfCalendars()
  {
    synchronized (this.lock) {
      return this.calendarsByName.size();
    }
  }

  public Set<JobKey> getJobKeys(GroupMatcher<JobKey> matcher)
  {
    Set outList = null;
    StringMatcher.StringOperatorName operator;
    String compareToValue;
    synchronized (this.lock)
    {
      operator = matcher.getCompareWithOperator();
      compareToValue = matcher.getCompareToValue();

      switch (1.$SwitchMap$org$quartz$impl$matchers$StringMatcher$StringOperatorName[operator.ordinal()]) {
      case 1:
        HashMap grpMap = (HashMap)this.jobsByGroup.get(compareToValue);
        if (grpMap != null) {
          outList = new HashSet();

          for (JobWrapper jw : grpMap.values())
          {
            if (jw != null)
              outList.add(jw.jobDetail.getKey());  }  } break;
      default:
        for (Map.Entry entry : this.jobsByGroup.entrySet()) {
          if ((operator.evaluate((String)entry.getKey(), compareToValue)) && (entry.getValue() != null)) {
            if (outList == null) {
              outList = new HashSet();
            }
            for (JobWrapper jobWrapper : ((HashMap)entry.getValue()).values()) {
              if (jobWrapper != null) {
                outList.add(jobWrapper.jobDetail.getKey());
              }
            }
          }
        }
      }
    }

    return outList == null ? Collections.emptySet() : outList;
  }

  public List<String> getCalendarNames()
  {
    synchronized (this.lock) {
      return new LinkedList(this.calendarsByName.keySet());
    }
  }

  public Set<TriggerKey> getTriggerKeys(GroupMatcher<TriggerKey> matcher)
  {
    Set outList = null;
    StringMatcher.StringOperatorName operator;
    String compareToValue;
    synchronized (this.lock)
    {
      operator = matcher.getCompareWithOperator();
      compareToValue = matcher.getCompareToValue();

      switch (1.$SwitchMap$org$quartz$impl$matchers$StringMatcher$StringOperatorName[operator.ordinal()]) {
      case 1:
        HashMap grpMap = (HashMap)this.triggersByGroup.get(compareToValue);
        if (grpMap != null) {
          outList = new HashSet();

          for (TriggerWrapper tw : grpMap.values())
          {
            if (tw != null)
              outList.add(tw.trigger.getKey());  }  } break;
      default:
        for (Map.Entry entry : this.triggersByGroup.entrySet()) {
          if ((operator.evaluate((String)entry.getKey(), compareToValue)) && (entry.getValue() != null)) {
            if (outList == null) {
              outList = new HashSet();
            }
            for (TriggerWrapper triggerWrapper : ((HashMap)entry.getValue()).values()) {
              if (triggerWrapper != null) {
                outList.add(triggerWrapper.trigger.getKey());
              }
            }
          }
        }
      }
    }

    return outList == null ? Collections.emptySet() : outList;
  }

  public List<String> getJobGroupNames()
  {
    List outList = null;

    synchronized (this.lock) {
      outList = new LinkedList(this.jobsByGroup.keySet());
    }

    return outList;
  }

  public List<String> getTriggerGroupNames()
  {
    LinkedList outList = null;

    synchronized (this.lock) {
      outList = new LinkedList(this.triggersByGroup.keySet());
    }

    return outList;
  }

  public List<OperableTrigger> getTriggersForJob(JobKey jobKey)
  {
    ArrayList trigList = new ArrayList();

    synchronized (this.lock) {
      for (int i = 0; i < this.triggers.size(); i++) {
        TriggerWrapper tw = (TriggerWrapper)this.triggers.get(i);
        if (tw.jobKey.equals(jobKey)) {
          trigList.add((OperableTrigger)tw.trigger.clone());
        }
      }
    }

    return trigList;
  }

  protected ArrayList<TriggerWrapper> getTriggerWrappersForJob(JobKey jobKey) {
    ArrayList trigList = new ArrayList();

    synchronized (this.lock) {
      for (int i = 0; i < this.triggers.size(); i++) {
        TriggerWrapper tw = (TriggerWrapper)this.triggers.get(i);
        if (tw.jobKey.equals(jobKey)) {
          trigList.add(tw);
        }
      }
    }

    return trigList;
  }

  protected ArrayList<TriggerWrapper> getTriggerWrappersForCalendar(String calName) {
    ArrayList trigList = new ArrayList();

    synchronized (this.lock) {
      for (int i = 0; i < this.triggers.size(); i++) {
        TriggerWrapper tw = (TriggerWrapper)this.triggers.get(i);
        String tcalName = tw.getTrigger().getCalendarName();
        if ((tcalName != null) && (tcalName.equals(calName))) {
          trigList.add(tw);
        }
      }
    }

    return trigList;
  }

  public void pauseTrigger(TriggerKey triggerKey)
  {
    synchronized (this.lock) {
      TriggerWrapper tw = (TriggerWrapper)this.triggersByKey.get(triggerKey);

      if ((tw == null) || (tw.trigger == null)) {
        return;
      }

      if (tw.state == 3) {
        return;
      }

      if (tw.state == 5)
        tw.state = 6;
      else {
        tw.state = 4;
      }

      this.timeTriggers.remove(tw);
    }
  }

  public List<String> pauseTriggers(GroupMatcher<TriggerKey> matcher)
  {
    List pausedGroups;
    synchronized (this.lock) {
      pausedGroups = new LinkedList();

      StringMatcher.StringOperatorName operator = matcher.getCompareWithOperator();
      switch (1.$SwitchMap$org$quartz$impl$matchers$StringMatcher$StringOperatorName[operator.ordinal()]) {
      case 1:
        if (this.pausedTriggerGroups.add(matcher.getCompareToValue()))
          pausedGroups.add(matcher.getCompareToValue()); break;
      default:
        for (String group : this.triggersByGroup.keySet()) {
          if ((operator.evaluate(group, matcher.getCompareToValue())) && 
            (this.pausedTriggerGroups.add(matcher.getCompareToValue()))) {
            pausedGroups.add(group);
          }
        }

      }

      for (String pausedGroup : pausedGroups) {
        Set keys = getTriggerKeys(GroupMatcher.triggerGroupEquals(pausedGroup));

        for (TriggerKey key : keys) {
          pauseTrigger(key);
        }
      }
    }

    return pausedGroups;
  }

  public void pauseJob(JobKey jobKey)
  {
    synchronized (this.lock) {
      List triggersOfJob = getTriggersForJob(jobKey);
      for (OperableTrigger trigger : triggersOfJob)
        pauseTrigger(trigger.getKey());
    }
  }

  public List<String> pauseJobs(GroupMatcher<JobKey> matcher)
  {
    List pausedGroups = new LinkedList();
    synchronized (this.lock)
    {
      StringMatcher.StringOperatorName operator = matcher.getCompareWithOperator();
      switch (1.$SwitchMap$org$quartz$impl$matchers$StringMatcher$StringOperatorName[operator.ordinal()]) {
      case 1:
        if (this.pausedJobGroups.add(matcher.getCompareToValue()))
          pausedGroups.add(matcher.getCompareToValue()); break;
      default:
        for (String group : this.jobsByGroup.keySet()) {
          if ((operator.evaluate(group, matcher.getCompareToValue())) && 
            (this.pausedJobGroups.add(group))) {
            pausedGroups.add(group);
          }
        }

      }

      for (String groupName : pausedGroups) {
        for (JobKey jobKey : getJobKeys(GroupMatcher.jobGroupEquals(groupName))) {
          List triggersOfJob = getTriggersForJob(jobKey);
          for (OperableTrigger trigger : triggersOfJob) {
            pauseTrigger(trigger.getKey());
          }
        }
      }
    }

    return pausedGroups;
  }

  public void resumeTrigger(TriggerKey triggerKey)
  {
    synchronized (this.lock) {
      TriggerWrapper tw = (TriggerWrapper)this.triggersByKey.get(triggerKey);

      if ((tw == null) || (tw.trigger == null)) {
        return;
      }

      OperableTrigger trig = tw.getTrigger();

      if ((tw.state != 4) && (tw.state != 6))
      {
        return;
      }

      if (this.blockedJobs.contains(trig.getJobKey()))
        tw.state = 5;
      else {
        tw.state = 0;
      }

      applyMisfire(tw);

      if (tw.state == 0)
        this.timeTriggers.add(tw);
    }
  }

  public List<String> resumeTriggers(GroupMatcher<TriggerKey> matcher)
  {
    Set groups = new HashSet();

    synchronized (this.lock) {
      Set keys = getTriggerKeys(matcher);

      for (TriggerKey triggerKey : keys) {
        groups.add(triggerKey.getGroup());
        if (this.triggersByKey.get(triggerKey) != null) {
          String jobGroup = ((TriggerWrapper)this.triggersByKey.get(triggerKey)).jobKey.getGroup();
          if (this.pausedJobGroups.contains(jobGroup));
        }
        else {
          resumeTrigger(triggerKey);
        }
      }
      for (String group : groups) {
        this.pausedTriggerGroups.remove(group);
      }
    }

    return new ArrayList(groups);
  }

  public void resumeJob(JobKey jobKey)
  {
    synchronized (this.lock) {
      List triggersOfJob = getTriggersForJob(jobKey);
      for (OperableTrigger trigger : triggersOfJob)
        resumeTrigger(trigger.getKey());
    }
  }

  public Collection<String> resumeJobs(GroupMatcher<JobKey> matcher)
  {
    Set resumedGroups = new HashSet();
    synchronized (this.lock) {
      Set keys = getJobKeys(matcher);

      for (String pausedJobGroup : this.pausedJobGroups) {
        if (matcher.getCompareWithOperator().evaluate(pausedJobGroup, matcher.getCompareToValue())) {
          resumedGroups.add(pausedJobGroup);
        }
      }

      for (String resumedGroup : resumedGroups) {
        this.pausedJobGroups.remove(resumedGroup);
      }

      for (JobKey key : keys) {
        List triggersOfJob = getTriggersForJob(key);
        for (OperableTrigger trigger : triggersOfJob) {
          resumeTrigger(trigger.getKey());
        }
      }
    }
    return resumedGroups;
  }

  public void pauseAll()
  {
    synchronized (this.lock) {
      List names = getTriggerGroupNames();

      for (String name : names)
        pauseTriggers(GroupMatcher.triggerGroupEquals(name));
    }
  }

  public void resumeAll()
  {
    synchronized (this.lock)
    {
      this.pausedJobGroups.clear();
      List names = getTriggerGroupNames();

      for (String name : names)
        resumeTriggers(GroupMatcher.triggerGroupEquals(name));
    }
  }

  protected boolean applyMisfire(TriggerWrapper tw)
  {
    long misfireTime = System.currentTimeMillis();
    if (getMisfireThreshold() > 0L) {
      misfireTime -= getMisfireThreshold();
    }

    Date tnft = tw.trigger.getNextFireTime();
    if ((tnft == null) || (tnft.getTime() > misfireTime) || (tw.trigger.getMisfireInstruction() == -1))
    {
      return false;
    }

    Calendar cal = null;
    if (tw.trigger.getCalendarName() != null) {
      cal = retrieveCalendar(tw.trigger.getCalendarName());
    }

    this.signaler.notifyTriggerListenersMisfired((OperableTrigger)tw.trigger.clone());

    tw.trigger.updateAfterMisfire(cal);

    if (tw.trigger.getNextFireTime() == null) {
      tw.state = 3;
      this.signaler.notifySchedulerListenersFinalized(tw.trigger);
      synchronized (this.lock) {
        this.timeTriggers.remove(tw);
      }
    } else if (tnft.equals(tw.trigger.getNextFireTime())) {
      return false;
    }

    return true;
  }

  protected String getFiredTriggerRecordId()
  {
    return String.valueOf(ftrCtr.incrementAndGet());
  }

  public List<OperableTrigger> acquireNextTriggers(long noLaterThan, int maxCount, long timeWindow)
  {
    synchronized (this.lock) {
      List result = new ArrayList();
      Set acquiredJobKeysForNoConcurrentExec = new HashSet();
      Set excludedTriggers = new HashSet();
      long firstAcquiredTriggerFireTime = 0L;

      if (this.timeTriggers.size() == 0)
        return result;
      while (true)
      {
        TriggerWrapper tw;
        try
        {
          tw = (TriggerWrapper)this.timeTriggers.first();
          if (tw == null)
            break;
          this.timeTriggers.remove(tw);
        } catch (NoSuchElementException nsee) {
          break;
        }

        if (tw.trigger.getNextFireTime() != null)
        {
          if ((firstAcquiredTriggerFireTime > 0L) && (tw.trigger.getNextFireTime().getTime() > firstAcquiredTriggerFireTime + timeWindow))
          {
            this.timeTriggers.add(tw);
          }
          else
          {
            if (applyMisfire(tw)) {
              if (tw.trigger.getNextFireTime() == null) continue;
              this.timeTriggers.add(tw); continue;
            }

            if (tw.getTrigger().getNextFireTime().getTime() > noLaterThan + timeWindow) {
              this.timeTriggers.add(tw);
            }
            else
            {
              JobKey jobKey = tw.trigger.getJobKey();
              JobDetail job = ((JobWrapper)this.jobsByKey.get(tw.trigger.getJobKey())).jobDetail;
              if (job.isConcurrentExectionDisallowed()) {
                if (acquiredJobKeysForNoConcurrentExec.contains(jobKey)) {
                  excludedTriggers.add(tw);
                  continue;
                }
                acquiredJobKeysForNoConcurrentExec.add(jobKey);
              }

              tw.state = 1;
              tw.trigger.setFireInstanceId(getFiredTriggerRecordId());
              OperableTrigger trig = (OperableTrigger)tw.trigger.clone();
              result.add(trig);
              if (firstAcquiredTriggerFireTime == 0L) {
                firstAcquiredTriggerFireTime = tw.trigger.getNextFireTime().getTime();
              }
              if (result.size() == maxCount) break;
            }
          }
        }
      }
      if (excludedTriggers.size() > 0)
        this.timeTriggers.addAll(excludedTriggers);
      return result;
    }
  }

  public void releaseAcquiredTrigger(OperableTrigger trigger)
  {
    synchronized (this.lock) {
      TriggerWrapper tw = (TriggerWrapper)this.triggersByKey.get(trigger.getKey());
      if ((tw != null) && (tw.state == 1)) {
        tw.state = 0;
        this.timeTriggers.add(tw);
      }
    }
  }

  public List<TriggerFiredResult> triggersFired(List<OperableTrigger> firedTriggers)
  {
    synchronized (this.lock) {
      List results = new ArrayList();

      for (OperableTrigger trigger : firedTriggers) {
        TriggerWrapper tw = (TriggerWrapper)this.triggersByKey.get(trigger.getKey());

        if ((tw != null) && (tw.trigger != null) && 
          (tw.state == 1))
        {
          Calendar cal = null;
          if (tw.trigger.getCalendarName() != null) { cal = retrieveCalendar(tw.trigger.getCalendarName());
            if (cal == null);
          }
          else {
            Date prevFireTime = trigger.getPreviousFireTime();

            this.timeTriggers.remove(tw);

            tw.trigger.triggered(cal);
            trigger.triggered(cal);

            tw.state = 0;

            TriggerFiredBundle bndle = new TriggerFiredBundle(retrieveJob(tw.jobKey), trigger, cal, false, new Date(), trigger.getPreviousFireTime(), prevFireTime, trigger.getNextFireTime());

            JobDetail job = bndle.getJobDetail();

            if (job.isConcurrentExectionDisallowed()) {
              ArrayList trigs = getTriggerWrappersForJob(job.getKey());
              Iterator itr = trigs.iterator();
              while (itr.hasNext()) {
                TriggerWrapper ttw = (TriggerWrapper)itr.next();
                if (ttw.state == 0) {
                  ttw.state = 5;
                }
                if (ttw.state == 4) {
                  ttw.state = 6;
                }
                this.timeTriggers.remove(ttw);
              }
              this.blockedJobs.add(job.getKey());
            } else if (tw.trigger.getNextFireTime() != null) {
              synchronized (this.lock) {
                this.timeTriggers.add(tw);
              }
            }

            results.add(new TriggerFiredResult(bndle));
          }
        }
      }
      return results;
    }
  }

  public void triggeredJobComplete(OperableTrigger trigger, JobDetail jobDetail, Trigger.CompletedExecutionInstruction triggerInstCode)
  {
    synchronized (this.lock)
    {
      JobWrapper jw = (JobWrapper)this.jobsByKey.get(jobDetail.getKey());
      TriggerWrapper tw = (TriggerWrapper)this.triggersByKey.get(trigger.getKey());

      if (jw != null) {
        JobDetail jd = jw.jobDetail;

        if (jd.isPersistJobDataAfterExecution()) {
          JobDataMap newData = jobDetail.getJobDataMap();
          if (newData != null) {
            newData = (JobDataMap)newData.clone();
            newData.clearDirtyFlag();
          }
          ((JobDetailImpl)jd).setJobDataMap(newData);
        }
        if (jd.isConcurrentExectionDisallowed()) {
          this.blockedJobs.remove(jd.getKey());
          ArrayList trigs = getTriggerWrappersForJob(jd.getKey());
          for (TriggerWrapper ttw : trigs) {
            if (ttw.state == 5) {
              ttw.state = 0;
              this.timeTriggers.add(ttw);
            }
            if (ttw.state == 6) {
              ttw.state = 4;
            }
          }
          this.signaler.signalSchedulingChange(0L);
        }
      } else {
        this.blockedJobs.remove(jobDetail.getKey());
      }

      if (tw != null)
        if (triggerInstCode == Trigger.CompletedExecutionInstruction.DELETE_TRIGGER)
        {
          if (trigger.getNextFireTime() == null)
          {
            if (tw.getTrigger().getNextFireTime() == null)
              removeTrigger(trigger.getKey());
          }
          else {
            removeTrigger(trigger.getKey());
            this.signaler.signalSchedulingChange(0L);
          }
        } else if (triggerInstCode == Trigger.CompletedExecutionInstruction.SET_TRIGGER_COMPLETE) {
          tw.state = 3;
          this.timeTriggers.remove(tw);
          this.signaler.signalSchedulingChange(0L);
        } else if (triggerInstCode == Trigger.CompletedExecutionInstruction.SET_TRIGGER_ERROR) {
          getLog().info("Trigger " + trigger.getKey() + " set to ERROR state.");
          tw.state = 7;
          this.signaler.signalSchedulingChange(0L);
        } else if (triggerInstCode == Trigger.CompletedExecutionInstruction.SET_ALL_JOB_TRIGGERS_ERROR) {
          getLog().info("All triggers of Job " + trigger.getJobKey() + " set to ERROR state.");

          setAllTriggersOfJobToState(trigger.getJobKey(), 7);
          this.signaler.signalSchedulingChange(0L);
        } else if (triggerInstCode == Trigger.CompletedExecutionInstruction.SET_ALL_JOB_TRIGGERS_COMPLETE) {
          setAllTriggersOfJobToState(trigger.getJobKey(), 3);
          this.signaler.signalSchedulingChange(0L);
        }
    }
  }

  protected void setAllTriggersOfJobToState(JobKey jobKey, int state)
  {
    ArrayList tws = getTriggerWrappersForJob(jobKey);
    Iterator itr = tws.iterator();
    while (itr.hasNext()) {
      TriggerWrapper tw = (TriggerWrapper)itr.next();
      tw.state = state;
      if (state != 0)
        this.timeTriggers.remove(tw);
    }
  }

  protected String peekTriggers()
  {
    StringBuffer str = new StringBuffer();
    TriggerWrapper tw = null;
    Iterator valueIter;
    synchronized (this.lock) {
      for (valueIter = this.triggersByKey.values().iterator(); valueIter.hasNext(); ) {
        tw = (TriggerWrapper)valueIter.next();
        str.append(tw.trigger.getKey().getName());
        str.append("/");
      }
    }
    str.append(" | ");

    synchronized (this.lock) {
      Iterator itr = this.timeTriggers.iterator();
      while (itr.hasNext()) {
        tw = (TriggerWrapper)itr.next();
        str.append(tw.trigger.getKey().getName());
        str.append("->");
      }
    }

    return str.toString();
  }

  public Set<String> getPausedTriggerGroups()
    throws JobPersistenceException
  {
    HashSet set = new HashSet();

    set.addAll(this.pausedTriggerGroups);

    return set;
  }

  public void setInstanceId(String schedInstId)
  {
  }

  public void setInstanceName(String schedName)
  {
  }

  public void setThreadPoolSize(int poolSize)
  {
  }

  public long getEstimatedTimeToReleaseAndAcquireTrigger() {
    return 5L;
  }

  public boolean isClustered() {
    return false;
  }
}