package org.quartz.utils;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Properties;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.sql.DataSource;
import javax.sql.XAConnection;
import javax.sql.XADataSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class JNDIConnectionProvider
  implements ConnectionProvider
{
  private String url;
  private Properties props;
  private Object datasource;
  private boolean alwaysLookup = false;

  private final Logger log = LoggerFactory.getLogger(getClass());

  public JNDIConnectionProvider(String jndiUrl, boolean alwaysLookup)
  {
    this.url = jndiUrl;
    this.alwaysLookup = alwaysLookup;
    init();
  }

  public JNDIConnectionProvider(String jndiUrl, Properties jndiProps, boolean alwaysLookup)
  {
    this.url = jndiUrl;
    this.props = jndiProps;
    this.alwaysLookup = alwaysLookup;
    init();
  }

  protected Logger getLog()
  {
    return this.log;
  }

  private void init()
  {
    if (!isAlwaysLookup()) {
      Context ctx = null;
      try {
        ctx = this.props != null ? new InitialContext(this.props) : new InitialContext();

        this.datasource = ((DataSource)ctx.lookup(this.url));
      } catch (Exception e) {
        getLog().error("Error looking up datasource: " + e.getMessage(), e);
      }
      finally {
        if (ctx != null) try {
            ctx.close();
          } catch (Exception ignore) {
          } 
      }
    }
  }

  public Connection getConnection() throws SQLException { Context ctx = null;
    try {
      Object ds = this.datasource;

      if ((ds == null) || (isAlwaysLookup())) {
        ctx = this.props != null ? new InitialContext(this.props) : new InitialContext();

        ds = ctx.lookup(this.url);
        if (!isAlwaysLookup()) {
          this.datasource = ds;
        }
      }

      if (ds == null)
        throw new SQLException("There is no object at the JNDI URL '" + this.url + "'");
      Connection localConnection;
      if ((ds instanceof XADataSource))
        return ((XADataSource)ds).getXAConnection().getConnection();
      if ((ds instanceof DataSource)) {
        return ((DataSource)ds).getConnection();
      }
      throw new SQLException("Object at JNDI URL '" + this.url + "' is not a DataSource.");
    }
    catch (Exception e) {
      this.datasource = null;
      throw new SQLException("Could not retrieve datasource via JNDI url '" + this.url + "' " + e.getClass().getName() + ": " + e.getMessage());
    }
    finally
    {
      if (ctx != null) try {
          ctx.close();
        }
        catch (Exception ignore) {
        } 
    } } 
  public boolean isAlwaysLookup() {
    return this.alwaysLookup;
  }

  public void setAlwaysLookup(boolean b) {
    this.alwaysLookup = b;
  }

  public void shutdown()
    throws SQLException
  {
  }
}