package org.quartz.xml;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLDecoder;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;
import javax.xml.namespace.NamespaceContext;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathException;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import org.quartz.CalendarIntervalScheduleBuilder;
import org.quartz.CronScheduleBuilder;
import org.quartz.DateBuilder.IntervalUnit;
import org.quartz.Job;
import org.quartz.JobBuilder;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobKey;
import org.quartz.ObjectAlreadyExistsException;
import org.quartz.ScheduleBuilder;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.SimpleScheduleBuilder;
import org.quartz.Trigger;
import org.quartz.TriggerBuilder;
import org.quartz.TriggerKey;
import org.quartz.impl.matchers.GroupMatcher;
import org.quartz.spi.ClassLoadHelper;
import org.quartz.spi.MutableTrigger;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class XMLSchedulingDataProcessor
  implements ErrorHandler
{
  public static final String QUARTZ_NS = "http://www.quartz-scheduler.org/xml/JobSchedulingData";
  public static final String QUARTZ_SCHEMA_WEB_URL = "http://www.quartz-scheduler.org/xml/job_scheduling_data_2_0.xsd";
  public static final String QUARTZ_XSD_PATH_IN_JAR = "org/quartz/xml/job_scheduling_data_2_0.xsd";
  public static final String QUARTZ_XML_DEFAULT_FILE_NAME = "quartz_data.xml";
  public static final String QUARTZ_SYSTEM_ID_JAR_PREFIX = "jar:";
  protected static final String XSD_DATE_FORMAT = "yyyy-MM-dd'T'hh:mm:ss";
  protected static final SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'hh:mm:ss");

  protected List<String> jobGroupsToDelete = new LinkedList();
  protected List<String> triggerGroupsToDelete = new LinkedList();
  protected List<JobKey> jobsToDelete = new LinkedList();
  protected List<TriggerKey> triggersToDelete = new LinkedList();

  protected List<JobDetail> loadedJobs = new LinkedList();
  protected List<MutableTrigger> loadedTriggers = new LinkedList();

  private boolean overWriteExistingData = true;
  private boolean ignoreDuplicates = false;

  protected Collection<Exception> validationExceptions = new ArrayList();
  protected ClassLoadHelper classLoadHelper;
  protected List<String> jobGroupsToNeverDelete = new LinkedList();
  protected List<String> triggerGroupsToNeverDelete = new LinkedList();

  private DocumentBuilder docBuilder = null;
  private XPath xpath = null;

  private final Logger log = LoggerFactory.getLogger(getClass());

  public XMLSchedulingDataProcessor(ClassLoadHelper clh)
    throws ParserConfigurationException
  {
    this.classLoadHelper = clh;
    initDocumentParser();
  }

  protected void initDocumentParser()
    throws ParserConfigurationException
  {
    DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();

    docBuilderFactory.setNamespaceAware(true);
    docBuilderFactory.setValidating(true);

    docBuilderFactory.setAttribute("http://java.sun.com/xml/jaxp/properties/schemaLanguage", "http://www.w3.org/2001/XMLSchema");

    docBuilderFactory.setAttribute("http://java.sun.com/xml/jaxp/properties/schemaSource", resolveSchemaSource());

    this.docBuilder = docBuilderFactory.newDocumentBuilder();

    this.docBuilder.setErrorHandler(this);

    NamespaceContext nsContext = new NamespaceContext()
    {
      public String getNamespaceURI(String prefix)
      {
        if (prefix == null)
          throw new IllegalArgumentException("Null prefix");
        if ("xml".equals(prefix))
          return "http://www.w3.org/XML/1998/namespace";
        if ("xmlns".equals(prefix)) {
          return "http://www.w3.org/2000/xmlns/";
        }
        if ("q".equals(prefix)) {
          return "http://www.quartz-scheduler.org/xml/JobSchedulingData";
        }
        return "";
      }

      public Iterator<?> getPrefixes(String namespaceURI)
      {
        throw new UnsupportedOperationException();
      }

      public String getPrefix(String namespaceURI)
      {
        throw new UnsupportedOperationException();
      }
    };
    this.xpath = XPathFactory.newInstance().newXPath();
    this.xpath.setNamespaceContext(nsContext);
  }

  protected Object resolveSchemaSource() {
    InputSource inputSource = null;

    InputStream is = null;
    try
    {
      is = this.classLoadHelper.getResourceAsStream("org/quartz/xml/job_scheduling_data_2_0.xsd");
    } finally {
      if (is != null) {
        inputSource = new InputSource(is);
        inputSource.setSystemId("http://www.quartz-scheduler.org/xml/job_scheduling_data_2_0.xsd");
        this.log.debug("Utilizing schema packaged in local quartz distribution jar.");
      }
      else {
        this.log.info("Unable to load local schema packaged in quartz distribution jar. Utilizing schema online at http://www.quartz-scheduler.org/xml/job_scheduling_data_2_0.xsd");
        return "http://www.quartz-scheduler.org/xml/job_scheduling_data_2_0.xsd"; }  } throw localObject;

    return inputSource;
  }

  public boolean isOverWriteExistingData()
  {
    return this.overWriteExistingData;
  }

  protected void setOverWriteExistingData(boolean overWriteExistingData)
  {
    this.overWriteExistingData = overWriteExistingData;
  }

  public boolean isIgnoreDuplicates()
  {
    return this.ignoreDuplicates;
  }

  public void setIgnoreDuplicates(boolean ignoreDuplicates)
  {
    this.ignoreDuplicates = ignoreDuplicates;
  }

  public void addJobGroupToNeverDelete(String group)
  {
    if (group != null)
      this.jobGroupsToNeverDelete.add(group);
  }

  public boolean removeJobGroupToNeverDelete(String group)
  {
    if (group != null)
      return this.jobGroupsToNeverDelete.remove(group);
    return false;
  }

  public List<String> getJobGroupsToNeverDelete()
  {
    return Collections.unmodifiableList(this.jobGroupsToDelete);
  }

  public void addTriggerGroupToNeverDelete(String group)
  {
    if (group != null)
      this.triggerGroupsToNeverDelete.add(group);
  }

  public boolean removeTriggerGroupToNeverDelete(String group)
  {
    if (group != null)
      return this.triggerGroupsToNeverDelete.remove(group);
    return false;
  }

  public List<String> getTriggerGroupsToNeverDelete()
  {
    return Collections.unmodifiableList(this.triggerGroupsToDelete);
  }

  protected void processFile()
    throws Exception
  {
    processFile("quartz_data.xml");
  }

  protected void processFile(String fileName)
    throws Exception
  {
    processFile(fileName, getSystemIdForFileName(fileName));
  }

  protected String getSystemIdForFileName(String fileName)
  {
    InputStream fileInputStream = null;
    try {
      String urlPath = null;

      File file = new File(fileName);
      if (!file.exists()) {
        URL url = getURL(fileName);
        if (url != null) {
          try {
            urlPath = URLDecoder.decode(url.getPath(), "UTF-8");
          } catch (UnsupportedEncodingException e) {
            this.log.warn("Unable to decode file path URL", e);
          }
          try {
            if (url != null)
              fileInputStream = url.openStream();
          } catch (IOException ignore) {
          }
        }
      } else {
        try {
          fileInputStream = new FileInputStream(file);
        }
        catch (FileNotFoundException ignore) {
        }
      }
      if (fileInputStream == null) {
        this.log.debug("Unable to resolve '" + fileName + "' to full path, so using it as is for system id.");
        return fileName;
      }
      return urlPath != null ? urlPath : file.getAbsolutePath();
    }
    finally {
      try {
        if (fileInputStream != null)
          fileInputStream.close();
      }
      catch (IOException ioe) {
        this.log.warn("Error closing jobs file: " + fileName, ioe);
      }
    }
  }

  protected URL getURL(String fileName)
  {
    return this.classLoadHelper.getResource(fileName);
  }

  protected void prepForProcessing()
  {
    clearValidationExceptions();

    setOverWriteExistingData(true);
    setIgnoreDuplicates(false);

    this.jobGroupsToDelete.clear();
    this.jobsToDelete.clear();
    this.triggerGroupsToDelete.clear();
    this.triggersToDelete.clear();

    this.loadedJobs.clear();
    this.loadedTriggers.clear();
  }

  protected void processFile(String fileName, String systemId)
    throws ValidationException, ParserConfigurationException, SAXException, IOException, SchedulerException, ClassNotFoundException, ParseException, XPathException
  {
    prepForProcessing();

    this.log.info("Parsing XML file: " + fileName + " with systemId: " + systemId);

    InputSource is = new InputSource(getInputStream(fileName));
    is.setSystemId(systemId);

    process(is);

    maybeThrowValidationException();
  }

  public void processStreamAndScheduleJobs(InputStream stream, String systemId, Scheduler sched)
    throws ValidationException, ParserConfigurationException, SAXException, XPathException, IOException, SchedulerException, ClassNotFoundException, ParseException
  {
    prepForProcessing();

    this.log.info("Parsing XML from stream with systemId: " + systemId);

    InputSource is = new InputSource(stream);
    is.setSystemId(systemId);

    process(is);
    executePreProcessCommands(sched);
    scheduleJobs(sched);

    maybeThrowValidationException();
  }

  protected void process(InputSource is)
    throws SAXException, IOException, ParseException, XPathException, ClassNotFoundException
  {
    Document document = this.docBuilder.parse(is);

    NodeList deleteJobGroupNodes = (NodeList)this.xpath.evaluate("/q:job-scheduling-data/q:pre-processing-commands/q:delete-jobs-in-group", document, XPathConstants.NODESET);

    this.log.debug("Found " + deleteJobGroupNodes.getLength() + " delete job group commands.");

    for (int i = 0; i < deleteJobGroupNodes.getLength(); i++) {
      Node node = deleteJobGroupNodes.item(i);
      String t = node.getTextContent();
      if ((t != null) && ((t = t.trim()).length() != 0))
      {
        this.jobGroupsToDelete.add(t);
      }
    }
    NodeList deleteTriggerGroupNodes = (NodeList)this.xpath.evaluate("/q:job-scheduling-data/q:pre-processing-commands/q:delete-triggers-in-group", document, XPathConstants.NODESET);

    this.log.debug("Found " + deleteTriggerGroupNodes.getLength() + " delete trigger group commands.");

    for (int i = 0; i < deleteTriggerGroupNodes.getLength(); i++) {
      Node node = deleteTriggerGroupNodes.item(i);
      String t = node.getTextContent();
      if ((t != null) && ((t = t.trim()).length() != 0))
      {
        this.triggerGroupsToDelete.add(t);
      }
    }
    NodeList deleteJobNodes = (NodeList)this.xpath.evaluate("/q:job-scheduling-data/q:pre-processing-commands/q:delete-job", document, XPathConstants.NODESET);

    this.log.debug("Found " + deleteJobNodes.getLength() + " delete job commands.");

    for (int i = 0; i < deleteJobNodes.getLength(); i++) {
      Node node = deleteJobNodes.item(i);

      String name = getTrimmedToNullString(this.xpath, "q:name", node);
      String group = getTrimmedToNullString(this.xpath, "q:group", node);

      if (name == null)
        throw new ParseException("Encountered a 'delete-job' command without a name specified.", -1);
      this.jobsToDelete.add(new JobKey(name, group));
    }

    NodeList deleteTriggerNodes = (NodeList)this.xpath.evaluate("/q:job-scheduling-data/q:pre-processing-commands/q:delete-trigger", document, XPathConstants.NODESET);

    this.log.debug("Found " + deleteTriggerNodes.getLength() + " delete trigger commands.");

    for (int i = 0; i < deleteTriggerNodes.getLength(); i++) {
      Node node = deleteTriggerNodes.item(i);

      String name = getTrimmedToNullString(this.xpath, "q:name", node);
      String group = getTrimmedToNullString(this.xpath, "q:group", node);

      if (name == null)
        throw new ParseException("Encountered a 'delete-trigger' command without a name specified.", -1);
      this.triggersToDelete.add(new TriggerKey(name, group));
    }

    Boolean overWrite = getBoolean(this.xpath, "/q:job-scheduling-data/q:processing-directives/q:overwrite-existing-data", document);

    if (overWrite == null) {
      this.log.debug("Directive 'overwrite-existing-data' not specified, defaulting to " + isOverWriteExistingData());
    }
    else {
      this.log.debug("Directive 'overwrite-existing-data' specified as: " + overWrite);
      setOverWriteExistingData(overWrite.booleanValue());
    }

    Boolean ignoreDupes = getBoolean(this.xpath, "/q:job-scheduling-data/q:processing-directives/q:ignore-duplicates", document);

    if (ignoreDupes == null) {
      this.log.debug("Directive 'ignore-duplicates' not specified, defaulting to " + isIgnoreDuplicates());
    }
    else {
      this.log.debug("Directive 'ignore-duplicates' specified as: " + ignoreDupes);
      setIgnoreDuplicates(ignoreDupes.booleanValue());
    }

    NodeList jobNodes = (NodeList)this.xpath.evaluate("/q:job-scheduling-data/q:schedule/q:job", document, XPathConstants.NODESET);

    this.log.debug("Found " + jobNodes.getLength() + " job definitions.");

    for (int i = 0; i < jobNodes.getLength(); i++) {
      Node jobDetailNode = jobNodes.item(i);
      String t = null;

      String jobName = getTrimmedToNullString(this.xpath, "q:name", jobDetailNode);
      String jobGroup = getTrimmedToNullString(this.xpath, "q:group", jobDetailNode);
      String jobDescription = getTrimmedToNullString(this.xpath, "q:description", jobDetailNode);
      String jobClassName = getTrimmedToNullString(this.xpath, "q:job-class", jobDetailNode);
      t = getTrimmedToNullString(this.xpath, "q:durability", jobDetailNode);
      boolean jobDurability = (t != null) && (t.equals("true"));
      t = getTrimmedToNullString(this.xpath, "q:recover", jobDetailNode);
      boolean jobRecoveryRequested = (t != null) && (t.equals("true"));

      Class jobClass = this.classLoadHelper.loadClass(jobClassName, Job.class);

      JobDetail jobDetail = JobBuilder.newJob(jobClass).withIdentity(jobName, jobGroup).withDescription(jobDescription).storeDurably(jobDurability).requestRecovery(jobRecoveryRequested).build();

      NodeList jobDataEntries = (NodeList)this.xpath.evaluate("q:job-data-map/q:entry", jobDetailNode, XPathConstants.NODESET);

      for (int k = 0; k < jobDataEntries.getLength(); k++) {
        Node entryNode = jobDataEntries.item(k);
        String key = getTrimmedToNullString(this.xpath, "q:key", entryNode);
        String value = getTrimmedToNullString(this.xpath, "q:value", entryNode);
        jobDetail.getJobDataMap().put(key, value);
      }

      if (this.log.isDebugEnabled()) {
        this.log.debug("Parsed job definition: " + jobDetail);
      }
      addJobToSchedule(jobDetail);
    }

    NodeList triggerEntries = (NodeList)this.xpath.evaluate("/q:job-scheduling-data/q:schedule/q:trigger/*", document, XPathConstants.NODESET);

    this.log.debug("Found " + triggerEntries.getLength() + " trigger definitions.");

    for (int j = 0; j < triggerEntries.getLength(); j++) {
      Node triggerNode = triggerEntries.item(j);
      String triggerName = getTrimmedToNullString(this.xpath, "q:name", triggerNode);
      String triggerGroup = getTrimmedToNullString(this.xpath, "q:group", triggerNode);
      String triggerDescription = getTrimmedToNullString(this.xpath, "q:description", triggerNode);
      String triggerMisfireInstructionConst = getTrimmedToNullString(this.xpath, "q:misfire-instruction", triggerNode);
      String triggerPriorityString = getTrimmedToNullString(this.xpath, "q:priority", triggerNode);
      String triggerCalendarRef = getTrimmedToNullString(this.xpath, "q:calendar-name", triggerNode);
      String triggerJobName = getTrimmedToNullString(this.xpath, "q:job-name", triggerNode);
      String triggerJobGroup = getTrimmedToNullString(this.xpath, "q:job-group", triggerNode);

      int triggerPriority = 5;
      if (triggerPriorityString != null) {
        triggerPriority = Integer.valueOf(triggerPriorityString).intValue();
      }
      String startTimeString = getTrimmedToNullString(this.xpath, "q:start-time", triggerNode);
      String startTimeFutureSecsString = getTrimmedToNullString(this.xpath, "q:start-time-seconds-in-future", triggerNode);
      String endTimeString = getTrimmedToNullString(this.xpath, "q:end-time", triggerNode);

      Date triggerStartTime = null;
      if (startTimeFutureSecsString != null)
        triggerStartTime = new Date(System.currentTimeMillis() + Long.valueOf(startTimeFutureSecsString).longValue() * 1000L);
      else
        triggerStartTime = (startTimeString == null) || (startTimeString.length() == 0) ? new Date() : dateFormat.parse(startTimeString);
      Date triggerEndTime = (endTimeString == null) || (endTimeString.length() == 0) ? null : dateFormat.parse(endTimeString);

      TriggerKey triggerKey = TriggerKey.triggerKey(triggerName, triggerGroup);

      ScheduleBuilder sched = null;

      if (triggerNode.getNodeName().equals("simple")) {
        String repeatCountString = getTrimmedToNullString(this.xpath, "q:repeat-count", triggerNode);
        String repeatIntervalString = getTrimmedToNullString(this.xpath, "q:repeat-interval", triggerNode);

        int repeatCount = repeatCountString == null ? -1 : Integer.parseInt(repeatCountString);
        long repeatInterval = repeatIntervalString == null ? 0L : Long.parseLong(repeatIntervalString);

        sched = SimpleScheduleBuilder.simpleSchedule().withIntervalInMilliseconds(repeatInterval).withRepeatCount(repeatCount);

        if ((triggerMisfireInstructionConst != null) && (triggerMisfireInstructionConst.length() != 0))
          if (triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_FIRE_NOW"))
            ((SimpleScheduleBuilder)sched).withMisfireHandlingInstructionFireNow();
          else if (triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_RESCHEDULE_NEXT_WITH_EXISTING_COUNT"))
            ((SimpleScheduleBuilder)sched).withMisfireHandlingInstructionNextWithExistingCount();
          else if (triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_RESCHEDULE_NEXT_WITH_REMAINING_COUNT"))
            ((SimpleScheduleBuilder)sched).withMisfireHandlingInstructionNextWithRemainingCount();
          else if (triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_RESCHEDULE_NOW_WITH_EXISTING_REPEAT_COUNT"))
            ((SimpleScheduleBuilder)sched).withMisfireHandlingInstructionNowWithExistingCount();
          else if (triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_RESCHEDULE_NOW_WITH_REMAINING_REPEAT_COUNT"))
            ((SimpleScheduleBuilder)sched).withMisfireHandlingInstructionNowWithRemainingCount();
          else if (!triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_SMART_POLICY"))
          {
            throw new ParseException("Unexpected/Unhandlable Misfire Instruction encountered '" + triggerMisfireInstructionConst + "', for trigger: " + triggerKey, -1);
          }
      } else if (triggerNode.getNodeName().equals("cron")) {
        String cronExpression = getTrimmedToNullString(this.xpath, "q:cron-expression", triggerNode);
        String timezoneString = getTrimmedToNullString(this.xpath, "q:time-zone", triggerNode);

        TimeZone tz = timezoneString == null ? null : TimeZone.getTimeZone(timezoneString);

        sched = CronScheduleBuilder.cronSchedule(cronExpression).inTimeZone(tz);

        if ((triggerMisfireInstructionConst != null) && (triggerMisfireInstructionConst.length() != 0))
          if (triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_DO_NOTHING"))
            ((CronScheduleBuilder)sched).withMisfireHandlingInstructionDoNothing();
          else if (triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_FIRE_ONCE_NOW"))
            ((CronScheduleBuilder)sched).withMisfireHandlingInstructionFireAndProceed();
          else if (!triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_SMART_POLICY"))
          {
            throw new ParseException("Unexpected/Unhandlable Misfire Instruction encountered '" + triggerMisfireInstructionConst + "', for trigger: " + triggerKey, -1);
          }
      } else if (triggerNode.getNodeName().equals("calendar-interval")) {
        String repeatIntervalString = getTrimmedToNullString(this.xpath, "q:repeat-interval", triggerNode);
        String repeatUnitString = getTrimmedToNullString(this.xpath, "q:repeat-interval-unit", triggerNode);

        int repeatInterval = Integer.parseInt(repeatIntervalString);

        DateBuilder.IntervalUnit repeatUnit = DateBuilder.IntervalUnit.valueOf(repeatUnitString);

        sched = CalendarIntervalScheduleBuilder.calendarIntervalSchedule().withInterval(repeatInterval, repeatUnit);

        if ((triggerMisfireInstructionConst != null) && (triggerMisfireInstructionConst.length() != 0))
          if (triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_DO_NOTHING"))
            ((CalendarIntervalScheduleBuilder)sched).withMisfireHandlingInstructionDoNothing();
          else if (triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_FIRE_ONCE_NOW"))
            ((CalendarIntervalScheduleBuilder)sched).withMisfireHandlingInstructionFireAndProceed();
          else if (!triggerMisfireInstructionConst.equals("MISFIRE_INSTRUCTION_SMART_POLICY"))
          {
            throw new ParseException("Unexpected/Unhandlable Misfire Instruction encountered '" + triggerMisfireInstructionConst + "', for trigger: " + triggerKey, -1);
          }
      } else {
        throw new ParseException("Unknown trigger type: " + triggerNode.getNodeName(), -1);
      }

      MutableTrigger trigger = (MutableTrigger)TriggerBuilder.newTrigger().withIdentity(triggerName, triggerGroup).withDescription(triggerDescription).forJob(triggerJobName, triggerJobGroup).startAt(triggerStartTime).endAt(triggerEndTime).withPriority(triggerPriority).modifiedByCalendar(triggerCalendarRef).withSchedule(sched).build();

      NodeList jobDataEntries = (NodeList)this.xpath.evaluate("q:job-data-map/q:entry", triggerNode, XPathConstants.NODESET);

      for (int k = 0; k < jobDataEntries.getLength(); k++) {
        Node entryNode = jobDataEntries.item(k);
        String key = getTrimmedToNullString(this.xpath, "q:key", entryNode);
        String value = getTrimmedToNullString(this.xpath, "q:value", entryNode);
        trigger.getJobDataMap().put(key, value);
      }

      if (this.log.isDebugEnabled()) {
        this.log.debug("Parsed trigger definition: " + trigger);
      }
      addTriggerToSchedule(trigger);
    }
  }

  protected String getTrimmedToNullString(XPath xpathToElement, String elementName, Node parentNode) throws XPathExpressionException {
    String str = (String)xpathToElement.evaluate(elementName, parentNode, XPathConstants.STRING);

    if (str != null) {
      str = str.trim();
    }
    if ((str != null) && (str.length() == 0)) {
      str = null;
    }
    return str;
  }

  protected Boolean getBoolean(XPath xpathToElement, String elementName, Document document) throws XPathExpressionException
  {
    Node directive = (Node)xpathToElement.evaluate(elementName, document, XPathConstants.NODE);

    if ((directive == null) || (directive.getTextContent() == null)) {
      return null;
    }
    String val = directive.getTextContent();
    if ((val.equalsIgnoreCase("true")) || (val.equalsIgnoreCase("yes")) || (val.equalsIgnoreCase("y"))) {
      return Boolean.TRUE;
    }
    return Boolean.FALSE;
  }

  public void processFileAndScheduleJobs(Scheduler sched, boolean overWriteExistingJobs)
    throws SchedulerException, Exception
  {
    String fileName = "quartz_data.xml";
    processFile(fileName, getSystemIdForFileName(fileName));

    setOverWriteExistingData(overWriteExistingJobs);
    executePreProcessCommands(sched);
    scheduleJobs(sched);
  }

  public void processFileAndScheduleJobs(String fileName, Scheduler sched)
    throws Exception
  {
    processFileAndScheduleJobs(fileName, getSystemIdForFileName(fileName), sched);
  }

  public void processFileAndScheduleJobs(String fileName, String systemId, Scheduler sched)
    throws Exception
  {
    processFile(fileName, systemId);
    executePreProcessCommands(sched);
    scheduleJobs(sched);
  }

  protected List<JobDetail> getLoadedJobs()
  {
    return Collections.unmodifiableList(this.loadedJobs);
  }

  protected List<MutableTrigger> getLoadedTriggers()
  {
    return Collections.unmodifiableList(this.loadedTriggers);
  }

  protected InputStream getInputStream(String fileName)
  {
    return this.classLoadHelper.getResourceAsStream(fileName);
  }

  protected void addJobToSchedule(JobDetail job) {
    this.loadedJobs.add(job);
  }

  protected void addTriggerToSchedule(MutableTrigger trigger) {
    this.loadedTriggers.add(trigger);
  }

  private Map<JobKey, List<MutableTrigger>> buildTriggersByFQJobNameMap(List<MutableTrigger> triggers)
  {
    Map triggersByFQJobName = new HashMap();

    for (MutableTrigger trigger : triggers) {
      List triggersOfJob = (List)triggersByFQJobName.get(trigger.getJobKey());
      if (triggersOfJob == null) {
        triggersOfJob = new LinkedList();
        triggersByFQJobName.put(trigger.getJobKey(), triggersOfJob);
      }
      triggersOfJob.add(trigger);
    }

    return triggersByFQJobName;
  }

  protected void executePreProcessCommands(Scheduler scheduler)
    throws SchedulerException
  {
    for (String group : this.jobGroupsToDelete) {
      if (group.equals("*")) {
        this.log.info("Deleting all jobs in ALL groups.");
        for (String groupName : scheduler.getJobGroupNames()) {
          if (!this.jobGroupsToNeverDelete.contains(groupName)) {
            for (JobKey key : scheduler.getJobKeys(GroupMatcher.jobGroupEquals(groupName))) {
              scheduler.deleteJob(key);
            }
          }
        }

      }
      else if (!this.jobGroupsToNeverDelete.contains(group)) {
        this.log.info("Deleting all jobs in group: {}", group);
        for (JobKey key : scheduler.getJobKeys(GroupMatcher.jobGroupEquals(group))) {
          scheduler.deleteJob(key);
        }
      }

    }

    for (String group : this.triggerGroupsToDelete) {
      if (group.equals("*")) {
        this.log.info("Deleting all triggers in ALL groups.");
        for (String groupName : scheduler.getTriggerGroupNames()) {
          if (!this.triggerGroupsToNeverDelete.contains(groupName)) {
            for (TriggerKey key : scheduler.getTriggerKeys(GroupMatcher.triggerGroupEquals(groupName))) {
              scheduler.unscheduleJob(key);
            }
          }
        }

      }
      else if (!this.triggerGroupsToNeverDelete.contains(group)) {
        this.log.info("Deleting all triggers in group: {}", group);
        for (TriggerKey key : scheduler.getTriggerKeys(GroupMatcher.triggerGroupEquals(group))) {
          scheduler.unscheduleJob(key);
        }
      }

    }

    for (JobKey key : this.jobsToDelete) {
      if (!this.jobGroupsToNeverDelete.contains(key.getGroup())) {
        this.log.info("Deleting job: {}", key);
        scheduler.deleteJob(key);
      }
    }

    for (TriggerKey key : this.triggersToDelete)
      if (!this.triggerGroupsToNeverDelete.contains(key.getGroup())) {
        this.log.info("Deleting trigger: {}", key);
        scheduler.unscheduleJob(key);
      }
  }

  protected void scheduleJobs(Scheduler sched)
    throws SchedulerException
  {
    List jobs = new LinkedList(getLoadedJobs());
    List triggers = new LinkedList(getLoadedTriggers());

    this.log.info("Adding " + jobs.size() + " jobs, " + triggers.size() + " triggers.");

    Map triggersByFQJobName = buildTriggersByFQJobNameMap(triggers);

    Iterator itr = jobs.iterator();
    while (itr.hasNext()) {
      JobDetail detail = (JobDetail)itr.next();
      itr.remove();

      JobDetail dupeJ = sched.getJobDetail(detail.getKey());

      if (dupeJ != null) {
        if ((!isOverWriteExistingData()) && (isIgnoreDuplicates())) {
          this.log.info("Not overwriting existing job: " + dupeJ.getKey());
        }
        else if ((!isOverWriteExistingData()) && (!isIgnoreDuplicates()))
          throw new ObjectAlreadyExistsException(detail);
      }
      else
      {
        if (dupeJ != null)
          this.log.info("Replacing job: " + detail.getKey());
        else {
          this.log.info("Adding job: " + detail.getKey());
        }

        List triggersOfJob = (List)triggersByFQJobName.get(detail.getKey());

        if ((!detail.isDurable()) && ((triggersOfJob == null) || (triggersOfJob.size() == 0))) {
          if (dupeJ == null) {
            throw new SchedulerException("A new job defined without any triggers must be durable: " + detail.getKey());
          }

          if ((dupeJ.isDurable()) && (sched.getTriggersOfJob(detail.getKey()).size() == 0))
          {
            throw new SchedulerException("Can't change existing durable job without triggers to non-durable: " + detail.getKey());
          }

        }

        if ((dupeJ != null) || (detail.isDurable())) {
          sched.addJob(detail, true);
        }
        else {
          boolean addJobWithFirstSchedule = true;

          Iterator titr = triggersOfJob.iterator();
          while (titr.hasNext()) {
            MutableTrigger trigger = (MutableTrigger)titr.next();
            triggers.remove(trigger);

            if (trigger.getStartTime() == null) {
              trigger.setStartTime(new Date());
            }

            Trigger dupeT = sched.getTrigger(trigger.getKey());
            if (dupeT != null) {
              if (isOverWriteExistingData()) {
                if (this.log.isDebugEnabled())
                  this.log.debug("Rescheduling job: " + trigger.getJobKey() + " with updated trigger: " + trigger.getKey());
              }
              else
              {
                if (isIgnoreDuplicates()) {
                  this.log.info("Not overwriting existing trigger: " + dupeT.getKey());
                  continue;
                }

                throw new ObjectAlreadyExistsException(trigger);
              }

              if (!dupeT.getJobKey().equals(trigger.getJobKey())) {
                this.log.warn("Possibly duplicately named ({}) triggers in jobs xml file! ", trigger.getKey());
              }

              sched.rescheduleJob(trigger.getKey(), trigger);
            } else {
              if (this.log.isDebugEnabled()) {
                this.log.debug("Scheduling job: " + trigger.getJobKey() + " with trigger: " + trigger.getKey());
              }

              try
              {
                if (addJobWithFirstSchedule) {
                  sched.scheduleJob(detail, trigger);
                  addJobWithFirstSchedule = false;
                }
                else {
                  sched.scheduleJob(trigger);
                }
              } catch (ObjectAlreadyExistsException e) {
                if (this.log.isDebugEnabled()) {
                  this.log.debug("Adding trigger: " + trigger.getKey() + " for job: " + detail.getKey() + " failed because the trigger already existed.  " + "This is likely due to a race condition between multiple instances " + "in the cluster.  Will try to reschedule instead.");
                }

                sched.rescheduleJob(trigger.getKey(), trigger);
              }
            }
          }
        }
      }
    }

    for (MutableTrigger trigger : triggers)
    {
      if (trigger.getStartTime() == null) {
        trigger.setStartTime(new Date());
      }

      Trigger dupeT = sched.getTrigger(trigger.getKey());
      if (dupeT != null) {
        if (isOverWriteExistingData()) {
          if (this.log.isDebugEnabled())
            this.log.debug("Rescheduling job: " + trigger.getJobKey() + " with updated trigger: " + trigger.getKey());
        }
        else
        {
          if (isIgnoreDuplicates()) {
            this.log.info("Not overwriting existing trigger: " + dupeT.getKey());
            continue;
          }

          throw new ObjectAlreadyExistsException(trigger);
        }

        if (!dupeT.getJobKey().equals(trigger.getJobKey())) {
          this.log.warn("Possibly duplicately named ({}) triggers in jobs xml file! ", trigger.getKey());
        }

        sched.rescheduleJob(trigger.getKey(), trigger);
      } else {
        if (this.log.isDebugEnabled()) {
          this.log.debug("Scheduling job: " + trigger.getJobKey() + " with trigger: " + trigger.getKey());
        }

        try
        {
          sched.scheduleJob(trigger);
        } catch (ObjectAlreadyExistsException e) {
          if (this.log.isDebugEnabled()) {
            this.log.debug("Adding trigger: " + trigger.getKey() + " for job: " + trigger.getJobKey() + " failed because the trigger already existed.  " + "This is likely due to a race condition between multiple instances " + "in the cluster.  Will try to reschedule instead.");
          }

          sched.rescheduleJob(trigger.getKey(), trigger);
        }
      }
    }
  }

  public void warning(SAXParseException e)
    throws SAXException
  {
    addValidationException(e);
  }

  public void error(SAXParseException e)
    throws SAXException
  {
    addValidationException(e);
  }

  public void fatalError(SAXParseException e)
    throws SAXException
  {
    addValidationException(e);
  }

  protected void addValidationException(SAXException e)
  {
    this.validationExceptions.add(e);
  }

  protected void clearValidationExceptions()
  {
    this.validationExceptions.clear();
  }

  protected void maybeThrowValidationException()
    throws ValidationException
  {
    if (this.validationExceptions.size() > 0)
      throw new ValidationException("Encountered " + this.validationExceptions.size() + " validation exceptions.", this.validationExceptions);
  }
}