package org.quartz.impl.jdbcjobstore;

import java.sql.Connection;
import java.util.HashSet;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public abstract class DBSemaphore
  implements Semaphore, Constants, StdJDBCConstants, TablePrefixAware
{
  private final Logger log = LoggerFactory.getLogger(getClass());

  ThreadLocal<HashSet<String>> lockOwners = new ThreadLocal();
  private String sql;
  private String insertSql;
  private String tablePrefix;
  private String schedName;
  private String expandedSQL;
  private String expandedInsertSQL;
  private String schedNameLiteral = null;

  public DBSemaphore(String tablePrefix, String schedName, String defaultSQL, String defaultInsertSQL)
  {
    this.tablePrefix = tablePrefix;
    this.schedName = schedName;
    setSQL(defaultSQL);
    setInsertSQL(defaultInsertSQL);
  }

  protected Logger getLog()
  {
    return this.log;
  }

  private HashSet<String> getThreadLocks() {
    HashSet threadLocks = (HashSet)this.lockOwners.get();
    if (threadLocks == null) {
      threadLocks = new HashSet();
      this.lockOwners.set(threadLocks);
    }
    return threadLocks;
  }

  protected abstract void executeSQL(Connection paramConnection, String paramString1, String paramString2, String paramString3)
    throws LockException;

  public boolean obtainLock(Connection conn, String lockName)
    throws LockException
  {
    lockName = lockName.intern();

    if (this.log.isDebugEnabled()) {
      this.log.debug("Lock '" + lockName + "' is desired by: " + Thread.currentThread().getName());
    }

    if (!isLockOwner(conn, lockName))
    {
      executeSQL(conn, lockName, this.expandedSQL, this.expandedInsertSQL);

      if (this.log.isDebugEnabled()) {
        this.log.debug("Lock '" + lockName + "' given to: " + Thread.currentThread().getName());
      }

      getThreadLocks().add(lockName);
    }
    else if (this.log.isDebugEnabled()) {
      this.log.debug("Lock '" + lockName + "' Is already owned by: " + Thread.currentThread().getName());
    }

    return true;
  }

  public void releaseLock(Connection conn, String lockName)
  {
    lockName = lockName.intern();

    if (isLockOwner(conn, lockName)) {
      if (getLog().isDebugEnabled()) {
        getLog().debug("Lock '" + lockName + "' returned by: " + Thread.currentThread().getName());
      }

      getThreadLocks().remove(lockName);
    }
    else if (getLog().isDebugEnabled()) {
      getLog().warn("Lock '" + lockName + "' attempt to return by: " + Thread.currentThread().getName() + " -- but not owner!", new Exception("stack-trace of wrongful returner"));
    }
  }

  public boolean isLockOwner(Connection conn, String lockName)
  {
    lockName = lockName.intern();

    return getThreadLocks().contains(lockName);
  }

  public boolean requiresConnection()
  {
    return true;
  }

  protected String getSQL() {
    return this.sql;
  }

  protected void setSQL(String sql) {
    if ((sql != null) && (sql.trim().length() != 0)) {
      this.sql = sql.trim();
    }

    setExpandedSQL();
  }

  protected void setInsertSQL(String insertSql) {
    if ((insertSql != null) && (insertSql.trim().length() != 0)) {
      this.insertSql = insertSql.trim();
    }

    setExpandedSQL();
  }

  private void setExpandedSQL() {
    if ((getTablePrefix() != null) && (getSchedName() != null) && (this.sql != null) && (this.insertSql != null)) {
      this.expandedSQL = Util.rtp(this.sql, getTablePrefix(), getSchedulerNameLiteral());
      this.expandedInsertSQL = Util.rtp(this.insertSql, getTablePrefix(), getSchedulerNameLiteral());
    }
  }

  protected String getSchedulerNameLiteral()
  {
    if (this.schedNameLiteral == null)
      this.schedNameLiteral = ("'" + this.schedName + "'");
    return this.schedNameLiteral;
  }

  public String getSchedName() {
    return this.schedName;
  }

  public void setSchedName(String schedName) {
    this.schedName = schedName;

    setExpandedSQL();
  }

  protected String getTablePrefix() {
    return this.tablePrefix;
  }

  public void setTablePrefix(String tablePrefix) {
    this.tablePrefix = tablePrefix;

    setExpandedSQL();
  }
}